/* 11.07.01:
 * MAJOR CHANGES:
 *
 * - substituted long int by LONG
 * - only the PV keywords, the CD matrix and
 *   the CRVAL values are now written in the
 *   ASTROM update section. The other keywords
 *   (FIELD_POS, MPIX) were used never.
 * - old PV keywords present in the header are
 *   now deleted before new ones are written.
 *   This allows the program to be run several
 *   times over the same image (see also next point).
 * - the photometric keywords are overwritten now
 *   in case they already existed, not created again
 * - introduces a new keyword in the control struct:
 *   MODE. If set to LDAC all processing of the PV
 *   keywords is done as up to now. Is this value set
 *   to WCSSTANDARD the PV keywords are written
 *   according to the "standard" defined Calabretta/
 *   Greisen (this differs in several points from Erik's
 *   implementation). Hence the images can be coadded
 *   by the swarp program now. This conversion from
 *   the LDAC to WCSSTANDARD style is realised in the new
 *   pvconvert subroutine.
 *
 * 19.07.01:
 * corrected errors in the WCSSTANARD case for the PV
 * polynomials. It turned out I can simply use the LDAC
 * polynomials converted to degrees and with a reshuffled
 * ordering if the scale keywords (CDELT, CD-matrix) are set 
 * to the unity matrix. This gives away the posibility
 * to make object overlays in skycat.
 *
 * 16.10.01:
 * Some elements of the code had to be included in an if statement
 * asking whether astrometric calibration should be added at all
 * (error when only photometric information was asked for)
 *
 * 19.10.01:
 * included the possibility to include the image seeing into the
 * weighting; if applied the seeing contributes with 1./(seeing**2)
 * to the WGTSCALE.
 *
 * 22.09.02:
 * If a file is not present the program only gives a warning and
 * continues. It no longer simply stops.
 *
 * 06.06.03:
 * for the moment I interpret the FITS images I read
 * (changed set_no_interp(NO) to set_no_interp(YES)). The
 * 'NO' case makes problems under HP.
 *
 * 26.06.2003:
 * - I included the possibility to do the updating
 *   only on images with a given IMAGEID. Hence, the task
 *   of Header updating can be parallelised now.
 *
 * - I initialised several variables when they are declared
 *   to avoid compiler warnings when optimisation flags are set.
 *
 * 09.02.2005:
 * - Headers of out images are now updated with Header information
 *   on the performed processing and a global version number.
 *
 * - I removed compiler warnings under gcc with -Wall -W -pedantic
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "fitscat.h"
#include "tsize.h"
#include "options_globals.h"
#include "utils_globals.h"
#include "image_io_globals.h"
#include "field_desc.h"
#include "fits_image_io_globals.h"
#include "relphotom_tab_globals.h"
#include "add_image_calibs_vers.h"
#include "global_vers.h"
#include "t_iso8601.h"

#define MAXLEV 1
#define MAXFILES 100
#define MAXCHAR 255

#define PI 3.14159265

static char	header[MAXCHAR];
static char     **exptime_names, **exptime_units;
static int      nexptimes;
static double   *exptime, *exptime_def;
static LONG     *lexptime;
static char     **out_names, **in_columns, **in_tables;
static int      nnames;


typedef enum {ALL, ASTROMETRY, PHOTOMETRY} ptype;

typedef struct {
   char conffile[MAXCHAR];
   char listfile[MAXFILES][MAXCHAR];
   char outfile[MAXFILES][MAXCHAR];
   char imageinpath[MAXCHAR];
   char imageoutpath[MAXCHAR];
   char mode[MAXCHAR];
   char newext[MAXCHAR], bckext[MAXCHAR], origext[MAXCHAR];
   int  ilist;
   int imageidvalue;
   char mag_scale[MAXCHAR];
   char cdelt1[MAXCHAR], cdelt2[MAXCHAR];
   char crval1[MAXCHAR], crval2[MAXCHAR];
   char crpix1[MAXCHAR], crpix2[MAXCHAR];
   char xwid[MAXCHAR], ywid[MAXCHAR];
   char exptime_keyw[TEXTVECLEN], exptime_unit[TEXTVECLEN];
   char out_name[TEXTVECLEN], in_table[TEXTVECLEN], in_column[TEXTVECLEN];
   char flxscale[MAXCHAR], wgtscale[MAXCHAR], fitsfile[MAXCHAR];
   char zeropoint[MAXCHAR];
   char bkgsig[MAXCHAR];
   char seeing[MAXCHAR];
   char imageid[MAXCHAR];
   bool delinput;
   bool use_seeing;
   bool use_imageid;
   ptype proc_type;
} control_struct;

control_struct control;

option opts[]={
         {"DEL_INPUT", BOOLEAN, &control.delinput, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "",
          {"YES", "NO", "", "", ""}, 0,
          "Remove the input file after copying to output image"},
         {"FITSFILE", TEXT, control.fitsfile, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "FITSFILE",
          {"", "", "", "", ""}, 0,
          "The name of the FIELDS table column containing the image filename"},
         {"MAG_SCALE", TEXT, control.mag_scale, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "MAG_APER",
          {"", "", "", "", ""}, 1,
          "The name of the magnitude estimator used for FLXSCALE"},
         {"FLXSCALE", TEXT, control.flxscale, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "FLXSCALE",
          {"", "", "", "", ""}, 0,
          "The actual FITS keyword for the FLXSCALE"},
         {"BACKGROUND_SIG", TEXT, control.bkgsig, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "SEXBKDEV",
          {"", "", "", "", ""}, 0,
          "The name of the FIELDS table column containing the background standard deviation value"},
         {"USE_SEEING", BOOLEAN, &control.use_seeing, 0, 0, 0.0, 0.0,
          0, 0.0, NO, "", {"YES", "NO", "", "", ""}, 0,
          "Is the image seeing included in weighting or not"},
         {"SEEING", TEXT, control.seeing, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "SEXSFWHM",
          {"", "", "", "", ""}, 0,
          "The name of the FIELDS table column containing the seeing of the images"},
         {"USE_IMAGEID", BOOLEAN, &control.use_imageid, 0, 0, 0.0, 0.0,
          0, 0.0, NO, "", {"YES", "NO", "", "", ""}, 0,
          "Is an IMAGEID used to decide whether to update or not"},
         {"IMAGEID", TEXT, control.imageid, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "IMAGEID",
          {"", "", "", "", ""}, 0,
          "The name of the FIELDS table column containing the IMAGEID of the images"},
         {"IMAGEIDVALUE", INT, &control.imageidvalue, 1, 40, 0.0, 0.0,
          1, 0.0, NO, "",
          {"", "", "", "", ""}, 1,
          "The IMAGEID of frames to be modified"},
         {"WGTSCALE", TEXT, control.wgtscale, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "WGTSCALE",
          {"", "", "", "", ""}, 0,
          "The actual FITS keyword for the WGTSCALE"},
         {"ZEROPNT", TEXT, control.zeropoint, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "ZEROPNT",
          {"", "", "", "", ""}, 0,
          "The actual FITS keyword for the ZEROPNT"},
         {"ORIG_EXT", TEXT, control.origext, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, ".fits",
          {"", "", "", "", ""}, 0,
          "The extension name of the input FITS files"},
         {"CDELT1", TEXT, control.cdelt1, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "CDELT1",
          {"", "", "", "", ""}, 0,
          "The name of the FIELDS table column containing the x coordinate pixel size"},
         {"CDELT2", TEXT, control.cdelt2, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "CDELT2",
          {"", "", "", "", ""}, 0,
          "The name of the FIELDS table column containing the y coordinate pixel size"},
         {"CRPIX1", TEXT, control.crpix1, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "CRPIX1",
          {"", "", "", "", ""}, 0,
          "The name of the FIELDS table column containing the RA reference pixel"},
         {"CRPIX2", TEXT, control.crpix2, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "CRPIX2",
          {"", "", "", "", ""}, 0,
          "The name of the FIELDS table column containing the DEC reference pixel "},
         {"CRVAL1", TEXT, control.crval1, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "CRVAL1",
          {"", "", "", "", ""}, 0,
          "The name of the FIELDS table column containing the RA of the image center"},
         {"CRVAL2", TEXT, control.crval2, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "CRVAL2",
          {"", "", "", "", ""}, 0,
          "The name of the FIELDS table column containing the DEC of the image center"},
         {"MODE", TEXT, control.mode, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "LDAC",
          {"", "", "", "", ""}, 0,
          "The name of the MODE for PV polynomials (LDAC or WCSSTANDARD)"},
            {"XWID", TEXT, control.xwid, 0, 0, 0.0, 0.0, 
              0, 0.0, NO, "MAPNAXS1",
             { "", "", "", "", ""}, 1,
              "Name of FIELDS table column containing the x field width"},
            {"YWID", TEXT, control.ywid, 0, 0, 0.0, 0.0, 
              0, 0.0, NO, "MAPNAXS2",
             { "", "", "", "", ""}, 1,
              "Name of FIELDS table column containing the y field width"},
         {"EXPTIME_KEYWORD", TEXTVEC, control.exptime_keyw, 0, 0, 0.0, 0.0,
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0,
          "Name of the FIELDS table column containign the exposure time"},
         {"EXPTIME_UNIT", TEXTVEC, control.exptime_unit, 0, 0, 0.0, 0.0,
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0,
          "The unit in which the EXPTIME_KEYWORD is to be expressed"},
         {"OUT_NAME", TEXTVEC, control.out_name, 0, 0, 0.0, 0.0,
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0, 
          "Series of FITS keywords copied from (IN_COLUMN, IN_TABLE) to the image header"},
         {"IN_COLUMN", TEXTVEC, control.in_column, 0, 0, 0.0, 0.0,
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0,
          "The name of the IN_TABLE column that should be copied to OUT_NAME in the FITS image header"},
         {"IN_TABLE", TEXTVEC, control.in_table, 0, 0, 0.0, 0.0,
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0,
          "The name of the  table comntaining IN_COLUMN which should be copied to the FITS image header"},
         {"", BOOLEAN, NULL, 0, 0, 0.0, 0.0, 0, 0.0, NO, "", {"", "", "", "", ""}, 0, ""},
};

void usage(char *me)
{
/****** usage ***************************************************
 *
 *  NAME	
 *      usage - Print out a message about how-to-use the program
 *
 *  SYNOPSIS
 *      void usage(char *me)
 *
 *  FUNCTION
 *      Print to stderr a message about how to use th main program
 *      The execution will stop and the exit code of the program is set
 *      to non-zero.
 *
 *  INPUTS
 *      me  - The calling name of the program
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 08-08-1997
 *
 ******************************************************************************/
    fprintf(stderr,"%s Version: %s; %s (%s)\n",NAME,VERS,__PIPEVERS__,DATE);
    fprintf(stderr,"Usage: %s	 -i incat1 [incat2 [incat3 ...]]\n", me);
    fprintf(stderr,"		 -o configfile\n");
    fprintf(stderr,"		[-a]  (add astrometry only)\n");
    fprintf(stderr,"		[-p]  (add photometry only)\n");
    fprintf(stderr,"		[-d image_directory] | \n");
    fprintf(stderr,"		   -di input_image_directory ");
    fprintf(stderr," -do output_image_directory]\n");
    fprintf(stderr,"		[-e extension_of_file_to_write]\n");
    fprintf(stderr,"		[-o extension_of_file_to_read]\n");
    exit(1);
}

void init(int argc, char *argv[])
{
/****** int ***************************************************
 *
 *  NAME	
 *      init - Initialize the program from argument and configuration file
 *
 *  SYNOPSIS
 *      void init(int argc, char *argv[])
 *
 *  FUNCTION
 *      This function reads the command line arguments and interprets the
 *      configuration file
 *
 *  INPUTS
 *      argc - Number of arguments passed
 *      argv - The argument char array
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *      The control structure is set with the argument list parameters
 *      and the configuration parameters
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 08-08-1997
 *
 ******************************************************************************/
    int i, j, k;

    if (argc == 1) usage(argv[0]);
/*
 *  Initialize command line options with some defaults
 */
    sprintf(control.conffile,"add_image_astrom.conf");
    sprintf(control.imageinpath,".");
    sprintf(control.imageoutpath,".");
    sprintf(control.newext,".new.fits");
    sprintf(control.bckext,".bck.fits");
    init_options(0, opts);
    control.proc_type = ALL;
/*
 *  Parse the command line
 */
    for (i=1; i<argc; i++) {
       if (argv[i][0] == '-') {
          switch(argv[i][1]) {
          case 'i': control.ilist= 0;
                    strcpy(control.listfile[control.ilist],argv[++i]);
                    while (i+1<argc && argv[i+1][0] != '-') {
                       strcpy(control.listfile[++control.ilist],argv[++i]);
                    }
                    ++control.ilist;
                    break;
          case 'd': if (argv[i][2] != '\0') {
                       switch(argv[i][2]) {
                       case 'i':
                         strcpy(control.imageinpath,argv[++i]);
                         break;
                       case 'o':
                         strcpy(control.imageoutpath,argv[++i]); 
                         break;
                       }
                    } else {
                      strcpy(control.imageinpath,argv[++i]); 
                      strcpy(control.imageoutpath, control.imageinpath);
                    }
                    break;
          case 'a': control.proc_type = ASTROMETRY;
                    break;
          case 'p': control.proc_type = PHOTOMETRY;
                    break;
          case 'e': strcpy(control.newext,argv[++i]);
                    break;
          case 'o': strcpy(control.bckext,argv[++i]);
                    break;
          case 'c': strcpy(control.conffile,argv[++i]);
                    break;
          case '@': print_options(argv[0]);
                    exit(0);
          default:  parse_options(control.conffile);
                    j = i++; k = i;
                    set_command_option(argv[j], argv[k]);
                    break;
          }
      } else {
          syserr(WARNING,"init","Command line syntax error\n");
          usage(argv[0]);
      }
   }
   finish_options(control.conffile);
   exptime_names = copy_textvec(control.exptime_keyw, &nexptimes);
   exptime_units = copy_textvec(control.exptime_unit, &i);
   if (i != nexptimes)
      syserr(FATAL, "init", "Number of %s parameters differs from %s\n",
              control.exptime_unit, control.exptime_keyw);
   ED_CALLOC(exptime_def, "init", double, nexptimes);
   for (i=0;i<nexptimes;i++)
      exptime_def[i] = atof(exptime_units[i]);
   out_names = copy_textvec(control.out_name, &nnames);
   in_columns = copy_textvec(control.in_column, &i);
   if (i != nnames)
      syserr(FATAL, "init", "Number of %s parameters differs from %s\n",
              control.out_name, control.in_column);
   in_tables = copy_textvec(control.in_table, &i);
   if (i != nnames)
      syserr(FATAL, "init", "Number of %s parameters differs from %s\n",
              control.out_name, control.in_table);
   for (i=0;i<nnames;i++)
      if (strcmp(in_tables[i], "FIELDS") != 0 &&
          strcmp(in_tables[i], "DISTORTIONS") != 0) 
         syserr(FATAL, "init", 
              "Can only copy information from FIELDS or DISTORTIONS tables not %s\n", in_tables[i]);
}

/* 
 * The function pvconvert converts a PV polynomial value
 * in the LDAC format to that specified in the FITS standard.
 * The apvvalue (expected to be in "relative" image pixel 
 * coordinates) is converted to radians.
 * The conversion mainly reshuffles the ordering within the
 * polynomial description.
 *
 * memory for anewpvstring and anewpvvalue has to be allocated
 * in the calling function.
 */
void pvconvert(char *apvstring, char *anewpvstring, 
               double apvvalue, double *anewpvvalue)
{
  int k, l; /* the first and second values caracterising
               the PV polynomial (PVk_l) */

  int degree, degx, degy, ncoff; 
  float ftmp;

  /* first check whether we have a PV string */
  if(strncmp(apvstring, "PV", 2) != 0)
  {
    syserr(FATAL, "no PV string given to pvconvert");
  }
  
  sscanf((const char *)apvstring, "PV%d_%d", &k, &l);

  /* has the polynomial a "valid" degree between 0 and 5?? */
  if((k < 1) || (l < 1) || (k > 20) || (l > 20))
  {
    syserr(FATAL, "only WCSSTANDARD polynomials with degree <= 2 supported right now");
  }

  /* determine the degree to that the currect PV term
   * belongs 
   */

  ftmp = 0.5*(sqrt(1.+8.*k)-1.);

  if((ftmp-(int)ftmp) > 1.0e-04)
  {
    degree = (int) ftmp;
  }
  else
  {
    degree = (int) (ftmp-1.);
  }

  /* determine the degree of the x and the y term
   */

  ncoff = (degree)*(degree+1)/2;
  degy = k-(ncoff+1);
  degx = degree - degy;

  /* calc the new pvvalue and write the new polynom
   * string (swarp wants degrees, not radians
   */
  *anewpvvalue = apvvalue * 180.0 / 3.14159;

  /* now bring 'k' to the swarp (WCSSTANDARD) order */

  k-=1;

  if(k>2) k++;
  if(k>10) k++;

  /* take into account that for the two
     component in the WCSSTANDARD case the
     powers for x and y are switched
  */

  if(l==2)
  {
    k = k+(degx-degy);
  }

  sprintf(anewpvstring, "PV%d_%d", l, k);  
}

int main(int argc, char *argv[])
{
/****** add_image_calibs ***************************************************
 *
 *  NAME	
 *       add_image_calibs - Add some calibration information to the images
 *
 *  SYNOPSIS
 *       add_image_calibs -i incat1 [incat2 [incat3 ...]]
 *                       [-a add astrometry only]
 *                       [-p add photometry only]
 *                       [-d path_to_file_files]
 *                       [-o extension of fits file to be read]
 *                       [-e extension of fits file to be written]
 *
 *  FUNCTION
 *      This routine adds some calibration parameters to the fits image
 *      headers. It can either add astrometric, photometric, or both
 *      calibration parameters. 
 *
 *      The astrometric parameters added to the fits header are the
 *      plate related polynomial coefficents of the astrometric plate
 *      deformation (PV?_?)
 *
 *      The photometric parameters added to the fits header are the
 *      relative photometric flux scale, and the relative weighting 
 *      per image. The flux scale is used to divide image fluxes to a
 *      unique system. The weight value needs to be used as multiplier
 *      in the plate to plate pixel averaging.
 *
 *  INPUTS
 *      i      - The input catalog list
 *      a|p    - Either astrometric or photometric calibration to be added
 *               Omission of these option defaults to adding both.
 *      d      - Full UNIX path name to the directory where the FITS files are
 *      e      - The extension of the FITS file to be created.*
 *      o      - The extension of the FITS file to be read. *
 *
 *  RESULTS
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *      The FITS file name read from the catalog consists of the filename
 *      and an extension. This extension is removed and the -e and -o
 *      arguments are pasted to the end of this FITS filename. The
 *      directory path given with -d is pasted in front of the FITS filename.
 *      The original FITS filename extension as stored in the catalog
 *      can be provided by the configuration parameter ORIG_EXT.
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 08-08-1997
 *
 ******************************************************************************/
    fitsfile_struct	*ffd, *ffn;
    char	infile[MAXCHAR], outfile[MAXCHAR], *t;
    char        tmpstring[MAXCHAR];
    char        keyname[MAXCHAR];
    field_params	**field_desc;
    float	*data;
    int		len;
    int		i, j, k, nfields;
    char	*filenames, tfile[MAXCHAR];
    catstruct	*incat;
    tabstruct	*in_ftab, *in_astromtab=NULL, *in_photomtab, *ctab;
    keystruct	*tempkey, *fitskey, *sigkey, *seeingkey;
    keystruct	**astromkeys=NULL, *ckey;
    keystruct   *imageidkey;
    float	*float_ptr, *a, zp_ave, *flxscale=NULL, *wgtscale=NULL;
    double	*bkgsig, *seeing=NULL, cdelt1_i=0., cdelt2_i=0.;
    double	flxave;
    double	ddd=0., dummy=0., raoff=0., decoff=0.;
    double      *double_ptr, keyvalue;
    double	cd1_1=0., cd1_2=0., cd2_1=0., cd2_2=0.;
    LONG        *imageidvalue=NULL;
    char	*ptr, nname[MAXCHAR];
    float	eps = 1.0e-6;
    float       *xx, yy, xxx, yyy;

    init(argc, argv);

    sprintf(header,"%s Version: %s (%s)",NAME,VERS,DATE);
    VPRINTF(NORMAL,"\n   %s\n\n",header);

    for (i=0;i<control.ilist;i++) {
/* 
 *     For each input catalog get the FITS filename list and addition
 *     parameters
 */
       if (!(incat = read_cat(control.listfile[i])))
          syserr(FATAL, "add_image_astrom", "Cannot open catalog %s\n", 
             control.listfile[i]);
       if (!(in_ftab = name_to_tab(incat, "FIELDS", 0)))
          syserr(FATAL, "add_image_astrom", "No table FIELDS in %s catalog\n",
             control.listfile[i]);
       field_desc = load_field_description(in_ftab, &nfields,
                            control.crval1, control.crval2,
                            control.cdelt1, control.cdelt2,
                            control.crpix1, control.crpix2,
                            control.xwid, control.ywid);

       ED_GETKEY(in_ftab, fitskey, control.fitsfile ,filenames, nfields, 
           "add_image_astrom");

/* 
 * Get IMAGEID information if necessary
 */
       if (control.use_imageid == YES)
       {
          if (!(imageidkey = read_key(in_ftab, control.imageid)))
             syserr(FATAL, "add_image_astrom", "Cannot find key %s\n", 
                control.imageid);

          switch (imageidkey->ttype)
          {
          case T_LONG:
            ED_GETVEC(imageidvalue, "main", imageidkey, 
                      imageidkey->name, LONG, LONG);
            break;
          case T_SHORT:
            ED_GETVEC(imageidvalue, "main", imageidkey, 
                      imageidkey->name, LONG, short);
            break;
          default:
             syserr(FATAL, "main", 
                    "Unsupported key type for %s. Only SHORT and LONG types supported.\n", imageidkey->name);        
            break;
          }
       }

       
/*
 *     Get the astrometric calibration parameters
 */
       if (control.proc_type == ALL || control.proc_type == ASTROMETRY) {
          if (!(in_astromtab = name_to_tab(incat, "DISTORTIONS", 0)))
             syserr(FATAL, "add_image_calibs", 
                "No table DISTORTIONS in %s catalog\n", control.listfile[i]);
          ED_CALLOC(astromkeys, "add_image_calibs", keystruct *, 
                    in_astromtab->nkey);
          read_keys(in_astromtab, NULL, astromkeys, in_astromtab->nkey, NULL);
       }
/*
 *     Get the photoemtric calibration parameters
 */
       if (control.proc_type == ALL || control.proc_type == PHOTOMETRY) 
       {
          /* load photometric information contained in the FIELDS table */

          ED_GETKEY(in_ftab, tempkey, exptime_names[0], exptime, nfields, 
              "add_image_photom");

          if (control.use_seeing == YES)
	  {
             ED_GETKEY(in_ftab, seeingkey, control.seeing, seeing, nfields, 
                    "add_image_photom");            
	  }

          if (tempkey->ttype == T_LONG) {
             lexptime = tempkey->ptr;
             ED_CALLOC(exptime, "load_photo_fields", double, (nfields));
             for (i=0;i<nfields;i++) exptime[i] = (double) lexptime[i];
          } else {
             exptime = tempkey->ptr;
          }

          ED_GETKEY(in_ftab,      sigkey, control.bkgsig, bkgsig, nfields, 
                    "add_image_photom");

          /* load photometric information contained in the 
             RELPHOTOM table
          */

          if (!(in_photomtab = name_to_tab(incat, "RELPHOTOM", 0)))
             syserr(FATAL, "add_image_calibs", 
                "No table RELPHOTOM in %s catalog\n", control.listfile[i]);
          len = strlen(control.mag_scale);
          if (control.mag_scale[len+1] != 0) {
             sprintf(nname,"%s(%-d)",control.mag_scale,
                                     control.mag_scale[len+1]);
          } else {
             strcpy(nname, control.mag_scale);
          }
          read_relphot_elem(in_photomtab, nname, &a, &xx, &zp_ave, &yy);
          ED_CALLOC(flxscale, "add_image_calibs", float, nfields);
          ED_CALLOC(wgtscale, "add_image_calibs", float, nfields);
          flxave = 0.0;
          for (i=0;i<nfields;i++) {
             flxscale[i] = pow(10.0,(-a[i]/2.5))*exptime_def[0]/exptime[i]; 
             if (fabs(flxscale[i]) < eps)
                syserr(FATAL, "add_image_calibs", 
                    "Flux scale value too small\n");
             if (fabs(bkgsig[i]) < eps)
                syserr(FATAL, "add_image_calibs", 
                    "Background Sigma value too small\n");
/* 
 *           So the inversion of the flscale needed to be implemented
 *           here too.
 */
             if(control.use_seeing == YES)
	     {
               wgtscale[i] = 1.0/(bkgsig[i]*flxscale[i]*seeing[i]*
                                  bkgsig[i]*flxscale[i]*seeing[i]);
	     }
             else
	     {
               wgtscale[i] = 1.0/(bkgsig[i]*flxscale[i]*
                                  bkgsig[i]*flxscale[i]);
	     }
          }
       }
/*
 *     Now put the values into the image headers
 */
       set_no_interp(NO);
       for (i=0;i<nfields;i++) {
          strncpy(tfile, &filenames[i*fitskey->nbytes],fitskey->nbytes);
/*
 *        Remove trailing spaces and .fits extension
 */
          for (j=fitskey->nbytes;j>=0;j--)
             if (tfile[j] == ' ') tfile[j] = '\0';
          if (control.origext[0] != '\0') {
             if (!(t=strstr(tfile, control.origext)))
                syserr(WARNING, "No extension %s in original filename %s\n",
                   control.origext, tfile);
             t[0] = '\0';
          }
          sprintf(infile,"%s/%s%s", control.imageinpath,tfile,control.bckext);
          sprintf(outfile,"%s/%s%s", control.imageoutpath,tfile,control.newext);
/*
 *        Open the FITS file and get the raw, uninterpreted data
 */
/*        
          Check first whether file exists; is not give a warning but
          continue
*/
	  if(ED_FCHECK("add_image_calibs", infile) == 0)
	  {
            /* check for IMAGEID */
            if((control.use_imageid == NO) || 
               ((control.use_imageid==YES) && (imageidvalue[i] == control.imageidvalue)))
	    {
              ffd = open_fits_file(infile);
              switch (ffd->bitpix) {
              case 16:
                  ffd->pix_type = T_SHORT;
                  break;
              case -32:
                  ffd->pix_type = T_FLOAT;
                  break;
              }
              data = interp_fits_image(ffd, NO); 
              
              if (control.proc_type == ALL || control.proc_type == ASTROMETRY) 
              {
              
                if ((cdelt1_i = r_fits_float(control.cdelt1, ffd->head, 0.0))==0.0)            {
                   xxx = r_fits_float("CD1_1", ffd->head, 0.0);
                   yyy = r_fits_float("CD1_2", ffd->head, 0.0);
                   if (fabs(xxx)>fabs(yyy)) {
                	  cdelt1_i = xxx;
                   } else {
                	  cdelt1_i = yyy;
                   }
                }
                if ((cdelt2_i = r_fits_float(control.cdelt2, ffd->head, 0.0))==0.0)            {
                   xxx = r_fits_float("CD2_2", ffd->head, 0.0);
                   yyy = r_fits_float("CD2_1", ffd->head, 0.0);
                   if (fabs(xxx)>fabs(yyy)) {
                	  cdelt2_i = xxx;
                   } else {
                	  cdelt2_i = yyy;
                   }
                }
              }
/*       
 *       Create the output FITS file
 */      
              ffn = new_fits();
              inherit_fits(ffn, ffd);
              ffn->pix_type = ffd->pix_type;

              /* add history comments */
              w_fits_str("HISTORY", ffn->head, "");
              sprintf(tmpstring, "add_image_calibs: %s", __PIPEVERS__);
              w_fits_str("HISTORY", ffn->head, (char *)tmpstring);
              sprintf(tmpstring, "add_image_calibs: called at %s", get_datetime_iso8601());
              w_fits_str("HISTORY", ffn->head, (char *)tmpstring);

/*       
 *       Selectively add the calibration information and correct cdelt
 *       dependencies
 */      
              if (control.proc_type == ALL || control.proc_type == ASTROMETRY) 
              {
/*       
 *       
 */      
                sprintf(tmpstring, "add_image_calibs: add astrometry imformation in %s format", control.mode);
                w_fits_str("HISTORY", ffn->head, (char *)tmpstring);
              
/*       
 * first delete PV polynomials already present.
 * As they can be LDAC and WCSSTANDARD type polynomials
 * we have to fully scan both components. We assume
 * a maximum degree of 5 what gives 23 for the maximum
 * coefficient (WCSSTANDARD style)
 */       

                 for(j=0; j<25; j++)
                 {
                   for(k=0; k<25; k++)
                   {
                     sprintf(tmpstring, "PV%d_%d", j, k);
                     d_fits(tmpstring, ffn->head);
                   }
                 }
              
                 for (j=0;j<in_astromtab->nkey;j++) 
                 {
/*       
 * we Only have to take over PV polynomial keywords
 */      
                    if(strncmp(astromkeys[j]->name, "PV", 2) == 0)
              	{
/*       
 * we only accept float and double type for PV polynomials.
 * short and double make no sense here. We also write out
 * everything in double type.
 */      
                      switch(astromkeys[j]->ttype) 
                      {
                      case T_FLOAT:
                         float_ptr = astromkeys[j]->ptr;
                         if (strcmp(astromkeys[j]->name, "PV1_1") != 0 &&
                      	 strcmp(astromkeys[j]->name, "PV1_2") != 0)
                         dummy = (double)(float_ptr[i] / field_desc[i]->cdelt1 * cdelt1_i);
                         if (strcmp(astromkeys[j]->name, "PV1_1") == 0) {
                      	raoff = (double) float_ptr[i];
                      	dummy = 0.0;
                         }
                         if (strcmp(astromkeys[j]->name, "PV1_2") == 0) {
                      	decoff  = (double) float_ptr[i];
                      	dummy = 0.0;
                         }
                         ddd = (double) float_ptr[i];
              
                         break;
                      case T_DOUBLE:
                         double_ptr = astromkeys[j]->ptr;
                         if (strcmp(astromkeys[j]->name, "PV1_1") != 0 &&
                      	 strcmp(astromkeys[j]->name, "PV1_2") != 0)
                         dummy = (double)(double_ptr[i] / field_desc[i]->cdelt1 * cdelt1_i);
                         if (strcmp(astromkeys[j]->name, "PV1_1") == 0) {
                      	raoff = double_ptr[i];
                      	dummy = 0.0;
                         }
                         if (strcmp(astromkeys[j]->name, "PV1_2") == 0) {
                      	decoff  = double_ptr[i];
                      	dummy = 0.0;
                         }
                         ddd = double_ptr[i];
                         break;
                      default:
                         syserr(FATAL, "add_image_calibs", 
                      		   "Unexpected type for PV polynomials in input table. Only float and double accepted.\n");
                         break;
                      }
              
/*       
 * now deal with the polynomial conversion if WCSSTANDARD
 * type polynomials are requested and write the keyword
 * to the header
 */      
              	   if(strcmp(control.mode, "LDAC") == 0)
              	   {
              	     strcpy(keyname, astromkeys[j]->name);
                         keyvalue=dummy;
              	   }
                       else
              	   {
                         pvconvert(astromkeys[j]->name, keyname, 
                                   dummy, &keyvalue);
              	   }
                       s_fits_double_c(keyname, ffn->head, keyvalue,
                      		   astromkeys[j]->comment);
              
                      if (strcmp(astromkeys[j]->name, "PV2_1") == 0) cd1_1 = ddd;
                      if (strcmp(astromkeys[j]->name, "PV2_2") == 0) cd1_2 = ddd;
                      if (strcmp(astromkeys[j]->name, "PV3_1") == 0) cd2_1 = ddd;
                      if (strcmp(astromkeys[j]->name, "PV3_2") == 0) cd2_2 = ddd;
                    }
                 }
              
                 cd1_1 = cd1_1 * 180.0 / PI;
                 cd1_2 = cd1_2 * 180.0 / PI;
                 cd2_1 = cd2_1 * 180.0 / PI;
                 cd2_2 = cd2_2 * 180.0 / PI;
                 field_desc[i]->crval1 += raoff/cos(field_desc[i]->crval2*PI/180.0)
                			    * 180.0 / PI;
                 field_desc[i]->crval2 += decoff * 180.0 / PI;
              
                 s_fits_double(control.crval1, ffn->head, field_desc[i]->crval1);
                 s_fits_double(control.crval2, ffn->head, field_desc[i]->crval2);
              
                 /* swarp does not interpret polynomials correctly if a CD
                    matrix is present. Hence in this case we have to delete
                    potential CD keywords.
                 */
                 if(strcmp(control.mode, "LDAC") == 0)
                 { 
                   s_fits_double("CD1_1",        ffn->head, cd1_1);
                   s_fits_double("CD1_2",        ffn->head, cd1_2);
                   s_fits_double("CD2_1",        ffn->head, cd2_1);
                   s_fits_double("CD2_2",        ffn->head, cd2_2);
                 }
                 else
                 {
                   d_fits("CD1_1", ffn->head);
                   d_fits("CD2_2", ffn->head);
                   d_fits("CD1_2", ffn->head);
                   d_fits("CD2_1", ffn->head);
                 } 
/*       
 *          Skycat needs to have these two keywords removed in orer to use
 *          the CD matrix!! Stupid!!
 */      
                 if(strcmp(control.mode, "LDAC") == 0)
                 {
                   d_fits("CDELT1", ffn->head);
                   d_fits("CDELT2", ffn->head);
                 }
                 else
                 /* The CDELT keywords in swarp are just used to scale the
                    parameters of the PV polynomials. As mine are scaled to
                    give directly degrees, the CDELT keywords have simply
                    to be set to 1. With this ALL the astrometric scaling is in
                    the PV polynomials - as a side effect the image cannot be
                    viewed (object overlays) with skycat.
                 */
                 {
                     s_fits_double("CDELT1", ffn->head, 1.0);
                     s_fits_double("CDELT2", ffn->head, 1.0);
                 }
              }
              if (control.proc_type == ALL || control.proc_type == PHOTOMETRY) 
              {
                 sprintf(tmpstring, "add_image_calibs: add photometry imformation");
                 w_fits_str("HISTORY", ffn->head, (char *)tmpstring);

                 s_fits_float_c(control.flxscale, ffn->head, flxscale[i], 
                                "Multiplication factor for frame flux");
                 s_fits_float_c(control.wgtscale, ffn->head, wgtscale[i], 
                                "Multiplication factor for weight map");
                 s_fits_float_c(control.zeropoint, ffn->head, (zp_ave), 
                                "Photometric Zero Point");
              }
              for (k=0;k<nnames;k++) {
                 if (!(ctab = name_to_tab(incat, in_tables[k], 0)))
                    syserr(FATAL, "add_image_calibs",
                       "No table %s in %s catalog\n", in_tables[k]);
                 if (!(ckey = name_to_key(ctab, in_columns[k])))
                    syserr(FATAL, "add_image_calibs",
                       "No key %s in %s catalog\n", in_columns[k]);
                 if (ckey->ptr == NULL) {
                    ED_GETKEY(ctab, ckey, in_columns[k], ptr, nfields,
                             "add_image_astrom");
                 } else {
                    ptr = ckey->ptr;
                 }
                 switch (ckey->ttype) {
                 case T_SHORT:
                       s_fits_short_c(out_names[k], ffn->head, 
                                      ((short int *)ptr)[i], ckey->comment);
                       break;
                 case T_LONG:
                       s_fits_int_c(out_names[k], ffn->head, 
                                    ((LONG *)ptr)[i], ckey->comment);
                       break;
                 case T_FLOAT:
                       s_fits_float_c(out_names[k], ffn->head, 
                                      ((float *)ptr)[i], ckey->comment);
                       break;
                 case T_DOUBLE:
                       s_fits_double_c(out_names[k], ffn->head, 
                                       ((double *)ptr)[i], ckey->comment);
                       break;
                 default:
                      syserr(FATAL, "add_image_calibs", 
                      "No provision for BYTE or STRING copying implemented yet");
                 }
              }
              w_fits_pad(ffn->head);
     /*       
      *       Now save this FITS file
      */      
              save_fits(ffn, (unsigned char *)data, outfile);
              ED_FREE(data);
              close_fits(ffd, control.delinput);
	    }
            else
	    {
              syserr(WARNING, "add_image_calibs", 
                "File %s has IMAGEID %ld\n", infile, imageidvalue[i]);              
	    }
	  }
	  else
	  {
            syserr(WARNING, "add_image_calibs", 
              "File %s not present\n", infile);
	  }
       }
    }
    return 0;
}

