/* 04.11.01:
 * first version
 *
 * 07.11.01:
 * - corrected error in the linear term of the conversion
 *
 * - Keywords are now written in LDAC or WCSSTANDARD form
 *   to a new Fits file.
 *
 * 14.01.02:
 * changed the position of the call 'set_no_interp(YES)'
 * before 'data = interp_fits_image(ffo, NO)'.
 * Doing it the other way around led to byte swapped
 * images on Linux machines
 *
 * 22.05.02:
 * the interpretation of cd1_2 and cd2_1 was wrong in all terms
 * except the linear ones. This was apparent in the SUBARU data
 * where these terms are dominant over cd1_1 and cd2_2.
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "fitscat.h"
#include "options_globals.h"
#include "utils_globals.h"
#include "image_io_globals.h"
#include "fits_image_io_globals.h"
#include "astrometrix_convert_vers.h"

#define MAXFILES 1
#define MAXCHAR 255

#define PI 3.14159265

static char	header[MAXCHAR];

typedef struct {
   char listfile[MAXFILES][MAXCHAR];
   char fitsinput[MAXCHAR];
   char fitsoutput[MAXCHAR];
   char type[MAXCHAR]; /* type of output polynomial: LDAC or WCSSTANDARD */
   char verbosity[MAXCHAR]; /* type of output polynomial: LDAC or WCSSTANDARD */
} control_struct;

control_struct control;

/* 
   The maximum index for the LDAC/swarp polynomials.
   For the moment the maximum degree is 3 and hence 
   the maximum index 10.
*/

#define MAXPOLINDEX 10

#define __USAGE__ "\n\
USAGE:\n\
      astrometrix_convert -i input catalog with DISTORTIONS table\n\
                        [ -f input Fits file ]\n\
                        [ -o output Fits file ]\n\
                        [ -t polynomial type (LDAC) ]\n\
                        [ -v verbosity level (NORMAL) ]\n\n\
DESCRIPTION:\n\
The program reads an input catalog that contains of a DISTORTIONS\n\
table containing PV polynomials in ASTROMETRIX format. Furthermore\n\
the table has to contain CD matrix, CRVAL1/2 and CRPIX1/2 (This\n\
catalog can be created with the \"astrometrix2ldac.sh\" script from\n\
the original .head astrometrix files. The rpogram\n\
converts these polynomials to LDAC/Drizzle format. The result together\n\
with the CD matrix, CRVAL1/2 and CRPIX1/2 is written to stdout.\n\
The order of ASTROMETRIX polynomials is limited to 3 right now.\n\
AUTHOR: Thomas Erben \n"


void usage(char *me)
{
/****** usage ***************************************************
 *
 *  NAME	
 *      usage - Print out a message about how-to-use the program
 *
 *  SYNOPSIS
 *      void usage(char *me)
 *
 *  FUNCTION
 *      Print to stderr a message about how to use th main program
 *      The execution will stop and the exit code of the program is set
 *      to non-zero.
 *
 *  INPUTS
 *      me  - The calling name of the program
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 08-08-1997
 *
 ******************************************************************************/
    fprintf(stderr,"%s Version: %s (%s)\n",NAME,VERS,DATE);
    fprintf(stderr,"%s", __USAGE__);
    exit(1);
}

void init(int argc, char *argv[])
{
/****** int ***************************************************
 *
 *  NAME	
 *      init - Initialize the program from argument and configuration file
 *
 *  SYNOPSIS
 *      void init(int argc, char *argv[])
 *
 *  FUNCTION
 *      This function reads the command line arguments and interprets the
 *      configuration file
 *
 *  INPUTS
 *      argc - Number of arguments passed
 *      argv - The argument char array
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *      The control structure is set with the argument list parameters
 *      and the configuration parameters
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 08-08-1997
 *
 ******************************************************************************/
    int i;

    if (argc == 1) usage(argv[0]);
/*
 *  Initialize command line options with some defaults
 */

    strcpy(control.fitsinput, "");
    strcpy(control.fitsoutput, "");
    strcpy(control.type, "LDAC");
    strcpy(control.verbosity, "NORMAL");

/*
 *  Parse the command line
 */
    for (i=1; i<argc; i++) {
       if (argv[i][0] == '-') {
          switch(argv[i][1]) {
          case 'i': strcpy(control.listfile[0],argv[++i]);
                    break;
          case 'f': strncpy(control.fitsinput,argv[++i],MAXCHAR-1);
                    break;
          case 'o': strncpy(control.fitsoutput,argv[++i],MAXCHAR-1);
                    break;
          case 't': strncpy(control.type,argv[++i],MAXCHAR-1);
                    break;
          case 'v': strncpy(control.verbosity,argv[++i],MAXCHAR-1);
                    break;
          default:  
                    break;
          }
      } else {
          syserr(WARNING,"init","Command line syntax error\n");
          usage(argv[0]);
      }
   }
}

int main(int argc, char *argv[])
{
/****** add_image_calibs ***************************************************
 *
 *  NAME	
 *       astrometrix_convert - convert astrometrix pol. to ldac pol.
 *
 *  SYNOPSIS
 *       see above
 *
 *  FUNCTION
 *       see above
 *
 *  INPUTS
 *      i      - The input catalog
 *
 *  RESULTS
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      Thomas Erben - dd 01.11.01
 *
 ******************************************************************************/
    int		ncoff, degree, degx, degy;
    float       ftmp;
    int         degmax, degreemax;
    int		i, j, k, l;
    catstruct	*incat;
    tabstruct	*in_astromtab;
    keystruct	**astromkeys;
    double      *double_ptr;
    double	cd1_1=0.0, cd1_2=0.0, cd2_1=0.0, cd2_2=0.0;
    double      crval1=0.0, crval2=0.0, crpix1=0.0, crpix2=0.0;
    double      **PVswarp, **PVldac;
    fitsfile_struct	*ffo, *ffn;
    float       *data;
    char        tmpstring[MAXCHAR];

    init(argc, argv);

    altverbose(control.verbosity);
    sprintf(header,"%s Version: %s (%s)",NAME,VERS,DATE);
    VPRINTF(NORMAL,"\n   %s\n\n",header);

/* 
 *     For each input catalog get the FITS filename list and addition
 *     parameters
 */
    i=0;
    if (!(incat = read_cat(control.listfile[i])))
       syserr(FATAL, "add_image_astrom", "Cannot open catalog %s\n", 
          control.listfile[i]);
/*
 *     Get the astrometric calibration parameters
 */
    if (!(in_astromtab = name_to_tab(incat, "DISTORTIONS", 0)))
    {
       syserr(FATAL, "add_image_calibs", 
          "No table DISTORTIONS in %s catalog\n", control.listfile[i]);
    }
    ED_CALLOC(astromkeys, "add_image_calibs", keystruct *, 
              in_astromtab->nkey);
    read_keys(in_astromtab, NULL, astromkeys, in_astromtab->nkey, NULL);

    ED_CALLOC(PVswarp, "add_image_astrometrix", double *, MAXPOLINDEX+1);
    ED_CALLOC(PVldac, "add_image_astrometrix", double *, MAXPOLINDEX+1);

    for(i=0; i<MAXPOLINDEX+1; i++)
    {
      ED_CALLOC(PVswarp[i],"add_image_astrometrix", double, MAXPOLINDEX+1); 
      ED_CALLOC(PVldac[i],"add_image_astrometrix", double, MAXPOLINDEX+1); 
    }
/* 
 * Get the CD-matrix, CRPIX and CRVAL
 */
  
    for (j=0;j<in_astromtab->nkey;j++) 
    {
       if(strncmp(astromkeys[j]->name, "CD1_1", 5) == 0)
	  {
         double_ptr = astromkeys[j]->ptr;
	    cd1_1 = double_ptr[0];
	  }
       if(strncmp(astromkeys[j]->name, "CD2_1", 5) == 0)
	  {
         double_ptr = astromkeys[j]->ptr;
	    cd2_1 = double_ptr[0];
	  }
       if(strncmp(astromkeys[j]->name, "CD1_2", 5) == 0)
	  {
         double_ptr = astromkeys[j]->ptr;
	    cd1_2 = double_ptr[0];
	  }
       if(strncmp(astromkeys[j]->name, "CD2_2", 5) == 0)
	  {
         double_ptr = astromkeys[j]->ptr;
	    cd2_2 = double_ptr[0];
	  }
       if(strncmp(astromkeys[j]->name, "CRPIX1", 6) == 0)
	  {
         double_ptr = astromkeys[j]->ptr;
	    crpix1 = double_ptr[0];
	  }
       if(strncmp(astromkeys[j]->name, "CRPIX2", 6) == 0)
	  {
         double_ptr = astromkeys[j]->ptr;
	    crpix2 = double_ptr[0];
	  }
       if(strncmp(astromkeys[j]->name, "CRVAL1", 6) == 0)
	  {
         double_ptr = astromkeys[j]->ptr;
	    crval1 = double_ptr[0];
	  }
       if(strncmp(astromkeys[j]->name, "CRVAL2", 6) == 0)
	  {
         double_ptr = astromkeys[j]->ptr;
	    crval2 = double_ptr[0];
	  }
    }
/* 
 * we ONLY have to take over PV polynomial keywords
 */

    degmax = 0;
    degreemax = 0;

    for (j=0;j<in_astromtab->nkey;j++) 
    {
       if(strncmp(astromkeys[j]->name, "PV", 2) == 0)
       {
 	 sscanf((const char *)astromkeys[j]->name, "PV%d_%d", &k, &l);
 	  
 	 /* has the polynomial a "valid" degree between 0 and 3?? */
 	 if((k < 0) || (l < 0) || (k > MAXPOLINDEX) || (l > MAXPOLINDEX))
 	 {
 	   syserr(FATAL, "add_image_astrometrix", 
              "only WCSSTANDARD polynomials with degree <= 3 supported right now\n");
 	 }
 	  
 	 /* first bring 'l' to the LDAC order if necessary */
 	  
         if(strcmp(control.type, "LDAC")==0)
	 {
 	   if(l>10) l--;
 	   if(l>2) l--;
 	  
 	   l+=1;
	 }

         if(l>degmax)
	 {
	    degmax=l;
	 }

 	 /* determine the degree to that the currect PV term
 	  * belongs. This is here necessary to account for the
          * switsching of indices in the 2-component in the SWARP
          * case.
 	  */

         if(strcmp(control.type, "LDAC")==0)
	 {
  	   if(k==2)
  	   {    	  
  	      ftmp = 0.5*(sqrt(1.+8.*l)-1.);
  	      
  	      if((ftmp-(int)ftmp) > 1.0e-04)
  	      {
  		degree = (int) ftmp;
  	      }
  	      else
  	      {
  		degree = (int) (ftmp-1.);
  	      }
  	      
  	      if(degreemax < degree)
  	      {
  		degreemax = degree;
  	      }
  
  	      /* determine the degree of the x and the y term
  	       */
  	      
  	      ncoff = (degree)*(degree+1)/2;
  
  	      degy = l-(ncoff+1);
  	      degx = degree - degy;	      
  
  	      /* take into account that for the two
  		 component in the SWARP case the
  		 powers for x and y are switched
  	      */
  
  	      l=l-(degy-degx);
  	      
  	    }
 	  }    	  
 	  
          double_ptr = astromkeys[j]->ptr;
	    PVswarp[l][k] = double_ptr[0];
       }
    }
  
    if(strcmp(control.type, "LDAC") == 0)
    {
      for(i=0; i <= degreemax; i++)
      {
  	if(i==0)
  	{
  	  PVldac[1][1] = PVswarp[1][1]*(PI/180.);
  
  	  PVldac[1][2] = PVswarp[1][2]*(PI/180.);
  	}
  	if(i==1)
  	{
  	  PVldac[2][1] = (cd1_1*PVswarp[2][1]+cd2_1*PVswarp[3][1])*
  			 (PI/180.);
  	  PVldac[3][1] = (cd2_2*PVswarp[3][1]+cd1_2*PVswarp[2][1])*
  			 (PI/180.);
  
  	  PVldac[2][2] = (cd1_1*PVswarp[2][2]+cd2_1*PVswarp[3][2])*
  			 (PI/180.);
  	  PVldac[3][2] = (cd2_2*PVswarp[3][2]+cd1_2*PVswarp[2][2])*
  			 (PI/180.);
  	}
  	if(i==2)
  	{
  	  PVldac[4][1] = (cd1_1*cd1_1*PVswarp[4][1]+
  			 cd2_1*cd2_1*PVswarp[6][1]+
  			 cd1_1*cd2_1*PVswarp[5][1])*(PI/180.);
  	  PVldac[5][1] = (2.*cd1_1*cd1_2*PVswarp[4][1]+
  			 2.*cd2_2*cd2_1*PVswarp[6][1]+
  			 (cd1_1*cd2_2+cd1_2*cd2_1)*PVswarp[5][1])*
  			 (PI/180.);
  	  PVldac[6][1] = (cd2_2*cd2_2*PVswarp[6][1]+
  			 cd1_2*cd1_2*PVswarp[4][1]+
  			 cd2_2*cd1_2*PVswarp[5][1])*(PI/180.);
  
  	  PVldac[4][2] = (cd1_1*cd1_1*PVswarp[4][2]+
  			 cd2_1*cd2_1*PVswarp[6][2]+
  			 cd1_1*cd2_1*PVswarp[5][2])*(PI/180.);
  	  PVldac[5][2] = (2.*cd1_1*cd1_2*PVswarp[4][2]+
  			 2.*cd2_2*cd2_1*PVswarp[6][2]+
  			 (cd1_1*cd2_2+cd2_1*cd1_2)*PVswarp[5][2])*
  			 (PI/180.);
  	  PVldac[6][2] = (cd2_2*cd2_2*PVswarp[6][2]+
  			 cd1_2*cd1_2*PVswarp[4][2]+
  			 cd2_2*cd1_2*PVswarp[5][2])*(PI/180.);
  	}
  	if(i==3)
  	{
  	  PVldac[7][1] = (cd1_1*cd1_1*cd1_1*PVswarp[7][1]+
  			 cd2_1*cd2_1*cd2_1*PVswarp[10][1]+
  			 cd1_1*cd1_1*cd2_1*PVswarp[8][1]+
  			 cd1_1*cd2_1*cd2_1*PVswarp[9][1])*
  			 (PI/180.);
  	  PVldac[10][1] = (cd2_2*cd2_2*cd2_2*PVswarp[10][1]+
  			  cd1_2*cd1_2*cd1_2*PVswarp[7][1]+
  			  cd2_2*cd2_2*cd1_2*PVswarp[9][1]+ 
  			  cd2_2*cd1_2*cd1_2*PVswarp[8][1])*
  			  (PI/180.);
  	  PVldac[8][1] = (3.*cd1_1*cd1_1*cd1_2*PVswarp[7][1]+
  			 3.*cd2_1*cd2_1*cd2_2*PVswarp[10][1]+
  			 (cd1_1*cd1_1*cd2_2+2.*cd1_1*cd2_1*cd1_2)*PVswarp[8][1]+
  			 (cd2_1*cd2_1*cd1_2+2.*cd2_2*cd2_1*cd1_1)*PVswarp[9][1])*
  			 (PI/180.);
  	  PVldac[9][1] = (3.*cd1_1*cd1_2*cd1_2*PVldac[7][1]+
  			 3.*cd2_2*cd2_2*cd2_1*PVldac[10][1]+
  			 (cd1_2*cd1_2*cd2_1+2.*cd1_1*cd1_2*cd2_2)*PVswarp[8][1]+
  			 (cd2_2*cd2_2*cd1_1+2.*cd2_2*cd2_1*cd1_2)*PVswarp[9][1])*
  			 (PI/180.);
  
  	  PVldac[7][2] = (cd1_1*cd1_1*cd1_1*PVswarp[7][2]+
  			 cd2_1*cd2_1*cd2_1*PVswarp[10][2]+
  			 cd1_1*cd1_1*cd2_1*PVswarp[8][2]+
  			 cd1_1*cd2_1*cd2_1*PVswarp[9][2])*
  			 (PI/180.);
  	  PVldac[10][2] = (cd2_2*cd2_2*cd2_2*PVswarp[10][2]+
  			  cd1_2*cd1_2*cd1_2*PVswarp[7][2]+
  			  cd2_2*cd2_2*cd1_2*PVswarp[9][2]+ 
  			  cd2_2*cd1_2*cd1_2*PVswarp[8][2])*
  			  (PI/180.);
  	  PVldac[8][2] = (3.*cd1_1*cd1_1*cd1_2*PVswarp[7][2]+
  			 3.*cd2_1*cd2_1*cd2_2*PVswarp[10][2]+
  			 (cd1_1*cd1_1*cd2_2+2.*cd1_1*cd1_2*cd2_1)*PVswarp[8][2]+
  			 (cd2_1*cd2_1*cd1_2+2.*cd2_2*cd2_1*cd1_1)*PVswarp[9][2])*
  			  (PI/180.);
  	  PVldac[9][2] = (3.*cd1_1*cd1_2*cd1_2*PVldac[7][2]+
  			 3.*cd2_2*cd2_2*cd2_1*PVldac[10][2]+
  			 (cd1_2*cd1_2*cd2_1+2.*cd1_1*cd1_2*cd2_2)*PVswarp[8][2]+
  			 (cd2_2*cd2_2*cd1_1+2.*cd2_2*cd1_2*cd2_1)*PVswarp[9][2])*
  			 (PI/180.);
  	}
      }
    }

    if(strcmp(control.fitsinput, "") != 0)
    {
       ffo = open_fits_file(control.fitsinput);
       switch (ffo->bitpix) 
       {
         case -32:
           ffo->pix_type = T_FLOAT;
           break;
	 default:
 	   syserr(FATAL, "add_image_astrometrix", 
              "only Float FITS images supported\n");
           break;
       }

       set_no_interp(YES);

       data = interp_fits_image(ffo, NO); 

       ffn = new_fits();
       inherit_fits(ffn, ffo);
       ffn->pix_type = ffo->pix_type;

       /* 
 	* first delete PV polynomials already present.
 	* As there can be LDAC and WCSSTANDARD type polynomials
 	* we have to fully scan both components. We assume
 	* a maximum degree of 5 what gives 23 for the maximum
 	* coefficient (WCSSTANDARD style)
       */
       for(j=0; j<25; j++)
       {
          for(k=0; k<25; k++)
	  {
             sprintf(tmpstring, "PV%d_%d", j, k);
             d_fits(tmpstring, ffn->head);
	  }
       }
       
       /* now update/add header keywords */
       if(strcmp(control.type, "LDAC") == 0)
       {
	 for(i=1; i<=2; i++)
	 {
	   for(j=1; j<= degmax;j++)
	   {
             sprintf(tmpstring, "PV%d_%d", j, i);
	     s_fits_double_c(tmpstring, ffn->head, 
                           PVldac[j][i], "");
	   }
	 }
       }
       else
       {
	 for(i=1; i<=2; i++)
	 {
	   for(j=0; j<= degmax;j++)
	   {
             sprintf(tmpstring, "PV%d_%d", i, j);
	     s_fits_double_c(tmpstring, ffn->head, 
                           PVswarp[j][i], "");
	   }
	 }
       }
       s_fits_double_c("CD1_1", ffn->head, cd1_1, "");
       s_fits_double_c("CD2_2", ffn->head, cd2_2, "");
       s_fits_double_c("CD1_2", ffn->head, cd1_2, "");
       s_fits_double_c("CD2_1", ffn->head, cd2_1, "");
       s_fits_double_c("CRVAL1", ffn->head, crval1, "");
       s_fits_double_c("CRVAL2", ffn->head, crval2, "");
       s_fits_double_c("CRPIX2", ffn->head, crpix2, "");
       s_fits_double_c("CRPIX1", ffn->head, crpix1, "");

       /* we delete existing CDELT keywords to avoid possible
          confusions in the interpretation of the headers
       */
       d_fits("CDELT1", ffn->head);
       d_fits("CDELT2", ffn->head);

       /* Now save the new FITS file */
       w_fits_pad(ffn->head);
       save_fits(ffn, (unsigned char *)data, control.fitsoutput);
       ED_FREE(data);       
       close_fits(ffo, NO);
    }  

    if(strcmp(control.type, "LDAC") == 0)
    {
      for(i=0; i< 2; i++)
      {
  	for(j=0; j<degmax; j++)
  	{
  	  VPRINTF(DEBUG,"PV%d_%d     %e\n",j+1, i+1, PVldac[j+1][i+1]);
  	}
      }
    }
    else
    {
      for(i=1; i<= 2; i++)
      {
  	for(j=0; j<=degmax; j++)
  	{
  	  VPRINTF(DEBUG,"PV%d_%d     %e\n",i , j, PVswarp[j][i]);
  	}
      }
    }

    VPRINTF(DEBUG, "CD1_1    %e\n", cd1_1);
    VPRINTF(DEBUG, "CD2_2    %e\n", cd2_2);
    VPRINTF(DEBUG, "CD2_1    %e\n", cd2_1);
    VPRINTF(DEBUG, "CD1_2    %e\n", cd1_2);

    VPRINTF(DEBUG, "CRVAL1    %e\n", crval1);
    VPRINTF(DEBUG, "CRVAL2    %e\n", crval2);
    VPRINTF(DEBUG, "CRPIX1    %e\n", crpix1);
    VPRINTF(DEBUG, "CRPIX2    %e\n", crpix2);

    /* free memory and bye */
    for(i=0; i<MAXPOLINDEX+1; i++)
    {
      ED_FREE(PVswarp[i]); 
      ED_FREE(PVldac[i]); 
    }

    ED_FREE(PVswarp);
    ED_FREE(PVldac);

    ED_FREE(astromkeys);

    
    return 0;
}
















