#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "options_globals.h"
#include "utils_globals.h"
#include "project_globals.h"
#include "thresh_globals.h"
#include "model_globals.h"
#include "astrom_io_globals.h"
#include "build_globals.h"

static void build_refpair(reflist_struct *refpair, model_struct *model)
/****** build_refpair ***************************************************
 *
 *  NAME	
 *      build_refpair - Build the reference information into the design matrix
 *
 *  SYNOPSIS
 *      void build_refpair(reflist_struct *refpair, model_struct *model)
 *
 *  FUNCTION
 *      Build up the least square equations for the determination of smoothly 
 *      varying plate parameters: 
 *      add all constraints resulting from reference stars in one field 
 *
 *  INPUTS
 *     refpair   - Pointer to the reference object information
 *     nref      - Number of refpair entries
 *     model     - The full model structure
 *
 *  RESULTS
 *     model     - The full model structure updated
 *
 *  RETURNS
 *     None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 17-07-1998
 *
 ******************************************************************************/
{
    int		i, j, k, f, c, jof;
    double	x, y;
    float	x0, y0, *work, siginv, siginv2;
    float	scale = sqrt(model->xscale * model->xscale + 
                             model->yscale * model->yscale);

    ED_CALLOC(work, "build_ref", float, (model->np+model->nf));
/*
 *  Project the reference stars onto current field center
 */

    for (i=0;i<refpair->npair;i++) 
    {
      x   = refpair->p[i]->ra;
      y   = refpair->p[i]->dec;
      f   = refpair->p[i]->f;
      c   = refpair->p[i]->c;
      jof = model->paroff[c];
      abs_to_rel(model->centers[0][c][f], 
                 model->centers[1][c][f], 0.0,&x,&y,&x,&y,1);


/*
 *    Compute current best estimate of position
 */
      bi_cubic(f, c, refpair->p[i]->xobj, refpair->p[i]->yobj,
               model, &work[jof]);
      x0 = model->params[0][model->np+f];
      y0 = model->params[1][model->np+f];

/*
 *    Increase the error on reference star to compensate for the fact that it 
 *    will enter for each colour channel? Obviously not a very clean handling,
 *    but we cannot afford a full overlapping plate solution.
 */
      siginv = 1.0 / sqrt((refpair->p[i]->eobj*scale)*
                          (refpair->p[i]->eobj*scale) +
                           refpair->p[i]->err * 
                           refpair->p[i]->err);
      siginv2 = siginv * siginv;
      for (j=jof;j<jof+model->parnum[c];j++) 
      {
        x0 += model->params[0][j] * work[j];
        y0 += model->params[1][j] * work[j];
      }

      x -= x0;
      y -= y0;
/*
 *    Compute normal matrix 
 */
      model->alpha[model->np+f][model->np+f] +=     siginv2;
      model->beta[0][model->np+f]            += x * siginv2;
      model->beta[1][model->np+f]            += y * siginv2;

      for (j=jof;j<jof+model->parnum[c];j++) 
      {
/*
 *      Could use symmetry for a factor of two speedup
 */
        model->alpha[j][model->np+f] += work[j] *     siginv2;
        model->alpha[model->np+f][j] += work[j] *     siginv2;
        model->beta[0][j]            += work[j] * x * siginv2;
        model->beta[1][j]              += work[j] * y * siginv2;
        for (k=jof;k<jof+model->parnum[c];k++) 
          model->alpha[j][k]           += work[j] * work[k] * siginv2;
      }
    }
    ED_FREE(work);

}
/**/
static void build_overpair(overlist_struct *overpair, model_struct *model)
/****** build_overpair ***************************************************
 *
 *  NAME	
 *      build_overpair - Build the overlap information into the design matrix
 *
 *  SYNOPSIS
 *      void build_overpair(overpair_struct *overpair, model_struct *model)
 *
 *  FUNCTION
 *      Build up the least square equations for the determination of smoothly 
 *      varying plate parameters: 
 *      add all constraints resulting from overlap stars
 *
 *  INPUTS
 *     overpair  - Pointer to the object overlap information
 *     model     - The full model structure
 *
 *  RESULTS
 *     model     - The full model structure updated
 *
 *  RETURNS
 *     None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 17-07-1998
 *
 ******************************************************************************/
{
    int		i, j, k, f1, f2, c1, c2, jof1, jof2;
    float	*work1, *work2, siginv, siginv2, off[2], off0[2];
    double	x1, y1, x2, y2;
    float	scale = sqrt(model->xscale * model->xscale + 
                             model->yscale * model->yscale);
/*
 *
 */
    ED_CALLOC(work1, "build_over", float, (model->np+model->nf));
    ED_CALLOC(work2, "build_over", float, (model->np+model->nf));
    for (i=0;i<overpair->npair;i++) {
      f1   = overpair->p[i]->f1;
      f2   = overpair->p[i]->f2;
      c1   = overpair->p[i]->c1;
      c2   = overpair->p[i]->c2;
      jof1 = model->paroff[c1];
      jof2 = model->paroff[c2];
/*
 *    Evaluate the difference of the gnomonic projections relative to both field
 *    centers for one representative point of the overlap (their mid-point is 
 *    used for this here). One should in principle evaluate the difference for 
 *    each star (using approximate corrections to derive an absolute position). 
 *    Given the small fieldsize in DENIS, the maximum error is however at the
 *    few millarcseconds level, which will (unfortunately...) not be a terribly 
 *    relevant accuracy scale.
 */
      x1 = x2 = (model->centers[0][c1][f1] + 
                 model->centers[0][c2][f2]) / 2.0;
      y1 = y2 = (model->centers[1][c1][f1] + 
                 model->centers[1][c2][f2]) / 2.0;
      abs_to_rel(model->centers[0][c1][f1],
                 model->centers[1][c1][f1],0.0,&x1,&y1,&x1,&y1,1);
      abs_to_rel(model->centers[0][c2][f2],
                 model->centers[1][c2][f2],0.0,&x2,&y2,&x2,&y2,1);
      off[0] = x1 - x2;
      off[1] = y1 - y2;
/*
 *    Evaluate current estimate of correction  
 */
      bi_cubic(f1, c1, overpair->p[i]->xobj1, overpair->p[i]->yobj1, 
               model, &work1[jof1]);
      bi_cubic(f2, c2, overpair->p[i]->xobj2, overpair->p[i]->yobj2, 
               model, &work2[jof2]);
      for (j=0;j<2;j++) {
        off0[j] = model->params[j][model->np+f1] - 
                  model->params[j][model->np+f2];
        for (k=jof1;k<jof1+model->parnum[c1];k++)
          off0[j] += +model->params[j][k]*work1[k];
        for (k=jof2;k<jof2+model->parnum[c2];k++)
          off0[j] -= model->params[j][k]*work2[k];
      }
      siginv = 1.0/sqrt(overpair->p[i]->eobj1 * overpair->p[i]->eobj1 + 
                        overpair->p[i]->eobj2 * overpair->p[i]->eobj2)/
                        fabs(scale);
      siginv2 = siginv * siginv;
/*
 *    Higher order terms vary smoothly and are thus determined globally
 */
      if (f1 != f2) {
/*
 *      (Field center)*(Field center) terms
 */
        model->alpha[model->np+f1][model->np+f1] += siginv2;
        model->alpha[model->np+f1][model->np+f2] -= siginv2;
        model->alpha[model->np+f2][model->np+f1] -= siginv2;
        model->alpha[model->np+f2][model->np+f2] += siginv2;
        for (k=0;k<2;k++) {
          model->beta[k][model->np+f1] +=  (off[k] - off0[k]) * siginv2;
          model->beta[k][model->np+f2] -=  (off[k] - off0[k]) * siginv2;
        }
/*
 *      (Field center)*(global parameters) terms
 */
        for (j=jof1;j<jof1+model->parnum[c1];j++) {
          model->alpha[j][model->np+f1] += work1[j] * siginv2;
          model->alpha[model->np+f1][j] += work1[j] * siginv2;
          model->alpha[j][model->np+f2] -= work1[j] * siginv2;
          model->alpha[model->np+f2][j] -= work1[j] * siginv2;
        }
        for (j=jof2;j<jof2+model->parnum[c2];j++) {
          model->alpha[j][model->np+f1] -= work2[j] * siginv2;
          model->alpha[model->np+f1][j] -= work2[j] * siginv2;
          model->alpha[j][model->np+f2] += work2[j] * siginv2;
          model->alpha[model->np+f2][j] += work2[j] * siginv2;
        }
      }
/*
 *    (global parameters)*(global parameters)
 *    IS THIS CORRECT IF IB1=IB2 !!!????
 *    yes, but not quite optimal for the terms which don't depend on F
 *    (0 = X-X...)... Also not numerically ideal, to say the least...
 */
      for (j=jof1;j<jof1+model->parnum[c1];j++) {
        for (k=jof1;k<jof1+model->parnum[c1];k++) 
          model->alpha[j][k] += work1[j] * work1[k] * siginv2;
        for (k=0;k<2;k++) 
          model->beta[k][j] += work1[j] * (off[k] - off0[k]) * siginv2;
      }
      for (j=jof2;j<jof2+model->parnum[c2];j++) {
        for (k=jof2;k<jof2+model->parnum[c2];k++) 
          model->alpha[j][k] += work2[j] * work2[k] * siginv2;
        for (k=0;k<2;k++) 
          model->beta[k][j] -= work2[j] * (off[k] - off0[k]) * siginv2;
      }
      for (j=jof1;j<jof1+model->parnum[c1];j++) {
        for (k=jof2;k<jof2+model->parnum[c2];k++) {
         model->alpha[j][k] -= work1[j] * work2[k] * siginv2;
          model->alpha[k][j] -= work1[j] * work2[k] * siginv2;
        }
      }
    }
    ED_FREE(work2);
    ED_FREE(work1);
}
/**/
static void zero(float **a, int dim1, int dim2)
/****** zero ***************************************************
 *
 *  NAME	
 *      zero - Fill the matrix with 0 values
 *
 *  SYNOPSIS
 *      void zero(float **a, int dim1, int dim2)
 *
 *  FUNCTION
 *      Make each element of the matrix equal to 0.0
 *
 *  INPUTS
 *      a    - The matrix
 *      dim1 - The first dimension size
 *      dim2 - The second dimension size
 *
 *  RESULTS
 *      a    - The matrix
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 17-07-1998
 *
 ******************************************************************************/
{
    int		i, j;
    float	*p;
/*
 *  Reset A
 */
    for (i=dim2;i>0;i--,a++) 
       for (j=dim1,p=*a;j>0;j--,p++) 
         (*p) = 0.0; 
}
/**/
void build_equations(model_struct *model, thresh_struct *thresholds,
                     int reject)
/****** build_equations ***************************************************
 *
 *  NAME	
 *      build_equations - Construct the Normal Equations matrices
 *
 *  SYNOPSIS
 *      void build_equations(model_struct *model, thresh_struct *thresholds,
 *                           int reject)
 *
 *
 *  FUNCTION
 *     Build up the least square equations.
 *     Depending on flag settings, the least square problem can be translated
 *     to the normal equations formulation and/or the design matrix 
 *     formulation. In both cases, two second members (for X and Y) are 
 *     accumulated for one matrix.
 *
 *  INPUTS
 *     model      - The full model structure
 *     thresholds - List of maximum distance to catalog position
 *     reject     - Preform object rejection or not
 *
 *  RESULTS
 *     model          - The full model structure updated
 *
 *  RETURNS
 *     None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 17-07-1998
 *
 ******************************************************************************/
{
/*
 *  Local variables
 */
    reflist_struct	*refpairs;
    overlist_struct	*overpairs;
/*
 *  Clean the design matrices
 */
    zero(model->alpha,model->nf+model->np,model->nf+model->np);
    zero(model->beta, model->nf+model->np,2);
/*
 *  There are two types of data to be processed:
 *    pairing information between extracted objects and reference stars and
 *    pairing information between extracted objects in the catalog
 *  Let's run through them both and build the design matrices on the way
 *
 *  First get the reference pairs
 */
    refpairs = new_refpairs();
    while (get_refpairs(refpairs, thresholds, model, reject, FALSE, FALSE)) 
    {
       build_refpair(refpairs, model);
    }
    free_refpairs(refpairs);
/*
 *  Then get all the internal pairing (inter catalog and cross catalog)
 */       
    overpairs = new_overpairs();
    while (get_overpairs(overpairs, thresholds, model, reject, FALSE, FALSE)) 
    {
       build_overpair(overpairs, model);
    }
    free_overpairs(overpairs);
}
