/* History: 05.01.2000:
   included some comments about the use of the program

   29.04.03:
   changed exit(0) to return 0 to avoid compiler warnings

   21.05.2004:
   included example call for UCAC2 zone file conversion.

   16.01.2005:
   I made the program more user friendly.

   06.05.2006:
   I removed a bug in the checking whether the maximum number of input
   catalogues was exceeded.  I increased this number to 1000
   catalogues.

   15.05.2006:
   I improved the program description.

   30.12.2012:
   I corrected a strncpy command that could possibly lead to a memory fault.
*/

#include <stdio.h>
#include <string.h>
#include "utils_globals.h"
#include "refs_io_types.h"
#include "refs_io_globals.h"
#include "global_vers.h"

#ifndef MAXSTRINGLENGTH
#define MAXSTRINGLENGTH 256
#endif

#ifndef MAXCATS
#define MAXCATS 1000
#endif

#define __PROGRAM__ "make_pmm: convert standardstar catalogues"
#define __VERS__ "1.0"

void usage()
{
    fprintf(stderr, "%s (V%s; %s)\n\n", __PROGRAM__, __VERS__, __PIPEVERS__) ;
    fprintf(stderr, "USAGE:\n");
    fprintf(stderr, "   make_pmm options\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "OPTIONS:\n");
    fprintf(stderr, "   -i incat1 [[incat2] incat3] ...]  (input catalogs)\n");
    fprintf(stderr, "   -d debug mode: only print catalogue entries to\n");
    fprintf(stderr, "                  screen but do not create *.pmm files\n");
    fprintf(stderr, "   -s catalog type: USNO or UCAC; default is USNO		  \n");
    fprintf(stderr, "\n");
    fprintf(stderr, "DESCRIPTION:\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "make_pmm transforms astrometric standard star catalogues to a special\n");
    fprintf(stderr, "format allowing a fast access to data of a specified region on the\n");
    fprintf(stderr, "sky. Furthermore it significantly compresses the information\n");
    fprintf(stderr, "content. Currently, two publicly avaailable catalogs are supported by\n");
    fprintf(stderr, "this program, the USNO-A2 and the UCAC2 catalogs. The raw USNO-A2\n");
    fprintf(stderr, "files come in the form of 24 files named zone????.cat where the four\n");
    fprintf(stderr, "'?'  represents digits from 0 to 9. The UCAC2 catalog consists of\n");
    fprintf(stderr, "288 files of the form z001-z288. Those files can be feeded into\n");
    fprintf(stderr, "the 'make_pmm' program and are transformed into files named\n");
    fprintf(stderr, "??????.pmm. The complete USNO-A2 catalogue results in more than\n");
    fprintf(stderr, "30000 .pmm files, the UCAC2 catalogue is kept in about 23000 of\n");
    fprintf(stderr, "them. USNO-A2 is compressed from 6GB to 3GB, UCAC2 from 2GB to \n");
    fprintf(stderr, "about 500M. The converted catalogues contain Ra, Dec, Mag1, Mag2.\n");
    fprintf(stderr, "In the USNO-A2 case Mag1 and Mag2 are the original Blue and Red\n");
    fprintf(stderr, "magnitudes, for UCAC2 both magnitude quantities represent the UCAC2 \n");
    fprintf(stderr, "'U2Rmag' quantity. The program outputs resulting files to the directory\n");
    fprintf(stderr, "where it is called. The program 'get_pmm_objects' can be used to\n");
    fprintf(stderr, "retrieve catalogue entries from the ??????.pmm files.\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "The program DOES NOT check the validity of input data.\n");
    fprintf(stderr, "\n");
    fprintf(stderr, "AUTHORS:\n");
    fprintf(stderr, "   Erik Deul     (deul@strw.leidenuniv.nl)\n");
    fprintf(stderr, "   Thomas Erben  (terben@astro.uni-bonn.de)\n");
 
    exit(1);
}

int main(int argc, char *argv[])
{
  int i;
  unsigned int j;
  unsigned int ncats = 0;
  char catnames[MAXSTRINGLENGTH][MAXCATS];
  char standardcat[MAXSTRINGLENGTH]="USNO";
  int debug = 0;

  if(argc < 3) {
    usage();
  }

  for (i=1; i<argc; i++) {
    if (argv[i][0] == '-') {
       switch(argv[i][1]) {
       case 'd': debug = 1;
                 break;
       case 's': strncpy(standardcat, argv[++i], MAXSTRINGLENGTH);
                 break;
       case 'i': ncats = 0;
                 strncpy(catnames[ncats], argv[++i], MAXSTRINGLENGTH - 1);
                 while (i + 1 < argc && argv[i + 1][0] != '-') {
                    strncpy(catnames[++ncats],argv[++i], MAXSTRINGLENGTH - 1);

                    if(ncats == MAXCATS) {
                      syserr(FATAL, "make_pmm", 
                        "Maximum number of input catalogs (%d) exceeded !!\n",
                        MAXCATS);
                      exit(1);
                    }
                 }
                 break;
       default:  usage();
                 break;
       }
    } else {
       syserr(WARNING,"init","Command line syntax error\n");
       usage();
    }
 }

  for (j = 0; j < ncats; j++) {
    /* a '1' as second argument in the pasrse_ccd calls runs the
       program in debug mode and the Ra, Dec values are printed to the
       screen only.  Set it to '0' to get the pmm files.
    */
    if(strcmp(standardcat, "USNO") == 0) {
      printf("converting USNO catalog %s\n", catnames[j]);
      parse_ccd_files(catnames[j], debug);
    } else {
      if(strcmp(standardcat, "UCAC") == 0) {
        printf("converting UCAC catalog %s\n", catnames[j]);
        parse_ucac2_files(catnames[j], debug);
      } else {
        syserr(FATAL, "make_pmm", 
               "unknown standardstar cat. format; known are USNO and UCAC");
      }
    }
  }
  return 0;
}
