#include <stdio.h>
#include <stdlib.h>
#include "utils_globals.h"
#include "options_globals.h"
#include "model_types.h"
#include "gaussj_globals.h"
#include "solve_globals.h"

static int compress_normal(model_struct *model, int *list)
/*-----------------------------------------------------------------------------
 *  Remove undeterminate variables from the normal equation system
 *  Arguments:
 *     A         REAL*4(*,*)  Normal matrix                            Modified
 *     NM        INTEGER      Total number of parameters               Input
 *     N         Integer      Number of determined parameters          Output
 *     B         Real*4       Second hand of normal equations          Modified
 *     LIST      Integer(*)   Mapping of full parameter list to fitted Output
 *                            parameter list
 *-----------------------------------------------------------------------------
 */
{
    int		i, j, n;
/*
 *  First produce a list of all constrained parameters.
 */
    for (i=0,n=0;i<model->np+model->nf;i++) {
      if (model->alpha[i][i] <= 0.0) 
        VPRINTF(NORMAL,"Diagonal element %d of covariance matrix is 0\n\n",i);
      else
        list[n++] = i;
    }
/*
 *  Now reshuffle the matrix to remove unconstrained variables. This is done
 *  in place so some care is needed.
 */
    for (j=0;j<n;j++) {
      for (i=0;i<n;i++) 
        model->alpha[i][j] = model->alpha[list[i]][list[j]];
      for (i=0;i<2;i++) 
        model->beta[i][j] = model->beta[i][list[j]];
    }
    return n;
}
/**/
static void expand_normal(model_struct *model, int *list, int n)
/*-----------------------------------------------------------------------------
 *  Remove undeterminate variables from the normal equation system
 *  Arguments:
 *     A         REAL*4(*,*)  Normal matrix                            Modified
 *     NM        INTEGER      Total number of parameters               Input
 *     N         Integer      Number of determined parameters          Input
 *     B         Real*4       Second hand of normal equations          Modified
 *     LIST      Integer(*)   Mapping of full parameter list to fitted Input
 *                            parameter list
 *-----------------------------------------------------------------------------
 */
{
    int		i, j, k;
/*
 *  Invert the (pseudo)mapping of fitted to full parameter list
 */
    for (i=model->np+model->nf, k=n-1; i--;) 
      if (list[k] < i) 
        list[i] = -1;
      else 
        list[i] = k--;
/*
 *  Reshuffle the matrix to restore unconstrained variables. This is done
 *  in place so some care is needed.
 */
    for (j=model->np+model->nf; j--;) {
      for (i=model->np+model->nf; i--;) {
        if (list[i] == -1) {
          model->alpha[i][j] = 0.0;
        } else {
          if (list[j] == -1) {
            model->alpha[i][j] = 0.0;
          } else {
            model->alpha[i][j] = model->alpha[list[i]][list[j]];
          }
        }
      }
      for (i=0;i<2;i++) {
        if (list[j] == -1)
          model->beta[i][j] = 0.0;
        else
          model->beta[i][j] = model->beta[i][list[j]];
      }
    }
}
/**/
static void fillin(float *r, int *flag, int n)
/* ----------------------------------------------------------------------
 * SAS	Internal routine
 *	Interpolate bad channel (if possible)
 * Arguments
 *	R	R*4(*)	Array to be interpolated
 *       FLAG    I       Flag array (zero for blanked pixels)
 *	N	I	Size of both arrays
 *	IMAX	I	Last pixel in array
 * Created	21-Mar-1986	S.Guilloteau
 *----------------------------------------------------------------------
 */
{
    int		i, j, k, iff, il;

/*
 *  Look for first non blanked pixel
 */
    for (i=0; i < n && flag[i] == -1; i++);
    iff = i;
    if (i == n) return;
    for (i=n-1; i >= 0 && flag[i] == -1; --i);
    il = i;
/*
 *  Use a constant for extrapolation
 */
    for (i=0;i<iff;i++) r[i]=r[iff];
    for (i=il+1;i<n;i++) r[i]=r[il];
/*
 *  Look for gaps, and use linear interpolation to bridge them
 */
    for (i=iff;i<il;i++) {
      if (flag[i+1] == -1) {
         for (j=i+1;flag[j] == -1;j++);	
         for (k=i+1;k<j;k++) 
           r[k] = (((float) (k-i)) * r[j] + ((float)(j-k))*r[i])/ 
                   ((float) (j-i));
         i = j;
      }
    }
}
/**/
void solve_equations(model_struct *model)
/*------------------------------------------------------------------------------
 * Determine the least square solution by direct inversion of the normal
 * equations.
 *
 * Parameters:
 *     NP      Integer        Number of free parameters
 *     ALPHA   Real*4(NP,NP)  Normal matrix on input, covariance on output
 *     BETA    Real*4(NP,2)   Second hand on input, solution on output
 *     NGLOB   Integer        Number of global (i.e. don't interpolate) 
 *                            parameters
 *------------------------------------------------------------------------------
 */
 {
    int		*list, n;
/*
 *  Remove obviously unconstrained parameters from the normal equation system,
 *  solve the reduced system,
 *  restore the full parameter set, and
 *  interpolate field center corrections
 */
    ED_CALLOC(list, "normal_solve", int, (model->np+model->nf));
    if (!(n = compress_normal(model, list)))
       syserr(FATAL, "solve_normal", "No parameter is constrained\n");
    else
       VPRINTF(NORMAL,"%d parameters are constrained out of %d\n",
               n,model->np+model->nf);
    gaussj(model->alpha, n, model->beta, 2);
    expand_normal(model, list, n);
    fillin(&model->beta[0][model->np],&list[model->np],model->nf);
    fillin(&model->beta[1][model->np],&list[model->np],model->nf);
    ED_FREE(list);
}
