#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include "fitspaste_globals.h"
#include "fitspaste_vers.h"
#include "options_globals.h"
#include "utils_globals.h"
#include "image_io_globals.h"
#include "fits_image_io_globals.h"
#include "tsize.h"

/* LAST MODIFICATION: 23.02.1998 (TE: fixed a bug in get_fitsline)
 */

/* 08.08.:
   fixed a bug in get_fits_line:
   the memcpy statement
           memcpy(&data[(x-segm[i].px0)
                  *t_size[T_FLOAT]], ldata, 
                  (afieldwith*t_size[T_FLOAT]));

   was replaced by:
           memcpy(&data[(x-segm[i].px0)
                  *t_size[T_FLOAT]], ldata, 
                  (segm[i].xwid*t_size[T_FLOAT]));

   The old copied too much.

   If no CDELT keywords are present on the old image, the
   CD1_1 and CD2_2 are now taken to get the pixelscale of
   the images.

   22.09.:
   several bug fixes to make the code work on AIX.

   28.09.:
   fixes a bug that caused damaged images when parts of the area
   (in the x-direction) were not covered by the input images.

   16.07.01:
   removed compiler warnings under DEC

   27.02.02:
   introduced the possibility to make the output image
   larger than the input image

   05.05.03:
   included the tsize.h file

   10.01.05:
   I removed compiler warnings under gcc with optimisation
   enabled (-O3).
*/

/*
 *    Global variables do not take as much memory
 */
static control_struct control;
static int clipd, clippix;
unsigned char *fdata;

void coepro(double ra, double dec, double raref, double decref, 
            double rpp, double *x, double *y);

char header[MAXCHAR];

typedef struct {
	int	px0, py0;
        int	xwid, ywid;
        fitsfile_struct *fds;
} segm_struct;

option my_opts[] = {
            {"BINSIZE", INT, &control.binsize, 0, 1000, 0.0, 0.0,
              1, 0.0, NO, "",
              {"", "", "", "", ""}, 0,
             "The size of the binning to use for creating output pixels"},
             {""},
            };

void usage(char *me)
{
    fprintf(stderr,"Usage: %s \n", me);
    fprintf(stderr,"	-i infits1 [infits2 [infits3 [...]]]\n");
    fprintf(stderr,"	-o outfits\n");
    fprintf(stderr,"	[-@] (get the configuration file parameters)\n");
    fprintf(stderr,"	[-c conf_file] \n");
    fprintf(stderr,"	[-d Ra Dec DRad DDec] \n\n");
    fprintf(stderr,"	[-x xmin xmax ymin ymax] \n\n");
    fprintf(stderr,"	[-g make output image larger than input if requested by '-x'] \n\n");
    fprintf(stderr,"	where Ra Dec are center coordinates (in degrees) of the output section\n");
    fprintf(stderr,"	and DRad and DDec are the widths (in arcsec) of the output section\n\n");
    fprintf(stderr,"	where xmin, xmax, ymin, ymax are the pixel extrena of the output section\n\n");
    exit(1);
}


void init(int argc, char *argv[], control_struct *c)
{
/*
 *  Default file name is calling name with .conf extension
 */
    int i, j, k;

    sprintf(c->conffile,"/denisoft/conf/associate.conf");
    clipd = clippix = FALSE;
    if (argc == 1) usage(argv[0]);
    init_options(0, my_opts);

    control.outlarge=0; /* do not enlarge output image by default
                           (even if requested via the '-x'/'-d' 
                           options */

    for (i=1; i<argc; i++) {
      if (argv[i][0] == '-') {
         switch(tolower((int)argv[i][1])) {
         case 'c': strcpy(c->conffile,argv[++i]);
                   break;
         case 'i': c->ilist = 0;
                   strcpy(c->fitsfile[c->ilist],argv[++i]);
                   while (i+1<argc && argv[i+1][0] != '-') {
                      strcpy(c->fitsfile[++c->ilist],argv[++i]);
                   }
                   ++c->ilist;
                   break;
        case 'o':  strcpy(control.outfile,argv[++i]);
                   break;
        case 'g':  control.outlarge=1;
                   break;
        case '@':  print_options(argv[0]);
                   exit(0);
        case 'd':  control.ra   = atof(argv[++i]);
                   control.dec  = atof(argv[++i]);
                   control.rad  = atof(argv[++i]);
                   control.decd = atof(argv[++i]);
                   clipd = TRUE;
                   break;
        case 'x':  control.xmin  = atof(argv[++i]);
                   control.xmax  = atof(argv[++i]);
                   control.ymin  = atof(argv[++i]);
                   control.ymax  = atof(argv[++i]);
                   clippix = TRUE;
                   break;
         default:  parse_options(c->conffile);
                   j = i++; k = i;
                   set_command_option(argv[j], argv[k]);
                   break;
         }
      } else {
         syserr(WARNING,"init","Command line syntax error\n");
         usage(argv[0]);
      }
   }
   finish_options(c->conffile);
}

/* TE: 23.02:
   - added parameter afieldwidth to the arguments and
     fixed a bug that occured when we are at the border of
     the field (had to do with the 'data' variable)
*/
unsigned char *get_fitsline(segm_struct *segm, int n, int x, int y,
                            int afieldwidth, int abinsize)
{
     int	i, j, k, yline, first=TRUE;
     unsigned char *ldata;
     float *fpointer;

     VPRINTF(VERBOSE,"Request at x:%d, y:%d\n",x,y);

/* Loop over abinsize lines */

     for(k=0; k< abinsize; k++)
     {
/*     
 *     First find the frame in which the line starts
 */    
       for (i=0;i<n;i++) 
       {
     	  if (segm[i].px0>= x && segm[i].px0-segm[i].xwid < x &&
     	      segm[i].py0>= (y+k) && segm[i].py0-segm[i].ywid < (y+k)) 
     	  {
     	     yline = segm[i].py0 - (y+k);

     	     VPRINTF(VERBOSE, "Reading line %d of start %s\n", yline, 
     			      segm[i].fds->name);
     	     ED_CALLOC(ldata, "get_fitsline", unsigned char,
     			     (segm[i].xwid*t_size[T_FLOAT]));
/*     
 *   	     Check if this file was already opened, if not then first close
 *   	     all other files (we are calling get_fitsline in a sequencial
 *   	     way, so we have passed the previous row of segments and are now
 *   	     on to the next row of segments. As we may not go through the
 *   	     first segment of a row (or it just does not exist) we must
 *   	     postpone the closing to the netx segment open. So keep track
 *   	     of this through the variable 'first'.
 */    
     	     if (!(segm[i].fds)) 
     	     {
     		first = FALSE;
     		for (j=0;j<n;j++) 
     		{
     		   if (segm[j].fds) 
     		   {
     		      VPRINTF(NORMAL, "Closing %s: \n", segm[j].fds->name);
     		      close_fits(segm[j].fds, NO);
     		      segm[j].fds = NULL;
     		   }
     		}
     		if (!(segm[i].fds = open_fits_file(control.fitsfile[i])))
     		  syserr(FATAL, "fitspaste", "Could not open file %s\n",
     		      control.fitsfile[i]);
     		segm[i].fds->pix_type = T_FLOAT;
     	     }
     	     read_fits_line(segm[i].fds, ldata, yline, NO);
     	     VPRINTF(VERBOSE, "Copying %d pixels from %d\n",
     		    (segm[i].xwid - (segm[i].px0-x)),
     		    segm[i].px0-x);
     	     memcpy(&fdata[(afieldwidth*k)*t_size[T_FLOAT]],&ldata[(segm[i].px0-x)*t_size[T_FLOAT]],
     		    ((segm[i].xwid-(segm[i].px0-x)))*t_size[T_FLOAT]);
     	     VPRINTF(VERBOSE, "Copied\n");
     	     ED_FREE(ldata);
     	     VPRINTF(VERBOSE, "Freed\n");
     	     break;
     	  }
       }
/*     
 *     No get all other frames that may contain the line
 */    
       for (i=0;i<n;i++) 
       {
     	  if (segm[i].px0 < x && 
     	      segm[i].py0>= (y+k) && segm[i].py0-segm[i].ywid < (y+k)) 
          {
     	     yline = segm[i].py0 - (y+k);

     	     VPRINTF(VERBOSE, "Reading line %d of next %s\n", yline, 
     			      segm[i].fds->name);
     	     ED_CALLOC(ldata, "get_fitsline", unsigned char,
     		       (segm[i].xwid*t_size[T_FLOAT]));
     	     if (!(segm[i].fds)) {
     		if (first) {
     		   for (j=0;j<n;j++) {
     		      if (segm[j].fds) {
     			 VPRINTF(NORMAL, "Closing %s: ", segm[j].fds->name);
     			 close_fits(segm[j].fds, NO);
     			 segm[j].fds = NULL;
     		      }
     		   }
     		   first = FALSE;
     		}
     		if (!(segm[i].fds = open_fits_file(control.fitsfile[i])))
     		  syserr(FATAL, "fitspaste", "Could not open file %s\n",
     		      control.fitsfile[i]);
     		segm[i].fds->pix_type = T_FLOAT;
     	     }
     	     read_fits_line(segm[i].fds, ldata, yline, NO);
     	     memcpy(&fdata[((afieldwidth*k)+(x-segm[i].px0))*t_size[T_FLOAT]], ldata, 
     		    (segm[i].xwid*t_size[T_FLOAT]));
             ED_FREE(ldata);
     	  }
       }
     }

     if (abinsize > 1) 
     {
        fpointer=(float *)fdata;

        for (j=0;j<afieldwidth;j++) 
        {
           for (k=1;k<abinsize;k++) 
	   {
	     fpointer[j] += fpointer[j+k*afieldwidth];
	   }
           fpointer[j] /= abinsize;
        }        
	for (j=0;j<afieldwidth;j++) 
        {
           fpointer[j] = fpointer[j*abinsize];
           for (k=1;k<abinsize;k++)
	   { 
               fpointer[j] += fpointer[j*abinsize+k];
	   }
           fpointer[j] /= abinsize;
        }
     }

     return fdata;
}

int main(int argc, char *argv[])
{
     fitsfile_struct **fds, *ffn=NULL;
     int	i, j;
     int	mincrpix1 = 999999, naxis1=0, maxcrpix1 = -999999;
     int	mincrpix2 = 999999, naxis2=0, maxcrpix2 = -999999;
     int        fieldwidth;
     int	outxwid, outywid;
     double	radelt, decdelt, scale;
     double	xmin, xmax, ymin, ymax;
     segm_struct	*segm;

     init(argc, argv, &control);

     sprintf(header,"%s Version: %s (%s)",NAME,VERS,DATE);
     VPRINTF(NORMAL,"\n   %s\n\n",header);
     ED_CALLOC(fds,  "fitspaste", fitsfile_struct*, control.ilist);
     ED_CALLOC(segm, "fitspaste", segm_struct, control.ilist);
     for (i=0;i<control.ilist;i++) {
        if ((fds[i] = open_fits_file(control.fitsfile[i])) == NULL)
           syserr(FATAL, "fitspaste", "Could not open file %s\n", 
                 control.fitsfile[i]);
        VPRINTF(NORMAL, "%s: , CRPIX1: %d, CRPIX2: %d\n",
                   fds[i]->name, fds[i]->px0, fds[i]->py0);
        fds[i]->pix_type = T_FLOAT;
        if (fds[i]->px0 > maxcrpix1) maxcrpix1 = fds[i]->px0;
        if (fds[i]->px0 < mincrpix1) {
           mincrpix1 = fds[i]->px0;
           naxis1 = fds[i]->xwid;
        }
        if (fds[i]->py0 > maxcrpix2) maxcrpix2 = fds[i]->py0;
        if (fds[i]->py0 < mincrpix2) {
           mincrpix2 = fds[i]->py0;
           naxis2 = fds[i]->ywid;
        }
/*
 *      Setup the output image and initialize ist dimensions
 */
        if (i == 0) {
           ffn = new_fits();
           inherit_fits(ffn, fds[0]);
           if (clipd || clippix) {
              if (control.binsize != 1)
                 syserr(FATAL, "fitspaste", 
                        "Cannot clip and bin simultaneously\n");
              if (clipd) {
                 radelt  = control.rad / (3600.0 * 
                           cos(control.dec*3.14159625/180.0)) / 2.0;
                 decdelt = control.decd / 3600.0 / 2.0;
                 scale   = fabs(ffn->dx);
                 coepro(control.ra+radelt, control.dec-decdelt, 
                        ffn->xcen, ffn->ycen, scale, &xmin, &ymin);
                 coepro(control.ra-radelt, control.dec+decdelt, 
                        ffn->xcen, ffn->ycen, scale, &xmax, &ymax);
              }
              if (clippix) {
                 xmin = control.xmin;
                 xmax = control.xmax;
                 ymin = control.ymin;
                 ymax = control.ymax;
              }
           }
        }
        segm[i].px0 = fds[i]->px0;
        segm[i].py0 = fds[i]->py0;
        segm[i].xwid= fds[i]->xwid;
        segm[i].ywid= fds[i]->ywid;
        close_fits(fds[i], NO);
     }
     mincrpix1 -= naxis1;
     mincrpix2 -= naxis2;
     outxwid = -mincrpix1 + maxcrpix1;
     fieldwidth = outxwid;
     outywid = -mincrpix2 + maxcrpix2;
     VPRINTF(NORMAL, "Size of output image: %d, %d\n\n", 
             outxwid/control.binsize, outywid/control.binsize);
     VPRINTF(NORMAL, "Ref pos output image: %d, %d\n\n", 
             maxcrpix1/control.binsize, maxcrpix2/control.binsize);
     if (clipd || clippix) {
        if ((maxcrpix1 < -(int) xmin) || 
            (maxcrpix2 < -(int) ymin) ||
            (mincrpix1 > -(int) xmax) || 
            (mincrpix2 > -(int) ymax)) {
/*
        if ((maxcrpix1 > -(int) xmin) || 
            (maxcrpix2 > -(int) ymin) ||
            (mincrpix1 < -(int) xmax) || 
            (mincrpix2 < -(int) ymax)) {
*/
           syserr(WARNING,"fitspaste", "Superimage boundaries: %d %d %d %d\n",
                          maxcrpix1, mincrpix1, maxcrpix2, mincrpix2);
           syserr(WARNING,"fitspaste", "Selection boundaries: %d %d %d %d\n",
                           -(int) xmin, -(int) (xmax), 
                           -(int) ymin, -(int) (ymax));
           syserr(WARNING,"fitspaste", 
                        "Selection area (partly) outside available area\n");
           if (((-(int) xmin) > maxcrpix1) && (control.outlarge==0)) xmin = -maxcrpix1;
           if (((-(int) xmax) < mincrpix1) && (control.outlarge==0)) xmax = -mincrpix1;
           if (((-(int) ymin) > maxcrpix2) && (control.outlarge==0)) ymin = -maxcrpix2;
           if (((-(int) ymax) < mincrpix2) && (control.outlarge==0)) ymax = -mincrpix2;
        }
        s_fits_int("NAXIS1", ffn->head, ((int) xmax-xmin+1));
        s_fits_int("NAXIS2", ffn->head, ((int) ymax-ymin+1));
        outxwid = ((int) xmax-xmin+1);
        outywid = ((int) ymax-ymin+1);
     } else {
        s_fits_int("NAXIS1", ffn->head, outxwid/control.binsize); 
        s_fits_int("NAXIS2", ffn->head, outywid/control.binsize);
        xmin = -maxcrpix1;
        ymin = -maxcrpix2;
        xmax = -mincrpix1;
        ymax = -mincrpix2;
     }
     VPRINTF(NORMAL,"Selection boundaries: %d %d %d %d\n\n",
                     (int) xmin, (int) (xmax), 
                     (int) ymin, (int) (ymax));
     s_fits_int("CRPIX1", ffn->head, -(int)xmin/control.binsize);
     s_fits_int("CRPIX2", ffn->head, -(int)ymin/control.binsize);
     if (fits_find("CDELT1", ffn->head)) {
       s_fits_float("CDELT1", ffn->head, fds[0]->dx*control.binsize);
       s_fits_float("CDELT2", ffn->head, fds[0]->dy*control.binsize);
     } else {
       s_fits_float("CD1_1", ffn->head, r_fits_float("CD1_1", ffn->head,1.0)*control.binsize);
       s_fits_float("CD2_2", ffn->head, r_fits_float("CD2_2", ffn->head,1.0)*control.binsize);
     }
     ffn->pix_type = T_FLOAT;
     ED_FOPEN(ffn->fd, "fitspaste", control.outfile, "w+");
     write_fits_header(ffn);

     ED_CALLOC(fdata, "main", unsigned char,
                           (fieldwidth*t_size[T_FLOAT]*control.binsize));

     for (i=(control.binsize-1);i<outywid;i+=control.binsize) 
     {
       /*data[0]=(float *)get_fitsline(segm,control.ilist, 
                 -xmin, -ymin-(i+k), fieldwidth, control.binsize);*/

       /* first set all elements of fdata to zer in case not the whole
          x-range of the final image is covered by the input images.
       */
       for(j=0; j<fieldwidth*t_size[T_FLOAT]*control.binsize; j++)
       {
	 fdata[j]=0;
       }

       get_fitsline(segm,control.ilist, 
                 -xmin, -ymin-i, fieldwidth, control.binsize);

        if (i % 256 == 0)
	{
           VPRINTF(NORMAL,"Processing line %d\n",i);
	}

        write_fits_line(ffn, fdata);
     }
     for (i=0;i<control.ilist;i++) {
        if (segm[i].fds)
	{
           close_fits(segm[i].fds, NO);
	}
     }
     padd_fits(ffn);
     close_fits(ffn, NO);
     ED_FREE(fds);
     ED_FREE(fdata);
     return 0;
}



