/*----------------------------------------------------------------------------
Code to combine 8 CCDs of the UH8K in a mosaic 
Jean-Charles Cuillandre jcc@cfht.hawaii.edu
Use of sub-functions from Mark Metzger code for uh8k (along with fitsio.c)
Faster code: does not "seek" on the upper half of the mosaic
Sample the overscan at the mid-height of the CCDs
Special handling of outputs - modular for the CFH12K
Now includes options: - binning factor selection
		      - overscan correction option
		      - output filename selection (default is uh8k.fits)

The geometry of the output amplifiers is:

 ----------A ----------A ----------A ----------A
|           |           |           |           |
|   CCD-2   |   CCD-3   |   CCD-5   |   CCD-4   |
|           |           |           |           |
 -----------------------------------------------
|           |           |           |           |
|   CCD-0   |   CCD-1   |   CCD-7   |   CCD-6   |
|           |           |           |           |
 A---------- A---------- A---------- A----------


01.08.00: modifeid the code to take all input images, output an int
          image for the ESO WFI
	  removed the bias possibility. We assume (without checking)
          that all inputimages have the same dimension. The used dimension
          of the input images is adapted to the required binning.

10.03.01: removed all DEC compiler warnings

24.04.03:
included the stdlib library to avoid warnings due to the
undefined 'exit' function.

08.07.2003:
fixed a bug when filling up the image to the 2880 byte block
(it was done with floats (that have 4 bytes) instead of characters
(having 1 byte))

13.07.2004:
corrected a bug in the padding (the number of bytes
to be padded is output.nx*output.ny*output.bytepix
instead of output.nx*output.ny).

29.01.2005:
add HISTORY information to the output headers.
----------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "immosaic_WFI.h"
#include "global_vers.h"

#define	MOSWBUF 2000000 /* 4Mb */

/* 12K raw data: 16 bits BSCALE=1 BZERO=32K */

extern struct IMAGE input[8],output;
FILE *open_file();
struct IMAGE read_FITS_header(),write_FITS_header(),add_FITS_header_text();
void give_date(char *);

/*----------------------------------------------------------------------------*/
int main(argc,argv)
int	argc;
char	*argv[];
{
int	nmospix;
int	i,j=0,k,l,z;
int     bitpix;
int	nxo,nyo,nxi,nyi,bin;
int     xdim, ydim;

float   sum;


PIXTYPE *pixtab;
PIXTYPE *im_out;
PIXTYPE *px;
char blockfill[blocksize];
char text[81], buf[81];
char *base;

	if(argc < 10) 
        {
	   printf("Syntax: immosaic X.*.fits binsize outbitpix [filename]\n");
	   printf("X = root of the filename (443165o)\n");
	   printf("with binsize >= 1\n");
	   printf("For bitpix only 16 and -32 are supported\n");
	   printf("Example: immosaic 443165o*.fits 8 16\n");
	   printf("         immosaic 443165o*MODF.fits 8 -32 443165o-B8.fits\n");
	   exit(1);
	}

        /* Get input images filename */
	for (i=0; i<8; i++) 
        {
            if     (i==0) j=8;
            else if(i==1) j=7;
            else if(i==2) j=6;
            else if(i==3) j=5;
            else if(i==4) j=1;
            else if(i==5) j=2;
            else if(i==6) j=3;
            else if(i==7) j=4;
	    strcpy(input[i].name, argv[j]);
	    input[i] = read_FITS_header(input[i]);
	}
        printf("Combining files:\n");
        printf("  %s  %s  %s  %s  \n",argv[1],argv[2],argv[3],argv[4]);
        printf("  %s  %s  %s  %s  \n\n",argv[8],argv[7],argv[6],argv[5]);

        /* Get binning factor */
	sscanf(argv[9],"%d",&bin);
        if (bin==0) 
        {
           printf("binsize must be >=1\n");
	   exit(1);
	}

	if (argc > 10)
	{
	  sscanf(argv[10],"%d",&bitpix);	  
	}

        /* Get output filename */
        if (argc == 12)
	{
           strcpy(output.name, argv[11]);
	}
        else
	{
           strcpy(output.name, "wfi.fits");
	}
	    
        /* Prepare output file */
	if(bitpix == 16)
	{
 	  output.bitpix=16;
	  output.bytepix=2;
	  output.offset=32768.0;
	}
	else
	{
 	  output.bitpix=-32;
	  output.bytepix=4;
	  output.offset=0.0;	  
	}

	output.scale=1.0;

	xdim=(int)(input[0].nx/bin)*bin;
	ydim=(int)(input[0].ny/bin)*bin;
	printf("using the dimension %d x %d of input images\n", xdim, ydim);

	nxo = output.nx = (4*xdim)/bin;
	nyo = output.ny = (2*ydim)/bin;
	output=write_FITS_header(input[0],output);

	/* add HISTORY information to the OUTPUT header */
        output=add_FITS_header_text("HISTORY"," ",output);
        sprintf(text,"imred: %s", __PIPEVERS__);
        output=add_FITS_header_text("HISTORY",text,output);
        give_date(buf);
        sprintf(text, "imred: called at %s", buf); 
        output=add_FITS_header_text("HISTORY",text,output);
        output=add_FITS_header_text("HISTORY","Input images:",output);
	for(i=0; i<8; i++)
	{
	    /* get basename from input images */  
	    base = strrchr (input[i].name, '/'); 
	    base = base ? base + 1 : (char *) input[i].name;
            snprintf(text, 70, "%s", base);
 	    output=add_FITS_header_text("HISTORY",text,output);  
	}

        /* input images size */
	nxi = input[0].nx; 
        nyi = input[0].ny;

        printf("Binning=%d => creating `%s' - %d x %d pixels\n",
               bin,output.name,nxo,nyo);


        /* The real images */
	nmospix = nxi * 4; /* 4 CCDs per half-bank */

        /* open files and go to data */
	for(i=0; i< 8; i++)
	{
	  input[i].imagefile=open_file(input[i].name,"rb");
	  fseek(input[i].imagefile, input[i].startdata, 0);
	}
        output.imagefile=open_file(output.name,"ab");
        fseek(output.imagefile, output.startdata, 0);
	
	pixtab = (PIXTYPE *)malloc(bin*nmospix*sizeof(PIXTYPE));
	im_out = (PIXTYPE *)malloc(nxo*sizeof(PIXTYPE));

			/*===================*/
			/* Lower half mosaic */
			/*===================*/

	for (i=0; i<nyo/2; i++) 
        {
	    px=pixtab;
            /* process "bin" lines */
	    for(k=0; k < bin; k++)
	    {
	      for(l=0; l<4; l++)
	      {
		read_file(&(input[l]), pixtab, nxi);
		pixtab+=nxi;
	      }
	    }

	    pixtab=px;

	    for (j=0; j<nxo; j++) 
            {
	        px = pixtab + j*bin;
	        sum = 0.0;
	        for (k=0; k<bin; k++) 
		{
	            for (l=0; l<bin; l++)
		    {
	                sum += (float)(*px++);
		    }
	            px += nmospix - bin;
		}
	        sum/=(float)(bin*bin);

	        im_out[j] = sum;
            }
		
            /* Write the line to the file */
     	    write_file(&(output), im_out, nxo);

	    if (i%2==0)
	    {
	       printf("Lower half: %d to %d\r",i,nyo/2);
	    }
	}
        printf("                                                \r");

                        /*===================*/
                        /* Upper half mosaic */
                        /*===================*/


	for (i=0; i<nyo/2; i++) 
	{
	    /* load buffers */
	    px=pixtab;
	    for(k=0; k < bin; k++)
	    {
	      for(l=4; l<8; l++)
	      {
		read_file(&(input[l]), pixtab, nxi);
		pixtab+=nxi;
	      }
	    }

	    pixtab=px;
            for (j=0; j<nxo; j++) 
	    {
                px = pixtab + j*bin;
                sum = 0;
                for (k=0; k<bin; k++) 
                {
                    for (l=0; l<bin; l++)
		    {
                        sum += (float)(*px++);
		    }
                    px += nmospix - bin;
                }
                sum/=(bin*bin);
                im_out[j] = (PIXTYPE)sum;
            }

            /* Write the line to the file */
     	    write_file(&(output), im_out, output.nx);
	    if (i%2==0)
	       printf("Upper half: %d to %d\r",i,nyo/2);
	}

        /* Fill in FITS file */
        z = blocksize-(output.nx*output.ny*output.bytepix)%blocksize;
        if(z) 
        {
           for(i=0; i<=z-1; i++)
	   {
              blockfill[i]=0.0;
	   }
           if((fwrite((char *)blockfill, z, 1, output.imagefile)) !=1 ) 
           {
             printf("\nCannot pad %d bytes to %s, exiting\n\n",z,output.name);
             exit(1);
           }
        }

	/* close files and bye */
	for(i=0; i<8; i++)
	{
	  fclose(input[i].imagefile);
	}
	fclose(output.imagefile);
	
	free(pixtab);
	free(im_out);

	return(0);
}
