#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "project_globals.h"
#include "options_globals.h"
#include "utils_globals.h"
#include "model_globals.h"
#include "compute_globals.h"

void compute_position(float *xobj, float *yobj, int *f, int *c, 
                      double *ra, double *dec, int nstars, 
                      model_struct *model)
/****** compute_position ***************************************************
 *
 *  NAME	
 *      compute_position - Compute the sky positions for a set of objects
 *
 *  SYNOPSIS
 *      void compute_position(float *xobj, float *yobj, int *f, int *c,
 *                            double *ra, double *dec, int nstars,
 *                            double ***centers,
 *                            float ***par, int npar, int *parnum, int *paroff)
 *
 *  FUNCTION
 *      Build up a list of catalog positions for all stars in one field
 *
 *  INPUTS
 *      xobj    - The x-coordinates of the input objects
 *      yobj    - The y-coordinates of the input objects
 *      f       - The frame numbers of the input objects
 *      c       - The camera numbers of the input objects
 *      nstars  - The number of objects to compute positions for
 *      model   - The full model structure
 *
 *  RESULTS
 *      ra      - List of derived Right Ascencions for the objects
 *      dec     - List of derived Declinations for the objects
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 17-07-1998
 *
 ******************************************************************************/
{
    int		i, j, jof;
    float	*work;
/*
 *  Apply the plate model for each star
 */
    ED_CALLOC(work, "compute_position", float, model->np);
    for (i=0;i<nstars;i++) {
       jof = model->paroff[c[i]];
       bi_cubic(f[i], c[i], xobj[i], yobj[i], model, &work[jof]);
       ra[i]  = model->params[0][model->np+f[i]];
       dec[i] = model->params[1][model->np+f[i]];
       for (j=jof;j<jof+model->parnum[c[i]];j++) {
          ra[i]  += model->params[0][j] * work[j];
          dec[i] += model->params[1][j] * work[j];
       }
/*
 *     Convert gnomonic coordinates to true RA and Dec
 */
       rel_to_abs(model->centers[0][c[i]][f[i]], 
                  model->centers[1][c[i]][f[i]],
                  0.0, &ra[i], &dec[i], &ra[i], &dec[i], 1);
       if (ra[i] < 0.0)     ra[i] += 2.0*DPI;
       if (ra[i] > 2.0*DPI) ra[i] -= 2.0*DPI;
    }
    ED_FREE(work);
}
/**/
void compute_errors(float *xobj, float *yobj, float *eobj, int *f, int *c,
                    float *sys_rms, float *rnd_rms, int nstars, 
                    model_struct *model)
/*-----------------------------------------------------------------------------
 *  DENIS: astrometric reduction
 *  Build up a list of position standard errors for all stars in one field
 * 
 *  * Parameters:
 *      F       INTEGER       Field number
 *      IB      INTEGER       Colour channel                              
 *      SYS_RMS REAL*4(*)     Standard error for individual stars in field 
 *                            (systematic part)
 *      RND_RMS REAL*4(*)     Standard error for individual stars in field
 *                            (random part)
 *      PIXPOS  REAL*4(4,*)   Pixel positions of stars
 *      NSTARS  INTEGER       Maximum number of stars in one field on input.
 *                            Number of stars in field on output
 *      PAR     REAL*4(2,*)   Plate model coefficients
 *      NPAR    INTEGER       Number of plate model coefficients
 *      NF      INTEGER       Number of fields in strip
 *      COV     REAL4(*,*)    Covariance matrix
 * -------------------------------------------------------------------------*/
{
    float	*afunc, **agrad;
    int		i, j, k, l, jof;
    double	sig2sys, sig2rand, rr;

    ED_CALLOC(afunc, "compute_errors", float, model->np);
    ED_CALLOC(agrad, "compute_errors", float *, 2);
    for (i=0;i<2;i++) 
       ED_CALLOC(agrad[i], "compute_errors", float, model->np);
    for (i=0;i<nstars;i++) {
       jof = model->paroff[c[i]];
       bi_cubic(f[i], c[i], xobj[i], yobj[i], model, afunc);
       bi_cubicgrad(f[i], c[i], xobj[i], yobj[i], model, agrad);
       for (j=0,sig2sys=sig2rand=0.0;j<2;j++) {
          for (k=1;k<model->parnum[c[i]];k++) 
             for (l=0;l<k;l++) 
                sig2sys += 2.0*afunc[k]*model->alpha[l+jof][k+jof]*afunc[l];
          for (k=0;k<model->parnum[c[i]];k++) {
             sig2sys +=  afunc[k]*model->alpha[k+jof][model->np+f[i]]*2.0 +
                         afunc[k]*model->alpha[k+jof][k+jof]*afunc[k];
          }
          sig2sys += model->alpha[model->np+f[i]][model->np+f[i]];
          rr = 0.0;
          for (k=0;k<model->parnum[c[i]];k++)
             rr += model->params[j][k+jof] * agrad[j][k];
          rr *= eobj[i];
          sig2rand += rr * rr;
       }
       sys_rms[i] = sqrt(sig2sys);
       rnd_rms[i] = sqrt(sig2rand);
    }
    for (i=0;i<2;i++) 
       ED_FREE(agrad[i]);
    ED_FREE(afunc);
    ED_FREE(agrad);
}
/**/
void compute_mom(float *xobj, float *yobj, int *f, int *c,
                 double *xx, double *yy, double *xy,
                 double *XX, double *YY, double *XY,
                 int nstars, model_struct *model)
/*-----------------------------------------------------------------------------
 * DENIS: astrometric reduction
 * Compute the true (i.e. in radians) second order moments for all stars in 
 * one field from the pixel ones.
 *
 * Parameters:
 *     F       INTEGER       Field number
 *     IB      INTEGER       Colour channel                              
 *     PIXPOS  REAL*4(4,*)   Pixel positions of stars (X,Y,Err,Number)
 *     PIXMOM  REAL*4(3,*)   Pixel second order moments (XX,XY,YY) 
 *     SKYMOM  REAL*4(3,*)   Sky units second order moments (XX,XY,YY)
 *     NSTARS  INTEGER       Number of stars in field 
 *     PAR     REAL*4(2,*)   Plate model coefficients
 *     NPAR    INTEGER       Number of plate model coefficients
 *     NF      INTEGER       Number of fields in strip
 *     GRAD    External      Subroutine which compute the gradients of the plate
 *                           model elements 
 *---------------------------------------------------------------------------*/
{
    float	**agrad, **vgrad;
    int		i, j, k, l, jof;

    ED_CALLOC(agrad, "compute_mom", float *, 2);
    ED_CALLOC(vgrad, "compute_mom", float *, 2);
    for (i=0;i<2;i++) {
      ED_CALLOC(agrad[i], "compute_mom", float, model->np);
      ED_CALLOC(vgrad[i], "compute_mom", float, 2);
    }
    for (i=0;i<nstars;i++) {
/*
 *    Compute normal coordinate gradient relative to pixel positions (i.e. local
 *    linear approximation of the astrometric transformation, 
 *    i.e. metric tensor).
 */
      jof = model->paroff[c[i]];
      bi_cubicgrad(f[i], c[i], xobj[i], yobj[i], model, agrad);
      for (j=0;j<2;j++) {
        for (k=0;k<2;k++) {
          vgrad[j][k] = 0.0;
          for (l=0;l<model->parnum[c[i]];l++) 
            vgrad[j][k] += model->params[j][l+jof]*agrad[k][l];
        }
      }
/*
 *    Use the metric tensor to convert the second order moments. 
 *    Given the small dimensionality formulas are explicited.
 */
      XX[i] =   vgrad[0][0]*vgrad[0][0]*xx[i] + 
              2*vgrad[0][0]*vgrad[0][1]*xy[i] +
                vgrad[0][1]*vgrad[0][1]*yy[i];
      XY[i] =   vgrad[0][0]*vgrad[1][0]*xx[i] + 
                vgrad[0][0]*vgrad[1][1]*xy[i] +
                vgrad[0][1]*vgrad[1][0]*xy[i] +
                vgrad[0][1]*vgrad[1][1]*yy[i];
      YY[i] =   vgrad[1][0]*vgrad[1][0]*xx[i] + 
              2*vgrad[1][0]*vgrad[1][1]*xy[i] +
                vgrad[1][1]*vgrad[1][1]*yy[i];
    }
    for (i=0;i<2;i++) {
      ED_FREE(agrad[i]);
      ED_FREE(vgrad[i]);
    }
    ED_FREE(agrad);
    ED_FREE(vgrad);
}
/**/
void compute_vec(float *xobj, float *yobj, int *f, int *c,
                 double *x, double *y,
                 double *X, double *Y,
                 int nstars, model_struct *model)
/*-----------------------------------------------------------------------------
 * DENIS: astrometric reduction
 * Compute the true (i.e. in radians) vector for all stars in 
 * one field from the pixel ones.
 *
 * Parameters:
 *     F       INTEGER       Field number
 *     IB      INTEGER       Colour channel                              
 *     PIXPOS  REAL*4(4,*)   Pixel positions of stars (X,Y,Err,Number)
 *     PIXMOM  REAL*4(3,*)   Pixel second order moments (XX,XY,YY) 
 *     SKYMOM  REAL*4(3,*)   Sky units second order moments (XX,XY,YY)
 *     NSTARS  INTEGER       Number of stars in field 
 *     PAR     REAL*4(2,*)   Plate model coefficients
 *     NPAR    INTEGER       Number of plate model coefficients
 *     NF      INTEGER       Number of fields in strip
 *     GRAD    External      Subroutine which compute the gradients of the plate
 *                           model elements 
 *---------------------------------------------------------------------------*/
{
    float	**agrad, **vgrad;
    int		i, j, k, l, jof;

    ED_CALLOC(agrad, "compute_vec", float *, 2);
    ED_CALLOC(vgrad, "compute_vec", float *, 2);
    for (i=0;i<2;i++) {
      ED_CALLOC(agrad[i], "compute_vec", float, model->np);
      ED_CALLOC(vgrad[i], "compute_vec", float, 2);
    }
    for (i=0;i<nstars;i++) {
/*
 *    Compute normal coordinate gradient relative to pixel positions (i.e. local
 *    linear approximation of the astrometric transformation, 
 *    i.e. metric tensor).
 */
      jof = model->paroff[c[i]];
      bi_cubicgrad(f[i], c[i], xobj[i], yobj[i], model, agrad);
      for (j=0;j<2;j++) {
        for (k=0;k<2;k++) {
          vgrad[j][k] = 0.0;
          for (l=0;l<model->parnum[c[i]];l++) 
            vgrad[j][k] += model->params[j][l+jof]*agrad[k][l];
        }
      }
/*
 *    Use the metric tensor to convert the second order moments. 
 *    Given the small dimensionality formulas are explicited.
 */
      X[i] =   vgrad[0][0]*x[i] + vgrad[0][1]*y[i];
      Y[i] =   vgrad[1][0]*x[i] + vgrad[1][1]*y[i];
    }
    for (i=0;i<2;i++) {
      ED_FREE(agrad[i]);
      ED_FREE(vgrad[i]);
    }
    ED_FREE(agrad);
    ED_FREE(vgrad);
}
/**/
void compute_mat(float *xobj, float *yobj, int *f, int *c,
                 double *xx, double *xy, double *yx, double *yy,
                 double *XX, double *XY, double *YX, double *YY,
                 int nstars, model_struct *model)
/*-----------------------------------------------------------------------------
 * DENIS: astrometric reduction
 * Compute the true (i.e. in radians) second order moments for all stars in 
 * one field from the pixel ones.
 *
 * Parameters:
 *     F       INTEGER       Field number
 *     IB      INTEGER       Colour channel                              
 *     PIXPOS  REAL*4(4,*)   Pixel positions of stars (X,Y,Err,Number)
 *     PIXMOM  REAL*4(3,*)   Pixel second order moments (XX,XY,YY) 
 *     SKYMOM  REAL*4(3,*)   Sky units second order moments (XX,XY,YY)
 *     NSTARS  INTEGER       Number of stars in field 
 *     PAR     REAL*4(2,*)   Plate model coefficients
 *     NPAR    INTEGER       Number of plate model coefficients
 *     NF      INTEGER       Number of fields in strip
 *     GRAD    External      Subroutine which compute the gradients of the plate
 *                           model elements 
 *---------------------------------------------------------------------------*/
{
    float	**agrad, **vgrad;
    int		i, j, k, l, jof;

    ED_CALLOC(agrad, "compute_mat", float *, 2);
    ED_CALLOC(vgrad, "compute_mat", float *, 2);
    for (i=0;i<2;i++) {
      ED_CALLOC(agrad[i], "compute_mat", float, model->np);
      ED_CALLOC(vgrad[i], "compute_mat", float, 2);
    }
    for (i=0;i<nstars;i++) {
/*
 *    Compute normal coordinate gradient relative to pixel positions (i.e. local
 *    linear approximation of the astrometric transformation, 
 *    i.e. metric tensor).
 */
      jof = model->paroff[c[i]];
      bi_cubicgrad(f[i], c[i], xobj[i], yobj[i], model, agrad);
      for (j=0;j<2;j++) {
        for (k=0;k<2;k++) {
          vgrad[j][k] = 0.0;
          for (l=0;l<model->parnum[c[i]];l++) 
            vgrad[j][k] += model->params[j][l+jof]*agrad[k][l];
        }
      }
/*
 *    Use the metric tensor to convert the second order moments. 
 *    Given the small dimensionality formulas are explicited.
 */
      XX[i] =   vgrad[0][0]*vgrad[0][0]*xx[i] + 
                vgrad[0][0]*vgrad[1][0]*xy[i] +
                vgrad[0][0]*vgrad[0][1]*yx[i] +
                vgrad[0][1]*vgrad[1][0]*yy[i];
      XY[i] =   vgrad[0][0]*vgrad[0][1]*xx[i] + 
                vgrad[0][0]*vgrad[1][1]*xy[i] +
                vgrad[0][1]*vgrad[0][1]*yx[i] +
                vgrad[0][1]*vgrad[1][1]*yy[i];
      YX[i] =   vgrad[0][0]*vgrad[1][0]*xx[i] + 
                vgrad[1][0]*vgrad[1][0]*xy[i] +
                vgrad[0][0]*vgrad[1][1]*yx[i] +
                vgrad[1][0]*vgrad[1][1]*yy[i];
      YY[i] =   vgrad[1][0]*vgrad[0][1]*xx[i] + 
                vgrad[1][0]*vgrad[1][1]*xy[i] +
                vgrad[0][1]*vgrad[1][1]*yx[i] +
                vgrad[1][1]*vgrad[1][1]*yy[i];
    }
    for (i=0;i<2;i++) {
      ED_FREE(agrad[i]);
      ED_FREE(vgrad[i]);
    }
    ED_FREE(agrad);
    ED_FREE(vgrad);
}
