#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "fitscat.h"
#include "sort_globals.h"
#include "utils_globals.h"
#include "field_desc.h"
#include "dist_globals.h"
#include "offsets_defs.h"
#include "options_globals.h"
#include "preastrom_globals.h"

/*
   11.03.2005:
   I changed the argument list of find_offsets to remove a not needed
   argument.

   27.03.2006:
   I renamed sorting and indexing functions to the new naming scheme
   described in the sort_globals.h file

   15.02.2010:
   I removed compiler warnings due to possibly undefined variables
*/

static void	offset(int, double, float *, float *, float *, float *),
		cloudsize(int, float, float, float *, float *, 
		          double *, double *);

static int	pairs(int *, float *, float *, int *, float, float,
                      int *, float *, float *, int *, float, float, 
                      int, int, double, double, 
                      float, float, float **, float **),
		boxcar(int, float, float, float *, float *, 
		       float, float, float *, float *, 
		       float *, float *);

static double	scatter(int, float, float, float *, float *),
		probab(int, int, int);


int find_offset(float *xa, float *ya, int na, float xa_off, float ya_off, 
                float *xb, float *yb, int nb, float xb_off, float yb_off, 
                double maxoff, double poserr, double maxdis,
                int xwid, int ywid, 
                float *xoff,  float *yoff, float *errx, float *erry,
                int field_nr, 
                int *obj_dist_nr, catstruct *distcat)
{

     float	xerr, yerr;	/* size of the search box in pairs() */
     float	pxerr, pyerr;	/* input parameter - estimated max. offset */
     float	cerr;		/* cerr: pos err in the reference catalog */
     int	xcelln, ycelln;	/* number of cells in the histogram */
     float	*xdist, *ydist;
     float	*xpdist, *ypdist;
     float	xpeak = 0.0, ypeak = 0.0;	
                                /* lower left corner of the boxcar at peak */	
     int	*idxa, *idxb;	/* indexes for sorting purposes */
     int	nd;		/* nd: number of distances to be stored */ 
     int	nbb;		/* nbb: estimated number of "widow" stars */
     int	nc;		/* nc: number of cells in the histogram */
     double	pf1, pf2;	/* pf: probability of false peak in histogram */
     float	xoffset = 0.0, yoffset = 0.0;/* the offset of the telescope */
     float	xcs, ycs;	/* (x/y)cs: histogam cellsize in arcseconds */
     int	ph;		/* ph: height of the histogram peak */
     double	spread;		/* scatter of the points in the peak */
     double	xcls, ycls;	/* the x and y size of the "cloud" */
     int	retcode = FALSE;

     pxerr  = maxoff;
     pyerr  = pxerr;
     cerr   = poserr;
     maxdis = maxdis;

     ED_MALLOC(xdist,  "find_offset", float, MAXD);
     ED_MALLOC(ydist,  "find_offset", float, MAXD);
     ED_MALLOC(xpdist, "find_offset", float, MAXD);
     ED_MALLOC(ypdist, "find_offset", float, MAXD);
     ED_MALLOC(idxa  , "find_offset", int  , na);
     ED_MALLOC(idxb  , "find_offset", int  , nb);
/* 
 *   set parameters 
 */
     xcs = cerr * K2;
     ycs = cerr * K2;
     VPRINTF(DEBUG,"X and Y dimension of the boxcar: xcs= %g ycs= %g \n",
             xcs, ycs);
     xcelln = ((int) (pxerr * K1/xcs)) * 2 + 1;
     VPRINTF(DEBUG,"The size of the search box in boxcar_size units:\n");
     VPRINTF(DEBUG,"xcelln= %d \n", xcelln);
     ycelln = ((int) (pyerr * K1/ycs)) * 2 + 1;
     VPRINTF(DEBUG,"ycelln= %d \n", ycelln);
     xerr = (xcelln * xcs)/2;
     yerr = (ycelln * ycs)/2;
     nd = 0;
/* 
 *   Sort then on x coordinate.
 */
     if (na != 0 && nb != 0) {
       findex(na, xa, idxa);
       findex(nb, xb, idxb);
       VPRINTF(VERBOSE,"na: %d, nb: %d\n",na,nb);

       nd = pairs(&na, xa, ya, idxa, xa_off, ya_off,
                  &nb, xb, yb, idxb, xb_off, yb_off,
                  xwid, ywid, ((double) xwid/10), ((double) ywid/10), 
                  xerr, yerr, &xdist, &ydist);

       VPRINTF(NORMAL,
               "Number of extracted_obj/reference_star pairs found: %d\n", nd);
     }
     if (nd > 0) {
/* 
 *      go through the coordinate-differences with a boxcar (size: xcs x ycs),
 *      and find the position of the peak density, and the height of the peak 
 */
        if (distcat) {
          *obj_dist_nr = nd;
          dist_objects_create(distcat, nd, field_nr, xdist, ydist);
        }
	ph = boxcar(nd, xcs, ycs, xdist, ydist, xerr, yerr,
                    xpdist, ypdist, &xpeak, &ypeak);
	VPRINTF(NORMAL,"Number of pairs at the peak: ph=%d \n", ph);
	VPRINTF(VERBOSE,"X and Y coordinate of the peak:\n");
	VPRINTF(VERBOSE,"xpeak= %g ypeak= %g \n", xpeak, ypeak);
/* 
 *      check peak  
 */
        if  (nb > na) {
           nbb = nd - ((int) nb* K3);
        } else {
           nbb = nd - ((int) na* K3);
        }
        nc = xcelln * ycelln;
        if (nbb >= ph) {
           pf1 = probab(nb, ph, nc);
           VPRINTF(VERBOSE,"The probability that it is only a noise peak:\n");
           VPRINTF(VERBOSE,"pf1=%g \n", pf1);
           if (pf1 > CRITP) {
              VPRINTF(NORMAL,"Warning! It might be a noise peak!\n");
              retcode = FALSE;
           }
        }
        nbb = nd;
        pf2 = probab(nbb, ph, nc);
        VPRINTF(VERBOSE,"The probability that it is only a noise peak:\n");
        VPRINTF(VERBOSE,"pf2=%g \n", pf2);
        if (pf2 > CRITP) {
           VPRINTF(NORMAL,"Warning! It might be a noise peak!\n");
           retcode = FALSE;
        } else {
           retcode = TRUE;
        }
/* 
 *      compute the x and y offsets, averaging the coordinate differences
 *      of the objects belonging to the peak, and print them - 
 *      offsets: r - e sense 
 */
        offset(ph, maxdis, xpdist, ypdist, &xoffset, &yoffset);
        VPRINTF(VERBOSE,"This is the final result:\n");
        VPRINTF(VERBOSE,"x_offset= %g y_offset= %g \n", xoffset, yoffset);
/* 
 *      compute the scatter of the distances belonging to the peak
 */
        if (ph > 1) {
           spread = scatter(ph, xoffset, yoffset, xpdist, ypdist);
           VPRINTF(VERBOSE,"This is the error in the offset:\n");
           VPRINTF(VERBOSE,"spread= %g \n", spread);
           cloudsize(ph, xoffset, yoffset, xpdist, ypdist, &xcls, &ycls);
           *errx = (float) xcls;
           *erry = (float) ycls;
        } else {
          *errx = maxdis;
          *erry = maxdis;
        }
     } else {
          VPRINTF(NORMAL,"No extracted stars within the error boxes\n");
          *errx = pxerr;
          *erry = pxerr;
          return FALSE;
     }
     VPRINTF(VERBOSE,"X_cloud_size= %f Y_cloud_size= %f\n", *errx, *erry);
     *xoff = xoffset;
     *yoff = yoffset;
     ED_FREE(xdist);
     ED_FREE(ydist);
     ED_FREE(xpdist);
     ED_FREE(ypdist);
     ED_FREE(idxa);
     ED_FREE(idxb);
     return retcode;
}

/*----------------------
  FIND UTILITY ROUTINES
  ----------------------------*/

/* coordinate-difference computing loop */

static int pairs(int *n_ne, float *xe, float *ye, int *idxe, 
                 float xoff, float yoff,
                 int *n_nr, float *xr, float *yr, int *idxr,
                 float xref, float yref,
                 int xwid, int ywid, double xmargin, double ymargin, 
                 float xerr, float yerr, float **xdist, float **ydist)
{
	int lastbeg, i, j, nu, nblk = 1;
	float xdif, ydif;
        float xmin = (float) xmargin, xmax = (float) xwid - (float) xmargin, 
              ymin = (float) ymargin, ymax = (float) ywid - (float) ymargin;

        VPRINTF(DEBUG,"ne: %d, nr: %d\n",*n_ne, *n_nr);

	nu = 0;
	lastbeg = 0;
	for (i = 0; i < *n_nr; i++) {
           for (j = lastbeg; j < *n_ne; j++) {
              if (xr[idxr[i]] > xmin && xr[idxr[i]] < xmax &&
                  xe[idxe[j]] > xmin && xe[idxe[j]] < xmax) {
                 xdif = (xr[idxr[i]] - xref) - (xe[idxe[j]] - xoff);
                 if (xdif > xerr) {
                    lastbeg = j;
                    continue;
                 }
                 if (xdif < -xerr)
                    break;
                 if (yr[idxr[i]] > ymin && yr[idxr[i]] < ymax &&
                     ye[idxe[j]] > ymin && ye[idxe[j]] < ymax) {
                    ydif = (yr[idxr[i]] - yref) - (ye[idxe[j]] - yoff);
                    if ((ydif > yerr) || (ydif < -yerr))
                       continue;
/* 
 *                  Record differences 
 */
                    (*xdist)[nu] = xdif;
                    (*ydist)[nu] = ydif;
                    nu++;
                    if (nu >= MAXD * nblk) {
                       nblk++;
                       ED_REALLOC(*xdist, "pairs", float, MAXD * nblk);
                       ED_REALLOC(*ydist, "pairs", float, MAXD * nblk);
                    }
                 }
              }
           }
        }
	return nu;
}

/* boxcaring through the coordinate-difference plane */

static int boxcar(int nd, float xcs, float ycs, float *xdist, float *ydist, 
                  float xerr, float yerr, float *xpdist, float *ypdist, 
                  float *xpeak, float *ypeak)
{
	int i, k, peak, lastp, firstk, k1peak, k2peak, ii;
	float lx, by;	/* left x and bottom y coordinate of the boxcar */
	float xd, yd;
        float bxstep, bystep;  /* boxcar stepsize in x and y direction */

	bxstep = xcs / (double) K4;
	bystep = ycs / (double) K4;

	VPRINTF(VERBOSE,"X and Y dimension of the search box (in pixels):\n");
	VPRINTF(VERBOSE,"xerr=%g yerr=%g\n", xerr, yerr); 
	VPRINTF(VERBOSE,"X and Y boxcar_step (in pixels):\n");
	VPRINTF(VERBOSE,"bxstep=%g bystep=%g\n", bxstep, bystep);

	lastp = 0;
/*
 *      To speed up the porcess of box-carring first sort the position
 *      distances arrays on the x-coordinate
 */
        if ( nd > 1) ffsort(nd, xdist, ydist);

        firstk = 0; k1peak = 0; k2peak = nd-1;
	for(lx = -xerr; lx <= (xerr - xcs); lx += bxstep) {
	for(by = -yerr; by <= (yerr - ycs); by += bystep) {
           peak = 0;
           for(k = firstk; k < nd; k++) {
              xd = xdist[k] - lx;
/*
 *            Set and check the x-coordinate bounds for fast processing
 *            We will never have to consult distances to the left of the
 *            current x-box.
 *            Nor do we have to look for distance to the right of the
 *            current x-box.
 */
              if (xd < 0.0) firstk = k;
              if (xd > xcs) break;

              yd = ydist[k] - by;
              if (xd < xcs && xd >= 0.0 && yd < ycs && yd >= 0.0) {
                 peak++;
                 if (peak >= MAXE) {
                    fprintf(stderr,"Too many distances - increase MAXE\n");
                    exit(1);
                 }
              }
           }
/*
 *         Keep track of the highest peak parameters
 */
           VPRINTF(DEBUG, "%f %f %d\n", lx, by, peak);
           if (peak > lastp) {
              *xpeak = lx;
              *ypeak = by;
              k1peak = firstk;
              k2peak = (k < nd)? k: nd-1;
              lastp = peak;
           }
	}
	}
/*
 *      Construct the list of distances for the highest peak box
 */
        ii = 0;
        for (i=k1peak; i<=k2peak;i++) {
           xd = xdist[i] - *xpeak;
           yd = ydist[i] - *ypeak;
           if (xd < xcs && xd >= 0.0 && yd < ycs && yd >= 0.0) {
              xpdist[ii] = xdist[i];
              ypdist[ii] = ydist[i];
              ii++;
           }
        }
	return lastp;
}

/* computing x and y offset, and the spread of the "cloud" */

static void offset(int ph, double maxdis, float *xpdist, float *ypdist, 
                   float *xoffset, float *yoffset)
{
	int i, count = 0;
        double del, xoff, yoff;

	*xoffset = xoff = 0.0;
	*yoffset = yoff = 0.0;
	for(i = 0; i < ph; i++) {
	   xoff += xpdist[i];
	   yoff += ypdist[i];
	}
	xoff /= ph;
	yoff /= ph;
	for(i = 0; i < ph; i++) {
           del = sqrt((xpdist[i] - xoff) * (xpdist[i] - xoff) +
                      (ypdist[i] - yoff) * (ypdist[i] - yoff));
           if (del <= 5.0 * maxdis) {
              *xoffset += xpdist[i];
              *yoffset += ypdist[i];
              count++;
           }
        }
        *xoffset /= count;
        *yoffset /= count;
}

/* compute the scatter of points belonging to the peak */

static double scatter(int ph, float xoffset, float yoffset, 
                      float *xpdist, float *ypdist)
{
	int i;
	double sigma, dx, dy;

	sigma = 0.0;
	for(i = 0; i < ph; i++) {
		dx = xoffset - xpdist[i];
		dy = yoffset - ypdist[i];
		sigma += dx * dx + dy * dy;
	}
	sigma = sqrt(sigma) / (double) ph;
	return sigma;
}

/* compute the size of the "cloud" */

static void cloudsize(int ph, float xoffset, float yoffset, 
                       float *xpdist, float *ypdist, double *xcls, double *ycls)
{
	int i;
	double fdx, fdy;

	*xcls = 0.0;
	*ycls = 0.0;
	for(i = 0; i < ph; i++) {
		fdx = fabs(xoffset - xpdist[i]);
		fdy = fabs(yoffset - ypdist[i]);
		if(fdx > *xcls) *xcls = fdx; 
		if(fdy > *ycls) *ycls = fdy;
	}
}


/* 
 *   compute the probability of a noise peak in the histogram 
 *   (binomial distribution) 
 */

static double probab(int nb, int ph, int nc)
{
	double p, pp;
	int nex;

	
/* p = ((double)fac(nb)/((double)fac(ph) * (double)fac(nb - ph))); */
	p = (double)divide_fac(nb,ph);
        pp = (double)fac(nb - ph);
        if (pp > 1.0e-50  && pp < 1.0e+50 &&
            p > 1.0e-50  && p < 1.0e+50) {
	   p = p * pow(((double)1.0 / (double) nc), (double) ph) / pp; 
           p = p * pow(((double)1.0-(1.0/(double) nc)), (double) (nb - ph))/pp;
	   nex = nc * K4 * K4;
	   pp = pow(((double)1.0 - p),(double) nex);
	   pp = 1.0 - pp;
        } else {
          pp = 0.0;
        }
	return pp;
}
