/* 
   the program get_pmm_objects extracts objects from
   a standardstar catalog directory generated with make_pmm.

   10.01.2005:
   I removed compiler warnings under gcc with optimisation
   enabled (-O3).

   15.05.2006:
   - I changed the welcome message of the program.
   - I added an option to print standard star catalogue
     entries directly usable with the ASTROMETRIX tool.
*/

#include <stdio.h>
#include <ctype.h>
#include <math.h>
#include <string.h>
#include "utils_globals.h"
#include "refs_io_defs.h"
#include "refs_io_types.h"
#include "refs_io_globals.h"
#include "global_vers.h"

#define MAXSTRINGLENGTH 255
#define DEG2RAD 3.14159/180.

#define __PROGRAMMVERSION__ "1.0"
#define __PROGRAMMNAME__ "get_pmm_objects"

#define __USAGE__ "\n\
USAGE:\n\
   get_pmm_objects -i dir. with the pmm catalogs\n\
                   -r RA  centre (deg)\n\
                   -d DEC centre (deg)\n\
                   -s size of the box for objects to be retrieved (deg)\n\
                   -a Create output compatible with a standard star catalogue\n\
                      directly usable with ASTROMETRIX. Ra and Dec are reported\n\
                      in sexagesimal notation and only one magnitude quantity is given.\n\
                      For the USNO-A2 catalogue, the Red Magnitude is printed.\n\n\
DESCRIPTION:\n\
The program get_pmm_objects retrieves objects from a pmm directory\n\
created with the make_pmm program. It lists all objects to stdout around\n\
RA and DEC in a square with sidelength set with the (-s) option.\n\n\
EXAMPLE:\n\
   get_pmm_objects -i /afs/mpa/data/erben/pmm1catalog/ -r 309.0 -d -29.0 -s 0.5 -a\n\n\
AUTHOR: Thomas Erben \n"

/* subroutine to convert degrees to sexagesimal notation */
char *decimaltoxms(double, char *);

int main(int argc, char **argv)
{
  /* variable declarations */

  char pmmdir[MAXCHAR];
  char hmsstring[MAXCHAR];
  char dmsstring[MAXCHAR];
  ref_objs    *refs;
  int i;
  int sexagesimal = 0;
  float racent=0.0, deccent=0.0, size=0.0;


  /* print start message: */
  fprintf(stderr, "%s (V%s; %s)\n\n", __PROGRAMMNAME__, __PROGRAMMVERSION__, __PIPEVERS__);
  if(argc < 8)
  {
    fprintf(stderr, "%s\n", __USAGE__);
    exit(1);
  }

  /* get the command line arguments: */
  for (i=1; i<argc; i++) 
  {
    if (argv[i][0] == '-') 
    {
       switch(tolower(argv[i][1])) 
       {
         case 'i': strncpy(pmmdir, argv[++i], MAXSTRINGLENGTH-1);
                   break;
         case 'r': racent=atof(argv[++i]);
                   break;
         case 'd': deccent=atof(argv[++i]);
                   break;
         case 's': size=atof(argv[++i]);
                   break;
         case 'a': sexagesimal = 1;
                   break;
         default:  
                   break;
       }
    } 
    else 
    {
       syserr(FATAL, "get_pmm_objects","Wrong Command line syntax\n");
    }
  }

  refs = get_refs(pmmdir, racent, deccent, size/cos(deccent*DEG2RAD), size); 

  for(i=0; i < refs->nobj; i++)
  {
    if(sexagesimal == 0)
    {
      fprintf(stdout, "%f %f %f %f\n", refs->ra[i], refs->dec[i],
                                      refs->rmag[i], refs->bmag[i]);
    }
    else
    {
      fprintf(stdout, "%s %s %f\n", decimaltoxms((double)(refs->ra[i]/15.0), hmsstring), 
                                       decimaltoxms((double)(refs->dec[i]), dmsstring),
                                       refs->rmag[i]);
    }
  }

  return 0;
}

/*-------------------------------------------------------------------------*/
/**
  @brief    converts an angle from decimal to sexagesimal notation
  @param    angle  the angle in degrees
  @param    hmsstring  the string for the result
  @return   a pointer to hmsstring.

  Space for hmsstring has to be allocated before the function
  is called.
 */
/*--------------------------------------------------------------------------*/

char *decimaltoxms(double angle, char *hmsstring)
{
	double 	h, m, s;
	int 	sign;

	sign = (angle < 0.0 ? -1 : 1);
	angle *= sign;
	angle = 60.0 * modf(angle, &h);
	s = 60.0 * modf(angle, &m);
	if (sign > 0) {
		sprintf(hmsstring, "%02d:%02d:%07.4f", (int) h, (int) m, s);
	} else {
		sprintf(hmsstring, "-%02d:%02d:%07.4f", (int) h, (int) m, s);
	}	
        return hmsstring;
}
