#include <math.h>
#include <stdio.h>
#include <string.h>
#include "utils_globals.h"
#include "options_globals.h"
#include "model_types.h"
#include "compute_globals.h"
#include "output_globals.h"


void print_param(model_struct *model)
/*------------------------------------------------------------------------------
 * Print plate model parameters (zero points and smooth parameters) in a
 * readable format.
 * Arguments:
 *     LUN      Integer   Logical unit for output
 *     NBAND    Integer   Total number of colour bands
 *     IBAND    Integer   Colour numbers                           
 *     JOFFSET  Integer   Offset for first smooth parameter of each band
 *     PARNUM   Integer   Number of smooth plate model parameter for each colour
 *     X        Real*4    Full parameter vector
 *     COEFFS   Real*4    Colour ordered parameter vector
 *     NF       Integer   Number of fields in strip
 *     CENTERS  Real*8    Nominal field centers                     
 *     NPAR     Integer   Total number of smooth parameters
 *     COVAR    Real*4    Covariance matrix of parameters
 *     NAME     Char      Name of parameters
 *     PRINTERR Logical   Print standard errors?
 *     FUNC     External  Model term computation routine
 *--------------------------------------------------------------------------*/
{
    int		i, j, ic;
    float	x, y, err;
    double	ra, dec;
/*
 *  Print distortion coefficients
 */
    VPRINTF(VERBOSE,"Global scaling parameter %d\n", model->mpix);
    for (ic=0;ic<model->ncam;ic++) {
      VPRINTF(VERBOSE,"Distortion coefficients for camera %d\n", ic);
      VPRINTF(VERBOSE,"     (arcseconds)            (radians)\n");
      for (j=0;j<model->parnum[ic];j++) {
        err = sqrt(model->alpha[j+model->paroff[ic]][j+model->paroff[ic]]);
        VPRINTF(VERBOSE,"%8.2f %8.2f +/-%5.2f %10.4g %10.4g +/-%9.3g %s\n",
                model->params[0][j+model->paroff[ic]]/3.141592*180*3600,
                model->params[1][j+model->paroff[ic]]/3.141592*180*3600,
                err/3.141592*180*3600,
                model->params[0][j+model->paroff[ic]],
                model->params[1][j+model->paroff[ic]],
                err, model->names[j+model->paroff[ic]]);
      }
    }
/*
 *  Print corrections to field zero points and true field centers
 */
    VPRINTF(VERBOSE,"!Corrections to plate center coordinates  (arcseconds)     (radians)         and plate center coordinates (radians)\n\n");
    for (i=0;i<model->nf;i++) {
      err = sqrt(model->alpha[i+model->np][i+model->np]);
      x = (float) model->crpix[0][0][i];
      y = (float) model->crpix[1][0][i];
      ic = 0;
      compute_position(&x, &y, &i, &ic, &ra, &dec, 1, model);
      VPRINTF(VERBOSE,"%4d %7.2f %7.2f %4.2f %9.2g %9.2g %6.1g %9.7f %10.7f %6.1g\n",
              i+1, model->params[0][i+model->np]*3600.0*180.0/3.14159,
              model->params[1][i+model->np]*3600.0*180.0/3.14159,
              err*3600*180/3.141592,
              model->params[0][i+model->np], model->params[1][i+model->np], err,
              ra, dec, err);
    }
}
/**/
void print_matrices(float **a, int n, float **b, int m)
{
    int         i, j;

    for (i=0;i<n;i++) {
      for (j=0;j<n;j++) {
        VPRINTF(VERBOSE,"%g ",a[j][i]);
      }
      for (j=0;j<m;j++)
        VPRINTF(VERBOSE,"%g ",b[j][i]);
      VPRINTF(VERBOSE,"\n");
    }
    VPRINTF(VERBOSE,"\n");
}

