/* History:
   05.01.2000 (TE):
   included setting of iira 
   in parse_ccd_files. 

   22.06.2001:
   function get_refs_l:
   initialised ra and dec to 0.0
   Not doing so led to a crash in the case
   where the refernece data is taken out of
   a file and not from the pmm directory.

   19.07.01:
   removed compiler warnings under DEC

   21.04.2003:
   There were problems changing the key names
   in a standard star catalog given in a file
   The strcpy commands for 'okeynames' in function
   get_fields and for 'fkeynames' in function
   get_refs_l led to core dumps. Therefore I disabled
   the possibility to change these names for the
   moment.

   05.05.03:
   included the tsize.h file

   30.06.2003:
   function get_refs_l:
   it is now checked with the 'stat' function whether
   the given path for the standardstar catalog is a 
   directory instead of relying on a finalising '/'.

   21.05.2004:
   - included conversion routine for the UCAC2 catalog
   - resolved long -> LONG etc. issues
   - fixed a bug in the byte swapping in the compact_pmm
     subroutine

   20.07.2004:
   I removed compiler warnings due to possibly
   undefined variables (compilation with optimisation
   turned on)

   16.01.2005:
   I fixed a bug of byte swapping in parse_ccd_files.

   29.01.2005:
   I removed compiler warnings due to the possible use
   of uninitialised variables.

   14.05.2005:
   I adapted the output format in several printf commands
   to '%d' if LONG is in fact an int variable.

   27.03.2006:
   I renamed sorting and indexing functions to the new naming scheme
   described in the sort_globals.h file

   01.05.2006:
   I removed the inclusion of precess_defs.h and precess_globals.h 
   (superfluous)

   06.05.2006:
   I removed superfluous numbers in the debug output (UCAC catalogue 
   debug option)
*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include "fitscat.h"
#include "tsize.h"
#include "options_globals.h"
#include "sort_globals.h"
#include "utils_globals.h"
#include "bin_search_globals.h"
#include "tabutil_globals.h"
#include "posixstat.h"
#include "refs_io_defs.h"
#include "refs_io_types.h"
#include "refs_io_globals.h"

unsigned char *compact_pmm(ULONG sra, ULONG sdec,
                  char s,char q, LONG bbb, LONG rrr)
/****** compact_pmm ***************************************************
 *
 *  NAME	
 *      compact_pmm - Convert a PMM object to compact format
 *
 *  SYNOPSIS
 *      unsigned char *compact_pmm(unsigned long int sra,
 *                                 unsigned long int sdec,
 *                                 char s,char q,long int bbb,long int rrr)
 *
 *
 *  FUNCTION
 *      This routine compresses the PMM object information to a 7 byte
 *      word. It does this by striping the high value Ra and Dec data
 *      to represent only 2^19 byte integers. This will retain the
 *      original position resolution when one splits the sky into 248x124
 *      tiles (each roughly 1.45 degrees wide). The s and q values are 
 *      stored as single bits, and the magnitudes are stored in a single
 *      byte each. The extranous values in magnitude are set to 0.
 *      The 65 bits layout is as followis:
 *      12345678901234567890123456789012345678901234567890123456
 *      RRRRRRRRRRRRRRRRDDDDDDDDDDDDDDDDRRRDDDsqbbbbbbbbrrrrrrrr
 *      lower part      lower part      up up     bmag     rmag
 *        Ra               Dec        part part
 *
 *  INPUTS
 *      sra  - Tiled value of Ra
 *      sdec - Tiled value of Dec
 *      s    - Sign flag for GSC counterpart
 *      q    - Quality flag
 *      bbb  - Scaled B Magnitude
 *      rrr  - Scaled R Magnitude
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      A newly allocated 7 byte character array with the compressed
 *      data stored in it.
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 03-04-1998
 *
 ******************************************************************************/
{
    unsigned char	*data, *rptr, *dptr;
    int		i;
#ifdef BSWAP
    unsigned char t;
#endif

    ED_CALLOC(data, "compact_pmm", unsigned char, 7);
    ED_CALLOC(rptr, "compact_pmm", unsigned char, 4);
    ED_CALLOC(dptr, "compact_pmm", unsigned char, 4);
    memcpy(rptr, &sra,  4);
    memcpy(dptr, &sdec, 4);
#ifdef BSWAP
    t = rptr[0];
    rptr[0] = rptr[3];
    rptr[3] = t;
    t = rptr[1];
    rptr[1] = rptr[2];
    rptr[2] = t;
    t = dptr[0];
    dptr[0] = dptr[3];
    dptr[3] = t;
    t = dptr[1];
    dptr[1] = dptr[2];
    dptr[2] = t;
#endif
/*
 *  Copy the least significant byte elements to the compressed data array.
 */
    for (i=0;i<2;i++) {
       data[i]   = rptr[3-i];
       data[i+2] = dptr[3-i];
    }
/*
 *  Now take only the least significant three bits from the third byte
 *  And store them in the fifth compressed data array.
 */
    data[4]  = ((rptr[1] << 5) >> 5);
    data[4] += ((dptr[1] << 5) >> 2);
/*
 *  Put the GSC flag and the quality flag in
 */
    data[4] += (s << 6);
    data[4] += (q << 7);
/*
 *  Save the magnitudes in the range 0 -> 255 and set anything outside this
 *  range to 0
 */
    data[5] = (unsigned char) (bbb > 256? 256: bbb);
    data[6] = (unsigned char) (rrr > 256? 256: rrr);

    ED_FREE(rptr);
    ED_FREE(dptr);

    return data;
}
/*L*/

void parse_ccd_files(char *file, int debug)
/****** parse_ccd_files ***************************************************
 *
 *  NAME	
 *      parse_ccd_files - Read and convert a USNO-A1.0 zone file
 *
 *  SYNOPSIS
 *      void parse_ccd_files(char *file)
 *
 *  FUNCTION
 *      Convert each element from the specified USNO-A1.0 zone file
 *      to compressed format and append the data to the new or existing
 *      tile files. The naming convention for the tile files is the
 *      row and column number if the tile on the sky. Row number one starts
 *      at Ra = 0.0, column number one starts at Dec = -90.0. Filenames
 *      are six characters wide, three characters (zero padded) for the
 *      row number and likewise form the column number. Tile filename
 *      postfix is .pmm
 *
 *  INPUTS
 *      file - Name of the USNO-A1.0 zone file
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *      The tile files for which objects have been read from the zone file
 *      will have been updated with new objects.
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 03-04-1998
 *
 ******************************************************************************/

{
    FILE		*fd, *lfd;
    char		*ptr, tile[MAXCHAR];
    int			size = 12;
    unsigned char	*data;
    char		s , q;
    LONG		ira, idec, iira, iidec, icod, fff, bbb, rrr;
    ULONG               sra,sdec;
#ifdef BSWAP
    unsigned char t;
#endif

    ED_FOPEN(fd, "parse_ccd_files", file, "r");
    ED_CALLOC(ptr, "parse_ccd_files", char, size);
    while (fread(ptr, 1, size, fd) == (size_t)size) {
#ifdef BSWAP
       t = ptr[0];
       ptr[0] = ptr[3];
       ptr[3] = t;
       t = ptr[1];
       ptr[1] = ptr[2];
       ptr[2] = t;
#endif
       memcpy(&ira, &ptr[0], 4);
       iira = (ira) / TILE_RES + 1;
       sra  = (ULONG) ira % TILE_RES;
#ifdef BSWAP
       t = ptr[4];
       ptr[4] = ptr[7];
       ptr[7] = t;
       t = ptr[5];
       ptr[5] = ptr[6];
       ptr[6] = t;
#endif
       memcpy(&idec, &ptr[4], 4);
       iidec= (idec ) / TILE_RES+ 1;
       sdec = (ULONG) idec % TILE_RES;
#ifdef BSWAP
       t = ptr[8];
       ptr[8] = ptr[11];
       ptr[11] = t;
       t = ptr[9];
       ptr[9] = ptr[10];
       ptr[10] = t;
#endif
       memcpy(&icod, &ptr[8], 4);
       s = 0;
       if (icod < 0) {
          icod = -icod;
          s = 1;
       }
       q = icod / 1000000000;
       fff = icod / 1000000 - q*1000;
       bbb = icod / 1000 - q*1000000 - fff*1000;
       rrr = icod - q*1000000000 - fff*1000000 - bbb*1000;
       data = compact_pmm(sra, sdec, s, q, bbb, rrr);
       if (debug) {
          fprintf(stderr, "Ra: %f, Dec: %f BMag: %f, RMag: %f\n", 
                  ((double) ira)/5400000.0,
                  ((double) idec)/360000-90.0,
                  ((float)bbb)/10.0, ((float)rrr)/10.0);
       }
#ifdef INT64       
       sprintf(tile, "%.3d%.3d.pmm", iira, iidec);
#else
       sprintf(tile, "%.3ld%.3ld.pmm", iira, iidec);
#endif
       if (!debug) {
         ED_FOPEN(lfd, "parse_ccd_files", tile, "a");
         ED_FWRITE(data, "parse_ccd_files", lfd, 7);
         ED_FCLOSE(lfd, "parse_ccd_files");
       }
       ED_FREE(data);
    }
}

/*L*/

void parse_ucac2_files(char *file, int debug)
/****** parse_ucac2_files ***************************************************
 *
 *  NAME	
 *      parse_ucac2_files - Read and convert a UCAC2 zone file
 *
 *  SYNOPSIS
 *      void parse_ucac2_files(char *file)
 *
 *  FUNCTION
 *      Convert each element from the specified UCAC2 zone file
 *      to compressed format and append the data to the new or existing
 *      tile files. The naming convention for the tile files is the
 *      row and column number if the tile on the sky. Row number one starts
 *      at Ra = 0.0, column number one starts at Dec = -90.0. Filenames
 *      are six characters wide, three characters (zero padded) for the
 *      row number and likewise form the column number. Tile filename
 *      postfix is .pmm
 *
 *      To use the same compression algorithm as for the USNO-A2 catalog,
 *      we first use the U2Rmag for the R AND the B magnitude. We convert
 *      UCAC2 Ra and Dec (given in mas units) to 10mas units by just
 *      dividing by 10 and loosing one digit of precision. 
 *
 *  INPUTS
 *      file - Name of the UCAC2 zone file
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *      The tile files for which objects have been read from the zone file
 *      will have been updated with new objects.
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      T. Erben - dd 20-05-2004
 *
 ******************************************************************************/

{
    FILE		*fd, *lfd;
    char		*ptr, tile[MAXCHAR];
    int			size = 44;
    unsigned char	*data;
    char		s , q;
    LONG		ira, idec, iira, iidec, 
                        bbb, rrr;
    ULONG               sra,sdec;
    float               fdec, fra, fmag;
    short               tmp;
#ifndef BSWAP
    unsigned char t;
#endif

    ED_FOPEN(fd, "parse_ucac2_files", file, "r");
    ED_CALLOC(ptr, "parse_ucac2_files", char, size);
    while (fread(ptr, 1, size, fd) == (size_t)size) {
#ifndef BSWAP
       t = ptr[0];
       ptr[0] = ptr[3];
       ptr[3] = t;
       t = ptr[1];
       ptr[1] = ptr[2];
       ptr[2] = t;
#endif
       memcpy(&ira, &ptr[0], 4);
       fra = (float)ira;
       /* convert Ra from 10 mas to
          the mas unit
       */
       fra = fra/15./3600./1000.;
       fra = fra*15.*3600.*100.;
       ira = (LONG)(fra+0.5);
       iira = (ira) / TILE_RES + 1;
       sra  = (ULONG) ira % TILE_RES;
#ifndef BSWAP
       t = ptr[4];
       ptr[4] = ptr[7];
       ptr[7] = t;
       t = ptr[5];
       ptr[5] = ptr[6];
       ptr[6] = t;
#endif
       memcpy(&idec, &ptr[4], 4);
       fdec = (float)idec;
       /* convert Dec from 10 mas to
          the mas unit and make it
          positive definite.
       */
       fdec = fdec/1000./3600.;
       fdec = (fdec+90.)*3600.*100.;
       idec = (LONG)(fdec+0.5);
       iidec= (idec ) / TILE_RES+ 1;
       sdec = (ULONG) idec % TILE_RES;
#ifndef BSWAP
       t = ptr[8];
       ptr[8] = ptr[9];
       ptr[9] = t;
#endif
       memcpy(&tmp, &ptr[8], 2);
       fmag = (float)tmp;
       /* convert mag from 0.01mag 
          to 0.1mag unit
       */
       fmag = fmag/100.;
       fmag = fmag*10.;
       /* R and B magnitude are the same in the
          UCAC2 case (their U2Rmag)
       */
       rrr = (LONG)(fmag + 0.5);
       bbb = rrr;
       s = 0;
       q = 1;
       data = compact_pmm(sra, sdec, s, q, bbb, rrr);
       if (debug) {
          fprintf(stderr, "Ra: %f, Dec: %f BMag: %f, RMag: %f\n", 
                  ((double) ira)/5400000.0,
                  ((double) idec)/360000-90.0,
                  ((float)bbb)/10.0, ((float)rrr)/10.0);
       }
#ifdef INT64
       sprintf(tile, "%.3d%.3d.pmm", iira, iidec);
#else
       sprintf(tile, "%.3ld%.3ld.pmm", iira, iidec);
#endif
       if (!debug) {
         ED_FOPEN(lfd, "parse_ucac2_files", tile, "a");
         ED_FWRITE(data, "parse_ucac2_files", lfd, 7);
         ED_FCLOSE(lfd, "parse_ucac2_files");
       }
       ED_FREE(data);
    }
}
/*L*/
void uncompact_pmm(unsigned char *data, 
                   ULONG *sra, ULONG *sdec,
                   char *s,char *q, LONG *bbb, LONG *rrr)
/****** uncompact_pmm ***************************************************
 *
 *  NAME	
 *      uncompact_pmm - Decompress the compressed PMM information
 *
 *  SYNOPSIS
 *      void uncompact_pmm(unsigned char *data,
 *                         unsigned long int *sra,unsigned long int *sdec,
 *                         char *s,char *q,long int *bbb,long int *rrr)
 *
 *  FUNCTION
 *      This routine decompresses the compression done by the compact_pmm
 *      routine. For details on the compression method read the comment
 *      section of compact_pmm.
 *
 *  INPUTS
 *      data - The 7 byte compressed data character array.
 *
 *  RESULTS
 *      sra  - Tiled value of Ra
 *      sdec - Tiled value of Dec
 *      s    - Sign flag for GSC counterpart
 *      q    - Quality flag
 *      bbb  - Scaled B Magnitude
 *      rrr  - Scaled R Magnitude
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 03-04-1998
 *
 ******************************************************************************/
{
    unsigned char	*rptr, *dptr, t;
    int		i;

    rptr = (unsigned char *)sra;
    dptr = (unsigned char *)sdec;
    for (i=0;i<2;i++) {
       rptr[3-i] = data[i];
       dptr[3-i] = data[i+2];
    }
    t = (data[4] << 5);
    rptr[1] = t >> 5;
    t = (data[4] << 2);
    dptr[1] = t >> 5;
#ifdef BSWAP
    t = rptr[0];
    rptr[0] = rptr[3];
    rptr[3] = t;
    t = rptr[1];
    rptr[1] = rptr[2];
    rptr[2] = t;
    t = dptr[0];
    dptr[0] = dptr[3];
    dptr[3] = t;
    t = dptr[1];
    dptr[1] = dptr[2];
    dptr[2] = t;
#endif
    t = (data[4] << 1);
    *s= t >> 7;
    *q      = data[4] >> 7;
    *bbb    = data[5];
    *rrr    = data[6];
    if (*bbb == 0) *bbb = 999;
    if (*rrr == 0) *rrr = 999;
}

/*L*/
static ref_objs *init_ref_objs()
/****i* init_ref_objs ***************************************************
 *
 *  NAME	
 *      init_ref_objs - Initialize the ref_objs structure
 *
 *  SYNOPSIS
 *      ref_objs *init_ref_objs()
 *
 *  FUNCTION
 *      This routine allocates the required memory to start storing
 *      reference catalog objects in. It initializes the arrays to a
 *      given length and sets the object counter to 0.
 *
 *  INPUTS
 *      None
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      A pointer to the newly allocated structure
 *
 *  MODIFIED GLOBALS
 *      REF_OBJ_BLKSIZE
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 03-04-1998
 *
 ******************************************************************************/
{
    ref_objs	*t;

    ED_CALLOC(t, "init_ref_objs", ref_objs, 1);
    t->nblk = 1;
    t->blksize = REF_OBJ_BLKSIZE;
    ED_CALLOC(t->ra,   "init_ref_objs", double, t->nblk * t->blksize);
    ED_CALLOC(t->dec,  "init_ref_objs", double, t->nblk * t->blksize);
    ED_CALLOC(t->s,    "init_ref_objs", char,   t->nblk * t->blksize);
    ED_CALLOC(t->q,    "init_ref_objs", char,   t->nblk * t->blksize);
    ED_CALLOC(t->bmag, "init_ref_objs", float,  t->nblk * t->blksize);
    ED_CALLOC(t->rmag, "init_ref_objs", float,  t->nblk * t->blksize);
    return t;
}
/*L*/
static void enlarge_ref_obj(ref_objs *t)
/****i* enlarge_ref_obj ***************************************************
 *
 *  NAME	
 *      enlarge_ref_obj - Allocate more memory for the ref_objs structure
 *
 *  SYNOPSIS
 *       void enlarge_ref_obj(ref_objs *t)
 *
 *  FUNCTION
 *       Add new memory segments to each of the data arrays in the
 *       ref_objs structure.
 *
 *  INPUTS
 *       t  - A pointer to the ref_objs structure to be enlarged
 *
 *  RESULTS
 *       t  - A pointer to the ref_objs structure to be enlarged
 *
 *  RETURNS
 *       None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 03-04-1998
 *
 ******************************************************************************/
{
    t->nblk++;
    ED_REALLOC(t->ra,   "init_ref_objs", double, t->nblk * t->blksize);
    ED_REALLOC(t->dec,  "init_ref_objs", double, t->nblk * t->blksize);
    ED_REALLOC(t->s,    "init_ref_objs", char,   t->nblk * t->blksize);
    ED_REALLOC(t->q,    "init_ref_objs", char,   t->nblk * t->blksize);
    ED_REALLOC(t->bmag, "init_ref_objs", float,  t->nblk * t->blksize);
    ED_REALLOC(t->rmag, "init_ref_objs", float,  t->nblk * t->blksize);
}
/*L*/
void free_ref_obj(ref_objs *t)
/****** free_ref_obj ***************************************************
 *
 *  NAME	
 *      free_ref_obj - Deacclocate the memory for the given ref_objs structure
 *
 *  SYNOPSIS
 *      void free_ref_obj(ref_objs *t)
 *
 *  FUNCTION
 *      Remove all memory requirements for the given ref_objs structure by
 *      freeing all array pointers and the structure itself.
 *
 *  INPUTS
 *      t - A pointer to the ref_objs structure
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *      Heap memory is freed.
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. DEul - dd 03-04-1998
 *
 ******************************************************************************/
{
    ED_FREE(t->ra);
    ED_FREE(t->dec);
    ED_FREE(t->s);
    ED_FREE(t->q);
    ED_FREE(t->bmag);
    ED_FREE(t->rmag);
    ED_FREE(t);
}
/*L*/
static void add_ref_obj(ref_objs *t, double ra, double dec, char s, char q,
                        float bmag, float rmag)
/****i* add_ref_obj ***************************************************
 *
 *  NAME	
 *      add_ref_obj - Add a new object to the ref_objs structure
 *
 *  SYNOPSIS
 *      void add_ref_obj(ref_objs *t, double ra, double dec, char s, char q,
 *                       float bmag, float rmag)
 *
 *  FUNCTION
 *      Add an element to each of the arrays in the structure and increment
 *      the object counter. Make sure that enough memory is allocated to
 *      append the current information. In not, enlarge it.
 *
 *  INPUTS
 *      t    - To be updated ref_objs structure
 *      ra   - Right Ascension of the object to be added
 *      dec  - Declination of the object to be added
 *      s    - GSC signal of the object to be added
 *      q    - Magnitude quality of the object to be added
 *      bmag - B Magnitude of the object to be added
 *      rmag - R Magnitude of the object to be added
 *
 *  RESULTS
 *      t    - Updated ref_objs structure
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 03-04-1998
 *
 ******************************************************************************/
{
    if (t->nobj == t->nblk * t->blksize - 1) enlarge_ref_obj(t);
    t->ra[t->nobj]   = ra; 
    t->dec[t->nobj]  = dec; 
    t->s[t->nobj]    = s; 
    t->q[t->nobj]    = q; 
    t->bmag[t->nobj] = bmag; 
    t->rmag[t->nobj] = rmag; 
    t->nobj++;
}

static int icat_first = TRUE;
static catstruct	*icat;
static tabstruct	*iftab, *itab;

#define ONKEYS 3
#define FNKEYS 8

static char    *okeynames[]={"Ra\0\0\0\0\0\0", "Dec\0\0\0\0\0\0", 
                             "Mag\0\0\0\0\0"};
static char    *fkeynames[]={"CRVAL1", "CRVAL2", "CDELT1", "CDELT2",
			     "NAXIS1\0\0", "NAXIS2\0\0", 
                             "OBJECT_POS", "OBJECT_COUNT"};
static double  *crval1, *crval2, *dx, *dy, *wx, *wy;
static int     nfields, *cridx2;
static float   tol;

/* 
 * get_fields - go through the field contents, call frame,
 * store the stars which are on the frame, return count of stored stars
 */

static void get_fields(ref_objs *t,
                       double minr, double maxr, double mind, double maxd)
{
    float	*rad=NULL, *decd=NULL, *mag=NULL;
    keystruct	**okeys;
    double	ra_tol, rra, rdec, ra, dec;
    int		i, jj, jjbeg, jjend, num;

    VPRINTF(DEBUG,"Tol=%f\n",tol);
    tol = tol * 1.01;
    ra  = (maxr+minr)/2.0;
    dec = (maxd+mind)/2.0;
    /*strcpy(okeynames[0], control.ref_ra);
    strcpy(okeynames[1], control.ref_dec);
    strcpy(okeynames[2], control.ref_mag);*/
    if (nfields > 1) {
       dbin_search_indexed(nfields,crval2,cridx2,(float)dec,tol,&jjbeg,&jjend,0);
    } else {
       jjbeg = jjend = 0;
    }
    for (jj=jjbeg;jj<=jjend;jj++) 
    {
      ra_tol = wx[cridx2[jj]] * fabs(dx[cridx2[jj]]);
      if ((crval1[cridx2[jj]]-ra_tol) < ra &&
        ra < (crval1[cridx2[jj]]+ra_tol)) {
        ED_MALLOC(okeys, "get_fields", keystruct *, ONKEYS);
        num = read_field(itab, okeynames, okeys, ONKEYS, cridx2[jj]+1, iftab);
        okeys[0]->nobj = num;
        okeys[1]->nobj = num;
        switch (okeys[0]->ttype) {
        case T_FLOAT:
            ED_GETVEC(rad,  "get_fields",okeys[0], okeynames[0], float, float);
            ED_GETVEC(decd, "get_fields",okeys[1], okeynames[1], float, float);
            break;
        case T_DOUBLE:
            ED_GETVEC(rad,  "get_fields",okeys[0], okeynames[0], float, double);
            ED_GETVEC(decd, "get_fields",okeys[1], okeynames[1], float, double);
            break;
        default: syserr(FATAL, "get_fields", "unsupported key type for Ra and/or Dec.");
            break;
        }
        mag    = okeys[2]->ptr;
        for (i=0;i<num;i++) 
        {
          if (minr <= rad[i]  && rad[i]  <= maxr &&
              mind <= decd[i] && decd[i] <= maxd) {
             rra  = (double) rad[i];
             rdec = (double) decd[i];
             add_ref_obj(t, rra, rdec, 'X', 'X', mag[i], mag[i]);
          }
        }
        ED_FREE(rad);
        ED_FREE(decd);
        ED_FREE(mag);
        ED_FREE(okeys);
      }
    }
}
/*L*/
static void get_refs_l(char *dir, ref_objs *t, 
                       double minr, double maxr, double mind, double maxd)
/****i* get_refs_l ***************************************************
 *
 *  NAME	
 *      get_refs_l - Get all PMM objects inside (minr,mind) - (maxr,maxd) box
 *
 *  SYNOPSIS
 *      void get_refs_l(char *dir, ref_objs *t,
 *                     double minr, double maxr, double mind, double maxd)
 *
 *  FUNCTION
 *      Read from the tile files all objects that are within the boundaries
 *      of the box confined by the corners (minr, mind) and (maxr, maxd).
 *      The tile files are to be found in the dir directory. The retrieved
 *      objects are stired in the ref_objs structure
 *
 *  INPUTS
 *      dir  - The path to the directory containing all tile files
 *      t    - A pointer to the ref_objs structure to be updated
 *      minr - Minimum Right Ascension of bounding box
 *      maxr - Maximim Right Ascension of bounding box
 *      mind - Minimum Declination of bounding box
 *      maxd - Maximum Declination of bounding box
 *
 *  RESULTS
 *      t   - A pointer to the updated ref_objs structure 
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 03-04-1998
 *
 ******************************************************************************/

{
    FILE		*lfd;
    char		tile[MAXCHAR];
    double		ra=0.0, dec=0.0;
    unsigned char	*data;
    float		b, r;
    char		s , q;
    LONG		bbb, rrr;
    int			min_r, max_r, min_d, max_d, i, j;
    int			MIN_R, MAX_R, MIN_D, MAX_D;
    ULONG               sra,sdec;
    int			PMM = FALSE;
    struct stat         finfo;
    keystruct   	**fkeys;
/*
 *  See if we have to read from the PMM dir of form a file
 */
    if(stat (dir, &finfo) < 0)
    {
      syserr(FATAL, "get_refs_l", "An error occured while checking file/directory: %s\n", dir);
    }
    else
    {
       if (S_ISDIR (finfo.st_mode)) {
          PMM = TRUE;
       } else {
          PMM = FALSE;
       }
    }
    if (PMM) {
/*
 *    Limit the range of the box to the available tile range: (1,1) - (248,124)
 */
      MIN_R = 1; 
      MIN_D = 1; 
      MAX_R = (360.0 * DEG_BINNING) / TILE_RES + 1;
      MAX_D = (180.0 * DEG_BINNING) / TILE_RES + 1;
/*
 *    Get the limiting tile numbers for row and column
 */
      min_r = MAX(MIN_R,MIN(MAX_R,(minr * DEG_BINNING) / TILE_RES + 1));
      min_d = MAX(MIN_D,MIN(MAX_D,((mind+90.0)*DEG_BINNING)/TILE_RES + 1));
      max_r = MIN(MAX_R,MAX(MIN_R,(maxr * DEG_BINNING) / TILE_RES + 1));
      max_d = MIN(MAX_D,MAX(MIN_D,((maxd+90.0)*DEG_BINNING)/TILE_RES + 1));
/*
 *    Step through all tiles that overlap the box and get the objects that
 *    fall precisely inside the box
 */
      for (i=min_r;i<=max_r;i++) {
         for (j=min_d;j<=max_d;j++) {
            sprintf(tile, "%s/%.3d%.3d.pmm", dir,i,j);
            ED_CALLOC(data, "get_pmm_l", unsigned char, 7);
            ED_FOPEN(lfd, "get_pmm_l", tile, "r");
            while (fread(data, sizeof(char), 7, lfd) == 7) {
               sra = sdec = 0;
               uncompact_pmm(data, &sra, &sdec, &s, &q, &bbb, &rrr);
               ra  = ((double) (sra +  (i-1)*TILE_RES))/ DEG_BINNING;
               dec = ((double) (sdec + (j-1)*TILE_RES))/ DEG_BINNING - 90.0;
               if (minr <= ra  && ra  <= maxr &&
                   mind <= dec && dec <= maxd) {
                  b = (float)bbb/10.0;
                  r = (float)rrr/10.0;
                  add_ref_obj(t, ra, dec, s, q, b, r);
               }
            }
            ED_FCLOSE(lfd, "get_pmm_l");
            ED_FREE(data);
         }
      }
    } else {
      if (icat_first) {
        ED_MALLOC(fkeys, "icat_query", keystruct *, FNKEYS);
        if ((icat = read_cat(dir)) == NULL) {
           syserr(FATAL, "icat_query", "Could not read catalog: %s\n",dir);
        }
        itab = name_to_tab(icat, "OBJECTS", 0);
        iftab = name_to_tab(icat, "FIELDS", 0);
        /*strcpy(fkeynames[4], control.ref_xwid);
	  strcpy(fkeynames[5], control.ref_ywid);*/
        read_keys(iftab, fkeynames, fkeys, FNKEYS, NULL);
        crval1 = fkeys[0]->ptr;
        crval2 = fkeys[1]->ptr;
        dx     = fkeys[2]->ptr;
        dy     = fkeys[3]->ptr;
        switch (fkeys[4]->ttype) {
        case T_LONG:
            ED_GETVEC(wx, "read_pairing",fkeys[4], fkeynames[4], double, int);
            ED_GETVEC(wy, "read_pairing",fkeys[5], fkeynames[5], double, int);
            break;
        case T_FLOAT:
            ED_GETVEC(wx, "read_pairing",fkeys[4], fkeynames[4], double, float);
            ED_GETVEC(wy, "read_pairing",fkeys[5], fkeynames[5], double, float);
            break;
        case T_DOUBLE:
            ED_GETVEC(wx, "read_pairing",fkeys[4], fkeynames[4], double,double);
            ED_GETVEC(wy, "read_pairing",fkeys[5], fkeynames[5], double,double);
            break;
        default: syserr(FATAL, "get_refs_l", "unsupported key type for crval and/or cdelt");
            break;
        }
        nfields = fkeys[0]->nobj;
        ED_FREE(fkeys);
        ED_CALLOC(cridx2, "main", int, nfields);
        tol = 0.0;
        for (i=0;i<nfields;i++) {
           tol = MAX(tol,wy[i]*fabs(dy[i]));
        }
        if (nfields > 1)
           dindex(nfields, crval2, cridx2);
        icat_first = FALSE;
      }
/*
 *    Dirty trick for the very unprobable singularities
 */
      if ( dec == 0.0 ) dec = -0.0000001;
      if ( dec == -90.0 ) dec = -89.9999999;
      if ( dec == 90.0 ) dec = 89.9999999;
      get_fields(t, minr, maxr, mind, maxd);
    }
}
/*L*/
ref_objs *get_refs(char *dir, double rac, double decc, float ras, float decs)
/****** get_pmm ***************************************************
 *
 *  NAME	
 *      get_refs - Get all PMM objects inside box (ras,decd) wide at (rac,dec)
 *
 *  SYNOPSIS
 *      ref_objs *get_refs (char *dir, double rac, double decc, 
 *                        float ras, float decs)
 *
 *  FUNCTION
 *      Get all the objects from the PMM tile files that fall inside the
 *      box with its center at (rac,decc) and which has sides of (ras, decs).
 *      Note that the box is square in Ra and Dec, no provision is taken
 *      for projections. The tile files reside in the dir directory, and
 *      the user is given a filled ref_objs structure.
 *
 *  INPUTS
 *      dir  - The directory path were the tile files reside
 *      rac  - Center Right Ascension of the box
 *      decc - Center Declination of the box
 *      ras  - Right Ascension width of the box
 *      decs - Declination width of the box
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      A pointer to the filled and allocated ref_objs structure
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 03-04-1998
 *
 ******************************************************************************/
{
    double	minra, maxra, mindec, maxdec;
    ref_objs	*refs;
/*
 *  Derive box corners
 */
    minra = rac - ras/2.0;
    maxra = rac + ras/2.0;
    mindec = decc - decs/2.0;
    maxdec = decc + decs/2.0;
    refs = init_ref_objs();
/*
 *  Check several conditions for the Ra = 0.0 or Ra = 360.0 line
 */
    if (minra < 0.0 && maxra < 0.0) {
/*
 *     Box is completely specified in negative Right Ascension, make is
 *     positive
 */
       minra += 360.0;
       maxra += 360.0;
    }
    if (minra < 0.0 && maxra >= 0.0) {
/*
 *     Box overlaps the Ra = 0.0 line. We must do two seperate requests
 */
       get_refs_l(dir, refs, minra+360.0, 360.0, mindec, maxdec);
       get_refs_l(dir, refs, 0.0, maxra, mindec, maxdec);
    } else {
       if (minra <= 360.0 && maxra > 360.0) {
/*
 *        Box overlaps the Ra = 360.0 line. We must do two seperate requests
 */
          get_refs_l(dir, refs, minra, 360.0, mindec, maxdec);
          get_refs_l(dir, refs, 0, maxra-360.0, mindec, maxdec);
       } else {
/*
 *        Normal request inside 0 <= Ra <= 360
 */
          get_refs_l(dir, refs, minra, maxra, mindec, maxdec);
       }
    }
    return refs;
}

void free_refs(ref_objs *r)
{
    ED_FREE(r->ra);
    ED_FREE(r->dec);
    ED_FREE(r->s);
    ED_FREE(r->q);
    ED_FREE(r->bmag);
    ED_FREE(r->rmag);
    ED_FREE(r);
}
