/* 14.07.00:
   - I changed the criteria to calculate the threshholds
     in the astrom iterations (routine calc_thresholds):
     MAX(3*var,max/2) -> MIN(3*var,0.9*max).
     This turned out to do a much better job when very high
     outliers are present.

   05.08:
   The cstring var. in print_threshholds is now allocated
   dynamically. Not doing this caused a crash when the string
   was too short (AIX).

   I changed back the limits I set on 14.07.
   The very high errors that made me change this do no longer
   occur after I fixed a bug in init_pairs on 04.08.

   28.08.:
   calculation of threshholds:
   for the overlap and intercam threshholds we calculate now
   variance and maximum only if there are more than two elements.

   04.09.00:
   I again changed back the threshh limits to the values of 14.07.
   I am still not satisfied with the rejection.   

   19.02.02:
   corrected function get_thresholds. The way the tmp array was filled
   before corrupted this pointer (by increasing it in the for loop) 
   and hence the return value was wrong

   20.07.2004:
   I removed compiler warnings due to possibly
   undefined variables (compilation with optimisation
   turned on)
*/

#include <math.h>      
#include <stdio.h>      
#include <stdlib.h>      
#include <string.h>      
#include "project_defs.h"
#include "options_globals.h"
#include "utils_globals.h"
#include "thresh_globals.h"
#include "astrom_globals.h"

static int	numcam;

int get_thresh_index(int c1, int c2, int f1, int f2)
{
    int		i, i1, i2;

    i1 = MIN(c1, c2);
    i2 = MAX(c1, c2);
    if (i1 < 0) {
      i = i2;
    } else {
      if (f1 == f2) {
        i = numcam+c2*numcam+c1;
      } else {
        i = numcam*numcam + numcam+c2*numcam+c1;
      }
    }
    return i;
}
/**/
thresh_struct *init_thresholds(int ncam)
/* ----------------------------------------------------------------------
 * DENIS astrometric package:
 * Initialize the outlier rejection thresholds to large values
 * residual statistics.
 * Arguments:
 *  NBAND      Integer    Number of colour bands                    Input
 *  THRESH     Real       Current outlier rejection threshold       Output
 *-----------------------------------------------------------------------
 */
{
    int		i;
    float	val = 4 * DPI;
    float	*ptr;
    thresh_struct	*t;
 
    ED_CALLOC(t,             "init_thresholds", thresh_struct, 1);
    numcam       = ncam;
    t->numthresh = ncam + ncam*ncam + ncam*ncam;
    ED_CALLOC(t->thresholds[0], "init_thresholds", float, t->numthresh);
    ED_CALLOC(t->thresholds[1], "init_thresholds", float, t->numthresh);
    for (i=t->numthresh, ptr=t->thresholds[0];i--;) *(ptr++) = val;
    for (i=t->numthresh, ptr=t->thresholds[1];i--;) *(ptr++) = val;
    t->icurr = 0; t->iprev = 1;
    return t;
}
/**/
void free_thresholds(thresh_struct *t)
{
    ED_FREE(t->thresholds[0]);
    ED_FREE(t->thresholds[1]);
    ED_FREE(t);
}
/**/
float threshold_reduction(thresh_struct *t)
/* --------------------------------------------------------------------------
 * DENIS astrometric package:
 * Determine the maximum relative reduction of all rejection thresholds 
 * from previous iteration
 * Arguments:
 *--------------------------------------------------------------------------
 */
{
    float	red, retval = 1.0, *t_prev, *t_curr;
    int		i;

    t_prev = t->thresholds[t->iprev];
    t_curr = t->thresholds[t->icurr];
    for (i=t->numthresh;i--;) {
      if (*(t_curr) == 0.0) {
        red = *(t_curr);
      } else {
        red = *(t_prev)/(*(t_curr));
      }
      t_prev++; t_curr++;
      retval = MAX(retval, red);
    }
    return retval;
}
/**/
float get_threshold(thresh_struct *t, int ic1, int ic2, int f1, int f2)
{
    return (t->thresholds[t->icurr][get_thresh_index(ic1, ic2, f1, f2)]);
}
/**/
float *get_thresholds(thresh_struct *t)
{
    float	*tmp, *tt;
    int		i;
 
    ED_CALLOC(tmp, "get_thresholds", float, t->numthresh);

    for (i=t->numthresh, tt=t->thresholds[t->icurr];i--;)
    {
      tmp[i] = tt[i];
    }

    return tmp;
}
void put_thresholds(thresh_struct *t, float *thresh)
{
    int		i;
    float	*tt;

    for (i=t->numthresh, tt=t->thresholds[t->icurr];i--;) 
    {
      *tt++ = *thresh++;
    }
}
/**/
void print_thresholds(thresh_struct *t, int check)
{
    int		i;
    char	*cstring=NULL;

    if(t->numthresh>0)
    {
      ED_CALLOC(cstring, "print_thresholds", char, t->numthresh*MAXCHAR);
    }

    if (! check) 
    {
      VPRINTF(NORMAL,"No outlier rejection will be used.\n");
    } 
    else 
    {
      for (i=0,strcpy(cstring,"");i<t->numthresh;i++) 
      {
        sprintf(cstring, "%s %f", cstring, 
                t->thresholds[t->icurr][i]/DPI*180.0*3600.0);
      }
      VPRINTF(NORMAL,"Outlier rejection above %s\n", cstring);
    }

    if(t->numthresh>0)
    {
      ED_FREE(cstring);
    }
}
/**/
void init_thresh_stat(thresh_struct *t)
{
    if (!t)
    {
      syserr(FATAL, "init_thresh_stat", "Threshold package not initialized\n");
    }
    ED_CALLOC(t->n,   "init_thresh_stat", int,    t->numthresh);
    ED_CALLOC(t->var, "init_thresh_stat", double, t->numthresh);
    ED_CALLOC(t->max, "init_thresh_stat", double, t->numthresh);
}
void finish_thresh_stat(thresh_struct *t)
{
    ED_FREE(t->n);
    ED_FREE(t->var);
    ED_FREE(t->max);
}
/**/
void save_thresh_stats(thresh_struct *t, float d, 
                       int c1, int c2, int f1, int f2)
{
    int		ic;
/*
 *  Save the camera based statistics
 */
    ic = get_thresh_index(c1, c2, f1, f2);
    t->var[ic] += d; 
    t->max[ic]  = MAX(t->max[ic],d);
    t->n[ic]++;
    return;
}
/**/
void calc_thresholds(thresh_struct *t)
/*
 * --------------------------------------------------------------------------
 *  DENIS astrometric package:
 *  Tries to determine reasonable outlier rejection thresholds from crude
 *  residual statistics.
 *  Arguments:
 *    THRESHOLDS Real        Outlier rejection thresholds              Input
 *    NBAND      Integer     Number of colour bands                    Input
 *    REF_RESI   Real        Residuals (rms and max) to ref stars      Output
 *    OVER_RESI  Real            "      to overlap stars               Output
 *    INTER_RESI Real            "      to intercolour stars           Output
 *--------------------------------------------------------------------------
 */
{
    int		ic, ic1, ic2, i;
    float	max_thresh;
    double	var, max;
/*
 * First attempt at determining threshold from the residuals. Probably needs
 * some further thought
 * First the reference statistics
 */
    for (ic1=0;ic1<numcam;ic1++) {
      ic = get_thresh_index(ic1, -1, -1, -1);
      var = sqrt(t->var[ic]/t->n[ic]);
      max = sqrt(t->max[ic]);
      VPRINTF(VERBOSE,
        "Rms and max ref stars residuals for camera %d: %f and %f\n", 
        ic, var/DPI*180*3600, max/DPI*180*3600);
      /*t->thresholds[t->icurr][ic] = MAX(3*var,max/2);*/
      t->thresholds[t->icurr][ic] = MIN(3*var,0.9*max);
    }
/*
 *  Now the overlap statistics (random camera, unequal frame)
 */
    for (ic1=0;ic1<numcam;ic1++) {
      for (ic2=0;ic2<numcam;ic2++) {
        ic = get_thresh_index(ic1, ic2, -1, 0);
        if (t->n[ic] > 2) 
        {
          var = sqrt(t->var[ic]/t->n[ic]);
          max = sqrt(t->max[ic]);
        } else {
          var = max = 0.0;
        }
        VPRINTF(VERBOSE,
          "Rms and max overlap res. cam. %d and %d (uneq frames): %f and %f\n", 
            ic1, ic2, var/DPI*180*3600, max/DPI*180*3600);
        /*t->thresholds[t->icurr][ic] = MAX(3*var,max/2);*/
        /*t->thresholds[t->icurr][ic] = 3.*var;*/
        t->thresholds[t->icurr][ic] = MIN(3*var,0.9*max);
      }
    }
/*
 *  Now the intercamera statistics (random camera, equal frame)
 */
    for (ic1=0,var=max=0.0;ic1<numcam;ic1++) {
      for (ic2=0;ic2<numcam;ic2++) {
        ic = get_thresh_index(ic1, ic2, -1, -1);
        if (t->n[ic] > 2) 
        {
          var = sqrt(t->var[ic]/t->n[ic]);
          max = sqrt(t->max[ic]);
        } else {
          var = max = 0.0;
        }
        VPRINTF(VERBOSE,
            "Rms and max overlap res. cam. %d and %d (eq frames): %f and %f\n", 
            ic1, ic2, var/DPI*180*3600, max/DPI*180*3600);
        /*t->thresholds[t->icurr][ic] = MAX(3*var, max/2);*/
        /*t->thresholds[t->icurr][ic] = 3.*var;*/
        t->thresholds[t->icurr][ic] = MIN(3*var,0.9*max);
      }
    }
/*
 *  Safety belt: compute maximum threshold, and ensure none of the others
 *  is less than half of it. This makes sure nothing too wrong is done when
 *  some identifications are wildly off.
 */
    for (i=0,max_thresh=0;i<t->numthresh;i++) 
      max_thresh=MAX(max_thresh, t->thresholds[t->icurr][i]);
    for (i=0;i<t->numthresh;i++) 
      if (t->thresholds[t->icurr][i] != 0.0 &&
          t->thresholds[t->icurr][i] < max_thresh/2.0) {
        VPRINTF(VERBOSE, "Threshold %d was lower than half the max threshold of %f and was brought back within limits\n", i, 
             max_thresh/DPI*180*3600);
        t->thresholds[t->icurr][i] = max_thresh/2.0;
      }
}
