/* The program imageheader gives FITS images
   a 'simple' header containing only the minimal
   necessary information necessary for a successfull
   astrometric and photometric calibration with our
   pipeline.
   The keywords present in the header are the following:
   - CRPIX?, CRVAL?, CD?_?, AIRMASS, GABODSID, IMAGEID,
     EQUINOX, EXPTIME, CTYPE, FILTER
   - 20 DUMMY keywords that can be replaced when necessary

   The 'old' headers containing a lot of 'useless' information
   that can lead to confusions is deleted.

   19.04.03: 
   deleted the PCOUNT and GCOUNT keywords as they are not
   permitted in the primary header according to the FITS
   standard.

   06.05.03:
   changed the variable type of gabodsid, imageid and eisid
   from 'short int' to 'int' because of core dump on HP system.

   08.05.03:
   if an OBJECT keyword is present in the old header it is
   transfered to the new header

   07.12.2004:
   we add default values for magnitude zeropoint (ZP) and
   extinction coefficient (COEFF). A -1.0 in both components
   says that the image has not yet been calibrated. 

   28.01.2005:
   - I added history information on the processing
   - old HISTORY information from previous image manipulations
     is transfered now.
 */
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#include "fitscat.h"
#include "options_globals.h"
#include "utils_globals.h"
#include "image_io_globals.h"
#include "field_desc.h"
#include "fits_image_io_globals.h"
#include "imageheader_vers.h"
#include "global_vers.h"
#include "t_iso8601.h"

#define MAXCHAR 255

static char	header[MAXCHAR];

typedef struct {
   char conffile[MAXCHAR];
   char infile[MAXCHAR];
   char outfile[MAXCHAR];
   char radecsys[MAXCHAR];
   char filter[MAXCHAR];
   double cd1_1, cd2_2;
   double cd1_2, cd2_1;
   double crval1, crval2;
   double crpix1, crpix2;
   double airmass; 
   double equinox;
   double exptime;
   int gabodsid, imageid, eisid;
} control_struct;

control_struct control;

option opts[]={
         {"CRVAL1", FLOAT, &control.crval1, 0, 0, 0.0, 360.0, 
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0,
          "Ra of the nominal pointing"},
         {"CRVAL2", FLOAT, &control.crval2, 0, 0, -90.0, 90.0, 
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0,
          "Dec of the nominal pointing"},
         {"CRPIX1", FLOAT, &control.crpix1, 0, 0, -30000.0, 30000.0, 
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0,
          "X pixel of optical axis"},
         {"CRPIX2", FLOAT, &control.crpix2, 0, 0, -30000.0, 30000.0, 
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0,
          "Y pixel of optical axis"},
         {"CD1_1", FLOAT, &control.cd1_1, 0, 0, -1.0, 1.0, 
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0,
          "CD1_1 matrix element"},
         {"CD2_2", FLOAT, &control.cd2_2, 0, 0, -1.0, 1.0, 
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0,
          "CD2_2 matrix element"},
         {"CD1_2", FLOAT, &control.cd1_2, 0, 0, -1.0, 1.0, 
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0,
          "CD1_2 matrix element"},
         {"CD2_1", FLOAT, &control.cd2_1, 0, 0, -1.0, 1.0, 
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0,
          "CD2_1 matrix element"},
         {"AIRMASS", FLOAT, &control.airmass, 0, 0, 1.0, 5.0, 
          0, 1.0, NO, "",
          {"", "", "", "", ""}, 0,
          "airmass of observation"},
         {"EQUINOX", FLOAT, &control.equinox, 0, 0, 0.0, 5000.0, 
          0, 2000.0, NO, "",
          {"", "", "", "", ""}, 0,
          "epoch of observation"},
         {"EXPTIME", FLOAT, &control.exptime, 0, 0, 0.0, 20000.0, 
          0, 0.0, NO, "",
          {"", "", "", "", ""}, 0,
          "epoch of observation"},
         {"RADECSYS", TEXT, control.radecsys, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "'FK5'",
          {"", "", "", "", ""}, 1,
          ""},
         {"FILTER", TEXT, control.filter, 0, 0, 0.0, 0.0, 
          0, 0.0, NO, "'R'",
          {"", "", "", "", ""}, 1,
          "Filter in that observation was done"},
         {"GABODSID", INT, &control.gabodsid, 0, 32767, 0., 0., 
          1, 0., NO, "",
          {"", "", "", "", ""}, 0,
          "GABODSID"},
         {"IMAGEID", INT, &control.imageid, 0, 100, 0., 0., 
          1, 0., NO, "",
          {"", "", "", "", ""}, 0,
          "IMAGEID"},
         {"EISID", INT, &control.eisid, 0, 32767, 0., 0., 
          0, 0, NO, "",
          {"", "", "", "", ""}, 0,
          "EISID"},
         {""},
};

void usage(char *me)
{
/****** usage ***************************************************
 *
 *  NAME	
 *      usage - Print out a message about how-to-use the program
 *
 *  SYNOPSIS
 *      void usage(char *me)
 *
 *  FUNCTION
 *      Print to stderr a message about how to use th main program
 *      The execution will stop and the exit code of the program is set
 *      to non-zero.
 *
 *  INPUTS
 *      me  - The calling name of the program
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      T. Erben - dd 15-04-2003
 *
 ******************************************************************************/
    fprintf(stderr,"%s Version: %s (%s)\n",NAME,VERS,DATE);
    fprintf(stderr,"Usage: %s	 -i inputimage\n", me);
    fprintf(stderr,"		 -o outputimage\n");
    fprintf(stderr,"		 -c configfile\n");
    exit(1);
}

void init(int argc, char *argv[])
{
/****** int ***************************************************
 *
 *  NAME	
 *      init - Initialize the program from argument and configuration file
 *
 *  SYNOPSIS
 *      void init(int argc, char *argv[])
 *
 *  FUNCTION
 *      This function reads the command line arguments and interprets the
 *      configuration file
 *
 *  INPUTS
 *      argc - Number of arguments passed
 *      argv - The argument char array
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *      The control structure is set with the argument list parameters
 *      and the configuration parameters
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 08-08-1997
 *
 ******************************************************************************/
    int i, j, k;

    if (argc == 1) usage(argv[0]);
/*
 *  Initialize command line options with some defaults
 */
    sprintf(control.conffile,"imageheader.conf");
    init_options(0, opts);
/*
 *  Parse the command line
 */
    for (i=1; i<argc; i++) {
       if (argv[i][0] == '-') {
          switch(argv[i][1]) {
          case 'i': strncpy(control.infile,argv[++i], MAXCHAR-1);
                    break;
          case 'o': strncpy(control.outfile,argv[++i], MAXCHAR-1);
                    break;
          case 'c': strncpy(control.conffile,argv[++i], MAXCHAR-1);
                    break;
          case '@': print_options(argv[0]);
                    exit(0);
          default:  parse_options(control.conffile);
                    j = i++; k = i;
                    if (argv[j] && argv[k])
                      set_command_option(argv[j], argv[k]);
                    else
                      usage(argv[0]);
                    break;
          }
      } else {
          syserr(WARNING,"init","Command line syntax error\n");
          usage(argv[0]);
      }
   }
   finish_options(control.conffile);
}

int main(int argc, char *argv[])
{
/****** imageheader ***************************************************
 *
 *  NAME	
 *       imageheader - Create a standard image header
 *
 *  SYNOPSIS
 *       imageheader -i inputimage
 *                   -o outputimage
 *                   -c conffile
 *
 *  FUNCTION
 *
 *  INPUTS
 *      i      - The input FITS image
 *      o      - The output FITS image
 *
 *  RESULTS
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      T. Erben - dd 15-04-2003
 *
 ******************************************************************************/
    fitsfile_struct	*ffi, *ffo;
    float	*data;
    char        name[20];
    char        *dummychar;
    char        tmpstring[80];
    int i;

    init(argc, argv);

    sprintf(header,"%s Version: %s (%s)",NAME,VERS,DATE);
    VPRINTF(NORMAL,"\n   %s\n\n",header);

/*
 *     Now put the values into the image headers
 */
    set_no_interp(YES);

    ffi = open_fits_file(control.infile);
    switch (ffi->bitpix) 
    {
    case 16:
        ffi->pix_type = T_SHORT;
        break;
    case -32:
        ffi->pix_type = T_FLOAT;
        break;
    }
    data = interp_fits_image(ffi, NO); 
            
/*       
 *       Create the output FITS file
 */      
    ffo = new_fits();
    init_fits(ffo);
    ffo->pix_type = ffi->pix_type;

    /* first create the new header */
    w_fits_bool  ("SIMPLE",  ffo->head, TRUE);
    w_fits_int   ("BITPIX",  ffo->head, (ffi->bitpix));
    w_fits_int   ("NAXIS",   ffo->head, 2);
    w_fits_int   ("NAXIS1",  ffo->head, (ffi->xwid));
    w_fits_int   ("NAXIS2",  ffo->head, (ffi->ywid));
    w_fits_double("BSCALE",  ffo->head, (ffi->bscale));
    w_fits_double("BZERO",   ffo->head, (ffi->bzero));
    w_fits_str   ("CTYPE1",  ffo->head, "'RA---TAN'");
    w_fits_str   ("CTYPE2",  ffo->head, "'DEC--TAN'");
    w_fits_double("CRPIX1",  ffo->head, control.crpix1);
    w_fits_double("CRPIX2",  ffo->head, control.crpix2);
    w_fits_double("CD1_1",   ffo->head, control.cd1_1);
    w_fits_double("CD2_2",   ffo->head, control.cd2_2);
    w_fits_double("CD1_2",   ffo->head, control.cd1_2);
    w_fits_double("CD2_1",   ffo->head, control.cd2_1);
    w_fits_double("CRVAL1",  ffo->head, control.crval1);
    w_fits_double("CRVAL2",  ffo->head, control.crval2);
    w_fits_str   ("RADECSYS",ffo->head, control.radecsys);
    w_fits_str   ("FILTER",  ffo->head, control.filter);
    w_fits_double("AIRMASS", ffo->head, control.airmass);
    w_fits_double("EXPTIME", ffo->head, control.exptime);
    w_fits_double("EQUINOX", ffo->head, control.equinox);
    w_fits_int   ("IMAGEID", ffo->head, (int)control.imageid);
    w_fits_int   ("GABODSID",ffo->head, (int)control.gabodsid);
    w_fits_int   ("EISID",   ffo->head, (int)control.eisid);

    /* see whether the old header has an OBJECT keyword;
       if yes, we transfer it to the new header
    */    
    dummychar=r_fits_str("OBJECT", ffi->head);

    if(dummychar[0] != '\0')
    {
      strcpy(name, "\'");
      strcat(name, dummychar);
      strcat(name, "\'");
      w_fits_str("OBJECT", ffo->head, name);
    }
    ED_FREE(dummychar);

    /* insert default values for magnitude zeropoint
       and extinction coefficient */
    w_fits_double("ZP", ffo->head, -1.0);
    w_fits_double("COEFF", ffo->head, -1.0);

    /* insert dummy keywords */
    for(i=0; i<20; i++)
    {
      sprintf(name, "DUMMY%d", i);
      w_fits_int(name, ffo->head, 0);
    }

    /* first go throuh the initial header and collect
       old HISTORY information
    */
    dummychar=r_fits_str("HISTORY", ffi->head);
    while(dummychar != NULL)
    {
      w_fits_str("HISTORY", ffo->head, dummychar);
      ED_FREE(dummychar);
      d_fits("HISTORY", ffi->head);
      dummychar=r_fits_str("HISTORY", ffi->head);
    }
    
    /* add history comments */
    w_fits_str("HISTORY", ffo->head, "");
    sprintf(tmpstring, "imageheader: %s", __PIPEVERS__);
    w_fits_str("HISTORY", ffo->head, (char *)tmpstring);
    sprintf(tmpstring, "imageheader: called at %s", get_datetime_iso8601());
    w_fits_str("HISTORY", ffo->head, (char *)tmpstring);
    sprintf(tmpstring, "imageheader: modified header from %s", control.infile);
    w_fits_str("HISTORY", ffo->head, (char *)tmpstring);
    w_fits_pad(ffo->head);
/*       
 *       Now save the new FITS file
 */      
    save_fits(ffo, (unsigned char *)data, control.outfile);
    ED_FREE(data);
    close_fits(ffi, NO);

    return 0;
}

