#include <stdio.h>
#include <math.h>
#include <ctype.h>
#include <string.h>

#include "bileave_globals.h"
#include "lists_globals.h"
#include "pos_globals.h"
#include "pos_defs.h"
#include "tabutil_globals.h"
#include "obj_obj_util.h"
#include "tsize.h"

/*
 * 05.05.03:
 * included the tsize.h file
 *
 * 29.01.2005:
 * - I added history information to the output catalog
 * - I added argc and argv to the argument list of save_catalog
 *   as they are needed to write HISTORY information to the
 *   output catalog.
 *
 * 11.03.2005:
 * changes to reflect the new calling syntax of init_catalogs
 *
 * 07.03.2006:
 * I removed compiler warnings connected with the opts structure
 *
 * 27.03.2006:
 * I renamed sorting and indexing functions to the new naming scheme
 * described in the sort_globals.h file
 */

/*
 *    Global variables do not take as much memory
 */
static control_struct control;

char header[MAXCHAR];


option my_opts[] = {
            {"RA", TEXT, control.ra, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "Ra",
              {"", "", "", "", ""}, 0,
              "The name of the XXXX table column containing the Right Ascension"},
            {"DEC", TEXT, control.dec, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "Dec",
              {"", "", "", "", ""}, 0,
              "The name of the XXXX table column containing the Declination"},
            {"BILEAVE", TEXT, control.bileave, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "BitInterLeaved",
              {"", "", "", "", ""}, 0,
              "The name of the XXXX table column recieving the Bit Interleave value"},
            {"", BOOLEAN, NULL, 0, 0, 0.0, 0.0, 0, 0.0, NO, "", 
             {"YES", "NO", "", "", ""}, 0,
              "dummy key to end structure my_opts"}
            };

void usage(char *me)
{
    fprintf(stderr,"Usage: %s \n", me);
    fprintf(stderr,"	-i incat1\n");
    fprintf(stderr,"	-o outcat1\n");
    fprintf(stderr,"	[-c conf_file] \n\n");
    fprintf(stderr,"	[-t table_name] \n\n");
    exit(1);
}


void init(int argc, char *argv[], control_struct *c)
{
/*
 *  Default file name is calling name with .conf extension
 */
    int i, j, k;

    sprintf(c->table_name,"OBJECTS");
    if (argc == 1) usage(argv[0]);
    init_options(0, my_opts);
    for (i=1; i<argc; i++) {
      if (argv[i][0] == '-') {
         switch(argv[i][1]) {
         case 'c': strcpy(c->conffile,argv[++i]);
                   break;
         case 't': strcpy(c->table_name,argv[++i]);
                   break;
         case 'i': strcpy(c->listfile,argv[++i]);
                   break;
         case 'o': strcpy(c->outfile,argv[++i]);
                   break;
         case '@': print_options(argv[0]);
                   exit(0);
         default:  parse_options(c->conffile);
                   j = i++; k = i;
                   set_command_option(argv[j], argv[k]);
                   break;
         }
      } else {
         syserr(WARNING,"init","Command line syntax error\n");
         usage(argv[0]);
      }
   }
   finish_options(c->conffile);
}
keystruct **bileave_out_keys(int *n)
{
    keystruct   **keys;
    int         nkeys = 2;

    ED_CALLOC(keys, "bileave_out_keys", keystruct *, nkeys);
    keys[0] = new_key(control.bileave);
    keys[0]->ttype = T_LONG;
    keys[0]->nbytes = t_size[T_LONG];
    strcpy(keys[0]->unit,"");
    sprintf(keys[0]->comment,"Bit Interleave Value");
    keys[1] = new_key("HASH_POS");
    keys[1]->ttype = T_LONG;
    keys[1]->nbytes = t_size[T_LONG];
    strcpy(keys[1]->unit,"");
    sprintf(keys[1]->comment,"Reference number to HASH table");

    *n = nkeys;
    return keys;
}


void save_catalog(catstruct *incat, int *bi_col, int *idxpos, int n,
                  char *filename, char *table_name, int argc, char **argv)
/****** save_catalog ***************************************************
 *
 *  NAME	
 *      save_catalog - Copy the input to output adding BIL obj-by-obj
 *
 *  SYNOPSIS
 *      void save_catalog(catstruct *incat, int *bi_col, int *idxpos, 
 *                        int n, char *filename, char *table_name)
 *
 *  FUNCTION
 *      This routine copies the input catalog to the output catalog
 *      adding the bit interleave value to the required output table.
 *      The copying is done on an object by object bases while using
 *      index that sorts the input catalog on position. Therefore the 
 *      output catalog is position sorted.
 *
 *  INPUTS
 *      incat      - Pointer to teh input catalog
 *      bi_col     - The bit interleave column to be written
 *      idxpos     - The index that sorts the input catalog on position
 *      n          - Number of input catalog objects
 *      filename   - Name of the output catalog
 *      table_name - Name of the table to recieve bi_col and that needs sorting
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 27-05-1998
 *
 ******************************************************************************/
{
    catstruct	*outcat;
    tabstruct	*tab, *intab, *outtab;
    keystruct	*key, *okey, **outkeys;
    int		i, j, k, nn, endpos, noutkeys;
    int		obj_prev, *obj_pos, *obj_cnt, *obj_bil, count;
    long int	bilscale, bilbclip;
    double	*obj_ra_min, *obj_dec_min, *obj_ra_max, *obj_dec_max, 
                side, sidee;
    posstruct	t_pos;

    outkeys = bileave_out_keys(&noutkeys);
    init_catalogs(incat, &outcat, filename,
                  table_name, NULL, NULL, 0,
                  outkeys, noutkeys, &intab, &outtab,
                  "bileave", argc, argv);
    tab = name_to_tab(incat, table_name, 0);
/*
 *  Start the copy process, keeping track of the hash information
 */
    ED_CALLOC(obj_pos,     "save_catalog", int,    n);
    ED_CALLOC(obj_cnt,     "save_catalog", int,    n);
    ED_CALLOC(obj_bil,     "save_catalog", int,    n);
    ED_CALLOC(obj_ra_min , "save_catalog", double, n);
    ED_CALLOC(obj_dec_min, "save_catalog", double, n);
    ED_CALLOC(obj_ra_max , "save_catalog", double, n);
    ED_CALLOC(obj_dec_max, "save_catalog", double, n);
    t_pos.msb  = 3; 
    t_pos.lsb  = 0;
    pos_to_radec(t_pos, &side, &sidee);
/*
    side = pos_box_side(32);
 */
    obj_prev = bi_col[idxpos[0]];
    k = 1; count = 0;
    for (i=0;i<n;i++) {
       read_obj_at(intab, tab, idxpos[i]);
       if (obj_prev != bi_col[idxpos[i]]) {
          obj_bil[k] = obj_prev;
          t_pos.msb  = obj_prev; 
          t_pos.lsb  = 0;
          pos_to_radec(t_pos, &obj_ra_min[k], &obj_dec_min[k]);
          obj_ra_max[k]  = obj_ra_min[k] + side;
          obj_dec_max[k] = obj_dec_min[k] + side;
          obj_prev = bi_col[idxpos[i]];
          obj_pos[k] = obj_pos[k-1] + obj_cnt[k-1];
          obj_cnt[k] = count;
          k++; count = 0;
       }
       count ++;
/*
 *     Because the order of the output table's keys is the same as that 
 *     for the input (see above) we can walk both key chains in parallel
 */
       key  = intab->key;
       okey = outtab->key;
       for (j=0;j<intab->nkey;j++) {
          memcpy(okey->ptr, key->ptr, key->nbytes);
          key = key->nextkey;
          okey = okey->nextkey;
       }
/* 
 *     Add the two additional output columns
 */
       memcpy(outkeys[0]->ptr, &bi_col[idxpos[i]], t_size[outkeys[0]->ttype]);
       memcpy(outkeys[1]->ptr, &k, t_size[outkeys[1]->ttype]);
       if ((i) % 2048 == 0)  
          VPRINTF(NORMAL,"Objects processed: %d from %d\n", i , n);
       write_obj(outtab);
    }
    end_writeobj(outcat, outtab);
/*
 *  Add the hashing table to the output catalog
 */
    obj_bil[k] = obj_prev;
    t_pos.msb  = obj_prev; 
    t_pos.lsb  = 0;
    pos_to_radec(t_pos, &obj_ra_min[k], &obj_dec_min[k]);
    obj_ra_max[k]  = obj_ra_min[k] + side;
    obj_dec_max[k] = obj_dec_min[k] + side;
    obj_pos[k] = obj_pos[k-1] + obj_cnt[k-1];
    obj_cnt[k] = count;

    outtab = new_tab("HASH");
    add_key_to_tab(outtab, "OBJECT_POS", k, &obj_pos[1], T_LONG, 1,
             "", "Position of the first object from that hash");
    add_key_to_tab(outtab, "OBJECT_COUNT", k, &obj_cnt[1], T_LONG, 1,
             "", "Number of objects in that hash");
    add_key_to_tab(outtab, "Ra_MIN", k, &obj_ra_min[1], T_DOUBLE, 1,
             "", "Minimum Right Ascension of hash area");
    add_key_to_tab(outtab, "Dec_MIN", k, &obj_dec_min[1], T_DOUBLE, 1,
             "", "Minimum Declination of hash area");
    add_key_to_tab(outtab, "Ra_MAX", k, &obj_ra_max[1], T_DOUBLE, 1,
             "", "Maximum Right Ascension of hash area");
    add_key_to_tab(outtab, "Dec_MAX", k, &obj_dec_max[1], T_DOUBLE, 1,
             "", "Maximum Declination of hash area");
    add_key_to_tab(outtab, control.bileave, k, &obj_bil[1], T_LONG, 1,
             "", "Bit Interleave value of hash area");
    if (add_tab(outtab, outcat, 0) != RETURN_OK)
       syserr(FATAL, "save_catalog", "Error adding table HASH to output catalog");
    endpos = fitsfind(outtab->headbuf, "END     ");
/*
 *  Check that the new parameters will fit in the allocated memory space
 */
    endpos += 2;
    if (endpos >= (nn=outtab->headnblock)*(FBSIZE/80)) {
       ED_REALLOC(outtab->headbuf, "save_catalog", char, (nn+1)*FBSIZE);
       memset(outtab->headbuf + n*FBSIZE, ' ', FBSIZE);
       outtab->headnblock++;
    }
    bilscale = accur; bilbclip = 32;
    fitsadd(outtab->headbuf, "BILSCALE", "Bit Interleave position scaling factor");
    fitswrite(outtab->headbuf, "BILSCALE", &bilscale, H_INT, T_LONG);
    fitsadd(outtab->headbuf, "BILBCLIP", "Bit Interleave bits clipped");
    fitswrite(outtab->headbuf, "BILBCLIP", &bilbclip, H_INT, T_LONG);
    save_tab(outcat, name_to_tab(outcat, "HASH", 0));
}

int main(int argc, char *argv[])

/****** bileave ***************************************************
 *
 *  NAME	
 *      bileave - Add a INT BitInterLeaved column to the table
 *
 *  SYNOPSIS
 *      bileave -i incat1 -o outcat1
 *                [-c configfile] [CONFOPTION]
 *
 *  FUNCTION
 *      This program derives and adds a bit interleaved column to any
 *      catalog that has a table containing two positional coordinates.
 *      In addition the output catalog is sorted using this bit
 *      interleaved column. The bit interleaving value is constructed
 *      out of the most significant part  (32 bits) of the full 64 bit 
 *      word, where the positional coordinates have been converted to
 *      integer using a scaling factor of 1000000.0 (accur in pos_def.h)
 *
 *  INPUTS
 *      -i incat1      - The name of input catalog
 *      -o outcat1     - The name of output catalogs
 *     [-c configfile] - The name of the configuration file. 
 *     [-t table_name] - The name of the table to be sorted
 *     [CONFOPTION]    - Pairs of configuration VARIABLE VALUE
 *                       sequences that allow dynamically
 *                       overwritting of configuration file values.
 *
 *  RESULTS
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 26-05-1998
 *
 ******************************************************************************/
{
     int	POSKEYS = 2, n;
     int	*idxpos, *bi_col;
     int	i;
     double	*ra, *dec;
     catstruct	*incat;
     tabstruct	*tab;
     keystruct	**keys;
     char	**keynames;

     hashstruct	*hash;

     init(argc, argv, &control);

     sprintf(header,"%s Version: %s (%s)",PROGRAM,VERS,DATE);
     VPRINTF(NORMAL,"\n   %s\n\n",header);
/*
 *   Open the table to be sorted abd read the position keys
 */
     if (!(incat = read_cat(control.listfile))) 
         syserr(FATAL, "bileave", "Could not open catalog %s\n", 
                   control.listfile);
     if (!(tab = name_to_tab(incat,control.table_name, 0)))
         syserr(FATAL, "bileave", "Could not find table %s\n", 
                   control.table_name);
     VPRINTF(NORMAL, "Reading position keys\n");
     ED_CALLOC(keys,     "make_pairing", keystruct *, POSKEYS);
     ED_CALLOC(keynames, "make_pairing", char *,      POSKEYS);
     for (i=0;i<POSKEYS;i++)
        ED_CALLOC(keynames[i], "make_pairing", char, MAXCHAR);
     strcpy(keynames[0], control.ra);
     strcpy(keynames[1], control.dec);
     read_keys(tab, keynames, keys, POSKEYS, NULL);
     for (i=0;i<POSKEYS;i++) 
        if (!keys[i])
          syserr(FATAL,"make_pairing", "No key %s in table\n", keynames[i]);
     ED_COPY_PTR(ra,   "make_pairing", keys[0], T_DOUBLE, "double");
     ED_COPY_PTR(dec,  "make_pairing", keys[1], T_DOUBLE, "double");
     n = keys[0]->nobj;
/*
 *   Create the hashing structure and index the most significant word of the
 *   full bit interleaved value
 */
     ED_CALLOC(idxpos, "main", int, n);
     ED_CALLOC(bi_col, "main", int, n);
     VPRINTF(NORMAL, "Making the hash table\n");
     hash = init_hash(ra, dec, n);
     VPRINTF(NORMAL, "Making the position index\n");
     for (i=0;i<n;i++) 
        bi_col[i] = hash->pos[i].msb;
     iindex(n, bi_col, idxpos);
/*
 *   Remove the data values from the memory to allow save_catalog to read
 *   in object-by-object mode
 */
     ED_FREE(keys[0]->ptr);
     ED_FREE(keys[1]->ptr);
     save_catalog(incat, bi_col, idxpos, n, 
                  control.outfile, control.table_name,
                  argc, argv);

     VPRINTF(NORMAL,"Done\n");
     return(0);
}
