 /*
 				main.c

*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*
*	Part of:	LDACTest
*
*	Author:		E.BERTIN, DeNIS/LDAC
*
*	Contents:	parsing and main loop.
*
*	Last modify:	21/04/97
*
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

/*
   04.01.05:
   removed compiler warnings (gcc with -O3 optimisation)

   07.03.06:
   removed compiler warning due to a %lf format string.
   (I replaced %lf with %f)
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>
#include	<string.h>

#include	"fitscat_defs.h"
#include	"fitscat.h"


#define		SYNTAX \
"ldacbutvec -i Catalog1 Catalog2 [-t <Table_name>] [-f FILTERKEY VALUE] [options]\n\n\
Options are:	-b (suppress printing of the banner)\n\
		-r (Print the row number)\n\
                -q (Quiet flag: defaulted to verbose!)\n"

catstruct	*incat, *outcat;
int		qflag;

/********************************** main ************************************/
int main(int argc, char *argv[])

/****** ldactoasc ***************************************************
 *
 *  NAME	
 *      ldactoasc - Convert binary table infor to ASCII formatted text
 *
 *  SYNOPSIS
 *      ldactoasc -i incat1 incat2 .. [-t tablename] [-s] [-q]
 *
 *  FUNCTION
 *      This program converts the binary information stored in a specified
 *      table into ASCII formatted text that can more easily be read. The
 *      Table column names and positions are given in a # commented header
 *      to allow easy interpretation and plotting through supermongo.
 *
 *  INPUTS
 *      -i incat1 incat2 .. - The list of input catalogs
 *     [-t tablename]       - The name of the table to be converted
 *     [-k key1 key2 ...]   - The lis of keynames to be printed
 *     [-b]                 - Suppress printing of a banner describing the
 *                            columns content.
 *     [-r]                 - Add to the output a dummy column, containing
 *                            the running row number.
 *     [-f]                 - Filter key and value
 *     [-q]                 - Make the program less verbose
 *
 *  RESULTS
 *     The output is send to standard output.
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E. Bertin - dd 13-06-1996
 *      E.R. Deul - dd Adapted to LDAC pipeline conventions
 *                     Added string support
 *	E. Bertin - dd 15-01-1997 Added handling of errors.
 *	E. Bertin - dd 21-01-1997 Added Banner and Row_pos flags.
 *
 ******************************************************************************/
  {
   tabstruct	*tab;
   static char	outfilename[MAXCHARS], tabname[MAXCHARS];
   static char  filterkey[MAXCHARS], filterval[MAXCHARS];
   int	a,ab,ae, ni,do_rowpos,do_banner,filter,*mask;
   int  i, j, k=0, n, kb,ke, ki, len;
   char **keynames=NULL;
   keystruct **keys=NULL;

  if (argc<2)
    {
    fprintf(OUTPUT, "				ldactoasc V1.2\n\n"
			"Purpose: convert a FITS table to ASCII\n");
    error(EXIT_FAILURE,"SYNTAX: ", SYNTAX);
    }

/*default parameters */
  do_rowpos = 0;
  do_banner = 1;
  qflag = 1;
  filter = 0;

  sprintf(tabname, "FIELDS");

  sprintf(outfilename, "default_test.cat");

  ab=ae=ni=0;
  kb=ke=ki=0;
  for (a=1; a<argc; a++)
    switch((int)tolower((int)argv[a][1]))
      {
      case 'b': do_banner = 0;
	        break;
      case 'i':	for(ab = ++a; (a<argc) && (argv[a][0]!='-'); a++);
		ae = a--;
		ni = ae - ab;
	        break;
      case 'r': do_rowpos = 1;
	        break;
      case 't':	sprintf(tabname, "%s", argv[++a]);
	        break;
      case 'f':	sprintf(filterkey, "%s", argv[++a]);
      		sprintf(filterval, "%s", argv[++a]);
                filter = 1;
	        break;
      case 'q': qflag = 1;
	        break;
      case 'k': for(kb = ++a; (a<argc) && (argv[a][0]!='-'); a++);
                ke = a--;
                ki = ke - kb;
                break;
      default : error(EXIT_FAILURE,"SYNTAX: ", SYNTAX);
      }

  if (!ni)
    error(EXIT_SUCCESS,"SYNTAX: ", SYNTAX);

/*Load the input catalogs*/
  QFPRINTF(OUTPUT, "Reading catalog(s)");
  if (ki) {
    QMALLOC(keynames,char *, (ki+filter))
    QMALLOC(keys,keystruct *, 1)
    for (k=kb;k<ke;k++) 
       keynames[k-kb] = argv[k];
       QMALLOC(keys[k-kb],keystruct, 1)
    if (filter) keynames[k-kb] = filterkey;
  } else  {
    error(EXIT_FAILURE, "*Error*: Please specify keynames","");
  }
  for (a=ab; a<ae; a++)
    {
    if (!(incat = read_cat(argv[a])))
      error(EXIT_FAILURE, "*Error*: catalog not found: ", argv[a]);
    if (!(tab = name_to_tab(incat, tabname, 0)))
      error(EXIT_FAILURE,"*Error*: found no such table: ", tabname);
    read_keys(tab, keynames, keys, ki+filter, NULL);
    if (keys[0]->naxis != 1)
      error(EXIT_FAILURE,"*Error*: Key not 2 dimensional: ", keys[0]->name);
    for (k=1;k<ki;k++) 
      {
      if (keys[k]->naxis != 1)
        error(EXIT_FAILURE,"*Error*: Key not 2 dimensional: ", keys[k]->name);
      if (keys[k]->naxisn[0] != keys[0]->naxisn[0])
        error(EXIT_FAILURE,"*Error*: Key length error : ", keys[k]->name);
      }
    }
    QCALLOC(mask, int, keys[0]->nobj);
    for (i=0;i<keys[0]->nobj;i++) {
       switch (keys[ki]->ttype) {
       case T_SHORT:
           if (((short *)keys[k]->ptr)[i] == (short) atoi(filterval) ||
              !filter) 
              mask[i] = 1;
           break;
       case T_LONG:
           if (((int *)keys[k]->ptr)[i] == atoi(filterval)||
              !filter) 
              mask[i] = 1;
           break;
       case T_FLOAT:
           if (((float *)keys[k]->ptr)[i] == (float) atof(filterval)||
              !filter) 
              mask[i] = 1;
           break;
       case T_DOUBLE:
           if (((double *)keys[k]->ptr)[i] == atof(filterval)||
              !filter) 
              mask[i] = 1;
           break;
       case T_BYTE:
           if (((unsigned char *)keys[k]->ptr)[i] == 
                (unsigned char) atoi(filterval)||
              !filter) 
              mask[i] = 1;
           break;
       case T_STRING:
           len = strlen(filterval);
           if (strncmp(&((char *)keys[k]->ptr)[i*keys[k]->naxisn[0]],
               filterval,len)==0|| !filter) 
              mask[i] = 1;
           break;
       }
    }
    n = 1;
    if (do_rowpos && do_banner) {
       fprintf(stdout, "# %3d %-15.15s %.47s\n", n++, 
                                 "(row_pos)", "running row");
    }
    for (k=0;k<ki;k++) {
       if (do_banner) {
       if (*keys[k]->unit)
          fprintf(stdout, "# %3d %-15.15s %-47.47s [%s]\n",
                      n, keys[k]->name,keys[k]->comment, keys[k]->unit);
       else
          fprintf(stdout, "# %3d %-15.15s %.47s\n",
                      n, keys[k]->name,keys[k]->comment);
       n += keys[k]->naxisn[1]?keys[k]->naxisn[1]:1;
       }
    }
    for (i=0;i<keys[0]->nobj;i++) {
       for (j=0;j<keys[0]->naxisn[0]&&mask[i];j++) {
          if (do_rowpos) {
             fprintf(stdout, "%d ", j+1);
          }
          for (k=0;k<ki;k++) {
             switch(keys[k]->ttype) {
             case T_SHORT:
                fprintf(stdout, *keys[k]->printf?keys[k]->printf:"%d ",
                            ((short *)keys[k]->ptr)[i*keys[k]->naxisn[0]+j]);
                break;
             case T_LONG:
                fprintf(stdout, *keys[k]->printf?keys[k]->printf:"%d ",
                            ((int *)keys[k]->ptr)[i*keys[k]->naxisn[0]+j]);
                break;
             case T_FLOAT:
                fprintf(stdout, *keys[k]->printf?keys[k]->printf:"%f ",
                            ((float *)keys[k]->ptr)[i*keys[k]->naxisn[0]+j]);
                break;
             case T_DOUBLE:
                fprintf(stdout, *keys[k]->printf?keys[k]->printf:"%f ",
                            ((double *)keys[k]->ptr)[i*keys[k]->naxisn[0]+j]);
                break;
             case T_BYTE: 
                if (keys[k]->htype==H_BOOL)
                   if(((char *)keys[k]->ptr)[i*keys[k]->naxisn[0]+j])
                     fprintf(stdout, "T ");
                   else
                     fprintf(stdout, "F ");
                else
                   fprintf(stdout, *keys[k]->printf?keys[k]->printf:"%d ",
                 (int) ((unsigned char *)keys[k]->ptr)[i*keys[k]->naxisn[0]+j]);
                break;
             case T_STRING: 
                break;
             }
          }
          putc('\n', stdout);
       }
    }

  QFPRINTF(OUTPUT, "All done");
  QPRINTF(OUTPUT,"\n");

  return EXIT_SUCCESS;
  }
