/*
 * newpeaks.c
 *
 * a hierarchical peak finding algorithm
 *
 * creates a sequence of progressively more coarsely smoothed images, finds
 * the peaks at each level of smoothing and links these peaks together into
 * lines according to criteria in neighbours.
 *
 * for each level
 *	* smooth
 *	* find peaks and load them into a list headed by peakhead
 *	  try to install the peaks one at a time onto one of lines in linked list
 *	  headed by linehead.  If we can't we create a new line and link it
 *	  onto the end of the linked list
 *	* find dead lines (ones which didn't get peak added to at last step)
 *	  and dispose of them; i.e. write out most significant point

April 1 1998: Added subimage peakfinding (HH)
April 29    : Fixed a memory leak problem (end of oldhfindpeaks()
              and a out of array-bounds problem when assigning fsub
              (begin of oldhfindpeaks()
May 28: declared thecat.zeropoint; not doing this led to problems
        with select and listhenk on DEC alpha machines.
Feb 1, 1999: adapted to the new hfindpeaks:
             included calculation of preliminary centroid position
                 (copy from newpeaks_local); deleted calculation of preliminary
                 ellipticity parameters
Mar 15, 1999: x[0] and x[1] are now calculated correctly (it was forgotten
              to add the offset from the part of the field that is just
              examined)
Apr 6, 1999: replaced calloc's be ED_CALLOC stuff.

Jun 1, 1999: fixed the bug that some obvious objects were missed.
             (routine disposeof)
23.01.01: removed unused variables

19.01.02: - included 'set_cat_opf(stdout)' to set explicitely the
            output catalog stream as is now necessary after the changes
            made in catio.c, fits.c
          - "removed" two free commands that caused a crash (have to search
            the memory leak)

10.01.05: I removed compiler warnings under gcc with optimisation
          enabled (-O3).
 */

#define usage "\n\n\n\
SYNOPSIS\n\
		newpeaks fitsfile [option...] >catalogue \n\
 			-r rf1 rf2	# range of filter radii (0.5 20.0)\n\
			-d dlnrf	# step in log(rf) (0.2)\n\
			-l flink	# linking parameter(1.0)\n\
			-n nu		# significance threshold (4.0)\n\
			-s sigma mode	# sky statistics\n\
                        -b blocksize    # size of subimages used for peak finding (256)\n\
\n\
DESCRIPTION\n\
		\"newpeaks\" --- a heirarchical object finder\n\
		fft gaussian filters an image with sequence of progressively\n\
		larger smoothing radius mexican hat filters and constructs peak\n\
		trajectories\n\
                The input image is divided in subimages of size blocksize,\n\
                specified by the -b option\n\
		These are output as an \"hcat\"; the actual job of\n\
		converting these to a catalogue of points being done\n\
		by the \"hcat2cat\" filter.\n\
		rf1, rf2 are min and max filter radii and we filter\n\
		with logarithmic steps in rf defined by dlnrf\n\
		Peaks at adjacent smoothing levels are connected if their\n\
		separation is less than flink * rf.\n\
		Use -s option to supply sky mode, sigma rather than have\n\
		them calculated from the image.\n\
\n\n\n"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <math.h>

#include "fits.h"
#include "error.h"
#include "myfft.h"
#include "arrays.h"
#include "stats_stuff.h"
#include "object_stuff.h"
#include "magic.h"
#include "catio.h"
#include "newpeaks.h"
#include "utils_globals.h"
#include "utils_macros.h"
/* #include "fitscat.h" */

static float	sigma1, sigma2, normfactor;

static	int	nlines = 0, npeaks = 0;
int             imin,imax,jmin,jmax;
int             blockstep1,blockstep2;
float           rf1init,rf2init,dlnrfinit;
int		calcstats;
int             nblock1,nblock2;
int             blocksize;
float           **fin;  
float           rf1init,rf2init,dlnrfinit,flink,nulimit,rfratio;
fstatsrec	srec;
int             nn1,nn2;
int             fieldwidth, fieldheight;

#define TINY	1.e-20

void oldhfindpeaks();

int main(int argc, char *argv[])
{
  int		arg = 2;
  int		com, comc;
  char		*comv[MAX_COMMENTS];
  /*int         i, j; */
  int		N1, N2, margin = 3;
  FILE		*fitsfile;
  cathead	thecat;
  float		magicsubstitute;
  
  /* defaults */
  rf1init = 0.5;
  rf2init = 20.0;
  dlnrfinit = 0.2;
  flink = 1.0;
  nulimit = 4.0;
  rfratio = 2.0;		/* mexican hat parameter */
  calcstats = 1;
  blocksize = 256;
  
  if (argv[1][0] == '-')
    error_exit(usage);
  
  while (arg < argc) {
    if (argv[arg][0] != '-')
      error_exit(usage);
    switch (argv[arg++][1]) {
    case 'r':
      if (EOF == sscanf(argv[arg++], "%f", &rf1init))
	error_exit(usage);
      if (EOF == sscanf(argv[arg++], "%f", &rf2init))
	error_exit(usage);
      break;
    case 'd':
      if (EOF == sscanf(argv[arg++], "%f", &dlnrfinit))
	error_exit(usage);
      break;
    case 'b':
      if (EOF == sscanf(argv[arg++], "%d", &blocksize))
	error_exit(usage);
      break;
    case 'l':
      if (EOF == sscanf(argv[arg++], "%f", &flink))
	error_exit(usage);
      break;
    case 'n':
      if (EOF == sscanf(argv[arg++], "%f", &nulimit))
	error_exit(usage);
      break;
    case 's':
      if (EOF == sscanf(argv[arg++], "%f", &(srec.sigma)))
	error_exit(usage);
      if (EOF == sscanf(argv[arg++], "%f", &(srec.fmode)))
	error_exit(usage);
      calcstats = 0;
      break;
    case 'u':
    default:
      error_exit(usage);
    }
  }

  strcpy(thecat.image, argv[1]);	
  fitsfile = fopen(thecat.image, "r");
  if (!fitsfile)
    error_exit("hfindpeaks: can't open fits file\n");
  set_fits_ipf(fitsfile);
  fread_fits(&fin, &N1, &N2, &comc, comv);

  fieldwidth = N1;
  fieldheight = N2;
  /* get normfactor */
  normfactor = 1.0;
  for (com = 0; com < comc; com++) {
    if (0 == strncmp("HISTORY = normalise fac=", comv[com], 23)) {
      sscanf(comv[com], "HISTORY = normalise fac= %f", &normfactor);
    }
  }

  if (calcstats)
    fdo_stats(fin, N1, N2, margin, &srec);
  thecat.srec = srec;
  magicsubstitute = srec.fmode;

  thecat.nu		= 0.0;
  thecat.rf		= 0.0;
  thecat.f_thresh 	= 0.0;
  thecat.N 		= N1;
  thecat.zeropoint      = 0.0;
  argsToString(argc, argv, thecat.com[0]);
  thecat.comc = 1;
  thecat.hassky = 0;

  set_cat_opf(stdout);
  write_cat_head(&thecat);
  
  substitute(fin, N1, N2, magicsubstitute);
  
  if (blocksize>N1) blocksize=(int)(N1/2);
  if (blocksize>N2) blocksize=(int)(N2/2);
  
  nblock1=(int)(N1/blocksize)+1;
  nblock2=(int)(N2/blocksize)+1;

  while((int)(blocksize-N1/nblock1) <75) nblock1++;
  while((int)(blocksize-N2/nblock2) <75) nblock2++;
  
  if (nblock1==1) {
    blockstep1=N1;
  } else {
    blockstep1=(int)(N1/nblock1)-(int)((blocksize-N1/nblock1)/(nblock1-1));
  }
  if (nblock2==1) {
    blockstep2=N2;
  } else {
    blockstep2=(int)(N2/nblock2)-(int)((blocksize-N2/nblock2)/(nblock2-1));
  }
  
  
  /*		fprintf(stderr,"%d %d\t%d %d\t%d %d\n",N1,N2,nblock1,nblock2,blockstep1,blockstep2); */
  
  /* fprintf(stderr,"\nnn1\tnn2\tjmin\tjmax\timin\timax\tstep1\tstep2\n"); */
  
  
  for(nn1=0;nn1<nblock1;nn1++) {
    for(nn2=0;nn2<nblock2;nn2++) {
      oldhfindpeaks();
    }
  }

  return 0;
}

void oldhfindpeaks()
{
  int           level = 0;
  int		icell, jcell, ncells, installed;
  peak		*peakhead, *thepeak = NULL, *nextpeak = NULL;
  line		*linehead = NULL, *theline, *nextline = NULL;
  line		***cell;
  int           i,j,ii,jj;
  float		**fsub, **fs;
  float         rf, rf1, rf2, dlnrf, dmax, sigma;
  fft_type	fk, fkcopy;
  
  allocFloatArray(&fs, blocksize, blocksize);
  allocFloatArray(&fsub,blocksize,blocksize);
  alloc_fft(&fk, blocksize, blocksize);
  alloc_fft(&fkcopy, blocksize, blocksize);
  
  for(ii=0;ii<=blocksize-1;ii++) {
    for(jj=0;jj<=blocksize-1;jj++) {
      /* it seems that sometimes ii and jj were out of bounds (TE) */
      if( ((nn2*blockstep2+ii)<fieldheight) && ((nn1*blockstep1+jj)<fieldwidth))
      {
        fsub[ii][jj]=fin[nn2*blockstep2+ii][nn1*blockstep1+jj];
      }
      else
      {
        fsub[ii][jj]=srec.fmode;
      }
    }
  }
  
  if (nn1==0) {
    jmin=0;
  } else {
    jmin=0.5*(blocksize-blockstep1);
  }
  if (nn1==nblock1-1) {
    jmax=blocksize;
  } else {
    jmax=blocksize-0.5*(blocksize-blockstep1);
  }
  
  if (nn2==0) {
    imin=0;
  } else {
    imin=0.5*(blocksize-blockstep2);
  }
  if (nn2==nblock2-1) {
    imax=blocksize;
  } else {
    imax=blocksize-0.5*(blocksize-blockstep2);
  }
  
  /* fprintf(stderr,"%d\t%d\t%d\t%d\t%d\t%d\t%d\t%d\n",nn1,nn2,jmin,jmax,imin,imax,blockstep1,blockstep2); */
  
  forward_fft(fsub, blocksize, blocksize, fk);
  
  rf1=rf1init; rf2=rf2init; dlnrf=dlnrfinit;
  level = 0;
  linehead = (line *) NULL;
  
  for (rf = rf1; rf < rf2; rf *= (1 + dlnrf)) {
    /*		fprintf(stderr, "starting level %d;  rf = %10.3f;  nlines = %10d;  npeaks = %10d\n", 
		level, rf, nlines, npeaks);*/
    sigma = srec.sigma / (2 * sqrt(PI) * rf);
    sigma *= sqrt(1 - 4 / (1 + rfratio * rfratio) + 1 / (rfratio * rfratio));
    dmax = rf * flink;
    ncells = ceil(blocksize / dmax);
    cell = makecell(linehead, dmax, ncells);	    
    sigma1 = 0.5 * rf * rf;
    sigma2 = rfratio * rfratio * sigma1;
    copy_fft(fk, blocksize, blocksize, fkcopy);
    filter(fkcopy, blocksize, blocksize, thefilterfunction);
    inverse_fft(fkcopy, blocksize, blocksize, fs);
    peakhead = getpeaks(blocksize, blocksize, fs, level, rf, sigma);
    thepeak = peakhead;
    while (thepeak) {		
      /* loop over new peaks */
      nextpeak = thepeak->next;	/* 'cos thepeak->next gets trashed */
      getcoords(thepeak, &icell, &jcell, dmax);
      installed = 0;
      for (i = icell - 1; i <= icell + 1; i++) {
	for (j = jcell - 1; j <= jcell + 1; j++) {
	  if (i < 0 || i >= ncells || j < 0 || j >= ncells)
	    continue;
	  if (install(thepeak, cell[i][j], dmax)) {
	    installed = 1;
	    break;
	  }
	}
	if (installed)
	  break;
      }
      if (!installed)
	addnewline(&linehead, thepeak);
      thepeak = nextpeak;
    }
    removedead(&linehead, level, nulimit);
    freecell(cell, ncells);
    level++;
  }
  
  /* dispose of any lines still alive at end */
  theline = linehead;
  while (theline) {
    nextline = theline->next;
    disposeof(theline, nulimit);
    theline = nextline;
  }

  /* (in my opinion a memory leak (TE) */
  /* allocFloatArray(&fs, blocksize, blocksize);
  allocFloatArray(&fsub,blocksize,blocksize);
  alloc_fft(&fk, blocksize, blocksize);
  alloc_fft(&fkcopy, blocksize, blocksize); */

  freeFloatArray(fs,blocksize,blocksize);
  freeFloatArray(fsub,blocksize,blocksize);
  free_fft(fk,blocksize,blocksize);
  free_fft(fkcopy,blocksize,blocksize);
  free(theline);
  free(nextline);
  /* free(linehead); */
  /* free(peakhead); */
  free(thepeak);
  free(nextpeak);
}




peak	*getpeaks(int N1, int N2, float **fs, int level, float rf, float sigma)
{
  int	i, j;
  peak	*thepeak = NULL, *newpeak, *peakhead = NULL;
  float	**frow, **ftop, **fbot, *fp, f;
  float Fp, Fpp;

  
  for (i = 1; i < N2 - 1; i++) {
    frow = fs + i;
		ftop = frow + 1;
		fbot = frow - 1;
		for (j = 1; j < N1 - 1; j++) {
		  fp = *frow + j;
		  f = *fp;
		  /* middle row */
		  if (*(fp-1) >= f)
		    continue;
		  if (*(fp+1) >= f)
		    continue;	
		  /* top row */
		  fp = *ftop + j;
		  if (*fp >= f)
		    continue;
		  if (*(fp-1) >= f)
		    continue;
		  if (*(fp+1) >= f)
		    continue;	
		  /* bottom row */
		  fp = *fbot + j;
		  if (*fp >= f)
		    continue;
		  if (*(fp-1) >= f)
		    continue;
		  if (*(fp+1) >= f)
		    continue;	
		  /*		  if (i<imin || i>=imax || j<jmin || j>=jmax) 
		    continue;
		    */
		  /* if it survived it is a peak */

                  ED_CALLOC(newpeak, "getpeaks", peak, 1);  
		  npeaks++;
		  newpeak->i = i;
		  newpeak->j = j;
		  newpeak->level = level;
		  newpeak->rf = rf;
		  newpeak->fs = fs[i][j] / normfactor;
		  newpeak->nu = fs[i][j] / sigma;
		  newpeak->e1 = 0.0;
		  newpeak->e2 = 0.0;

  		  newpeak->x[0] = newpeak->j + 0.5;
  		  newpeak->x[1] = newpeak->i + 0.5;
  		  Fp  = 0.5 * (fs[i][j+1] - fs[i][j-1]);
  		  Fpp = fs[i][j+1] - 2 * fs[i][j] + fs[i][j-1];
  		  if (fabs(Fpp) > TINY)
  		  {
  			newpeak->x[0] -= Fp / Fpp;
  		  }
  		  Fp  = 0.5 * (fs[i+1][j] - fs[i-1][j]);
  		  Fpp = fs[i+1][j] - 2 * fs[i][j] + fs[i-1][j];
  		  if (fabs(Fpp) > TINY)
  		  {
  			newpeak->x[1] -= Fp / Fpp;
  		  }

		  if (!peakhead)
		    peakhead = thepeak = newpeak;
		  else
		    thepeak = thepeak->next = newpeak;
		  		  
		}
  }
  return (peakhead);
}



int	install(peak *thepeak, line *linehead, float dmax)
{
	line	*theline;
	int	installed;
	
	installed = 0;
	theline = linehead;
	while (theline) {
		if (neighbours(theline->tail, thepeak, dmax)) {	/* add thepeak to the list */
			thepeak->next = NULL;
			theline->tail = theline->tail->next = thepeak;
			installed = 1;
			break;
		}
		theline = theline->cellmate;
	}
	return (installed);
}




void	addnewline(line **lineheadptr, peak *thepeak)
{
	line	*newline;

        ED_CALLOC(newline, "addnewline", line, 1);
	newline->head = newline->tail = thepeak;
	thepeak->next = NULL;
	if (*lineheadptr)
		(*lineheadptr)->prev = newline;
	newline->next = *lineheadptr;
	*lineheadptr = newline;
	nlines++;
}



int	neighbours(peak *peak1, peak *peak2, float dmax)
{
	double	di, dj, dd;
	
	di = peak1->i - peak2->i;
	dj = peak1->j - peak2->j;
	dd = di * di + dj * dj;
	return (dd <= dmax * dmax);
}

void	removedead(line **lineheadptr, int level, float nulimit)
{
	line	*theline, *nextline;
	
	theline = *lineheadptr;
	while (theline) {
		nextline = theline->next;
		if (theline->tail->level != level) {	/* a dead one */
			if (theline == *lineheadptr) {
				*lineheadptr = nextline;
			} else {
				theline->prev->next = nextline;
			}
			if (nextline)
				nextline->prev = theline->prev;
			disposeof(theline, nulimit);
		}
		theline = nextline;
	}
}



void	disposeof(line *theline, float nulimit)
{
  peak	*thepeak, *nextpeak;
  float	maxsig = 0.0;
  int jpeakmax = 0, ipeakmax = 0;
  int writeout = 0;
  
  thepeak = theline->head;
  while (thepeak) {
    nextpeak = thepeak->next;
    if (thepeak->nu > maxsig) 
    {
      maxsig = thepeak->nu;
      jpeakmax = thepeak->j;
      ipeakmax = thepeak->i;
    }
    thepeak = nextpeak;
  }

  if ((maxsig > nulimit) && (jpeakmax >=jmin) &&
      (jpeakmax <jmax) && (ipeakmax >=imin) && (ipeakmax <imax))
  {
    writeout = 1;
  }

  thepeak = theline->head;
  while (thepeak) {
    nextpeak = thepeak->next;
    if (writeout == 1) 
    {
      thepeak->j += nn1*blockstep1;
      thepeak->i += nn2*blockstep2;
      thepeak->x[0] += nn1*blockstep1;
      thepeak->x[1] += nn2*blockstep2;
      fwrite(thepeak, sizeof(peak), 1, stdout);
    }
    free(thepeak);
    npeaks--;
    thepeak = nextpeak;
  }
  free(theline);
  nlines--;
}




line	***makecell(line *linehead, float dmax, int ncells)
{
	line 	***cell, *theline;
	int	icell, jcell;
		
        ED_CALLOC(cell, "makecell", line **, ncells+10); 

	for (icell = 0; icell < ncells+10; icell++)
	{
          ED_CALLOC(cell[icell], "makecell", line *, ncells+10);
	}
	theline = linehead;
	
	while (theline) {
	  getcoords(theline->tail, &icell, &jcell, dmax);
	  if (icell < 0 || icell > ncells || jcell < 0 || jcell > ncells) {
	    fprintf(stderr, "whoops\n");
	    exit(-1);
	  }
	  theline->cellmate = cell[icell][jcell];
	  cell[icell][jcell] = theline;
	  theline = theline->next;
	}
	return (cell);
}



void	freecell(line ***cell, int ncells)
{
	int icell;
	for (icell = 0; icell < ncells+10; icell++) {
		free(cell[icell]);
	}
	free(cell);
}


void	getcoords(peak *thepeak, int *icell, int *jcell, float dmax)
{
	*icell = floor(thepeak->i / dmax);
	*jcell = floor(thepeak->j / dmax);
}


#define ZMAX	10

/*
 * mexican hat filter function 
 */
float	thefilterfunction(float ki, float kj)
{
	float	kk, z1, z2;
		
	kk = ki * ki + kj * kj;
	z1 = kk * sigma1;
	if (z1 > ZMAX)
		return (0.0);
	z2 = kk * sigma2;
	if (z2 > ZMAX) {
		return (exp(-z1));
	} else {
		return (exp(-z1) - exp(-z2));
	}
}

#undef ZMAX

