// Program to determine a weighting factor for galaxies 
// based on the corrected ellipticities. N next neighbours
// in a user defined space around a galaxy are considered
// and the sigma of these neighbours are calculated. This 
// is the weight later used.
//
#include <iostream>
#include <string.h>
#include <ctype.h>
#include "ANN.h"
#include "fitscat.h"
#include "tsize.h"
#include "utils_globals.h"
#include "tabutil_globals.h"

using namespace std;

// HISTORY:
// 22.10.1999: project started
//             
// TODO: check that all memory if freed at the end.
//
// 15.07.2001:
// included "tabutil_globals.h" to declare add_key_to tab
//
// 05.05.03:
// included the tsize.h file
//
// 10.01.05:
// I removed compiler warnings under gcc with optimisation
// enabled (-O3).
//
// 29.01.2005:
// I now add HISTORY information to the output catalog.
//
// 13.08.2008:
// I changed the inclusion of iostream.h to iostream (the first
// is no longer supported by gcc 4.3)

#define __PROGRAMMVERSION__ "1.0"
#define __PROGRAMMNAME__ "sigmae"

#define __USAGE__ "\n\
USAGE:\n\
       sigmae -i input catalog\n\
              -o output catalog\n\
              -c space coordinates\n\
            [ -e ellipticity quantities (e1, e2) ]\n\
            [ -t table with the objects (OBJECTS) ]\n\
            [ -s name for the Sigma key (Sigmae) ]\n\
            [ -n number of nearest neighbours (10) ]\n\n\
DESCRIPTION:\n\
EXAMPLE:\n\
  sigmae -i in.cat -o out.cat -c rg mag -e e1iso e2iso -n 20\n\n\
This would calculate Sigmae with the ellipticity quantities e1iso and\n\
e2iso using 20 nearest neighbours in the rg, mag plane.\n\n\
AUTHOR:\n\
      Thomas Erben\n"

#ifndef MAXSTRINGLENGTH
#define MAXSTRINGLENGTH 256
#endif


int main(int argc, char **argv)
{
  char incatname[MAXSTRINGLENGTH];
  char outcatname[MAXSTRINGLENGTH];
  char tablename[MAXSTRINGLENGTH]="OBJECTS";
  char e1keyname[MAXSTRINGLENGTH]="e1";
  char e2keyname[MAXSTRINGLENGTH]="e2";
  char sigmakeyname[MAXSTRINGLENGTH]="Sigmae";
  char tmpstring[MAXSTRINGLENGTH];

  char **keynames;
  int startcoords = 0;
  int ncoords=0;
 
  catstruct *incat;
  catstruct *outcat;
  tabstruct *tab;
  keystruct **keys;

  double **keyptr;
  double *e1, *e2;
  double *sigmae;

  // variables fro the next neighbour search
  int nobj;  
  int nneighbours = 10;
  ANNpointArray data_pts;
  ANNpoint      query_pt;
  ANNidxArray   nn_idx;
  ANNdistArray  dists;
  ANNkd_tree    *the_tree;

  // other variables
  int i, j;

  
  // read command line arguments
  for (i=1; i<argc; i++) 
  {
    if (argv[i][0] == '-') 
    {
       switch(tolower(argv[i][1])) 
       {
         case 'i': strncpy(incatname, argv[++i], MAXSTRINGLENGTH-1);
                   break;
         case 'o': strncpy(outcatname, argv[++i], MAXSTRINGLENGTH-1);
                   break;
         case 't': strncpy(tablename, argv[++i], MAXSTRINGLENGTH-1);
                   break;
         case 's': strncpy(sigmakeyname, argv[++i], MAXSTRINGLENGTH-1);
                   break;
         case 'n': nneighbours = atoi(argv[++i]);
                   break;
         case 'c': startcoords=i+1;
                   while (i+1<argc && argv[i+1][0] != '-') 
                   {
                      i++;
                      ncoords++;
                   }
                   break;
         case 'e': strncpy(e1keyname, argv[++i], MAXSTRINGLENGTH-1);
                   strncpy(e2keyname, argv[++i], MAXSTRINGLENGTH-1);
                   break;
 
         default:  
                   break;
       }
    } 
    else 
    {
       syserr(FATAL, "main", "Wrong Command line syntax\n");
    }
  }

  // print start message:
  cout << "\n" << __PROGRAMMNAME__ << "  V" << __PROGRAMMVERSION__ 
       << " (" << __DATE__ << ";" << __TIME__ << ")" << endl << endl;

  if(argc < 6)
  {
    cout << __USAGE__ << endl;
    exit(1);
  }

  if(ncoords < 1)
  {
    syserr(FATAL, "main", "no coordinate given\n");
  }

  // we need the corrdinate keys and the ellipticity keys
  ED_CALLOC(keynames, "main", char *, ncoords+2);
  for (i=0;i<ncoords+2;i++) 
  {
    ED_CALLOC(keynames[i], "main", char, MAXSTRINGLENGTH);
  }

  for(i=0; i< ncoords; i++)
  {
    strncpy(keynames[i], argv[startcoords+i], MAXSTRINGLENGTH-1);
  }

  strncpy(keynames[ncoords], e1keyname, MAXSTRINGLENGTH-1);
  strncpy(keynames[ncoords+1], e2keyname, MAXSTRINGLENGTH-1);

  // read catalog and quantities:
  if((incat=read_cat(incatname))==NULL)
  {
    syserr(FATAL, "main", "Error opening input catalog\n");
  }

  outcat=new_cat(1);
  inherit_cat(incat, outcat);
  copy_tabs(incat,outcat);
  
  if((tab=name_to_tab(outcat,tablename,0))==NULL)
  {
    syserr(FATAL, "main", "No %s table in catalog\n", tablename);
  }
  
  // read keys:
  ED_CALLOC(keys, "main", keystruct *, ncoords+2);
  ED_CALLOC(keyptr, "main", double *, ncoords+2);

  read_keys(tab, keynames, keys, ncoords+2, NULL);
  
  for(j=0; j<ncoords+2; j++)
  {
    if(!keys[j])
    {
      syserr(FATAL, "main", "No key %s in table\n", keynames[j]);
    }

    switch (keys[j]->ttype)
    {
    case T_FLOAT:
      ED_GETVEC(keyptr[j], "main", keys[j], 
                keys[j]->name, double, float);  
      break;
    case T_DOUBLE:
      ED_GETVEC(keyptr[j], "main", keys[j], 
                keys[j]->name, double, double);
      break;
    case T_LONG:
      ED_GETVEC(keyptr[j], "main", keys[j], 
                keys[j]->name, double, long);
      break;
    case T_SHORT:
      ED_GETVEC(keyptr[j], "main", keys[j], 
                keys[j]->name, double, short);
      break;
    default:
       syserr(FATAL, "main", "Unsupported key type key type in %s\n", keys[j]->name);        
      break;
    }
  }

  e1 = keyptr[ncoords];
  e2 = keyptr[ncoords+1];
  nobj = keys[0]->nobj;

  //for(i=0; i<nobj; i++)
  //{
  //  cout << keyptr[0][i] << " " << keyptr[1][i] << " " << e1[i] << " " << e2[i] << endl;
  //}

  // prepare stuff for the next neighbour search:
  query_pt = annAllocPt(ncoords);
  data_pts = annAllocPts(nobj, ncoords);
  nn_idx = new ANNidx[nneighbours];
  dists = new ANNdist[nneighbours];
  ED_CALLOC(sigmae, "main", double, nobj);
  
  // copy the coordinates into the required matrix structure
  // of the ANNpointArray

  for(i=0; i< nobj; i++)
  {
    for(j=0; j< ncoords; j++)
    {
      data_pts[i][j] = keyptr[j][i];
    }
  }

  the_tree = new ANNkd_tree((ANNpointArray)(data_pts), nobj, ncoords);

  for(i=0; i< nobj; i++)
  {
    for(j=0; j<ncoords; j++)
    {
      query_pt[j] = keyptr[j][i];
    }

    the_tree->annkSearch(query_pt, nneighbours, nn_idx, dists, 0.0);

    for(j=0; j< nneighbours; j++)
    {
      sigmae[i] += e1[nn_idx[j]]*e1[nn_idx[j]]+e2[nn_idx[j]]*e2[nn_idx[j]];
    }
    sigmae[i] = sqrt(sigmae[i]/nneighbours);

  }

  sprintf(tmpstring, "Variance of eps with %s and %s", e1keyname, e2keyname);
  add_key_to_tab(tab, sigmakeyname, nobj, sigmae, T_DOUBLE,
                        1, "", tmpstring);
  
  // free memory and bye
  ED_FREE(data_pts);
  delete [] nn_idx;
  delete [] dists;

  //annDeallocPt(query_pt);
  //annDeallocPts(data_pts);

  delete the_tree;

  for (i=0;i<ncoords;i++) 
  {
    ED_FREE(keynames[i]);
  }
  ED_FREE(keynames);

  /* add history information to the output catalog */
  historyto_cat(outcat, "sigmae", argc, argv);

  save_cat(outcat, outcatname);
  free_cat(incat, 1);
  free_cat(outcat, 1);
    
  return 0;
}  
