/* 
   08.08.00:
   added a pad_fits to the no_interp==YES case
   in save_fits_image. FITS images now have the correct 
   size.

   08.05.00:
   routine "read_fits_line":
   I made sure that the value given to the fseek function
   is of "long" type.

   16.04.2003:
   replaced all L_ENDIAN by BSWAP what is used in all
   the other modules to indicate low endian number
   representation

   19.04.2003:
   I made integers written to FITS headers right-justified
   (functions w_fits_short_c and w_fits_int_c)

   05.05.03:
   included the tsize.h file

   13.05.03:
   - substituted "long" and "unsigned long" by the LONG
     and ULONG macros
   - to all calls to the swapbyte function a wrong
     pointer was given for the data array 

   25.05.2004:
   routine END_line:  
   if the END keyword does not exist we give a warning and add
   one instead of leaving with a FATAL error. Our way
   of doing this assumes that the header->currline variable
   is at the end of the header (Is this safe always?).

   05.07.2004:
   added a pad_fits to the no_interp==NO case
   in save_fits_image. FITS images now have the correct 
   size also in this case (same change as for the
   no_interp==YES case on 08.08.2000).

   19.07.2004:
   - function d_fits:
     if a FITS keyword is deleted all the following keys
     are now moved one line up instead of filling the 
     deleted line with spaces.
   - I placed the '/' (separator between FITS keyword and
     comment for the SIMPLE keyword one space right.
   - I removed compiler warnings due to possibly 
     uninitialised variables.

   16.01.2005:
   - I removed a cast to a (void *) pointer in a pointer
     arithmetic expression (probably this gave no error
     as the original type of the pointer (char *) and
     (void *) have the same size.

   28.01.2005:
   - I added support for HISTORY and COMMENT keywords in
     w_fits_str_c
   - I added support for HISTORY and COMMENT keywords in
     r_fits_str. If no result is found, the NULL pointer
     is returned now.

   09.02.2005:
   routine write_fits_image:
   - the bzero value was not taken into account accurately when writing
     out FITS images.
   - For SHORT type images the swapbytes command was not at the
     correct place.

   30.12.2012:
   I removed compiler warnings on set but never used variables
*/

#ifndef TIFF_IO
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#include <errno.h>

#include "image_io_globals.h"
#include "fits_image_io_globals.h"
#include "utils_globals.h"
#include "casts_globals.h"
#include "decomp_globals.h"
#include "tsize.h"

#define MAXCHAR 255

int fits_find_pos;
static bool no_interp = NO;

void set_no_interp(bool val)
{
    no_interp = val;
}

/*
 * FITS header reading utilities
 */

char *fits_find(char *key, fitshead_struct *header)
{
    int i;

    for (i=0;i<header->nblocks*36;i++) 
    {
       if (!strncmp(&header->ptr[80*i],key,strlen(key))) 
       {
          fits_find_pos = i;
          return &header->ptr[80*i];
       }
    }
    return NULL;
} 

char *r_fits_bool(char *key, fitshead_struct *header)
{
    char *s, *tkey;
    int i;

    ED_CALLOC(s, "r_fits_bool", char, MAXCHAR);
    if ((tkey = fits_find(key,header))) {
       for (i=0;i<21;i++)
          if (tkey[10+i] != ' ')
             break;
       strncpy(s,&tkey[10+i], 21-i-1);
       s[21-i-1] = '\0';
    } else
       s[0] = '\0';
    return s;
}


int r_fits_int(char *key, fitshead_struct *header, int idef)
{
    char s[MAXCHAR], *tkey;
    int i;

    if ((tkey = fits_find(key,header))) 
      i = atoi(strncpy(s, &tkey[10], 21));
    else
      i = idef;
    return i;
}

int r_fits_short(char *key, fitshead_struct *header, short int idef)
{
    char s[MAXCHAR], *tkey;
    short int i;

    if ((tkey = fits_find(key,header))) 
      i = atoi(strncpy(s, &tkey[10], 21));
    else
      i = idef;
    return i;
}

double r_fits_float(char *key, fitshead_struct *header, double ddef)
{
    char *tkey;
    double f;
    int i;

    if((tkey = fits_find(key,header))) {
      for (i=0;i<21;i++) {
          if (tkey[10+i] == 'd' || tkey[10+i] == 'D') {
             tkey[10+i] = 'E';
          }
      }
      f = atof(tkey+10);
    } else {
      f = ddef;
    }
    return f;
}

char *r_fits_str(char *key, fitshead_struct *header)
{
    char *s, *tkey;
    int i, j;

    if ((tkey = fits_find(key,header))) 
    {
       ED_CALLOC(s, "r_fits_str", char, MAXCHAR);
       /* deal with the special COMMENT and HISTORY
          keyywords */
       if(strncmp(key, "HISTORY", 7) == 0 || strncmp(key, "COMMENT", 7) == 0 )
       {
	   strncpy(s, &tkey[8], 70);
       }
       else
       {
         for (i=0;i<21;i++)
            if (tkey[10+i] != '\'' && tkey[10+i] != ' ')
               break;
         for (j=19;j>0;j--)
            if (tkey[10+j] != '\'' && tkey[10+j] != ' ')
               break;
         strncpy(s,&tkey[10+i], j-i+1);
         s[j-i+1] = '\0';
       }
    } else
       s = NULL;
    return s;
}
/*
 * FITS header writing utilities
 */

static void w_fits_check(fitshead_struct *header)
{
    if (header->curline == (FITSBLKSIZ/FITSCARDLEN) * header->nblocks) {
       header->nblocks++;
       ED_REALLOC(header->ptr, "w_fits_check", char, 
                     header->nblocks * FITSBLKSIZ);
    }
}

int d_fits(char *key, fitshead_struct *header)
{
    char	*tkey;

    if ((tkey = fits_find(key, header)))
    {
      memmove(header->ptr+fits_find_pos*FITSCARDLEN,
              header->ptr+(fits_find_pos+1)*FITSCARDLEN,
              (header->curline-fits_find_pos)*FITSCARDLEN);

      if ((header->curline-1) == (FITSBLKSIZ/FITSCARDLEN) * header->nblocks) 
      {
         header->nblocks--;
         ED_REALLOC(header->ptr, "w_fits_check", char, 
                     header->nblocks * FITSBLKSIZ);
      }      
      header->curline--;
    }
    return 0;
}

int w_fits_bool_c(char *key, fitshead_struct *header, int value, char *comment)
{
    int		i;
    char	s[MAXCHAR];

    w_fits_check(header);
    for (i=0;i<FITSCARDLEN;i++) header->ptr[header->curline*FITSCARDLEN+i]=' ';
    strncpy(&header->ptr[header->curline*FITSCARDLEN],key,strlen(key));
    header->ptr[header->curline*FITSCARDLEN+8]='=';
    if (value) {
       sprintf(s,"T");
    } else {
       sprintf(s,"F");
    }
    i = strlen(s);
    strncpy(&header->ptr[header->curline*FITSCARDLEN+30-i],s,i);
    header->ptr[header->curline*FITSCARDLEN+32]='/';
    header->ptr[header->curline*FITSCARDLEN+33]=' ';
    if (comment != NULL) {
       i = strlen(comment);
       if (i>45) i = 45;
       strncpy(&header->ptr[header->curline*FITSCARDLEN+34],comment,i);
    }
    header->curline++;

    return 0;
}

int w_fits_bool(char *key, fitshead_struct *header, int value)
{
    return(w_fits_bool_c(key, header, value, NULL));
}

int s_fits_bool_c(char *key, fitshead_struct *header, int value, char
*comment)
{
    int         save_currline, retcode = FALSE;
    char        *here;

    if ((here = fits_find(key, header)) != NULL) {
       save_currline = header->curline;
       header->curline = fits_find_pos;
       w_fits_bool_c(key, header, value, comment);
       header->curline = save_currline;
       retcode = TRUE;
    } else {
       w_fits_bool_c(key, header, value, comment);
    }
    return retcode;
}

int s_fits_bool(char *key, fitshead_struct *header, int value)
{
    return(s_fits_bool_c(key, header, value, NULL));
}

int w_fits_int_c(char *key, fitshead_struct *header, int value, char *comment)
{
    int i;
    char s[MAXCHAR];

    w_fits_check(header);
    for (i=0;i<FITSCARDLEN;i++) header->ptr[header->curline*FITSCARDLEN+i]=' ';
    strncpy(&header->ptr[header->curline*FITSCARDLEN],key,strlen(key));
    header->ptr[header->curline*FITSCARDLEN+8]='=';
    sprintf(s,"%20d",value);
    i = strlen(s);
    strncpy(&header->ptr[header->curline*FITSCARDLEN+30-i],s,i);
    header->ptr[header->curline*FITSCARDLEN+32]='/';
    header->ptr[header->curline*FITSCARDLEN+33]=' ';
    if (comment != NULL) {
       i = strlen(comment);
       if (i>45) i = 45;
       strncpy(&header->ptr[header->curline*FITSCARDLEN+34],comment,i);
    }
    header->curline++;

    return 0;
}

int w_fits_int(char *key, fitshead_struct *header, int value)
{
    return(w_fits_int_c(key, header, value, NULL));
}

int s_fits_int_c(char *key, fitshead_struct *header, int value, char
*comment)
{
    int         save_currline, retcode = FALSE;
    char        *here;

    if ((here = fits_find(key, header)) != NULL) {
       save_currline = header->curline;
       header->curline = fits_find_pos;
       w_fits_int_c(key, header, value, comment);
       header->curline = save_currline;
       retcode = TRUE;
    } else {
       w_fits_int_c(key, header, value,comment);
    }
    return retcode;
}

int s_fits_int(char *key, fitshead_struct *header, int value)
{
    return(s_fits_int_c(key, header, value, NULL));
}


int w_fits_short_c(char *key, fitshead_struct *header, short int value, char *comment)
{
    short int i;
    char s[MAXCHAR];

    w_fits_check(header);
    for (i=0;i<FITSCARDLEN;i++) header->ptr[header->curline*FITSCARDLEN+i]=' ';
    strncpy(&header->ptr[header->curline*FITSCARDLEN],key,strlen(key));
    header->ptr[header->curline*FITSCARDLEN+8]='=';
    sprintf(s,"%20d",value);
    i = strlen(s);
    strncpy(&header->ptr[header->curline*FITSCARDLEN+30-i],s,i);
    header->ptr[header->curline*FITSCARDLEN+32]='/';
    header->ptr[header->curline*FITSCARDLEN+33]=' ';
    if (comment != NULL) {
       i = strlen(comment);
       if (i>45) i = 45;
       strncpy(&header->ptr[header->curline*FITSCARDLEN+34],comment,i);
    }
    header->curline++;

    return 0;
}

int w_fits_short(char *key, fitshead_struct *header, short int value)
{
    return(w_fits_int_c(key, header, value, NULL));
}

int s_fits_short_c(char *key, fitshead_struct *header, short int value,
                   char *comment)
{
    int         save_currline, retcode = FALSE;
    char        *here;

    if ((here = fits_find(key, header)) != NULL) {
       save_currline = header->curline;
       header->curline = fits_find_pos;
       w_fits_short_c(key, header, value, comment);
       header->curline = save_currline;
       retcode = TRUE;
    } else {
       w_fits_short_c(key, header, value, comment);
    }
    return retcode;
}

int s_fits_short(char *key, fitshead_struct *header, short int value)
{
    return(s_fits_int_c(key, header, value, NULL));
}

int w_fits_double_c(char *key, fitshead_struct *header, double value,
                    char *comment)
{
    int i;
    char s[MAXCHAR];

    w_fits_check(header);
    for (i=0;i<FITSCARDLEN;i++) header->ptr[header->curline*FITSCARDLEN+i]=' ';
    strncpy(&header->ptr[header->curline*FITSCARDLEN],key,strlen(key));
    header->ptr[header->curline*FITSCARDLEN+8]='=';
    if ((fabs(value) - (double) ((int)fabs(value))) < 1.0E-24) {
       sprintf(s,"%8.12G.",value);
    } else {
       sprintf(s,"%8.12G",value);
    }
    i = strlen(s);
    strncpy(&header->ptr[header->curline*FITSCARDLEN+30-i],s,i);
    header->ptr[header->curline*FITSCARDLEN+32]='/';
    header->ptr[header->curline*FITSCARDLEN+33]=' ';
    if (comment != NULL) {
       i = strlen(comment);
       if (i>45) i = 45;
       strncpy(&header->ptr[header->curline*FITSCARDLEN+34],comment,i);
    }
    header->curline++;

    return 0;
}

int w_fits_double(char *key, fitshead_struct *header, double value)
{
    return (w_fits_double_c(key, header, value, NULL));
}

int s_fits_double_c(char *key, fitshead_struct *header, double value, char
*comment)
{
    int		save_currline, retcode = FALSE;
    char	*here;

    if ((here = fits_find(key, header)) != NULL) {
       save_currline = header->curline;
       header->curline = fits_find_pos;
       w_fits_double_c(key, header, value, comment);
       header->curline = save_currline;
       retcode = TRUE;
    } else {
       w_fits_double_c(key, header, value, comment);
    }
    return retcode;
}
int s_fits_double(char *key, fitshead_struct *header, double value)
{
    return (s_fits_double_c(key, header, value, NULL));
}

int w_fits_float_c(char *key, fitshead_struct *header, float value, 
                   char *comment)
{
    double	val = value;

    return (w_fits_double_c(key, header, val, comment));
}

int w_fits_float(char *key, fitshead_struct *header, float value)
{
    return (w_fits_float_c(key, header, value, NULL));
}

int s_fits_float(char *key, fitshead_struct *header, float value)
{
    double	val = value;

    return (s_fits_double(key, header, val));
}
int s_fits_float_c(char *key, fitshead_struct *header, float value, char
*comment)
{
    double	val = value;

    return (s_fits_double_c(key, header, val, comment));
}

int w_fits_str_c(char *key, fitshead_struct *header, char *value,
                 char *comment)
{
    int i;

    w_fits_check(header);
    for (i=0;i<FITSCARDLEN;i++) header->ptr[header->curline*FITSCARDLEN+i]=' ';
    strncpy(&header->ptr[header->curline*FITSCARDLEN],key,strlen(key));
    if (value != NULL) 
    {
       /* deal with the special case of HISTORY and COMMENT
          keywords */
       if(strncmp(key, "HISTORY", 7)==0 || strncmp(key, "COMMENT", 7)==0)
       {
         strncpy(&header->ptr[header->curline*FITSCARDLEN+8],value,
                 MIN(70,strlen(value)));
       }
       else
       {
         header->ptr[header->curline*FITSCARDLEN+8]='=';
         i = strlen(value);
         strncpy(&header->ptr[header->curline*FITSCARDLEN+30-i],value,i);
         header->ptr[header->curline*FITSCARDLEN+32]='/';
         header->ptr[header->curline*FITSCARDLEN+33]=' ';
       }
    }
    if (comment != NULL) {
       i = strlen(comment);
       if (i>45) i = 45;
       strncpy(&header->ptr[header->curline*FITSCARDLEN+34],comment,i);
    }
    header->curline++;
    return 0;
}

int w_fits_str(char *key, fitshead_struct *header, char *value)
{
    return(w_fits_str_c(key, header, value, NULL));
}

void w_fits_pad(fitshead_struct *header)
{
    int i;

    w_fits_check(header);
    for (i=0;i<FITSCARDLEN;i++) header->ptr[header->curline*FITSCARDLEN+i]=' ';
    strncpy(&header->ptr[header->curline*FITSCARDLEN],"END",3);
    header->curline++;
    while (header->curline < (FITSBLKSIZ/FITSCARDLEN) * header->nblocks) {
       for (i=0;i<FITSCARDLEN;i++) 
          header->ptr[header->curline*FITSCARDLEN+i]=' ';
       header->curline++;
    }
}       


unsigned char *alloc_fits_map(int x, int y, t_type pixtype)
{
    LONG nbytes;
    unsigned char	*d;

    nbytes  = x * y * t_size[pixtype];
    ED_MALLOC(d, "alloc_fits_map", unsigned char, nbytes);

    return d;
}

/*
 * Other FITS utility routines
 */

void END_line(fitshead_struct *header)
{
    if (fits_find("END     ",header) != NULL)
    {
       header->curline = fits_find_pos;
    } 
    else 
    {
       /* Probably we assume here that we are at the end of the file
          with header->curline (TE: 25.05.2004)
       */
       syserr(WARNING,"END_line", "No END line in image header - adding one\n");
       w_fits_pad(header);
    }
}

int interp_fits_header(fitsfile_struct *ffs)
{
/****** interp_fits_header ***************************************************
 *
 *  NAME	
 *      interp_fits_header - Process information from the FITS header
 *
 *  SYNOPSIS 
 *      int interp_fits_header(fitsfile_struct ffs)
 *
 *  FUNCTION
 *      This routine interprets the textual information in the provided
 *      fits header into the structure elements. It requires the following
 *      FITS keywords to be present in the header of the image:
 *      SIMPLE|XTENSION, BITPIX, NAXIS, CRVAL1, CRVAL2, CRPIX1, CRPIX2, 
 *      CDELT1, CDELT2, and BSCALE, BZERO. When some of these keywaords
 *      are not present reasonable default values are provided.
 *
 *  INPUTS
 *      ffs - The fitsfile structure with a filled header pointer
 *
 *  RESULTS
 *      ffs - The elements of the structure are filled 
 *
 *  RETURNS
 *      TRUE  - Upon successful completeion
 *      FALSE - When a failure occurs
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *      Reasonable default values may be useless for further processing.
 *
 *  AUTHOR
 *      E.R. Deul - dd 26-11-1996
 *
 ******************************************************************************/
    int		compressed = FALSE;

    if (strcmp(r_fits_bool("SIMPLE", ffs->head),"T") != 0) {
       if (strcmp(r_fits_str("XTENSION", ffs->head),"COMPRESSED") != 0) {
          syserr(FATAL,"interp_fits_header",
                       "File contains no SIMPLE, nor XTENSION keywords\n");
       } else {
          compressed = TRUE;
       }
    } else {
       if (r_fits_int("NAXIS   ",ffs->head,0) == 0) {
/*
 *        Allow for the case when the main header is just the main header
 *        with an extention header containing the real information
 */
          return FALSE;
       }
       compressed = FALSE;
    }
    if (r_fits_int("NAXIS   ",ffs->head,0) != 2)
       syserr(FATAL,"interp_fits_header",
                    "Can handle only 2 dimensional FITS images\n");
    ffs->bitpix = r_fits_int  ("BITPIX  ",ffs->head,16);
    ffs->bit_sign=r_fits_int  ("BITSGN  ",ffs->head,1);
    ffs->xwid   = r_fits_int  ("NAXIS1  ",ffs->head,0);
    ffs->ywid   = r_fits_int  ("NAXIS2  ",ffs->head,0);
    ffs->xcen   = r_fits_float("CRVAL1  ",ffs->head,0.0);
    ffs->ycen   = r_fits_float("CRVAL2  ",ffs->head,0.0);
    ffs->bscale = r_fits_float("BSCALE  ",ffs->head,1.0);
    ffs->bzero  = r_fits_float("BZERO   ",ffs->head,0.0);
    ffs->px0    = (int)r_fits_float("CRPIX1  ",ffs->head,(float)ffs->xwid/2);
    ffs->py0    = (int)r_fits_float("CRPIX2  ",ffs->head,(float)ffs->ywid/2);
    ffs->dx     = r_fits_float("CDELT1  ",ffs->head,1.0);
    ffs->dy     = r_fits_float("CDELT2  ",ffs->head,1.0);
/*
 *  Check DENIS specific compression or not
 */
    if (compressed) {
       if (strcmp(r_fits_str("IMAGECOD",ffs->head),"BASEBYTE") == 0) {
          ffs->comp_type = BASE_BYTE;
       }
       if (strcmp(r_fits_str("IMAGECOD",ffs->head),"PREVPIX") == 0) {
          ffs->comp_type = PREV_PIX;
       }
       ffs->code_len = r_fits_int("IMAGELEN",ffs->head,0);
       if (ffs->comp_type == 0) {
          syserr(FATAL,"interp_fits_header",
                       "Cannot uncompress this type of image\n");
       }
    } else {
       ffs->comp_type = NO_COMP;
    }

    END_line(ffs->head);
    return TRUE;
}


FILE *read_fits_header(fitsfile_struct *ff)
{
/****** read_fits_header ***************************************************
 *
 *  NAME	
 *     read_fits_header - Read the primary header blocks 
 *
 *  SYNOPSIS
 *      int read_fits_header
 *
 *  FUNCTION
 *      This routine reads the primary FITS header blocks and puts them
 *      into the structure. The file will be scanned until an image header
 *      is detected.
 *
 *  INPUTS
 *      The structure with an opened file handle
 *
 *  RESULTS
 *      The structure now has the full header info in the pointer
 *
 *  RETURNS
 *      TRUE  - When operation was successful
 *      FALSE - In case of failure
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 25-11-1996
 *
 ******************************************************************************/
    int i, done;

    done = FALSE;
    while (!done) {
       ED_FREE(ff->head->ptr);
       ED_MALLOC(ff->head->ptr, "read_fits_header", char, FITSBLKSIZ);
       if (fread(ff->head->ptr,1,FITSBLKSIZ,ff->fd) != FITSBLKSIZ)
          syserr(FATAL,"read_fits_header","Error reading fits file header\n");
/*
 *     Read required number of fits headers
 */
       ff->head->nblocks = 1;
       for (i=1; !fits_find("END     ",ff->head); i++) {
          ED_REALLOC(ff->head->ptr,"read_fits_header", char, (i+1)*FITSBLKSIZ);
          if (fread(&ff->head->ptr[i*FITSBLKSIZ],1,FITSBLKSIZ,ff->fd) != 
                      FITSBLKSIZ)
            syserr(FATAL,"read_fits_header","Error reading fits file header\n");
          ff->head->nblocks++;
       }
       ff->head->curpos = 0;
       done = interp_fits_header(ff);
    }
    return ff->fd;
}


/*
 *  More global FITS routines 
 */


fitsfile_struct *open_fits_file(char *file)
/****** open_fits_file ***************************************************
 *
 *  NAME	
 *      open_fits_file - Open a FITS image file
 *
 *  SYNOPSIS
 *      fitsfile_struct *open_fits_file(char *file)
 *
 *  FUNCTION
 *      This routine opens a FITS image file and reads the appropriate
 *      header information. After sucessfull completion it returns the
 *      a file handle filled with information
 *
 *  INPUTS
 *      file - The name of the fitsfile
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      The fits file handle is returns on success. On failure a NULL
 *      pointer is returned
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R Deul - 25-11-1996
 *
 ******************************************************************************/
{
    fitsfile_struct *ff;

    ff = new_fits();

    VPRINTF(NORMAL,"Reading file %s\n",file);
    ED_FOPEN(ff->fd, "read_fits_header", file, "r");
    if (ff->fd ==  NULL)
       syserr(FATAL, "open_fits_file", "Could not open FITS file %s\n",
          file);
    strcpy(ff->name, file);
    read_fits_header(ff);

    return ff;
}

void read_fits_line(fitsfile_struct *ff, unsigned char *map, int line, 
                    bool invert)
{
/****** read_fits_line ***************************************************
 *
 *  NAME	
 *      read_fits_line - Read a line of data and convert types
 *
 *  SYNOPSIS
 *      void read_fits_line(fitsfile_struct *ff, char *map, int line,
 *                          bool invert)
 *
 *  FUNCTION
 *      This routine reads a line of image data and convert the FITS pixel
 *      representation to the internal map representation. The conversion
 *      is done form any type to any type.
 *
 *  INPUTS
 *      ff     - The fitsfile structure describing the image
 *      line   - The line number to read
 *      invert - Invert bytes
 *
 *  RESULTS
 *      map - A pointer to the start of the line to store the read result
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 26-11-1996
 *
 ******************************************************************************/
    short int	pix_size = (short)(fabs(ff->bitpix)/8);
    unsigned char	*b, *b0;
    int		i, nx = ff->xwid;
    double	bscale = ff->bscale, bzero = ff->bzero;

    ED_MALLOC(b, "read_fits_line", unsigned char, nx * pix_size);
    b0 = b;

    ED_FSEEK(ff->fd, "read_fits_line", 
           (LONG)((LONG)pix_size*(LONG)nx*(LONG)line+
                  (LONG)ff->head->nblocks*FITSBLKSIZ));
    ED_FREAD(b, "read_fits_line", ff->fd, pix_size*nx);
    switch(ff->pix_type) {
    case T_BYTE: 
       switch (ff->bitpix) {
       case PIX_BYTE:
          for (i=0;i<nx;i++,map++,b++) {
             char aa = *b * bscale + bzero;
             memcpy(map, &aa, sizeof(char));
          }
          break;
       case PIX_SHORT: 
          for (i=0;i<nx;i++,map++,b+=sizeof(short)) {
             char aa = (char) short_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(char));
          }
          break;
       case PIX_LONG:
          for (i=0;i<nx;i++,map++,b+=sizeof(LONG)) {
             char aa = (char) int_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(char));
          }
          break;
       case PIX_FLOAT:
          for (i=0;i<nx;i++,map++,b+=sizeof(float)) {
             char aa = (char) float_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(char));
          }
          break;
       case PIX_DOUBLE:
          for (i=0;i<nx;i++,map++,b+=sizeof(double)) {
             char aa = (char) double_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(char));
          }
          break;
       }
       break;
    case T_SHORT : 
/*
 *     The processing of short type of pixels
 */
       switch (ff->bitpix) {
       case PIX_BYTE:
          for (i=0;i<nx;i++,map+=sizeof(short),b++) {
             short aa = (short) *b * bscale + bzero;
             memcpy(map, &aa, sizeof(short));
          }
          break;
       case PIX_SHORT: 
          for (i=0;i<nx;i++,map+=sizeof(short),b+=sizeof(short)) {
             short aa = short_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(short));
          }
          break;
       case PIX_LONG:
          for (i=0;i<nx;i++,map+=sizeof(short),b+=sizeof(LONG)) {
             short aa = (short) int_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(short));
          }
          break;
       case PIX_FLOAT:
          for (i=0;i<nx;i++,map+=sizeof(short),b+=sizeof(float)) {
             short aa = (short) float_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(short));
          }
          break;
       case PIX_DOUBLE:
          for (i=0;i<nx;i++,map+=sizeof(short),b+=sizeof(double)) {
             short aa = (short) double_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(short));
          }
          break;
       }
       break;
    case T_LONG  : 
/*
 *     The processing of long type of pixels
 */
       switch (ff->bitpix) {
       case PIX_BYTE:
          for (i=0;i<nx;i++,map+=sizeof(LONG),b++) {
             LONG aa = (LONG) *b * bscale + bzero;
             memcpy(map, &aa, sizeof(LONG));
          }
          break;
       case PIX_SHORT: 
          for (i=0;i<nx;i++, map+=sizeof(LONG),b+=sizeof(short)) {
             LONG aa = (LONG) short_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(LONG));
          }
          break;
       case PIX_LONG:
          for (i=0;i<nx;i++,map+=sizeof(LONG),b+=sizeof(LONG)) {
             LONG aa = int_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(LONG));
          }
          break;
       case PIX_FLOAT:
          for (i=0;i<nx;i++,map+=sizeof(LONG),b+=sizeof(float)) {
             LONG aa = (LONG) float_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(LONG));
          }
          break;
       case PIX_DOUBLE:
          for (i=0;i<nx;i++,map+=sizeof(LONG),b+=sizeof(double)) {
             LONG aa = (LONG) double_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(LONG));
          }
          break;
       }
       break;
    case T_FLOAT : 
/*
 *     The processing of float type of pixels
 */
       switch (ff->bitpix) {
       case PIX_BYTE:
          if (invert == YES) {
             for (i=0;i<nx;i++) b[i] = b[i] ^ 0xFF;
          }
          for (i=0;i<nx;i++,map+=sizeof(float),b++) {
             float aa = ((float) *b) * bscale + bzero;
             memcpy(map, &aa, sizeof(float));
          }
          break;
       case PIX_SHORT: 
          for (i=0;i<nx;i++, map+=sizeof(float),b+=sizeof(short)) {
             float aa = (float) short_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(float));
          }
          break;
       case PIX_LONG:
          for (i=0;i<nx;i++, map+=sizeof(float),b+=sizeof(LONG)) {
             float aa = (float) int_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(float));
          }
          break;
       case PIX_FLOAT:
          for (i=0;i<nx;i++, map+=sizeof(float),b+=sizeof(float)) {
             float aa = float_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(float));
          }
          break;
       case PIX_DOUBLE:
          for (i=0;i<nx;i++, map+=sizeof(float),b+=sizeof(double)) {
             float aa = (float) double_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(float));
          }
          break;
       }
       break;
    case T_DOUBLE: 
/*
 *     The processing of double type of pixels
 */
       switch (ff->bitpix) {
       case PIX_BYTE:
          for (i=0;i<nx;i++,map+=sizeof(double),b++) {
             double aa = (double) *b * bscale + bzero;
             memcpy(map, &aa, sizeof(double));
          }
          break;
       case PIX_SHORT: 
          for (i=0;i<nx;i++, map+=sizeof(double),b+=sizeof(short)) {
             double aa = (double) short_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(double));
          }
          break;
       case PIX_LONG:
          for (i=0;i<nx;i++, map+=sizeof(double),b+=sizeof(LONG)) {
             double aa = (double) int_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(double));
          }
          break;
       case PIX_FLOAT:
          for (i=0;i<nx;i++, map+=sizeof(double),b+=sizeof(float)) {
             double aa = (double) float_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(double));
          }
          break;
       case PIX_DOUBLE:
          for (i=0;i<nx;i++, map+=sizeof(double),b+=sizeof(double)) {
             double aa = double_cast(b) * bscale + bzero;
             memcpy(map, &aa, sizeof(double));
          }
          break;
       }
       break;
    default:
         syserr(FATAL,"read_fits_line","Unknown BITPIX type!\n");
    }
    ED_FREE(b0);
}
/**/
void read_fits_image(fitsfile_struct *ff, unsigned char *map)
{
/****** read_fits_image ***************************************************
 *
 *  NAME	
 *      read_fits_image - Read data and convert types
 *
 *  SYNOPSIS
 *      void read_fits_image(fitsfile_struct *ff, char *map)
 *
 *  FUNCTION
 *      This routine reads image data and convert the FITS pixel
 *      representation to the internal map representation. The conversion
 *      is done form any type to any type.
 *
 *  INPUTS
 *      ff     - The fitsfile structure describing the image
 *
 *  RESULTS
 *      map - A pointer to the start of the line to store the read result
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 26-11-1996
 *
 ******************************************************************************/
    short int	pix_size = (short)(fabs(ff->bitpix)/8);
    unsigned char	*b, *b0;
    unsigned char	*bptr;
    short int		*sptr;
    int			*lptr;
    float		*fptr;
    double		*dptr;
    int			i, nx = ff->xwid, ny = ff->ywid;
    int			scoop = nx*ny;
    double		bs = ff->bscale, bz = ff->bzero;

    ED_MALLOC(b, "read_fits_line", unsigned char, scoop * pix_size);
    b0=b;
    ED_FSEEK(ff->fd, "read_fits_line", ff->head->nblocks*FITSBLKSIZ);
    ED_FREAD(b, "read_fits_line", ff->fd, (scoop * pix_size));
    switch(ff->pix_type) {
    case T_BYTE: 
      bptr = map;
      switch (ff->bitpix) {
        {
        case PIX_BYTE:
          if (ff->bit_sign)
            for (i=scoop; i--;) *(bptr++) = *(b++)*bs + bz;
          else
            for (i=scoop; i--;) *(bptr++) = *((unsigned char *)b++)*bs + bz;
          break;
        case PIX_SHORT:
#ifdef BSWAP
          swapbytes(b0, 2, scoop);
#endif
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(short))
              *(bptr++) = *((short *)b)*bs + bz;
          else
            for (i=scoop; i--; b += sizeof(unsigned short))
              *(bptr++) = *((unsigned short *)b)*bs + bz;
          break;
        case PIX_LONG:
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(LONG))
              *(bptr++) = *((LONG *)b)*bs + bz;
          else
            for (i=scoop; i--; b += sizeof(ULONG))
              *(bptr++) = *((ULONG *)b)*bs + bz;
            break;
        case PIX_FLOAT:
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          for (i=scoop; i--; b += sizeof(float))
            *(bptr++) = *((float *)b)*bs + bz;
          break;
        case PIX_DOUBLE:
#ifdef BSWAP
          swapbytes(b0, 8, scoop);
#endif
          for (i=scoop; i--; b += sizeof(double))
            *(bptr++) = *((double *)b)*bs + bz;
          break;
        default:
          syserr(FATAL,"read_fits_image","Unknown BITPIX type!\n");
          break;
        }
      }
      break;
    case T_SHORT : 
/*
 *     The processing of short type of pixels
 */
      sptr = (short *)map;
      switch (ff->bitpix) {
        {
        case PIX_BYTE:
          if (ff->bit_sign)
            for (i=scoop; i--;) *(sptr++) = *(b++)*bs + bz;
          else
            for (i=scoop; i--;) *(sptr++) = *((unsigned char *)b++)*bs + bz;
          break;
        case PIX_SHORT:
#ifdef BSWAP
          swapbytes(b0, 2, scoop);
#endif
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(short))
              *(sptr++) = *((short *)b)*bs + bz;
          else
            for (i=scoop; i--; b += sizeof(unsigned short))
              *(sptr++) = *((unsigned short *)b)*bs + bz;
          break;
        case PIX_LONG:
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(LONG))
              *(sptr++) = *((LONG *)b)*bs + bz;
          else
            for (i=scoop; i--; b += sizeof(ULONG))
              *(sptr++) = *((ULONG *)b)*bs + bz;
            break;
        case PIX_FLOAT:
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          for (i=scoop; i--; b += sizeof(float))
            *(sptr++) = *((float *)b)*bs + bz;
          break;
        case PIX_DOUBLE:
#ifdef BSWAP
          swapbytes(b0, 8, scoop);
#endif
          for (i=scoop; i--; b += sizeof(double))
            *(sptr++) = *((double *)b)*bs + bz;
          break;
        default:
          syserr(FATAL,"read_fits_image","Unknown BITPIX type!\n");
          break;
        }
      }
      break;
    case T_LONG  : 
/*
 *     The processing of long type of pixels
 */
      lptr = (int *)map;
      switch (ff->bitpix) {
        {
        case PIX_BYTE:
          if (ff->bit_sign)
            for (i=scoop; i--;) *(lptr++) = *(b++)*bs + bz;
          else
            for (i=scoop; i--;) *(lptr++) = *((unsigned char *)b++)*bs + bz;
          break;
        case PIX_SHORT:
#ifdef BSWAP
          swapbytes(b0, 2, scoop);
#endif
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(short))
              *(lptr++) = *((short *)b)*bs + bz;
          else
            for (i=scoop; i--; b += sizeof(unsigned short))
              *(lptr++) = *((unsigned short *)b)*bs + bz;
          break;
        case PIX_LONG:
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(LONG))
              *(lptr++) = *((LONG *)b)*bs + bz;
          else
            for (i=scoop; i--; b += sizeof(ULONG))
              *(lptr++) = *((ULONG *)b)*bs + bz;
            break;
        case PIX_FLOAT:
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          for (i=scoop; i--; b += sizeof(float))
            *(lptr++) = *((float *)b)*bs + bz;
          break;
        case PIX_DOUBLE:
#ifdef BSWAP
          swapbytes(b0, 8, scoop);
#endif
          for (i=scoop; i--; b += sizeof(double))
            *(lptr++) = *((double *)b)*bs + bz;
          break;
        default:
          syserr(FATAL,"read_fits_image","Unknown BITPIX type!\n");
          break;
        }
      }
      break;
    case T_FLOAT : 
/*
 *     The processing of float type of pixels
 */
      fptr = (float *)map;
      switch (ff->bitpix) {
        {
        case PIX_BYTE:
          if (ff->bit_sign)
            for (i=scoop; i--;) *(fptr++) = *(b++)*bs + bz;
          else
            for (i=scoop; i--;) *(fptr++) = *((unsigned char *)b++)*bs + bz;
          break;
        case PIX_SHORT:
#ifdef BSWAP
          swapbytes(b0, 2, scoop);
#endif
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(short))
              *(fptr++) = *((short *)b)*bs + bz;
          else
            for (i=scoop; i--; b += sizeof(unsigned short))
              *(fptr++) = *((unsigned short *)b)*bs + bz;
          break;
        case PIX_LONG:
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(LONG))
              *(fptr++) = *((LONG *)b)*bs + bz;
          else
            for (i=scoop; i--; b += sizeof(ULONG))
              *(fptr++) = *((ULONG *)b)*bs + bz;
            break;
        case PIX_FLOAT:
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          for (i=scoop; i--; b += sizeof(float))
            *(fptr++) = *((float *)b)*bs + bz;
          break;
        case PIX_DOUBLE:
#ifdef BSWAP
          swapbytes(b0, 8, scoop);
#endif
          for (i=scoop; i--; b += sizeof(double))
            *(fptr++) = *((double *)b)*bs + bz;
          break;
        default:
          syserr(FATAL,"read_fits_image","Unknown BITPIX type!\n");
          break;
        }
      }
      break;
    case T_DOUBLE: 
/*
 *     The processing of double type of pixels
 */
      dptr = (double *)map;
      switch (ff->bitpix) {
        {
        case PIX_BYTE:
          if (ff->bit_sign)
            for (i=scoop; i--;) *(dptr++) = *(b++)*bs + bz;
          else
            for (i=scoop; i--;) *(dptr++) = *((unsigned char *)b++)*bs + bz;
          break;
        case PIX_SHORT:
#ifdef BSWAP
          swapbytes(b0, 2, scoop);
#endif
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(short))
              *(dptr++) = *((short *)b)*bs + bz;
          else
            for (i=scoop; i--; b += sizeof(unsigned short))
              *(dptr++) = *((unsigned short *)b)*bs + bz;
          break;
        case PIX_LONG:
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(LONG))
              *(dptr++) = *((LONG *)b)*bs + bz;
          else
            for (i=scoop; i--; b += sizeof(ULONG))
              *(dptr++) = *((ULONG *)b)*bs + bz;
            break;
        case PIX_FLOAT:
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          for (i=scoop; i--; b += sizeof(float))
            *(dptr++) = *((float *)b)*bs + bz;
          break;
        case PIX_DOUBLE:
#ifdef BSWAP
          swapbytes(b0, 8, scoop);
#endif
          for (i=scoop; i--; b += sizeof(double))
            *(dptr++) = *((double *)b)*bs + bz;
          break;
        default:
          syserr(FATAL,"read_fits_image","Unknown BITPIX type!\n");
          break;
        }
      }
      break;
    default:
      syserr(FATAL,"read_fits_image","Unknown BITPIX type!\n");
    }
    ED_FREE(b0);
}

void *interp_fits_image(fitsfile_struct *ffs, bool invert)
{
    unsigned char	*data;
    LONG	nbytes;


    if (no_interp == YES) {
       nbytes = ffs->xwid * t_size[ffs->pix_type] * ffs->ywid;
       ED_MALLOC(data, "interp_fits_image", unsigned char, nbytes);
       ED_FREAD(data, "interp_fits_image", ffs->fd, nbytes);
    } else {   
    if (ffs->comp_type != NO_COMP) {
       data = read_compressed_image(ffs);
    } else {
      data = alloc_fits_map(ffs->xwid, ffs->ywid, ffs->pix_type);
      read_fits_image(ffs, data);
/*
      for (i=0;i<ffs->ywid;i++) {
        xoff = i * ffs->xwid * t_size[ffs->pix_type];
        read_fits_line(ffs, &data[xoff], i, invert);
      }
 */
    }
    }
    return data;
}

void padd_fits(fitsfile_struct *ffs)
{
    LONG nb;
    char *b;

    nb = (((LONG)fabs(ffs->bitpix/8))*((LONG)ffs->xwid)*
          ((LONG)ffs->ywid));
    nb = 2880 - nb % 2880;
    ED_CALLOC(b, "padd_fits", char, nb);
    ED_FWRITE(b, "padd_fits", ffs->fd, nb);
    ED_FREE(b);
}
void close_fits(fitsfile_struct *ffs, bool delete)
{
    ED_FCLOSE(ffs->fd, "close_fits");
    if (delete == YES)
       unlink(ffs->name);
    ED_FREE(ffs->head->ptr);
    ED_FREE(ffs->head);
    ED_FREE(ffs);
}

fitsfile_struct *new_fits()
{
    fitsfile_struct	*ff;

    ED_CALLOC(ff,       "new_fits", fitsfile_struct, 1);
    ED_CALLOC(ff->head, "new_fits", fitshead_struct, 1);
    ff->bscale = 1.0;
    ff->bzero = 0.0;
    ff->fd = NULL;
    return ff;
}

void inherit_fits(fitsfile_struct *new, fitsfile_struct *old)
{
    int		i;
    static char simpleline[80] = {
"SIMPLE  =                    T  / Make life simple                              "};

    if (old->fd != NULL) {
       new->fd        = NULL;
       ED_CALLOC(new->head->ptr, "inherit_fits", char, 
                 old->head->nblocks*FITSBLKSIZ);
       for (i=0;i<old->head->nblocks;i++) {
          memcpy(&new->head->ptr[i*FITSBLKSIZ], 
                 &old->head->ptr[i*FITSBLKSIZ], 
                 FITSBLKSIZ);
       }
       strncpy(&new->head->ptr[0],simpleline, 80);
          
       new->head->nblocks = old->head->nblocks;
       interp_fits_header(new);
    } else {
       syserr(FATAL, "inherit_fits", "Originating fits file is not opened\n");
    }
}

int write_fits_header(fitsfile_struct *ff)
{
    interp_fits_header(ff);
    ED_FWRITE(ff->head->ptr, "write_fits_header", ff->fd,
             ff->head->nblocks*FITSBLKSIZ);

    return 0;
}

void write_fits_line(fitsfile_struct *ff, unsigned char *map)
{
/****** write_fits_line ***************************************************
 *
 *  NAME	
 *      write_fits_line - Write a line of data and convert types
 *
 *  SYNOPSIS
 *      void write_fits_line(fitsfile_struct *ff, char *map)
 *
 *  FUNCTION
 *      This routine writes a line of image data and convert the FITS pixel
 *      representation from the internal map representation. The conversion
 *      is done form any type to any type.
 *
 *  INPUTS
 *      ff  - The fitsfile structure describing the image
 *      map - A pointer to the start of the line to store the read result
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 28-11-1996
 *
 ******************************************************************************/
    short int	pix_size = (short)(fabs(ff->bitpix)/8);
    unsigned char	*b, *b0;
    int		i, nx = ff->xwid;
    double	bscale = ff->bscale, bzero = ff->bzero;

    ED_MALLOC(b, "read_fits_line", unsigned char, nx * pix_size);
    b0 = b;

    switch(ff->pix_type) { 
    unsigned char bb[8];
    case T_BYTE: 
       switch (ff->bitpix) {
       case PIX_BYTE:
          for (i=0;i<nx;i++,map++,b++) {
             bb[0] = (*map - bzero) / bscale;
             memcpy(b, bb, sizeof(char));
          }
          break;
       case PIX_SHORT: 
          for (i=0;i<nx;i++,map++,b+=sizeof(short)) {
             cast_short(bb,(*map - bzero) / bscale);
             memcpy(b, &bb, sizeof(short));
          }
          break;
       case PIX_LONG:
          for (i=0;i<nx;i++,map++,b+=sizeof(LONG)) {
             cast_int(bb,(*map - bzero) / bscale);
             memcpy(b, &bb, sizeof(LONG));
          }
          break;
       case PIX_FLOAT:
          for (i=0;i<nx;i++,map++,b+=sizeof(float)) {
             cast_float(bb,(*map - bzero) / bscale);
             memcpy(b, &bb, sizeof(float));
          }
          break;
       case PIX_DOUBLE:
          for (i=0;i<nx;i++,map++,b+=sizeof(double)) {
             cast_double(bb, (*map - bzero) / bscale);
             memcpy(b, &bb, sizeof(char));
          }
          break;
       }
       break;
    case T_SHORT : 
/*
 *     The processing of short type of pixels
 */
       switch (ff->bitpix) {
       short aa;
       case PIX_BYTE:
          for (i=0;i<nx;i++,map+=sizeof(short),b++) {
             memcpy(&aa, map, sizeof(short));
             bb[0] = ((aa - bzero)/bscale);
             memcpy(b, bb, sizeof(char));
          }
          break;
       case PIX_SHORT: 
          for (i=0;i<nx;i++,map+=sizeof(short),b+=sizeof(short)) {
             memcpy(&aa, map, sizeof(short));
             cast_short(bb, (short) (aa - bzero)/bscale);
             memcpy(b, bb, sizeof(short));
          }
          break;
       case PIX_LONG:
          for (i=0;i<nx;i++,map+=sizeof(short),b+=sizeof(LONG)) {
             memcpy(&aa, map, sizeof(short));
             cast_int(bb, (LONG) (aa - bzero) / bscale);
             memcpy(b, bb, sizeof(LONG));
          }
          break;
       case PIX_FLOAT:
          for (i=0;i<nx;i++,map+=sizeof(short),b+=sizeof(float)) {
             memcpy(&aa, map, sizeof(short));
             cast_float(bb, (float) (aa - bzero) / bscale);
             memcpy(b, bb, sizeof(float));
          }
          break;
       case PIX_DOUBLE:
          for (i=0;i<nx;i++,map+=sizeof(short),b+=sizeof(double)) {
             memcpy(&aa, map, sizeof(short));
             cast_float(bb, (double) (aa - bzero) / bscale);
             memcpy(b, bb, sizeof(double));
          }
          break;
       }
       break;
    case T_LONG  : 
/*
 *     The processing of long type of pixels
 */
       switch (ff->bitpix) {
       LONG aa;
       case PIX_BYTE:
          for (i=0;i<nx;i++,map+=sizeof(LONG),b++) {
             memcpy(&aa, map, sizeof(LONG));
             bb[0] = ((aa - bzero)/bscale);
             memcpy(b, bb, sizeof(char));
          }
          break;
       case PIX_SHORT: 
          for (i=0;i<nx;i++, map+=sizeof(LONG),b+=sizeof(short)) {
             short cc;
             memcpy(&aa, map, sizeof(LONG));
             cc = ((aa - bzero)/bscale);
             cast_short(bb, cc);
             memcpy(b, bb, sizeof(short));
          }
          break;
       case PIX_LONG:
          for (i=0;i<nx;i++,map+=sizeof(LONG),b+=sizeof(LONG)) {
             LONG cc;
             memcpy(&aa, map, sizeof(LONG));
             cc = ((aa - bzero)/bscale);
             cast_int(bb, cc);
             memcpy(b, bb, sizeof(LONG));
          }
          break;
       case PIX_FLOAT:
          for (i=0;i<nx;i++,map+=sizeof(LONG),b+=sizeof(float)) {
             float cc;
             memcpy(&aa, map, sizeof(LONG));
             cc = ((aa - bzero)/bscale);
             cast_float(bb, cc);
             memcpy(b, bb, sizeof(float));
          }
          break;
       case PIX_DOUBLE:
          for (i=0;i<nx;i++,map+=sizeof(LONG),b+=sizeof(double)) {
             double cc;
             memcpy(&aa, map, sizeof(LONG));
             cc = ((aa - bzero)/bscale);
             cast_double(bb, cc);
             memcpy(b, bb, sizeof(double));
          }
          break;
       }
       break;
    case T_FLOAT : 
/*
 *     The processing of float type of pixels
 */
       switch (ff->bitpix) {
       float aa;
       case PIX_BYTE:
          for (i=0;i<nx;i++,map+=sizeof(float),b++) {
             memcpy(&aa, map, sizeof(float));
             bb[0] = ((aa - bzero)/bscale);
             memcpy(b, bb, sizeof(char));
          }
          break;
       case PIX_SHORT: 
          for (i=0;i<nx;i++, map+=sizeof(float),b+=sizeof(short)) { 
             short cc;
             memcpy(&aa, map, sizeof(float));
             cc = ((aa - bzero)/bscale);
             cast_short(bb, cc);
             memcpy(b, bb, sizeof(short));
          }
          break;
       case PIX_LONG:
          for (i=0;i<nx;i++, map+=sizeof(float),b+=sizeof(LONG)) {
             LONG cc;
             memcpy(&aa, map, sizeof(float));
             cc = ((aa - bzero)/bscale);
             cast_int(bb, cc);
             memcpy(b, bb, sizeof(LONG));
          }
          break;
       case PIX_FLOAT:
          for (i=0;i<nx;i++, map+=sizeof(float),b+=sizeof(float)) {
             float cc;
             memcpy(&aa, map, sizeof(float));
             cc = ((aa - bzero)/bscale);
             cast_float(bb, cc);
             memcpy(b, bb, sizeof(float));
          }
          break;
       case PIX_DOUBLE:
          for (i=0;i<nx;i++, map+=sizeof(float),b+=sizeof(double)) {
             double cc;
             memcpy(&aa, map, sizeof(float));
             cc = ((aa - bzero)/bscale);
             cast_double(bb, cc);
             memcpy(b, bb, sizeof(double));
          }
          break;
       }
       break;
    case T_DOUBLE: 
/*
 *     The processing of double type of pixels
 */
       switch (ff->bitpix) {
       double aa;
       case PIX_BYTE:
          for (i=0;i<nx;i++,map+=sizeof(double),b++) {
             memcpy(&aa, map, sizeof(double));
             bb[0] = ((aa - bzero)/bscale);
             memcpy(b, bb, sizeof(char));
          }
          break;
       case PIX_SHORT: 
          for (i=0;i<nx;i++, map+=sizeof(double),b+=sizeof(short)) {
             short cc;
             memcpy(&aa, map, sizeof(double));
             cc = ((aa - bzero)/bscale);
             cast_short(bb, cc);
             memcpy(b, bb, sizeof(short));
          }
          break;
       case PIX_LONG:
          for (i=0;i<nx;i++, map+=sizeof(double),b+=sizeof(LONG)) {
             LONG cc;
             memcpy(&aa, map, sizeof(double));
             cc = ((aa - bzero)/bscale);
             cast_int(bb, cc);
             memcpy(b, bb, sizeof(LONG));
          }
          break;
       case PIX_FLOAT:
          for (i=0;i<nx;i++, map+=sizeof(double),b+=sizeof(float)) {
             float cc;
             memcpy(&aa, map, sizeof(double));
             cc = ((aa - bzero)/bscale);
             cast_float(bb, cc);
             memcpy(b, bb, sizeof(float));
          }
          break;
       case PIX_DOUBLE:
          for (i=0;i<nx;i++, map+=sizeof(double),b+=sizeof(double)) {
             double cc;
             memcpy(&aa, map, sizeof(double));
             cc = ((aa - bzero)/bscale);
             cast_double(bb, cc);
             memcpy(b, bb, sizeof(double));
          }
          break;
       }
       break;
    default:
         syserr(FATAL,"read_fits_line","Unknown BITPIX type!\n");
    }
    ED_FWRITE(b0, "write_fits_line", ff->fd, nx * pix_size);
    ED_FREE(b0);
}

void write_fits_image(fitsfile_struct *ff, unsigned char *map)
{
/****** write_fits_image ***************************************************
 *
 *  NAME	
 *      write_fits_image - Write a line of data and convert types
 *
 *  SYNOPSIS
 *      void write_fits_image(fitsfile_struct *ff, char *map)
 *
 *  FUNCTION
 *      This routine writes a image data and convert the FITS pixel
 *      representation from the internal map representation. The conversion
 *      is done form any type to any type.
 *
 *  INPUTS
 *      ff  - The fitsfile structure describing the image
 *      map - A pointer to the start of the line to store the read result
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 28-11-1996
 *
 ******************************************************************************/
    short int	pix_size = (short)(t_size[ff->pix_type]);
    unsigned char	*b, *b0;
    unsigned char	*bptr;
    short int		*sptr;
    int			*lptr;
    float		*fptr;
    double		*dptr;
    int		i, nx = ff->xwid, ny = ff->ywid;
    int		scoop = nx * ny;
    double	bs = ff->bscale, bz = ff->bzero;

    ED_MALLOC(b, "read_fits_line", unsigned char, scoop * pix_size);
    b0 = b;
    switch(ff->pix_type) {
    case T_BYTE:
      bptr = map;
      switch (ff->bitpix) {
        case PIX_BYTE:
          if (ff->bit_sign)
            for (i=scoop; i--;) *(b++) = (*(bptr++) - bz)/bs;
          else
            for (i=scoop; i--;) *((unsigned char *)b++) = (*(bptr++) - bz)/bs;
          break;
        case PIX_SHORT:
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(short))
              *((short *)b) = (*(bptr++) - bz)/bs;
          else
            for (i=scoop; i--; b += sizeof(unsigned short))
	    {
              *((unsigned short *)b) = (*(bptr++)-bz)/bs;
	    }
#ifdef BSWAP
          swapbytes(b0, 2, scoop);
#endif
          break;
        case PIX_LONG:
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(LONG))
              *((LONG *)b) = (*(bptr++) - bz)/bs;
          else
            for (i=scoop; i--; b += sizeof(ULONG))
              *((ULONG *)b) = (*(bptr++)-bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          break;
        case PIX_FLOAT:
          for (i=scoop; i--; b += sizeof(float))
            *((float *)b) = (*(bptr++) - bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          break;
        case PIX_DOUBLE:
          for (i=scoop; i--; b += sizeof(double))
            *((double *)b) = (*(bptr++) - bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 8, scoop);
#endif
          break;
        default:
          syserr(FATAL,"read_fits_line","Unknown BITPIX type!\n");
      }
      break;
    case T_SHORT:
      sptr = (short *)map;
      switch (ff->bitpix) {
        case PIX_BYTE:
          if (ff->bit_sign)
            for (i=scoop; i--;) *(b++) = (*(sptr++) -bz)/bs;
          else
            for (i=scoop; i--;) *((unsigned char *)b++) = (*(sptr++) - bz)/bs;
          break;
        case PIX_SHORT:
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(short))
	      {
              *((short *)b) = (*(sptr++) - bz)/bs;
	      }
          else
            for (i=scoop; i--; b += sizeof(unsigned short))
	    {
              *((unsigned short *)b) = (*(sptr++)-bz)/bs;
	    }
#ifdef BSWAP
          swapbytes(b0, 2, scoop);
#endif
          break;
        case PIX_LONG:
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(LONG))
              *((LONG *)b) = (*(sptr++) - bz)/bs;
          else
            for (i=scoop; i--; b += sizeof(ULONG))
              *((ULONG *)b) = (*(sptr++)-bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          break;
        case PIX_FLOAT:
          for (i=scoop; i--; b += sizeof(float))
            *((float *)b) = (*(sptr++) - bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          break;
        case PIX_DOUBLE:
          for (i=scoop; i--; b += sizeof(double))
            *((double *)b) = (*(sptr++) - bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 8, scoop);
#endif
          break;
        default:
          syserr(FATAL,"read_fits_line","Unknown BITPIX type!\n");
      }
      break;
    case T_LONG  : 
/*
 *     The processing of long type of pixels
 */
      lptr = (int *)map;
      switch (ff->bitpix) {
        case PIX_BYTE:
          if (ff->bit_sign)
            for (i=scoop; i--;) *(b++) = (*(lptr++) -bz)/bs;
          else
            for (i=scoop; i--;) *((unsigned char *)b++) = (*(lptr++) - bz)/bs;
          break;
        case PIX_SHORT:
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(short))
              *((short *)b) = (*(lptr++) - bz)/bs;
          else
            for (i=scoop; i--; b += sizeof(unsigned short))
	    {
              *((unsigned short *)b) = (*(lptr++)-bz)/bs;
	    }
#ifdef BSWAP
          swapbytes(b0, 2, scoop);
#endif
          break;
        case PIX_LONG:
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(LONG))
              *((LONG *)b) = (*(lptr++) - bz)/bs;
          else
            for (i=scoop; i--; b += sizeof(ULONG))
              *((ULONG *)b) = (*(lptr++)-bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          break;
        case PIX_FLOAT:
          for (i=scoop; i--; b += sizeof(float))
            *((float *)b) = (*(lptr++) - bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          break;
        case PIX_DOUBLE:
          for (i=scoop; i--; b += sizeof(double))
            *((double *)b) = (*(lptr++) - bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 8, scoop);
#endif
          break;
        default:
          syserr(FATAL,"read_fits_line","Unknown BITPIX type!\n");
      }
      break;
    case T_FLOAT : 
/*
 *     The processing of float type of pixels
 */
      fptr = (float *)map;
      switch (ff->bitpix) {
        case PIX_BYTE:
          if (ff->bit_sign)
            for (i=scoop; i--;) *(b++) = (*(fptr++) -bz)/bs;
          else
            for (i=scoop; i--;) *((unsigned char *)b++) = (*(fptr++) - bz)/bs;
          break;
        case PIX_SHORT:
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(short))
              *((short *)b) = (*(fptr++) - bz)/bs;
          else
            for (i=scoop; i--; b += sizeof(unsigned short))
	    {
              *((unsigned short *)b) = (*(fptr++)-bz)/bs;
	    }
#ifdef BSWAP
          swapbytes(b0, 2, scoop);
#endif
          break;
        case PIX_LONG:
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(LONG))
              *((LONG *)b) = (*(fptr++) - bz)/bs;
          else
            for (i=scoop; i--; b += sizeof(ULONG))
              *((ULONG *)b) = (*(fptr++)-bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          break;
        case PIX_FLOAT:
          for (i=scoop; i--; b += sizeof(float))
            *((float *)b) = (*(fptr++) - bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          break;
        case PIX_DOUBLE:
          for (i=scoop; i--; b += sizeof(double))
            *((double *)b) = (*(fptr++) - bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 8, scoop);
#endif
          break;
        default:
          syserr(FATAL,"read_fits_line","Unknown BITPIX type!\n");
      }
      break;
    case T_DOUBLE: 
/*
 *     The processing of double type of pixels
 */
      dptr = (double *)map;
      switch (ff->bitpix) {
        case PIX_BYTE:
          if (ff->bit_sign)
            for (i=scoop; i--;) *(b++) = (*(dptr++) -bz)/bs;
          else
            for (i=scoop; i--;) *((unsigned char *)b++) = (*(dptr++) - bz)/bs;
          break;
        case PIX_SHORT:
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(short))
              *((short *)b) = (*(dptr++) - bz)/bs;
          else
            for (i=scoop; i--; b += sizeof(unsigned short))
	    {
              *((unsigned short *)b) = (*(dptr++)-bz)/bs;
	    }
#ifdef BSWAP
          swapbytes(b0, 2, scoop);
#endif
          break;
        case PIX_LONG:
          if (ff->bit_sign)
            for (i=scoop; i--; b += sizeof(LONG))
              *((LONG *)b) = (*(dptr++) - bz)/bs;
          else
            for (i=scoop; i--; b += sizeof(ULONG))
              *((ULONG *)b) = (*(dptr++)-bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          break;
        case PIX_FLOAT:
          for (i=scoop; i--; b += sizeof(float))
            *((float *)b) = (*(dptr++) - bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 4, scoop);
#endif
          break;
        case PIX_DOUBLE:
          for (i=scoop; i--; b += sizeof(double))
            *((double *)b) = (*(dptr++) - bz)/bs;
#ifdef BSWAP
          swapbytes(b0, 8, scoop);
#endif
          break;
        default:
          syserr(FATAL,"read_fits_line","Unknown BITPIX type!\n");
      }
      break;
    default:
         syserr(FATAL,"read_fits_line","Unknown BITPIX type!\n");
    }
    ED_FWRITE(b0, "write_fits_line", ff->fd, (scoop * pix_size));
    ED_FREE(b0);
}


int save_fits_image(fitsfile_struct *ffs, unsigned char *data)
{
    LONG	nbytes;
    
    if (no_interp == YES) 
    {
       nbytes = ffs->xwid * t_size[ffs->pix_type] * ffs->ywid;
       ED_FWRITE(data, "save_fits_image", ffs->fd, nbytes);
       padd_fits(ffs);
    } else {
       write_fits_image(ffs, data);
       padd_fits(ffs);
/*
    for (i=0;i<ffs->ywid;i++) {
       xoff = i * ffs->xwid * t_size[ffs->pix_type];
       write_fits_line(ffs, &data[xoff]);
    }
 */
    }
    return 0;
}

void save_fits(fitsfile_struct *ff, unsigned char *data, char *file)
{
    ED_FOPEN(ff->fd, "save_fits", file, "w+");

    VPRINTF(NORMAL,"Saving file %s\n", file);

    write_fits_header(ff);

    save_fits_image(ff, data);

    ED_FCLOSE(ff->fd, "save_fits");
}

void init_fits(fitsfile_struct *ff)
{
    ED_CALLOC(ff->head->ptr, "init_fits", char, FITSBLKSIZ);
    ff->head->curline = 0;
    ff->head->nblocks  = 1;
}
#endif
