#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "precess_defs.h"
#include "precess_globals.h"
#include "utils_globals.h"
/*
 * AST_PRECESS -- Precess coordinates from epoch1 to epoch2.
 *
 * The method used here is based on the new IAU system described in the
 * supplement to the 1984 Astronomical Almanac.  The precession is
 * done in two steps; precess epoch1 to the standard epoch J2000.0 and then
 * precess from the standard epoch to epoch2.  The precession between
 * any two dates is done this way because the rotation matrix coefficients
 * are given relative to the standard epoch.
 */
void ast_rotmatrix (double , double [4][4]);
double ast_julday (double );

void precess (double ra1, double dec1, double epoch1, 
                  double *ra2, double *dec2, double epoch2)
{
    double	r0[4], r1[4], p[4][4];
/*
 *  If the input epoch is 0 or undefined then assume the input epoch
 *  is the same as the output epoch.  If the two epochs are the same
 *  then return the coordinates from epoch1.
 */
    if ((epoch1 == 0.) || fabs(epoch1 - epoch2) < 1.e-5) {
	    *ra2 = ra1;
	    *dec2 = dec1;
	    return;
	}
/*
 *      Rectangular equitorial coordinates (direction cosines).
 */
	*ra2 = ra1 * 3.14159625 /180.;
	*dec2 = dec1 * 3.14159625 /180.;

	r0[1] = cos (*ra2) * cos (*dec2);
	r0[2] = sin (*ra2) * cos (*dec2);
	r0[3] = sin (*dec2);
/*
 *  If epoch1 is not the standard epoch then precess to the standard
 *  epoch.
 */
	if (epoch1 != 2000.) {
	    ast_rotmatrix (epoch1, p);
/* 
 *  Note that we multiply by the inverse of p which is the
 *  transpose of p.
 */
	    r1[1] = p[1][1] * r0[1] + p[1][2] * r0[2] + p[1][3] * r0[3];
	    r1[2] = p[2][1] * r0[1] + p[2][2] * r0[2] + p[2][3] * r0[3];
	    r1[3] = p[3][1] * r0[1] + p[3][2] * r0[2] + p[3][3] * r0[3];
	    r0[1] = r1[1];
	    r0[2] = r1[2];
	    r0[3] = r1[3];
	}
/*
 * If epoch2 is not the standard epoch then precess from the standard
 * epoch to the desired epoch.
 */
	if (epoch2 != 2000.) {
	    ast_rotmatrix (epoch2, p);
	    r1[1] = p[1][1] * r0[1] + p[2][1] * r0[2] + p[3][1] * r0[3];
	    r1[2] = p[1][2] * r0[1] + p[2][2] * r0[2] + p[3][2] * r0[3];
	    r1[3] = p[1][3] * r0[1] + p[2][3] * r0[2] + p[3][3] * r0[3];
	    r0[1] = r1[1];
	    r0[2] = r1[2];
	    r0[3] = r1[3];
	}
/*
 *  Convert from radians to hours and degrees.
 */
	*ra2 = atan2 (r0[2], r0[1]) * 180. / 3.14159625;
	*dec2 = asin (r0[3]) * 180. / 3.14159625;
	if (*ra2 < 0.)
	    *ra2 = *ra2 + 24;
}

/*
 * ROTMATRIX -- Compute the precession rotation matrix from the standard epoch
 * J2000.0 to the specified epoch.
 */

void ast_rotmatrix (double epoch, double p[4][4])
{
   double	t, a, b, c, ca, cb, cc, sa, sb, sc;
   double	ast_julday();
/*
 * The rotation matrix coefficients are polynomials in time measured
 * in Julian centuries from the standard epoch.  The coefficients are
 * in degrees.
 */

	t = (ast_julday (epoch) - 2451545.0) / 36525.0;

	a = t * (0.6406161 + t * (0.0000839 + t * 0.0000050));
	b = t * (0.6406161 + t * (0.0003041 + t * 0.0000051));
	c = t * (0.5567530 - t * (0.0001185 + t * 0.0000116));
/*
 * Compute the cosines and sines once for efficiency.
 */
	ca = cos (a * 3.14159625 / 180.);
	sa = sin (a * 3.14159625 / 180.);
	cb = cos (b * 3.14159625 / 180.);
	sb = sin (b * 3.14159625 / 180.);
	cc = cos (c * 3.14159625 / 180.);
	sc = sin (c * 3.14159625 / 180.);
/*
 * Compute the rotation matrix from the sines and cosines.
 */
	p[1][1] = ca * cb * cc - sa * sb;
	p[2][1] = -sa * cb * cc - ca * sb;
	p[3][1] = -cb * sc;
	p[1][2] = ca * sb * cc + sa * cb;
	p[2][2] = -sa * sb * cc + ca * cb;
	p[3][2] = -sb * sc;
	p[1][3] = ca * sc;
	p[2][3] = -sa * sc;
	p[3][3] = cc;
}

/*
 * AST_JULDAY -- Convert epoch to Julian day.
 */

double ast_julday (double epoch)
{
     int     year, century;
     double  jd;

      year = (int) (epoch) - 1;
      century = year / 100;
      jd = 1721425.5 +
          365 * year - century + (int) (year / 4) + (int) (century / 4);
      jd = jd + (epoch - (int)(epoch)) * 365.25;

      return (jd);
}
