/*
   05.05.03:
   included the tsize.h file

   13.03.2005:
   I added two arguments to save_relphotom_elem to
   store information on photometric frames in the
   RELPHOTOM table.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fitscat.h"
#include "tsize.h"
#include "relphotom_tab_globals.h"
#include "options_globals.h"
#include "utils_globals.h"
#include "tabutil_globals.h"

typedef struct struct_relphot {
    char		*name;
    char		*comment;
    float		*data;
    int			depth;
    struct struct_relphot	*next, *prev;
} relphot_struct;

relphot_struct *reldata;
/**/
void init_relphotom_tab()
{
/****** init_relphotom_tab ***************************************************
 *
 *  NAME	
 *      init_relphotom_tab - Initialize the relphot table structure
 *
 *  SYNOPSIS
 *      void init_relphotom_tab()
 *
 *  FUNCTION
 *      This routine initializes the RELPHOTOM table structure, opens
 *      an I/O port to write line by line the results to the catalog and
 *      does the required pre-work to allow initialization.
 *
 *  INPUTS
 *      None
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *      A global pointer to the top of the data structure is created.
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 14-04-1998
 *
 ******************************************************************************/
/*
 *  Create the RELPHOTOM output table
 */
    reldata = NULL;
}
/**/
void save_relphot_elem(char *name, float *zps, float *zp_sdevs,
                       float *framephot, float *corrzp, 
                       float zp_aver, float chi2, int n)
{
/****** save_relphot_elem ***************************************************
 *
 *  NAME	
 *      save_relphot_elem - Save a RELPHOTOM column set
 *
 *  SYNOPSIS
 *      void save_relphot_elem(char *name, int n, float *zps, float *zp_sdevs,
 *                     float zp_aver, float chi2)
 *
 *  FUNCTION
 *      Store the data into new elements of the relphotom global structure
 *
 *  INPUTS
 *      name     - The name of the magnitude estimator
 *      zps      - The zeropoints ordered by field number
 *      zp_sdevs - The zeropoint standard deviations ordered by field number
 *      zp_aver  - Average zero piont for the whole set
 *      chi2     - The CHI2 of the fit
 *      n        - The number of fields in the input catalog
 *
 *  RESULTS
 *     None
 *
 *  RETURNS
 *     None
 *
 *  MODIFIED GLOBALS
 *     The global relphot_struct is updated with the information
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 14-04-1998
 *
 ******************************************************************************/
    relphot_struct *prev, *curr;
    int		len, i, depth;
    char	lname[MAXCHAR], *t;

    curr = reldata;
    if (!curr) {
       prev = NULL;
    } else {
       while (curr) {
          prev = curr;
          curr = curr->next;
       }
    }
/*
 *   Save the zeropoints
 */
    ED_CALLOC(curr, "save_relphot_elem", relphot_struct, 1);
    curr->prev = prev;
    if (prev) prev->next = curr;
    else reldata = curr;
    strcpy(lname, name);
    if ((t=strchr(lname, '('))) {
       t[0] = '\0';
       strtok(name, "()");
       depth = atoi(strtok(NULL, "()"));
    } else {
       depth = 0;
    }
    len = strlen(lname);
    curr->depth = depth;
    ED_CALLOC(curr->name, "save_relphot_elem", char, (len+8));
    sprintf(curr->name, "%s_ZP", lname);
    ED_CALLOC(curr->comment, "save_relphot_elem", char, MAXCHAR);
    sprintf(curr->comment, "Relative Zero Point");
    ED_CALLOC(curr->data, "save_relphot_elem", float, n);
    for (i=0;i<n;i++)
       curr->data[i] = zps[i];
    prev = curr;
/*
 *   Save the zeropoints standard deviations
 */
    ED_CALLOC(curr, "save_relphot_elem", relphot_struct, 1);
    curr->prev = prev;
    if (prev) prev->next = curr;
    curr->depth = depth;
    ED_CALLOC(curr->name, "save_relphot_elem", char, (len+8));
    sprintf(curr->name, "%s_ZP_SDEV", lname);
    ED_CALLOC(curr->comment, "save_relphot_elem", char, MAXCHAR);
    sprintf(curr->comment, "Relative Zero Point Standard Deviation");
    ED_CALLOC(curr->data, "save_relphot_elem", float, n);
    for (i=0;i<n;i++)
       curr->data[i] = zp_sdevs[i];
    prev = curr;
/*
 *   Save the zero point average
 */
    ED_CALLOC(curr, "save_relphot_elem", relphot_struct, 1);
    curr->prev = prev;
    if (prev) prev->next = curr;
    curr->depth = depth;
    ED_CALLOC(curr->name, "save_relphot_elem", char, (len+8));
    sprintf(curr->name, "%s_ZP_AVER", lname);
    ED_CALLOC(curr->comment, "save_relphot_elem", char, MAXCHAR);
    sprintf(curr->comment, "Relative Zero Point Average");
    ED_CALLOC(curr->data, "save_relphot_elem", float, n);
    for (i=0;i<n;i++)
       curr->data[i] = zp_aver;
    prev = curr;
/*
 *   Save the CHI 2 of the fit
 */
    ED_CALLOC(curr, "save_relphot_elem", relphot_struct, 1);
    curr->prev = prev;
    if (prev) prev->next = curr;
    curr->depth = depth;
    ED_CALLOC(curr->name, "save_relphot_elem", char, (len+8));
    sprintf(curr->name, "%s_CHI2", lname);
    ED_CALLOC(curr->comment, "save_relphot_elem", char, MAXCHAR);
    sprintf(curr->comment, "Relative Zero Point CHI square of fit");
    ED_CALLOC(curr->data, "save_relphot_elem", float, n);
    for (i=0;i<n;i++)
       curr->data[i] = chi2;
    prev=curr;
/*
 *   Save the framephot varialble (it indicated whether a frame was
 *   within the photometric pegs (1=YES)
 */
    ED_CALLOC(curr, "save_relphot_elem", relphot_struct, 1);
    curr->prev = prev;
    if (prev) prev->next = curr;
    curr->depth = depth;
    ED_CALLOC(curr->name, "save_relphot_elem", char, (len+8));
    sprintf(curr->name, "%s_PHOT", lname);
    ED_CALLOC(curr->comment, "save_relphot_elem", char, MAXCHAR);
    sprintf(curr->comment, "Photometric Peg?");
    ED_CALLOC(curr->data, "save_relphot_elem", float, n);
    for (i=0;i<n;i++)
       curr->data[i] = framephot[i];
    prev = curr;
/*
 *   Save the corrected photometric zeropoint. For photometric
 *   frames this is the zeropoint at airmass 0 corrected for
 *   extinction and relative zeropoint.
 */
    ED_CALLOC(curr, "save_relphot_elem", relphot_struct, 1);
    curr->prev = prev;
    if (prev) prev->next = curr;
    curr->depth = depth;
    ED_CALLOC(curr->name, "save_relphot_elem", char, (len+8));
    sprintf(curr->name, "%s_CORRZP", lname);
    ED_CALLOC(curr->comment, "save_relphot_elem", char, MAXCHAR);
    sprintf(curr->comment, "Corrected ZP for Photometric Pegs");
    ED_CALLOC(curr->data, "save_relphot_elem", float, n);
    for (i=0;i<n;i++)
       curr->data[i] = corrzp[i];
}
/**/
void read_relphot_elem(tabstruct *tab, char *name, float **zps, 
                       float **zp_sdevs, float *zp_aver, float *chi2)
{
/****** read_relphot_elem ***************************************************
 *
 *  NAME	
 *      read_relphot_elem - Save a RELPHOTOM column set
 *
 *  SYNOPSIS
 *      void read_relphot_elem(tabstruct *tab, char *name, int n, float **zps, 
 *                     float **zp_sdevs, float *zp_aver, float *chi2)
 *
 *  FUNCTION
 *      read the data into the calling arguments
 *
 *  INPUTS
 *      tab      - Pointer to the RELPHOTOM table structure
 *      name     - The name of the magnitude estimator
 *
 *  RESULTS
 *      zps      - The zeropoints ordered by field number
 *      zp_sdevs - The zeropoint standard deviations ordered by field number
 *      zp_aver  - Average zero piont for the whole set
 *      chi2     - The CHI2 of the fit
 *
 *  RETURNS
 *     None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 14-04-1998
 *
 ******************************************************************************/
    int		n, depth;
    char	nname[MAXCHAR], lname[MAXCHAR], *t;
    float	*xx;
    keystruct	*key;

/*
 *   Read the zeropoints
 */
    strcpy(lname, name);
    if ((t=strchr(lname, '('))) {
       t[0] = '\0';
       strtok(name, "()");
       depth = atoi(strtok(NULL, "()"));
    } else {
       depth = 0;
    }
    sprintf(nname, "%s_ZP", lname);
    ED_GETKEY(tab, key, nname, (*zps), n, "read_relphotom_tab");
    if (depth) {
       nname[strlen(nname)+1] = depth;
       ED_GETVEC((*zps), "read_relphotom_tab", key, nname, float, float);
    }
    sprintf(nname, "%s_ZP_SDEV", lname);
    ED_GETKEY(tab, key, nname, (*zp_sdevs), n, "read_relphotom_tab");
    if (depth) {
       nname[strlen(nname)+1] = depth;
       ED_GETVEC((*zp_sdevs), "read_relphotom_tab", key, nname, float, float);
    }
    sprintf(nname, "%s_ZP_AVER", lname);
    ED_GETKEY(tab, key, nname, xx, n, "read_relphotom_tab");
    if (depth) {
       nname[strlen(nname)+1] = depth;
       ED_GETVEC(xx, "read_relphotom_tab", key, nname, float, float);
    }
    *zp_aver = xx[0];
    sprintf(nname, "%s_CHI2", lname);
    ED_GETKEY(tab, key, nname, xx, n, "read_relphotom_tab");
    if (depth) {
       nname[strlen(nname)+1] = depth;
       ED_GETVEC(xx, "read_relphotom_tab", key, nname, float, float);
    }
    *chi2 = xx[0];
}
/**/
void save_relphotom_tab(catstruct *cat, int n)
{
/****** save_relphotom_tab ***************************************************
 *
 *  NAME	
 *      save_relphotom_tab - Save the RELPHOTOM table to the catalog
 *
 *  SYNOPSIS
 *      void save_relphotom_tab(catstruct *cat, int n)
 *
 *  FUNCTION
 *      Copy the globally stored information to the actual table structure
 *      that can be stored in the catalog on disk
 *
 *  INPUTS
 *      cat  - Pointer to the output catalog
 *      n    - The number of rows in the RELPHOTOM table
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *      The catalog structure is updated with the RELPHOTOM data
 *      and table information
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 14-04-1998
 *
 ******************************************************************************/
    tabstruct		*tab;
    relphot_struct 	*curr, *tcurr;
    float		*data;
    char		vec_name[MAXCHAR];
    int			max_depth, i;

    tab =  new_tab("RELPHOTOM");
    curr = reldata;
    while (curr) {
/*
 *     For regular scalar data just put it into the table
 */
       if (curr->depth == 0) {
          add_key_to_tab(tab, curr->name, n, curr->data, T_FLOAT, 1, "", 
                curr->comment);
       } else {
/*
 *        For vector data that is hit the first time (depth > 0) get all
 *        the associated elements and merge them into a single output
 *        data array. In the mean time mark the used elements.
 */
          if (curr->depth > 0) {
             strcpy(vec_name, curr->name);
             max_depth = curr->depth;
/*   
 *           First determine the final depth of the vector
 */
             tcurr= curr;
             while (tcurr) {
                if (!strcmp(vec_name, tcurr->name))
                   max_depth = MAX(max_depth, tcurr->depth);
                tcurr = tcurr->next;
             }
             tcurr = reldata;
             ED_CALLOC(data, "save_relphotom_tab", float, (n*max_depth));
             while (tcurr) {
                if (!strcmp(vec_name, tcurr->name)) {
                   for (i=0;i<n;i++) 
                      data[i*max_depth+tcurr->depth-1] = tcurr->data[i];
                   tcurr->depth = -1;
                }
                tcurr = tcurr->next;
             }
             add_key_to_tab(tab,vec_name,n,data,T_FLOAT,max_depth,"", 
                      curr->comment);
          }
       }
       curr = curr->next;
    }
    update_tab(tab);
    if (add_tab(tab, cat, 0) == RETURN_ERROR) 
       syserr(FATAL, "store_relphotom_tab", "Error add table RELPHOTOM\n");
}
