/* History:
   11.01.2000:
   function init_outcat:
   keynames is now from type char **.
   The old definition led to a core dump in read_keys.

   17.10.2001:
   - removed the superfluous OBSTIME keyword from the control 
     parameter list.
   - removed compiler warnings under DEC.

   05.05.03:
   included the tsize.h file

   10.01.05:
   removed compiler warnings under gcc with optimisation
   (-O3) enabled.

   29.01.2005:
   - I added history information to the output catalog
   - I added argc and argv to the arguments of function
     init_outcat to write HISTORY info to output catalogs.

   13.03.2005:
   I added default values for EXPTIME and EXPTIME_UNIT
   to avoid core dumps if these quantities are not defined.
   Changes to avoid compiler warnings with '-Wall -W -pedantic'
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#define NBANDS 10

#include "fitscat.h"
#include "tsize.h"
#include "tabutil_globals.h"
#include "utils_globals.h"
#include "utils_macros.h"
#include "options_globals.h"
#include "aplphotom_globals.h"
#include "aplphotom_vers.h"
#include "relphotom_tab_globals.h"

#define STRING_LEN 16

static float	*final_zp, *final_coef, *final_zp_sdev, zp_ave, *exptime_def;
static char	*final_names;
static int	nfinal;
static float	**phot, **photerr, **mag, **magerr, **magsyserr, **relphot;
static short int	no_relphot;
static int      iphots, nphots;
static short int	*fpos;
static char	**phot_names, **photerr_names;
static char	**mag_names, **magerr_names, **magsyserr_names;
static double	*airmass;
static int	*chan_nr;
static char     **exptime_names, **exptime_values;
static int      nexptimes;
static double   *exptime;
static long int *lexptime;

char	header[MAXCHAR];

control_struct control;

option my_opts[] = {
            {"ZP_ESTIMATES", FARRAY, &control.zp_estimate, 0, 0, 0.0, 100.0,
              0, 0.08, NO, "",
             {"", "", "", "", ""}, 0,
             "Estimate of the instrumental zeropoint (# = number of input cats)"},
            {"ZP_ERRORS", FARRAY, &control.zp_error, 0, 0, 0.0, 100.0,
              0, 0.12, NO, "",
             {"", "", "", "", ""}, 0, 
             "Estimate of the instr zeropoint error (# = number of input cats)"},
            {"COEFS", FARRAY, &control.coef, 0, 0, 0.0, 100.0,
              0, 0.03, NO, "",
             {"", "", "", "", ""}, 0, 
             "List of extinction coefficients (# = number of input cats)"},
            {"COEFSERR", FARRAY, &control.coeferr, 0, 0, 0.0, 100.0,
              0, 0.01, NO, "",
             {"", "", "", "", ""}, 0,
             "List of extinction coef errors (# = number of input cats)"},
            {"DUMMY_MAG", FLOAT, &control.dummag, 0, 0, 0.0, 1000.0,
              0, 99.0, NO, "",
             {"", "", "", "", ""}, 0,
             "Value of the magnitude for blanked objects"},
            {"AIRMASS", TEXT, &control.airmass, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "2MUAIRM",
             {"", "", "", "", ""}, 0,
             "Name of the FIELDS table column containing the airmass value"},
            {"EXPTIME_KEYWORD", TEXTVEC, &control.exptime_keyw, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "EXPTIME",
             {"", "", "", "", ""}, 0,
             "Name of the FIELDS table column containign the exposure time"},
            {"EXPTIME_UNIT", TEXTVEC, &control.exptime_unit, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "1.0",
             {"", "", "", "", ""}, 0,
             "Name of the FIELDS table column containign the exposure time"},
            {"CORR_RELATIVE", BOOLEAN, &control.relative, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "",
             {"YES", "NO", "", "", ""}, 0, 
             "Correct only with relative part from RELPHOTOM table (uses ZP_ESTIMATE) for absolute"},
            {"FIX_MAG", BOOLEAN, &control.fix_mag, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "",
             {"YES", "NO", "", "", ""}, 0, 
             "Do we use one magnitude estimator for all relative photometry"},
            {"FIX_MAGNAME", TEXT, &control.fix_magname, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "MAG_AUTO",
             {"", "", "", "", ""}, 0,
             "Name of the magnitude estimator to be used as the fixator"},
            {"PHOT", TEXTVEC, &control.phots, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "FLUX_AUTO",
             {"", "", "", "", ""}, 0,
             "Name of the OBJECTS table column containing the FLUX estimator"},
            {"PHOTERR", TEXTVEC, &control.photerrs, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "FLUXERR_AUTO",
             {"", "", "", "", ""}, 0,
             "Name of the OBJECTS table column containing the FLUX error estimator"},
            {"MAG", TEXTVEC, &control.mags, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "MAG_AUTO",
             {"", "", "", "", ""}, 0, 
             "Name of the OBJECTS table column receiving the Magnitude estimator"},
            {"MAGERR", TEXTVEC, &control.magerrs, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "MAGERR_AUTO",
             {"", "", "", "", ""}, 0,
             "Name of the OBJECTS table column receiving the Mag error estimator"},
            {"MAGSYSERR", TEXTVEC, &control.magsyserrs, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "MAGSYSERR_AUTO",
             {"", "", "", "", ""}, 0, 
             "Name of the OBJECTS table column receiving the Mag systematic error"},
            {"", BOOLEAN, NULL, 0, 0, 0.0, 0.0, 0, 0.0, NO, "",
             {"YES", "NO", "", "", ""}, 0,
              "dummy key to end structure my_opts"}
};

static int CHECK;

/**/
void usage(char *me)
{
    fprintf(stderr,"Usage: %s \n",me);
    fprintf(stderr,"                 -i incat1 [[incat2] incat3] ... \n");
    fprintf(stderr,"                 -o outcat1 [[outcat2] outcat3] ... \n");
    fprintf(stderr,"                [-p photom_cat]\n");
    fprintf(stderr,"                 -c conffile\n");
    exit(1);
}
 
void init(int argc, char *argv[])
{
   int i, j, k;

   if (argc == 1) usage(argv[0]);
   init_options(0, my_opts);
   control.photomfile[0] = '\0';
   for (i=1; i<argc; i++) {
     if (argv[i][0] == '-') {
        switch(argv[i][1]) {
        case 'c': strcpy(control.conffile,argv[++i]);
                  break;
        case 'p': strcpy(control.photomfile,argv[++i]);
                  break;
        case 'b': strcpy(control.chan_name, argv[++i]);
                  break;
        case 'i': control.ilist = 0;
                  strcpy(control.listfile[control.ilist],argv[++i]);
                  while (i+1<argc && argv[i+1][0] != '-') {
                     strcpy(control.listfile[++control.ilist],argv[++i]);
                  }
                  ++control.ilist;
                  break;
        case 'o': j = 0;
                  strcpy(control.outfile[j],argv[++i]);
                  while (i+1<argc && argv[i+1][0] != '-') {
                     strcpy(control.outfile[++j],argv[++i]);
                  }
                  if (j+1 != control.ilist) {
                     syserr(FATAL, "init", "Input & Output lists have different length");
                  }
                  break;
        case '@': print_options(argv[0]);
                  exit(0);
        default:  parse_options(control.conffile);
                  j = i++; k = i;
                  set_command_option(argv[j], argv[k]);
                  break;
        }
     } else {
        syserr(WARNING,"init","Command line syntax error\n");
        usage(argv[0]);
     }
  }
  finish_options(control.conffile);
  phot_names    = copy_textvec(control.phots,    &nphots);
  photerr_names = copy_textvec(control.photerrs, &i);
  if (i != nphots)
     syserr(FATAL, "init", "Number if %s parameters differs from %s\n",
             control.photerrs, control.phots);
  mag_names     = copy_textvec(control.mags,     &i);
  if (i != nphots)
     syserr(FATAL, "init", "Number if %s parameters differs from %s\n",
             control.mags, control.phots);
  magerr_names  = copy_textvec(control.magerrs,  &i);
  if (i != nphots)
     syserr(FATAL, "init", "Number if %s parameters differs from %s\n",
             control.magerrs, control.phots);
  magsyserr_names  = copy_textvec(control.magsyserrs,  &i);
  if (i != nphots)
     syserr(FATAL, "init", "Number if %s parameters differs from %s\n",
             control.magsyserrs, control.phots);
  exptime_names = copy_textvec(control.exptime_keyw, &nexptimes);
  exptime_values= copy_textvec(control.exptime_unit, &i);
  if ((i != nexptimes) || (i==0))
    syserr(FATAL, "init", "Number of %s parameters differ from %s\n",
       control.exptime_keyw, control.exptime_unit);
  ED_CALLOC(exptime_def, "init", float, nexptimes);
  for (i=0;i<nexptimes;i++)
     exptime_def[i] = atof(exptime_values[i]);
}

/**/
int load_strip_phots(char *name, char *photname, char *photerrname, 
                                 char *magname, int nfields)
{
/****** load_strip_phots ***************************************************
 *
 *  NAME	
 *      load_strip_phots - Load the photometric parameters
 *
 *  SYNOPSIS
 *      int load_strip_phots(char *name)
 *
 *  FUNCTION
 *      This routine loads all the required photometric parameters. 
 *
 *  INPUTS
 *      name        - The name of the catalog to load
 *      photname    - The photometric parameter to load
 *      photerrname - The photometric error parameter to load
 *      magname     - The magnitude parameter to load
 *      nfields     - The number of fields in the catalog
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      The number of objects read
 *
 *  MODIFIED GLOBALS
 *      The global variables phot, photerr, airmass, and exptime 
 *      are filled.
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *
 ******************************************************************************/
    catstruct	*in_cat;
    tabstruct	*in_tab, *phottab;
    keystruct	**keys, *key;
    int		nobjs, i, len;
    char	**keynames, nname[MAXCHAR];
    float	*fptr, *xx, yy;

    int		NKEYS = 3, NPHOT = 1, NPHOTERR = 2;

    VPRINTF(NORMAL,"Loading photometric parameter %s\n", photname);
    if ((in_cat = read_cat(name)) == NULL) 
       syserr(FATAL, "load_strip_phots", "Could not read catalog: %s\n",
              name);
    if (!(in_tab = name_to_tab(in_cat, "OBJECTS", 0))) 
       syserr(FATAL, "load_strip_phots", "No table OBJECTS in %s catalog\n", 
             name);

    if (!(key = name_to_key(in_tab, photname)))
       syserr(FATAL, "load_strip_phots", "No Key %s in OBJECT table\n", 
             photname);

    iphots = 1;
    if (key->naxis > 1)
       syserr(FATAL, "load_strip_phots", 
            "Too large a dimension (%d) for %s (maximum of 1 only)\n",
             key->naxis, photname);
    for (i=0;i<key->naxis;i++) iphots *= key->naxisn[i];

    ED_MALLOC(keys,     "load_strip_phots", keystruct *, NKEYS);
    ED_MALLOC(keynames, "load_strip_phots", char *,      NKEYS);
    for (i=0;i<NKEYS;i++)
       ED_MALLOC(keynames[i], "load_strip_phots", char, MAXCHAR);
    strcpy(keynames[0],        "FIELD_POS");
    strcpy(keynames[NPHOT],    photname);
    strcpy(keynames[NPHOTERR], photerrname);

    read_keys(in_tab, keynames, keys, NKEYS, NULL);

    for (i=0;i<NKEYS;i++) {
       if (!keys[i]) 
          syserr(FATAL,"load_strip_phots", "No key %s in table\n",
              keynames[i]);
    }
    fpos    = keys[0]->ptr;
    nobjs   = keys[0]->nobj;

    ED_CALLOC(phot,    "load_strip_phots", float *, iphots);
    ED_CALLOC(photerr, "load_strip_phots", float *, iphots);
    for (i=0;i<iphots;i++) {
       len = strlen(keynames[NPHOT]);
       keynames[NPHOT][len+1] = i+1;
       len = strlen(keynames[NPHOTERR]);
       keynames[NPHOTERR][len+1] = i+1;
       ED_GETVEC(fptr,    "load_strip_phots", keys[NPHOT], 
                  keynames[NPHOT], float, float);
       phot[i] = fptr;
       ED_GETVEC(fptr, "load_strip_phots", keys[NPHOTERR], 
                  keynames[NPHOTERR], float, float);
       photerr[i] = fptr;
    }
    ED_FREE(keynames[0]);
    for (i=1;i<NKEYS;i++) {
       ED_FREE(keynames[i]);
       ED_FREE(keys[i]->ptr);
    }
    ED_FREE(keynames);
    ED_FREE(keys);
/*
 *  See if there is a relphotom and use it when available
 */
    ED_CALLOC(relphot, "load_strip_phots", float *, iphots);
    if ((phottab = name_to_tab(in_cat, "RELPHOTOM", 0)) != NULL) {
/*
 *     So read the appropriate columns
 */
       no_relphot = FALSE;
       for (i=0;i<iphots;i++) {
          if (control.fix_mag == YES) {
             sprintf(nname,control.fix_magname);
          } else {
             if (iphots>1) 
                sprintf(nname,"%s(%-d)",magname,i+1);
             else
                strcpy(nname,magname);
          }
          len = strlen(nname);
          VPRINTF(VERBOSE,"Loading relative photometry for %s\n", nname);
          read_relphot_elem(phottab, nname, &relphot[i], &xx, &zp_ave, &yy);
       }
       if (no_relphot) {
          for (i=0;i<iphots;i++) {
             ED_CALLOC(relphot[i], "load_strip_phots", float, nfields);
          }
       }
    } else {
       for (i=0;i<iphots;i++) {
          ED_CALLOC(relphot[i], "load_strip_phots", float, nfields);
       }
       no_relphot = TRUE;
    }

    ED_CALLOC(mag,       "load_strip_phots", float *, iphots);
    ED_CALLOC(magerr,    "load_strip_phots", float *, iphots);
    ED_CALLOC(magsyserr, "load_strip_phots", float *, iphots);
    for (i=0;i<iphots;i++) {
       ED_CALLOC(mag[i],       "load_strip_phots", float, nobjs);
       ED_CALLOC(magerr[i],    "load_strip_phots", float, nobjs);
       ED_CALLOC(magsyserr[i], "load_strip_phots", float, nobjs);
    }
    close_cat(in_cat);

    return nobjs;
}

/**/
catstruct *init_outcat(char *inname, int *nobjs, int argc, char **argv)
{
/****** init_outcat ***************************************************
 *
 *  NAME	
 *      init_outcat - Initialize the output catalog
 *
 *  SYNOPSIS
 *      catstruct *init_outcat(char *inname)
 *
 *  FUNCTION
 *      This function opens the output catalog and copies the input
 *      catalog information.
 *
 *  INPUTS
 *      inname  - The name of the input catalog
 *
 *  RESULTS
 *      nobjs   - The number of input fields
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *      Disk file with the pasted information
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul  - dd 26-12-1996
 *
 ******************************************************************************/

    catstruct	*in_cat, *out_cat;
    tabstruct	*in_ftab;
    keystruct	**keys;
    char	**keynames;
    int         i, NKEYS = 3;

    VPRINTF(NORMAL,"Opening catalog %s\n", inname);

    ED_CALLOC(keynames, "init_outcat", char *, NKEYS);

    for(i=0; i<NKEYS; i++)
    {
      ED_CALLOC(keynames[i], "init_outcat", char, MAXCHAR);
    }

    strcpy(keynames[0],control.airmass);
    strcpy(keynames[1],"CHANNEL_NR");
    strcpy(keynames[2],exptime_names[0]);

    VPRINTF(NORMAL,"Reading catalog %s\n", inname);
    if ((in_cat = read_cat(inname)) == NULL) 
    {
       syserr(FATAL,"init_outcat","Could not read catalog: %s\n",inname);
    }
/*
 *  Now initialize the output catalog
 */

    out_cat = new_cat(1);
    inherit_cat(in_cat, out_cat);
    historyto_cat(out_cat, "aplphotom", argc, argv);
/*
 *  Copy all tables
 */
    copy_tabs(in_cat, out_cat);

    if (!(in_ftab = name_to_tab(in_cat, "FIELDS", 0)))
    {
       syserr(FATAL, "init_outcat", "No table FIELDS in %s catalog\n", 
             inname);
    }

    ED_MALLOC(keys, "init_outcat", keystruct *, NKEYS);
    read_keys(in_ftab, keynames, keys, NKEYS, NULL);

    for (i=0;i<NKEYS;i++) 
    {
       if (!keys[i]) 
          syserr(FATAL,"init_outcat", "No key %s in table\n",
              keynames[i]);
    }

    airmass = keys[0]->ptr;
    chan_nr = keys[1]->ptr;
    *nobjs  = keys[0]->nobj;
    if (keys[2]->ttype == T_LONG) {
       lexptime = keys[2]->ptr;
       ED_CALLOC(exptime, "load_photo_fields", double, (*nobjs));
       for (i=0;i<*nobjs;i++) exptime[i] = (double) lexptime[i];
    } else {
       exptime = keys[2]->ptr;
    }

    close_cat(in_cat);

    nfinal = 0;
    if (CHECK == TRUE) 
    {
       if (control.chan_nr != chan_nr[0]) 
        {
          syserr(FATAL,"init_outcat",
              "Calibration data for channel %d, while strip data for %d\n",
              control.chan_nr, chan_nr[0]);
       }
    }

    for(i=0; i<NKEYS; i++)
    {
      ED_FREE(keynames[i]);
    }

    ED_FREE(keynames);

    return out_cat;
}
/**/

void load_photom_tab(char *name, char *photname)
{
    catstruct	*photo_cat;
    tabstruct	*photo_tab;
    int		i, j, jj, k, len,nobjs;
    int		NKEYS = 6;
    keystruct	**keys;
    char        **keynames, **nname, *nptr, *zptr, *sptr, *cptr, *uptr;

    VPRINTF(NORMAL,"Loading photometric calibration table\n");

    if ((photo_cat = read_cat(name)) == NULL) {
       syserr(WARNING, "load_photom_tab", "Could not read catalog: %s\n",
              name);
       syserr(WARNING, "load_photom_tab", 
                       "Reverting to configuration parameters\n");
       CHECK = FALSE;
       name[0] = '\0';
       control.chan_nr = chan_nr[0]; 
    } else {
       CHECK = TRUE;
       if (!(photo_tab =  name_to_tab(photo_cat, "PHOTOM", 0)))
          syserr(FATAL, "No table PHOTOM in %s catalog\n", name);
       ED_CALLOC(keys,     "load_photom_tab", keystruct *, NKEYS);
       ED_CALLOC(keynames, "load_photom_tab", char *,      NKEYS);
       for (i=0;i<NKEYS;i++)
          ED_MALLOC(keynames[i], "load_photom_tab", char, MAXCHAR);
       strcpy(keynames[0],"CHANNEL_NR");
       strcpy(keynames[1],"PHOT_NAME");
       strcpy(keynames[2],"FINAL_ZP");
       strcpy(keynames[3],"FINAL_ZP_SDEV");
       strcpy(keynames[4],"FINAL_COEF");
       strcpy(keynames[5],"EXPTIME_UNIT");
       read_keys(photo_tab, keynames, keys, NKEYS, NULL);
       for (i=0;i<NKEYS;i++) {
          if (!keys[i])
             syserr(FATAL,"load_photom_tab", "No key %s in table\n",
                 keynames[i]);
       }
       if ((nobjs = keys[0]->nobj) == 0) {
          CHECK = FALSE;
          name[0] = '\0';
          control.chan_nr = chan_nr[0];
       } else {
/*
 *        Find out which rows to use for the current photometric parameter
 */
          chan_nr  = keys[0]->ptr;
          control.chan_nr = chan_nr[0];
          if (nfinal == 0) {
             ED_CALLOC(final_zp,      "load_photom_tab", float, nobjs);
             ED_CALLOC(final_zp_sdev, "load_photom_tab", float, nobjs);
             ED_CALLOC(final_coef,    "load_photom_tab", float, nobjs);
             ED_CALLOC(exptime_def,   "load_photom_tab", float, nobjs);
             ED_CALLOC(final_names,   "load_photom_tab", char ,
                       (nobjs*STRING_LEN));
          }
          ED_CALLOC(nname, "load_photom_tab", char *, iphots);
          jj = 0;
          for (i=0;i<iphots;i++) {
             ED_CALLOC(nname[i],   "load_photom_tab", char,   MAXCHAR);
             if (control.fix_mag == YES) {
                sprintf(nname[i],control.fix_magname);
             } else {
                if (iphots>1)
                   sprintf(nname[i],"%s(%-d)",photname,i+1);
                else
                   strcpy(nname[i],photname);
             }
/*
             if (iphots>1) 
                sprintf(nname[i],"%s(%-d)",photname,i+1);
             else
                strcpy(nname[i],photname);
 */
             len = strlen(nname[i]);
             nptr = keys[1]->ptr;
             zptr = keys[2]->ptr;
             sptr = keys[3]->ptr;
             cptr = keys[4]->ptr;
             uptr = keys[5]->ptr;
             for (j=0;j<nobjs;j++) {
                if (strncmp(nptr, nname[i], len) == 0) {
                   memcpy(&final_zp[nfinal+jj],      zptr, keys[2]->nbytes);
                   memcpy(&final_zp_sdev[nfinal+jj], sptr, keys[3]->nbytes);
                   memcpy(&final_coef[nfinal+jj],    cptr, keys[4]->nbytes);
                   memcpy(&exptime_def[nfinal+jj],   uptr, keys[5]->nbytes);
                   jj++;
/*
 *                 First hit is enough 
 */
                   break;
                }
                nptr += keys[1]->nbytes;
                zptr += keys[2]->nbytes;
                sptr += keys[3]->nbytes;
                cptr += keys[4]->nbytes;
                uptr += keys[5]->nbytes;
             }
          }
          for (k=0;k<iphots;k++) 
             strcpy(&final_names[(nfinal+k)*STRING_LEN], nname[k]);
          nfinal += iphots;
          ED_FREE(nname);
       }
       for (j=0;j<NKEYS;j++) {
          ED_FREE(keynames[j]);
       }
       ED_FREE(keynames);
       close_cat(photo_cat);
    }
}

/**/
void derive_mags(int n, int chan)
{
/****** derive_mags ***************************************************
 *
 *  NAME	
 *      derive_mags - Calculates the magnitude using the derive photom calib.
 *
 *  SYNOPSIS
 *      void derive_mags(int n, int chan)
 *
 *  FUNCTION
 *      This routine calculates the magnitudes of each star in the strip
 *      using the photometric calibration as derived elsewere. The
 *      photometric calibration parameters are stored in the variables:
 *       control.coef and zp
 *      The raw flux values that are converted to magnitudes are in:
 *       phot and mag
 *
 *  INPUTS
 *      n     - Number of objects to derive magnitudes for   
 *      chan  - The channel number for which to derive the magnitudes
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *      control.coef - The extinction coefficients for all channels
 *      zp           - The derived zero-points for all channes/all phot pars
 *      phot         - The raw photometric parameters
 *      mag          - The derived magnitudes
 *
 *  NOTES
 *      This function is for survey nights only.
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul  - dd 26-12-1996
 *
 ******************************************************************************/
    int		i, k;
    double	coef, zp, zps;
    
    VPRINTF(NORMAL,"Deriving magnitudes\n");
    for (k=0;k<iphots;k++) {
       if (CHECK) {
          coef = final_coef[nfinal-iphots+k];
          zp   = final_zp[nfinal-iphots+k];
          zps  = final_zp_sdev[nfinal-iphots+k];
          if (fabs(coef-control.coef[chan]) > 3.0 * control.coef[chan] ||
             coef < 0.0)
             coef = control.coef[chan];
       } else {
/*
 *        Because of the way the relative photometry is derived, the
 *        extinction coefficient is effective removed from the fluxes
 */
          if (no_relphot) {
             zp   = control.zp_estimate[chan];
             coef = control.coef[chan];
          } else {
             if (control.relative == YES) {
                zp   = control.zp_estimate[chan];
             } else {
                zp   = zp_ave;
             }
             coef = 0.0;
          }
          zps  = control.zp_error[chan];
       }
       for (i=0;i<n;i++) {
          if (phot[k][i]>0.0) {
             if (!exptime_def) {
               mag[k][i] = zp - airmass[fpos[i]-1] * coef - 
                  2.5 * log10(phot[k][i]) + relphot[k][fpos[i]-1];
             } else {
               mag[k][i] = zp - airmass[fpos[i]-1] * coef - 
                  2.5 * log10(phot[k][i]/exptime[fpos[i]-1]*exptime_def[0]) +
                       relphot[k][fpos[i]-1];
             }
             VPRINTF(DEBUG, "Mag: %f, ZP: %f, Exptime: %f, ExpDef: %f, relphot: %f\n", mag[k][i], zp, exptime[fpos[i]-1], exptime_def);
             magerr[k][i] = 1.086 * photerr[k][i] / phot[k][i];
          } else {
             mag[k][i] = magerr[k][i] = control.dummag;
          }
          magsyserr[k][i] = zps;
       }
    }
}

void save_mags(catstruct *cat, char *photname, char *photerrname,
                               char *magname, char *magerrname, 
                               char *magsyserrname)
/****** save_mags ***************************************************
 *
 *  NAME	
 *      save_mags - Save the derived magnitudes
 *
 *  SYNOPSIS
 *      void save_mags(catstruct *cat, char *photname, char *photerrname,
 *                                     char *magname, char *magerrname,
 *                                     char *magsyserrname)
 *
 *  FUNCTION
 *      This routine saves the derived magniudes to the output catalog
 *
 *  INPUTS
 *      cat            - Pointer to the output catalog structure
 *      photname       - The name of the photometric parameter
 *      photerrname    - The name of the photometric error parameter
 *      magname        - The name of the magnitude parameter
 *      magerrname     - The name of the magnitude error parameter
 *      magsyserrname  - The name of the magnitude systematic error parameter
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *     E.R. Deul - dd 29-07-1997
 *
 ******************************************************************************/
{
    int		i, j, len;
    tabstruct	*tab;
    keystruct	*key1, *key2;
    char	from_name[MAXCHAR], to_name[MAXCHAR];
    float	**fptr=NULL, *lptr;

    VPRINTF(NORMAL,"Saving magnitudes %s\n", magname);
    if (!(tab = name_to_tab(cat, "OBJECTS", 0))) 
       syserr(FATAL, "save_mags", "No table OBJECTS in output catalog\n");
    for (j=0;j<3;j++) {
       if (j==0) {
          strcpy(from_name, photname);
          strcpy(to_name,   magname);
          fptr = mag;
       }
       if (j==1) {
          strcpy(from_name, photerrname);
          strcpy(to_name,   magerrname);
          fptr = magerr;
       }
       if (j==2) {
          strcpy(from_name, photerrname);
          strcpy(to_name,   magsyserrname);
          fptr = magsyserr;
       }
       if (!(key1 = name_to_key(tab, to_name))) {
/*
 *        The output key does not yet exist, create it based on info from the
 *        input key.
 */
          key1 = new_key(to_name);
          if (!(key2 = name_to_key(tab, from_name))) 
             syserr(FATAL, "save_mags", 
                "No key %s in table OBJECTS of output catalog\n", from_name);
          *key1 = *key2;
          strcpy(key1->name,to_name);
          sprintf(key1->comment,"Magnitude(s) derived from %s", from_name);
          strcpy(key1->unit,"Mag");
          if (key2->naxis) ED_CALLOC(key1->naxisn,"save_mags",int,key2->naxis);
          for (i=0;i<key1->naxis;i++) key1->naxisn[i] = key2->naxisn[i];
          key1->ptr = NULL;
/*
 *        And add it to the table
 */
          add_key(key1, tab, 0);
       } 
       if (!key1->ptr) {
          ED_CALLOC(key1->ptr, "save_mags", char, (key1->nbytes*key1->nobj));
       }
       for (i=0;i<iphots;i++) {
          len = strlen(to_name);
          if (len+1 >= MAXCHAR)
             syserr(FATAL,"save_mags","Name %s too long\n", to_name);
          to_name[len+1] = i+1;
          lptr = fptr[i]; 
          ED_PUTVEC(lptr, "save_mags", key1, to_name, float, float); 
       }
    }
}

void save_final_cat(catstruct *outcat, char *photom_name, char *name)
{
    catstruct	*photo_cat;
    tabstruct	*photo_tab;

    if (control.photomfile[0] != '\0' &&
        (photo_cat = read_cat(photom_name)) != NULL) {
       copy_tabs(photo_cat, outcat);
    } else {
       if (nfinal != 0) {
          photo_tab = new_tab("PHOTOM");
          add_key_to_tab(photo_tab, "PHOT_NAME", nfinal, final_names, 
                      T_STRING, STRING_LEN, "", "Photometric parameter");
          add_key_to_tab(photo_tab, "FINAL_ZP",nfinal,final_zp, T_FLOAT, 1, "",
                      "Final Calibration Zero Point");
          add_key_to_tab(photo_tab, "FINAL_COEF",nfinal,final_coef, T_FLOAT, 1, 
                      "", "Final Calibration Extinction Coefficient");
          add_key_to_tab(photo_tab, "EXPTIME_UNIT",nfinal,exptime_def,T_FLOAT,1, 
                      "", "Final Calibration Exposure Time Unit");
          if (add_tab(photo_tab, outcat, 0) != RETURN_OK)
             syserr(FATAL, "save_final_cat", "Error saving table PHOTOM\n");
       }
    }
    save_cat(outcat, name);
}
/**/
int main(int argc, char *argv[])
/****** aplphotom ***************************************************
 *
 *  NAME	
 *      aplphotom - Apply photometric calibration
 *
 *  SYNOPSIS
 *      aplphotom -i incat1 incat2 ... -o outcat2 outcat2 ...
 *                -p photom_cat [-c configfile] [CONFOPTION]
 *
 *  FUNCTION
 *      This program applies the photometric calibration that is stored
 *      in the photom_cat catalog to the input catalogs. The output
 *      catalogs are related one-to-one to the input catalogs, thus outcat1
 *      is the photometrically corrected version of incat1, etc.
 *
 *  INPUTS
 *      -i incat1 incat2 ...   - The list of input catalogs
 *      -o outcat1 outcat2 ... - The list of output catalogs
 *      -p photom_cat          - The name of the catalog that contains
 *                               the PHOTOM table te be used in the
 *                               photometric application
 *     [-c configfile]         - The name of the configuration file. The default
 *                               name is /denisoft/conf/aplphotom.conf
 *     [CONFOPTION]            - Pairs of configuration VARIABLE VALUE
 *                               sequences that allow dynamically
 *                               overwritting of configuration file values.
 *
 *  RESULTS
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *      The photometric calibration process has not been fully implemented
 *      yet.
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 01-08-1996
 *
 ******************************************************************************/
{ 
    int		i, j, n, nfields;
    catstruct	*outcat;

    init(argc, argv);

    sprintf(header,"%s Version: %s (%s)",NAME,VERS,DATE);
    VPRINTF(NORMAL,"\n   %s\n\n",header);

    for (i=0;i<control.ilist;i++) 
    {
       outcat = init_outcat(control.listfile[i], &nfields, argc, argv);

       for (j=0;j<nphots;j++) {
          n = load_strip_phots(control.listfile[i],
               phot_names[j], photerr_names[j], mag_names[j], nfields);
          if (control.photomfile[0] != '\0') {
             load_photom_tab(control.photomfile, phot_names[j]);
          }
          derive_mags(n, control.chan_nr);
          save_mags(outcat, phot_names[j], photerr_names[j], 
                    mag_names[j], magerr_names[j], magsyserr_names[j]);
       }
       VPRINTF(NORMAL,"Saving catalog %s\n", control.outfile[i]);
       save_final_cat(outcat,control.photomfile,control.outfile[i]);
    }
    return 0;
}
