#include <stdio.h>
#include <math.h>
#include "solve.h"
#include "svdvar.h"
#include "lists_globals.h"
#include "utils_globals.h"
#include "options_globals.h"
#include "sv_globals.h"

/* 21.06.01:
   removed compiler warnings under DEC

   12.09.01:
   - fixed a bug in choldc:
     the inner loop only had to run from the matrix
     diagonal element (i) on, hence: 
     for (j=1;j<=n;j++) {  --> for (j=i;j<=n;j++) {
   - replaced all unsingned long definitions with the
     ULONG macro

   10.01.05:
   removed compiler warnings under gcc with optimisation
   (-O3) enabled.

   13.03.2005:
   - I renamed the function solve to relphotsolve as this function is
     especially for the relphotom program
   - In relphotsolve I now artificially set the first relative zeropoint
     to zero. Hence, the system to solve is no longer degenerate.

   09.03.2006:
   I removed the last argument from the NR free_... functions.
   That argument is not used in the functions.
*/

/*
 * Many routines from the numerical recipes C-2nd edition are included here
 * It concerns the choleski method and the sparse maxtrix biconjugate
 * gradient methods.
 */

int choldc(double **a, int n, float *p)
/****** choldc ***************************************************
 *
 *  NAME	
 *     choldc - Choleski decomposition
 *
 *  SYNOPSIS
 *     int choldc(double **a, int n, float *p)
 *
 *  FUNCTION
 *     Decompose the square, symmetric, positive definite matrix according
 *     to the Choleski method. If the matrix was not symmetric, positive
 *     definite, report this with an error return.
 *
 *  INPUTS
 *     a   - The matrix to decompose
 *     n   - THe dimension of the matrix
 *
 *  RESULTS
 *     a   - The decomposed matrix 
 *     p   - Diagonal elements of a
 *
 *  RETURNS
 *     TRUE  - If the decomposition succeeded
 *     FALSE - When the matrix is not symmetric, positive definite
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      Numerical Recipies 2nd edition
 *
 ******************************************************************************/
{
    int i, j, k;
    double sum;
    
    for (i=1;i<=n;i++) {
      for (j=i;j<=n;j++) {
        sum = a[i][j];
        for (k=i-1;k>=1;k--) 
          sum -= a[i][k] * a[j][k];
        if (i == j) {
          if (sum <= 0.0) {
            return 0;
          }
          p[i] = sqrt(sum);
        } else {
           a[j][i] = sum/p[i];
        }
      }
    }
    return -1;
}


void cholsl(double **a, int n, float *p, float *b, float *x)
/****** cholsl ***************************************************
 *
 *  NAME	
 *      cholsl - Solve a Choleski decomposed set of linear equations
 *
 *  SYNOPSIS
 *      void cholsl(double **a, int n, float *p, float *b, float *x)
 *
 *  FUNCTION
 *      Solve the set of equations that have been Choleski decomposed
 *      with choldc. Equation solved is A.x=b
 *
 *  INPUTS
 *      a  - The Choleski decomposed matrix
 *      n  - The dimension of the matrix
 *      p  - The diagonal elements of the matrix
 *      b  - The input right-hand side vector
 *
 *  RESULTS
 *      x  - The solution vector
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      Numerical Recipies 2nd edition
 *
 ******************************************************************************/
{
    int i,k;
    double sum;
  
    for (i=1; i<=n; i++) {
      sum = b[i];
      for (k=i-1;k>=1;k--) 
          sum -= a[i][k] * x[k];
      x[i]=sum/p[i];
    }
    for (i=n;i>=1;i--) {
      sum = x[i];
      for (k=i+1;k<=n;k++) 
        sum -= a[k][i] * x[k];
      x[i] = sum/p[i];
    }
}

/*
 * A series of sparse matrix routines
 */
int         sparse;
double      **A;
double      *sa;
int         *ija;

int dsprsin(double **a, int n, float thresh, ULONG nmax)
/****** dsprsin ***************************************************
 *
 *  NAME	
 *      dsprsin - Convert a matrix into sparse notations (double version)
 *
 *  SYNOPSIS
 *      int dsprsin(double **a, int n, float thresh, unsigned long nmax)
 *
 *  FUNCTION
 *      Convert a square matrix into the sparse matrix notation. In the course
 *      of converting the values less the thresh are considered zero
 *      and the matrix representation is store in two vectors. When the
 *      conversion hits the case where the number of to be stored elements
 *      exceeds nmax, there is no sense in performing a sparse conversion.
 *
 *  INPUTS
 *      a      - The square sparse matrix
 *      n      - The dimension of the square matrix
 *      thresh - The threshold below which a matrix element is considered zero
 *      nmax   - The size of the sparse matrix storage arrays
 *
 *  RESULTS
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *      sparse - True when matrix successfully converted to sparse storage
 *      sa     - A predeclared array that stores the non-zero values
 *      ija    - A predeclared array that stores the indices
 *      aa     - A copy of the input matrix when not converted to sparse
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      Numerical Recipies 2nd edition
 *
 ******************************************************************************/
{
   int i,j;
   ULONG k;

   for (j=1;j<=n;j++) sa[j]=a[j][j];
   ija[1]=n+2;
   k=n+1;
   for (i=1;i<=n;i++) {
      for (j=1;j<=n;j++) {
         if (fabs(a[i][j]) >= thresh && i != j) {
            if (++k > nmax) {
               syserr(WARNING,"sprsin", "nmax(%d) too small %d\n",nmax,k);
               sparse = 0;
/*
 *             Save input matrix in global storage
 */
               A = dmatrix(1, n, 1, n);
               for (i=1;i<=n;i++) 
                  for (j=1;j<=n;j++) 
                     A[i][j] = a[i][j];
               return (sparse);
            }
            sa[k]=a[i][j];
            ija[k]=j;
         }
      }
      ija[i+1]=k+1;
   }
   sparse = -1;
   return (sparse);
}

void dsprsax(double sa[], int ija[], double x[], double b[], int n)
/****** dsprsax ***************************************************
 *
 *  NAME	
 *      dsprsax - Multiply sparce matrix with a vector
 *
 *  SYNOPSIS
 *      void dsprsax(double sa[], int ija[], double x[], double b[], int n)
 *
 *  FUNCTION
 *      Multiply a matrix in row-index sparse storage arrays by a vector
 *
 *  INPUTS
 *      sa  - The matrix elements in sparse storage form
 *      ija - The matrix indices in sparse storage form
 *      x   - The vector to use in the multiplication
 *      n   - The dimension of the matrix/vector
 *
 *  RESULTS
 *      b   - The result of the A.x multiplication
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      Numerical Recipies 2nd edition
 *
 ******************************************************************************/
{
   int i,k;

   if (ija[1] != n+2) 
      syserr(FATAL, "dsprsax","mismatched vector and matrix\n");
   for (i=1;i<=n;i++) {
      b[i]=sa[i]*x[i];
      for (k=ija[i];k<=ija[i+1]-1;k++) b[i] += sa[k]*x[ija[k]];
   }
}

void dsprstx(double sa[], int ija[], double x[], double b[], int n)
/****** dsprstx ***************************************************
 *
 *  NAME	
 *      dsprstx - Multiply a transposed sparse matrix with a vector
 *
 *  SYNOPSIS
 *      void dsprstx(double sa[], int ija[], double x[], double b[], int n)
 *
 *  FUNCTION
 *      Multiply transpose of a matrix in row-index sparse storage arrays by 
 *      a vector
 *
 *  INPUTS
 *      sa  - The matrix elements in sparse storage form
 *      ija - The matrix indices in sparse storage form
 *      x   - The vector to use in the multiplication
 *      n   - The dimension of the matrix/vector
 *
 *  RESULTS
 *      b   - The result of the A^t.x multiplication
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *
 ******************************************************************************/
{
   int i,j,k;

   if (ija[1] != n+2) 
      syserr(FATAL, "dsprstx", "mismatched vector and matrix\n");
   for (i=1;i<=n;i++) b[i]=sa[i]*x[i];
   for (i=1;i<=n;i++) {
      for (k=ija[i];k<=ija[i+1]-1;k++) {
         j=ija[k];
         b[j] += sa[k]*x[i];
      }
   }
}
/*
 * Utility routines for the use with linbcg, biconjugate gradient routine
 * Routines for sparse matrix handling
 */
void atimes(int n, double x[], double r[], int itrnsp)
/****** atimes ***************************************************
 *
 *  NAME	
 *      atimes - Multiply a matrix or its transpose by a vector
 *
 *  SYNOPSIS
 *      void atimes(int n, double x[], double r[], int itrnsp)
 *
 *  FUNCTION
 *      Multiply a matrix or its transpose which can be in ordinary
 *      storage form or in sparse storage form with a vector.
 *
 *  INPUTS
 *      n      - The dimension of the matrix/vector
 *      x      - The vector with which the multiplications should be done
 *      itrnsp - The transpose identification
 *
 *  RESULTS
 *      r    - The result of the multiplication
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *      Uses the global variables A, sa, ija, sparse
 *
 *  BUGS
 *
 *  AUTHOR
 *      Numerical Recipies 2nd edition
 *
 ******************************************************************************/
{
   int  i,j;

   if (sparse) {
     if (itrnsp) 
       dsprstx(sa,ija,x,r,n);
     else 
       dsprsax(sa,ija,x,r,n);
   } else {
     if (itrnsp) {
       for (i=1;i<=n;i++)
         r[i] = x[1] * A[1][i];
         for (j=2;j<=n;j++)
           r[i] += x[j] * A[j][i];
     } else {
       for (i=1;i<=n;i++)
         r[i] = x[1] * A[i][1];
         for (j=2;j<=n;j++)
           r[i] += x[j] * A[i][j];
     }
   }
}


void asolve(int n, double b[], double x[], int itrnsp)
/****** asolve ***************************************************
 *
 *  NAME	
 *      asolve - Solve the A~.x=b problem or A~^t.x=b one
 *
 *  SYNOPSIS
 *      void asolve(int n, double b[], double x[], int itrnsp)
 *
 *  FUNCTION
 *      Solve the A~.x=b problem or A~^t.x=b one for some preconditioner
 *      matrix A~ which can either be in ordinary or sparse storage form.
 *
 *  INPUTS
 *      n      - The dimension of the matrix/vector
 *      b      - The vector wiht which the solution should be make
 *      itrnsp - The transpose identification
 *
 *  RESULTS
 *      x    - The result of the solution
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *
 ******************************************************************************/
{
   int i;

   if (sparse) {
     for(i=1;i<=n;i++) 
       x[i]=(sa[i] != 0.0 ? b[i]/sa[i] : b[i]);
   } else {
     for(i=1;i<=n;i++)
       x[i]=(A[i][i] != 0.0 ? b[i]/A[i][i] : b[i]);
   }
}

double snrm(ULONG n, double sx[], int itol)
/****** snrm ***************************************************
 *
 *  NAME	
 *      snrm - Compute one of two norms
 *
 *  SYNOPSIS
 *      double snrm(unsigned long n, double sx[], int itol)
 *
 *  FUNCTION
 *      Utility function to linbcg to compute the norm of either
 *      vector magnitude norm or largest component norm
 *
 *  INPUTS
 *      n    - Dimension of vector
 *      sx   - The vector
 *      itol - Type of norm to be derived
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      The norm requested as a double
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      Numerical Recipies 2nd edition
 *
 ******************************************************************************/
{
   ULONG i,isamax;
   double ans;

   if (itol <= 3) {
      ans = 0.0;
      for (i=1;i<=n;i++) ans += sx[i]*sx[i];
      return sqrt(ans);
   } else {
      isamax=1;
      for (i=1;i<=n;i++) {
         if (fabs(sx[i]) > fabs(sx[isamax])) isamax=i;
      }
      return fabs(sx[isamax]);
   }
}
#define NRANSI
#define EPS 1.0e-14

void linbcg(ULONG n, double b[], double x[], int itol, double tol,
   int itmax, int *iter, double *err)
/****** linbcg ***************************************************
 *
 *  NAME	
 *      linbcg - Solve A.x=b by iterative biconjugate gradient method
 *
 *  SYNOPSIS
 *      void linbcg(unsigned long n, double b[], double x[], int itol,
 *                  double tol, int itmax, int *iter, double *err)
 *
 *  FUNCTION
 *      Solves the A.x = b set of linear equations using the iterative
 *      biconjugate gradient method. The matrix is assumed to be a
 *      square matrix, which possibly has been converted to sparse storage
 *      form using the local routine dsprsin.
 *
 *  INPUTS
 *      n     - The dimension of the square matrix
 *      b     - The vector of input data
 *      x     - The solution vector
 *      itol  - The type of stopping condition:  
 *              1: |A.x-b|/|b| < tol
 *              2: |A~^-1.(A.x-b)|/|A~^-1.b| < tol
 *              3: own estimate of error in x/|x| < tol
 *              4: largest component of error /|x| < tol
 *      tol   - Maximum allowed convergence tolerance
 *      itmax - Maximum number of allowed iterations
 *
 *  RESULTS
 *      iter  - The actual number of iterations used
 *      err   - The estimated error (to be compared with tol if iter<itmax)
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      Numerical Recipies 2nd edition
 *
 ******************************************************************************/
{
   ULONG j;
   double ak,akden,bk,bkden=0.0,bknum,bnrm=0.0,dxnrm,xnrm,zm1nrm,znrm;
   double *p,*pp,*r,*rr,*z,*zz;

   p=dvector(1,n);
   pp=dvector(1,n);
   r=dvector(1,n);
   rr=dvector(1,n);
   z=dvector(1,n);
   zz=dvector(1,n);

   *iter=0;
   atimes(n,x,r,0);
   for (j=1;j<=n;j++) {
      r[j]=b[j]-r[j];
      rr[j]=r[j];
   }
   znrm=1.0;
   if (itol == 1) bnrm=snrm(n,b,itol);
   else if (itol == 2) {
      asolve(n,b,z,0);
      bnrm=snrm(n,z,itol);
   }
   else if (itol == 3 || itol == 4) {
      asolve(n,b,z,0);
      bnrm=snrm(n,z,itol);
      asolve(n,r,z,0);
      znrm=snrm(n,z,itol);
   } else nrerror("illegal itol in linbcg");
   asolve(n,r,z,0);
   while (*iter <= itmax) {
      ++(*iter);
      zm1nrm=znrm;
      asolve(n,rr,zz,1);
      for (bknum=0.0,j=1;j<=n;j++) bknum += z[j]*rr[j];
      if (*iter == 1) {
         for (j=1;j<=n;j++) {
            p[j]=z[j];
            pp[j]=zz[j];
         }
      }
      else {
         bk=bknum/bkden;
         for (j=1;j<=n;j++) {
            p[j]=bk*p[j]+z[j];
            pp[j]=bk*pp[j]+zz[j];
         }
      }
      bkden=bknum;
      atimes(n,p,z,0);
      for (akden=0.0,j=1;j<=n;j++) akden += z[j]*pp[j];
      ak=bknum/akden;
      atimes(n,pp,zz,1);
      for (j=1;j<=n;j++) {
         x[j] += ak*p[j];
         r[j] -= ak*z[j];
         rr[j] -= ak*zz[j];
      }
      asolve(n,r,z,0);
      if (itol == 1 || itol == 2) {
         znrm=1.0;
         *err=snrm(n,r,itol)/bnrm;
      } else if (itol == 3 || itol == 4) {
         znrm=snrm(n,z,itol);
         if (fabs(zm1nrm-znrm) > EPS*znrm) {
            dxnrm=fabs(ak)*snrm(n,p,itol);
            *err=znrm/fabs(zm1nrm-znrm)*dxnrm;
         } else {
            *err=znrm/bnrm;
            continue;
         }
         xnrm=snrm(n,x,itol);
         if (*err <= 0.5*xnrm) *err /= xnrm;
         else {
            *err=znrm/bnrm;
            continue;
         }
      }
   if (*err <= tol) break;
   }

   free_dvector(p,1);
   free_dvector(pp,1);
   free_dvector(r,1);
   free_dvector(rr,1);
   free_dvector(z,1);
   free_dvector(zz,1);
}

#undef EPS
#undef NRANSI

float relphotsolve(float *x, float *sigx, int *ix, int *iy,
           int nobs, int nc,
           float *c, int *ic, float **C, char *method)
{
/****** solve ***************************************************
 *
 *  NAME	
 *     relphotsolve - Solve the set of linear equations for relative
 *                    photometric zeropoints
 *
 *  SYNOPSIS
 *     float solve(float *x, float *sigx, int *ix, int *iy, int nobs, int nc,
 *                 float *c, int *ic, float **C, char method)
 *
 *  FUNCTION
 *     Solve the set of linear equations using a number of possible methods.
 *     The set of equations is given in the form of the least squares value
 *     (x) and its weight (sigx) for measurements between two parameter of
 *     the to be solved set (ix, iy).
 *
 *  INPUTS
 *     x      - LSQ values for observation points
 *     sigx   - Weight of LSQ values for observations points
 *     ix     - Index to parameter number (from)
 *     iy     - Index to parameter number (to)
 *     nobs   - Number of observations in above arrays
 *     nc     - Number of parameters to derive
 *     method - Solution method to use: BICONGRAD, CHOLESKI, SVD
 *
 *  RESULTS
 *     c      - Resulting parameter values
 *     ic     - Index of parameters in c
 *     C      - Covariance matrix
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R.Deul - dd 15-09-1999
 *
 ******************************************************************************/
    int      i,j,nparms;
    float    *b;
    float   **U,*W,**V,wmin,wmax;
    double   **AA, *B, *Cc, err;
    int      iter;
/*
 *  Now create the necessary data storage items
 */
    nparms = nc;
    sa     = dvector(1,3*nobs+nc);
    ija    = ivector(1,3*nobs+nc);
    AA     = dmatrix(1,nparms,1,nparms);
    B      = dvector(1,nparms);
    Cc     = dvector(1,nparms);
/*
 *  Initialize to clear
 */
    for (i=1;i<nparms+1;i++) {
       for (j=1;j<nparms+1;j++) {
          AA[i][j] = 0.0;
       }
       B[i] = 0.0;
       Cc[i] = 0.0;
    }
    for (i=1;i<nobs+1;i++) {
/*
 *     Fill with an observation
 */
       AA[ix[i]][ix[i]]  += 1.0/sigx[i];
       AA[ix[i]][iy[i]]  -= 1.0/sigx[i];
       AA[iy[i]][ix[i]]  -= 1.0/sigx[i];
       AA[iy[i]][iy[i]]  += 1.0/sigx[i];
       B[ix[i]]         -= x[i]/sigx[i];
       B[iy[i]]         += x[i]/sigx[i];
    }

/*
 * set the first relative photometric zeropoint artificially
 * to zero to make the system non-degenerate
 */
    AA[1][1] += 1.0;

/*
 *  Go for the individual methods
 */
    if (!strcmpcu(method, "BICONGRAD")) {
       dsprsin(AA, nparms, 1.0e-28, 3*nobs+nc);
       linbcg(nparms, B, Cc, 4, 1.0e-14, 10, &iter, &err);
       for (j=1;j<nparms+1;j++) c[j] = Cc[j];
    } else {
      b = vector(1,nparms);
      W = vector(1,nparms);
      for (i=1;i<nparms+1;i++) {
        b[i] = B[i];
      }
      if (!strcmpcu(method, "CHOLESKI")) {
/*
 *       Cholesky decompose matrix and solve
 */
         if (choldc(AA, nparms, W)) {
            cholsl(AA, nparms, W, b, c);
         } else {
            syserr(FATAL, "solve", 
               "Matrix is not symmetric, positive definite. Use other method to solve\n");
         }
      } else {
         if (!strcmpcu(method, "SVD")) {
/*
 *          Single Value Decompose the A matrix
 */
            U      = matrix(1,nparms,1,nparms);
            V      = matrix(1,nparms,1,nparms);
            for (i=1;i<nparms+1;i++) {
               for (j=1;j<nparms+1;j++) {
                  U[i][j] = AA[i][j];
               }
            }
            svdcmp(U,nparms,nparms,W,V);
/*
 *          Check for representable dynamic range
 */
            wmax=0.0;
            for (i=1;i<nparms+1;i++) {
               if (W[i]>wmax) wmax=W[i];
            }
            wmin=wmax*1.0e-6;
            for (i=1;i<nparms+1;i++) {
               if (W[i]<wmin) W[i]=0.0;
            }
/* 
 *          Solve the b = Ax problem
 */
            svbksb(U,W,V,nparms,nparms,b,c);
/*
 *          Derive the Covariance matrix
 */
            svdvar(V, nparms, W, C);
            free_matrix(U,1,nparms,1);
            free_matrix(V,1,nparms,1);
         } else {
            syserr(FATAL, "solve", 
                "Unknown method %s, should be BICONGRAD, CHOLESKI, or SVD\n",
                method);
         }
      }
      free_vector(b,1);
      free_vector(W,1);
    }
    for (j=1;j<nparms+1;j++) {
       ic[j] = j;
    }
    free_dvector(sa,1);
    free_ivector(ija,1);
    free_dvector(Cc,1);
    free_dvector(B,1);
    free_dmatrix(AA,1,nparms,1);
    return 0.;
}

