#!/usr/bin/env python

"""
starcol.py - calculate magnitudes to estimate stellar colours

SYNOPSIS:
   starcol.py  AB / Vega    (is the result in AB or Vega mags)
               filter_1     (filter_1 for which you want mags)
               filter_2     (filter_2 for which you want mags)
               .
               .
               filter_n         (filter_n for which you want mags)
               output_all_file  (file listing the results for all stars)
               output_main_file (file listing results for Main Sequence stars)

DESCRIPTION:
   The program calculates AB or Vega magnitudes for stellar spectra
   from the Pickles library for the specified filters.  The absolute
   magnitudes have no meaning but the differences (colors) do. They
   can be used to verify photometric calibration of multi-colour
   observations for instance. We create one output file listing
   magnitudes for all stars in the Pickels library and a second one
   containing only main sequnece stars.

EXAMPLE:
   python starcol.py Vega WFI_U38 WFI_B123 WFI_V WFI_R WFI_I \
                     WFI_all.txt WFI_main.txt

   Calculate magnitudes for the filter U38, B123, V, R and I from
   WFI@MPG/ESO2.2m in the Vega system. Results are written for all
   stars to the file 'WFI_all.txt' and for main sequence sources to
   'WFI_main.txt'.
   
AUTHOR:
   Thomas Erben (terben@astro.uni-bonn.de)
"""

import sys

import bpz_tools
import cosmology

# A list of all stellar Pickles SEDs:
picklesall_seds = ['a0iii', 'a0i', 'a0iv', 'a0v', 'a2i', 'a2v', 'a3iii',
                   'a3v', 'a47iv', 'a5iii', 'a5v', 'a7iii', 'a7v', 'b0i',
                   'b0v', 'b12iii', 'b1i', 'b1v', 'b2ii', 'b2iv', 'b3iii',
                   'b3i', 'b3v', 'b57v', 'b5iii', 'b5ii', 'b5i', 'b6iv',
                   'b8i', 'b8v', 'b9iii', 'b9v', 'f02iv', 'f0iii', 'f0ii',
                   'f0i', 'f0v', 'f2iii', 'f2ii', 'f2v', 'f5iii', 'f5i',
                   'f5iv', 'f5v', 'f6v', 'f8i', 'f8iv', 'f8v', 'g0iii',
                   'g0i', 'g0iv', 'g0v', 'g2i', 'g2iv', 'g2v', 'g5iii',
                   'g5ii', 'g5i', 'g5iv', 'g5v', 'g8iii', 'g8i', 'g8iv',
                   'g8v', 'k01ii', 'k0iii', 'k0iv', 'k0v', 'k1iii', 'k1iv',
                   'k2iii', 'k2i', 'k2v', 'k34ii', 'k3iii', 'k3i', 'k3iv',
                   'k3v', 'k4iii', 'k4i', 'k4v', 'k5iii', 'k5v', 'k7v',
                   'm0iii', 'm0v', 'm10iii', 'm1iii', 'm1v', 'm2iii', 'm2i',
                   'm2p5v', 'm2v', 'm3iii', 'm3ii', 'm3v', 'm4iii', 'm4v',
                   'm5iii', 'm5v', 'm6iii', 'm6v', 'm7iii', 'm8iii', 'm9iii',
                   'o5v', 'o8iii', 'o9v', 'rf6v', 'rf8v', 'rg0v', 'rg5iii',
                   'rg5v', 'rk0iii', 'rk0v', 'rk1iii', 'rk2iii', 'rk3iii',
                   'rk4iii', 'rk5iii', 'wf5v', 'wf8v', 'wg0v', 'wg5iii',
                   'wg5v', 'wg8iii', 'wk0iii', 'wk1iii', 'wk2iii', 'wk3iii',
                   'wk4iii']

# A list of main sequence star templates:
picklesmain_seds = ['a0v', 'a2v', 'a3v', 'a5v', 'a7v', 'b0v', 'b1v', 'b3v',
                    'b57v', 'b8v', 'b9v', 'f0v', 'f2v', 'f5v', 'f6v', 'f8v',
                    'g0v', 'g2v', 'g5v', 'g8v', 'k0v', 'k2v', 'k3v', 'k4v',
                    'k5v', 'k7v', 'm0v', 'm1v', 'm2p5v', 'm2v', 'm3v', 'm4v',
                    'm5v', 'm6v', 'o5v', 'o9v', 'rf6v', 'rf8v', 'rg0v',
                    'rg5v', 'rk0v', 'wf5v', 'wf8v', 'wg0v', 'wg5v']

# very crude test for valid command line arguments:
if len(sys.argv) < 6:
    print __doc__
    sys.exit(1)

outfileall = open(sys.argv[-2], 'w')
outfilemain = open(sys.argv[-1], 'w')

# print some comment lines:
for actufile in [ outfileall, outfilemain ]:
    actufile.write("# Artificial magnitudes of stars in the Pickles\n")
    actufile.write("# stellar library; Only the differences (colours)\n")
    actufile.write("# of the stars are meaningful!\n")
    actufile.write("# \n")
    actufile.write("# This file contains magnitudes for ")
    if actufile == outfileall:
        actufile.write("all stars\n")
    else:
        actufile.write("main sequence stars\n")
        
    actufile.write("# in the Pickles library\n")
    actufile.write("# \n")
    actufile.write("# Included filters:\n")
    for i in range(2, len(sys.argv) - 2):
        actufile.write("# %s\n" % (sys.argv[i]))

    actufile.write("#\n")
    actufile.write("# Magnitude System: %s\n" % (sys.argv[1]))
    
# The calculation of magnitudes and a possible AB correction are
# straight forward calls to BPZ library functions:
for sed in picklesall_seds:
    if sys.argv[1] == "Vega":
        magstring = "%.2f" % (bpz_tools.ABtoVega(15.0, sys.argv[2]))
    else:
        magstring = "15.00"
        
    for i in range(3, len(sys.argv) - 2):
        mobs = cosmology.reobs(sed, 15.0, 0.0,
                               sys.argv[2], 0.0, sys.argv[i])

        if sys.argv[1] == "Vega":
            magstring = "%s %.2f" % (magstring,
                                     bpz_tools.ABtoVega(mobs, sys.argv[i]))
        else:
            magstring = "%s %.2f" % (magstring, mobs)

    magstring = magstring + "\n"        
    outfileall.write(magstring)

    if sed in picklesmain_seds:
        outfilemain.write(magstring)

outfileall.close()
outfilemain.close()

