#!/usr/bin/env python

# script to mask borders in the lower and upper parts of DECam@CTIO images
# for blooming effects. We observe that stripes in the upper and lower
# parts of DECam chips become useless if complete columns suffer from
# saturation of very bright stars. We 'solve' this problem by generously
# masking areas in the lower and upper parts of affected chips.

# HISTORY COMMENTS:
#
# 28.07.2014:
# project started
#
# 29.07.2014:
# I changed the algorithim to check whether a line is fully saturated.
# It suffices to consider sums of the image columns. This speeds up the
# whole script significantrly.

import sys
import numpy as np
import datetime
import pyfits

# main directory
MD = sys.argv[1]
# science directory
SD = sys.argv[2]
# weight directory
WD = sys.argv[3]
# image base
image = sys.argv[4]
# image ending
ending = sys.argv[5]
# weight ending
wending = sys.argv[6]
# saturation level
satlevkey = sys.argv[7]

# number of chips in DECam images:
nchips = 62

# the stripe width that is masked in upper and lower parts of chips affected
# by beright saturated stars:
width = 100

# default saturation level:
satlev_default = 60000.0

# minimum number of pixels that must be saturated in a row to initiate
# weight masking
nsaturated = 3900

for chip in xrange(nchips):
    imagepath = "%s/%s/%s_%d%s.fits" % (MD, SD, image, chip + 1, ending)

    try:
        hdulist = pyfits.open(imagepath)
        scidata = hdulist[0].data

        # get saturation level:
        header = hdulist[0].header

        try:
            satlev = header[satlevkey]
        except KeyError:
            print "Keyword %s not found in %s." % (satlevkey, imagepath)
            print "I use the default value for saturated pixels: %d" % \
                (satlev_default)

            satlev = satlev_default
            
    except IOError:
        print "I cannot open file %s. I skip it!" % (imagepath)

    # We sum the 2D array along the columns and check whether there
    # is one of the sum larger than the saturation level times the
    # number of pixels in the column.    
    summed = scidata.sum(axis=0)

    if np.max(summed) > satlev * nsaturated:
        weight = "%s/%s/%s_%d%s.weight.fits" % \
            (MD, WD, image, chip + 1, wending)

        print "Correcting %s for defects in chip %s" % (weight, chip + 1)
        
        try:
            # modify the weight data by putting zeros to the first and
            # last 'width' lines:
            currdate = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
            hdulist_weight = pyfits.open(weight, mode='update')
            weightdata = hdulist_weight[0].data

            weight_size = scidata.shape[0]
            weightline = np.ones(weight_size, dtype=np.float32)
            weightline[0:width] = 0.0
            weightline[(weight_size - 1 - width):(weight_size - 1)] = 0.0
            hdulist_weight[0].data = np.array(weightdata * \
                 weightline.reshape((weightdata.shape[0], 1)),
                                    dtype=np.float32)
            header_weight = hdulist_weight[0].header
            header_weight['HISTORY'] = ""
            header_weight['HISTORY'] = "remove_DECAM_blooming_effect: " \
                "called at %s" % (currdate)
            header_weight['HISTORY'] = "weight was corrected for " \
                "blooming defects in the science data."
            hdulist_weight.flush()
            hdulist_weight.close()

        except IOError:
            print "I cannot open weight file %s. I skip it!" % (weight)

    # close science image:        
    hdulist.close()
