#!/usr/bin/env python

"""
SCRIPT NAME:
  detect_OMEGACAM_ramp.py

  identify and mark 'ramps' in OMEGACAM data

ARGUMENTS:
  1. main directory     (e.g. /data/thomas/OMEGACAM/run_11_11_f)
  2. science directory  (e.g. SCIENCE_i_SDSS)
  3. weight directory   (e.g. WEIGHTS)
  3. image ending       (e.g. OFCSF)
  4. file with list of images to consider

  The 4th argument is a list of basenames from images. If images
  have names such as OMEGA.2013-05-28T02:45:18.759_4OFCSF.fits or
  OMEGA.2013-05-28T02:40:34.308_11OFCSF.sub.fits the list should
  look like:

  .
  .
  OMEGA.2013-05-28T02:45:18.759
  OMEGA.2013-05-28T02:40:34.308
  .
  .

DESCRIPTION:
  OMEGACAM data show, in an unpredictable way, 'ramps' in the upper
  part of the lowest chip-row (THELI chips 1-8) and/or in the lower
  parts of the highest chip-row (THELI chips 25-32). If ramps are
  present they are so in all exposures of an OB. This script tries to
  detect these ramps and, if necessary, modify weight images to give
  the ramp-regions zero weight. The detection is done in the following
  way:
  
  - The script should be given simultaneously all images of an OB to
    have a more robust ramp detection than with individual frames.
  - From each chip in corresponding rows we estimate an y median vector
    along the x-axis.
  - To the median vector we fit a straight line to regions that can be
    affected by ramps. We register chips whose line slope exceeds a
    certain threshold.
  - If more than x% of the chips from the provided images are registered
    we assume presence of a ramp and modify the weight images.
 
  Free parameters of the algorithm are the slope threshold and the percentage
  of chips that need to exceed the slope threshold.

AUTHOR:
  Thomas Erben (terben@astro.uni-bonn.de)
"""

# HISTORY COMMENTS:
#
# 05.08.2015:
# project started: Thomas Erben

import sys
import os
import glob
import multiprocessing
import itertools
import numpy as np
import datetime as dt
import astropy.io.fits as aif


def calc_ramp_slopes(imagedata):
    """
    The function calculates slopes of lines fitted around 'ramp regions'.

    Input:
    A list with four entries:
    - imagebase: basename of the image to test
    - dirname: The directory where images reside
    - ending: image ending
    - ramplocation: 'low' or 'up'. 'low' tests for ramps in the first
                    image row (THELI chips 1-8) and 'up' in the last
                    image rtow (THELI chips 28-32)

    Output:
    A list of line slopes around ramp locations
    """

    imagebase = imagedata[0]
    dirname = imagedata[1]
    ending = imagedata[2]
    ramplocation = imagedata[3]

    print "Testing %s ramp region for image %s/%s" % \
        (ramplocation, dirname, imagebase)

    width = 200
    if ramplocation == 'low':
        chips = range(1, 9)
        lowfitlimit = -width - 49
        highfitlimit = -50
    else:
        chips = range(25, 33)
        lowfitlimit = 51
        highfitlimit = 50 + width

    x = np.arange(1, width) # x-values for line fits    
    resultslopes = []    
    for chip in chips:
        image = "%s/%s_%d%s.fits" % (dirname, imagebase, chip, ending)
        try:
            data, header = aif.getdata(image, header=True)

            # do nothing if the chip is marked as bad:
            if 'BADCCD' not in header or header['BADCCD'] == 0:
                # determine essential part of median vector from the
                # image rows
                median = np.median(data, axis=1)[lowfitlimit:highfitlimit]

                # fit a line to the median and append the line slope
                # to the results list:
                resultslopes.append(np.polyfit(x, median, 1)[0])

        except IOError:
            print "I cannot open file %s. I skip it!" % (image)

    return resultslopes

def modify_weights(imagedata):
    """
    The function modifies weights of science images with identified
    ramps.

    Input:
    A list with four entries:
    - imagebase: basename of the weight image to modify
    - dirname: The directory where images reside
    - ending: image ending
    - ramplocation: 'low' or 'up'. 'low' tests for ramps in the first
                    image row (THELI chips 1-8) and 'up' in the last
                    image rtow (THELI chips 28-32)

    Output:
    Weight images are modified in place. The function has no
    further output.    
    """

    imagebase = imagedata[0]
    dirname = imagedata[1]
    ending = imagedata[2]
    ramplocation = imagedata[3]

    print "Modifying weight in %s ramp region for image %s" % \
        (ramplocation, imagebase)
    
    # we set weights in a generous region of width pixels around
    # the ramp regions to zero:
    width = 300
    if ramplocation == 'low':
        chips = range(1, 9)
        lowweightlimit = -1 - width
        highweightlimit = -1
    else:
        chips = range(25, 33)
        lowweightlimit = 0
        highweightlimit = width

    for chip in chips:
        image = "%s/%s_%d%s.weight.fits" % (dirname, imagebase, chip, ending)
        try:
            data, header = aif.getdata(image, header=True)

            data[lowweightlimit:highweightlimit,:] = 0

            # add History information to the header:
            currdate = dt.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
            header['HISTORY'] = ""
            header['HISTORY'] = "%s: called at %s" % \
                (os.path.basename(sys.argv[0]), currdate)
            header['HISTORY'] = "weight modified for ramp effect"
            
            # just overwrite the old file:
            aif.writeto(image, data, header, clobber=True)

        except IOError:
            print "I cannot open weight %s. I skip it!" % (image)
    
# Here the main script starts:
if __name__ == "__main__":

    if len(sys.argv) != 6:
        print __doc__
        sys.exit(1)

    # main directory
    MD = sys.argv[1]
    # science directory
    SD = sys.argv[2]
    # weight directory
    WD = sys.argv[3]
    # image ending
    ending = sys.argv[4]
    # image list
    inputfile = sys.argv[5]

    # the algorithm parameters (see docstring):
    slope_threshold = 0.1
    percentage_threshold = 0.5

    try:
        inputlist = open(inputfile, "r")
    except IOError:
        print "Error: cannot open file:", inputlist
        sys.exit(1)

    # parameters except those varying for each input catalogue:
    science_path = "%s/%s" % (MD, SD)
    weight_path = "%s/%s" % (MD, WD)

    baselist_science = [ science_path, ending ]
    baselist_weight = [ weight_path, ending ]
 
    processlist = []    

    # Initialise process pool:
    pool = multiprocessing.Pool()

    for location in [ 'low', 'up' ]:
        
        if location == 'low':
            print "Testing for a ramp in the lowest chip row\n"
        else:
            print "Testing for a ramp in the highest chip row\n"

        processlist = []
        inputlist.seek(0)
        for line in inputlist:
            processlist.append(line.strip().split() + \
                                   baselist_science + [ location ])
    
        # do the testing of images in parallel if possible:

        # execute the ramp-slope calculation with a 'pool-map' command:
        slopes_unflat = pool.map(calc_ramp_slopes, processlist)

        # flatten the list:
        slopes = np.array(list(itertools.chain.from_iterable(slopes_unflat)))

        if slopes.size > 0:
            if np.sum(np.abs(slopes) > slope_threshold) / \
                    (float)(slopes.size) > percentage_threshold:
                # we found a ramp!
                if location == 'low':                    
                    print '\nramp identified in lowest chip row!\n'
                else:
                    print '\nramp identified in highest chip row!'

                print "I modify weights now!\n"
    
                processlist = []
                inputlist.seek(0)
                for line in inputlist:
                    processlist.append(line.strip().split() + \
                                           baselist_weight + [ location ])
            
                # do the weight modifications again in parallel:
                pool.map(modify_weights, processlist)    

# clean up and bye:
inputlist.close()
sys.exit(0)
