#!/usr/bin/env python

# The script tries to detect horizontal electronic defects in
# OMEGACAM@VST data. These defects show up especially prominently in
# short ecposures (their signal seems to be constant regardless of the
# actual exposure time of the images). If they occur they do so in
# complete lines of the mosaic. We try to detect them by simply
# estimating the S/N on the lines of an OMEGACAM mosaic. We do so
# individually on all images belonging to a line. If the feature shows
# up in at least 5 images we mark all pixels identified with zeros in
# corresponding weight images. We do so with a safety margin around
# identified bad pixel lines.

# HISTORY COMMENTS:
#
# 12.02.2014:
# I ensure that the output weight data is 32-bit float (It was 64-bit
# by default on 64-bit machines)
#
# 04.08.2015:
# I replaced the pyfits module with astropy. pyfits will no longer
# be supported in the future!

import sys
import numpy as np
import datetime
import astropy.io.fits as aif

# main directory
MD = sys.argv[1]
# science directory
SD = sys.argv[2]
# weight directory
WD = sys.argv[3]
# image base
image = sys.argv[4]
# image ending
ending = sys.argv[5]
# weight ending
wending = sys.argv[6]

# The S/N limit for horizontal electric stripes
snratio = 300

# each bad pixels row is surrounded by a 'bad pixel'border of
# 'nborder' lines:
nborder = 10

# I do not really understand why the convolution below needs
# this factor of 2. Without it the safety border is only half the
# size of 'norder' 
borderarray = np.ones(nborder * 2)

# The script is currently hardwired to the geometry of
# OMEGACAM (8 x 4 chips)
nrows = 4
ncols = 8

for row in xrange(nrows):
    # we need to detect a horizontal stripes in at lease 'minpresent'
    # images of a row:
    minpresent = 5

    j=1
    for i in xrange(row * ncols + 1, (row + 1) * ncols + 1):
        imagepath = "%s/%s/%s_%d%s.fits" % (MD, SD, image, i, ending)

        try:
            hdulist = aif.open(imagepath)
            scidata = hdulist[0].data
            summed = scidata.sum(axis=1)
            stddev = scidata.std(axis=1)
            ratio = summed / stddev

            if j == 1:
                final = np.where(ratio > snratio, 1, 0)
            else:
                final += np.where(ratio > snratio, 1, 0)

            j = j + 1
            hdulist.close()

        except IOError:
            print "I cannot open file %s. I skip it!" % (imagepath)
            minpresent = minpresent - 1
    
    final = np.where(final >= minpresent, 1, 0)

    # modify weight if any horizontal stripes were detected:

    # final.any() is true when at least one the values in final
    # is unequal to zero:
    if final.any():
        # we realise the security margin around detected bad pixel
        # lines with a convolution. After the following operation,
        # finalconv contains zeros for the bad pixel lines.
        finalconv = np.where(np.convolve(final,
                                         borderarray, mode='same') > 0, 0, 1)

        for i in xrange(row * ncols + 1, (row + 1) * ncols + 1):
            weight = "%s/%s/%s_%d%s.weight.fits" % (MD, WD, image, i, wending)

            print "Correcting %s for defects" % (weight)

            try:
                # modify the weight data by putting zeros to the detected
                # horizontal stripes. Also write a history comment about
                # the change to the header
                currdate = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
                hdulist = aif.open(weight, mode='update')
                weightdata = hdulist[0].data
                # note that 'finalconv' is a one-dimensional array with the
                # y-dimension of the images. It is 'broadcasted' to the
                # 2-dimensional weight image with which it is multiplied.
                # Note also that we need to force 32-bits for the image data
                # array:                
                hdulist[0].data = np.array(weightdata * \
                     finalconv.reshape((weightdata.shape[0],1)),
                                           dtype=np.float32)
                header = hdulist[0].header
                header['HISTORY'] = ""
                header['HISTORY'] = "remove_electric_effect: called at %s" % \
                    (currdate)
                header['HISTORY'] = "weight was corrected for horizontal " \
                    "electronic defects"
                hdulist.flush()
                hdulist.close()

            except IOError:
                print "I cannot open weight file %s. I skip it!" % (weight)

