#!/bin/bash -u

# ----------------------------------------------------------------
# File Name:           convert_mask.sh
# Author:              Thomas Erben (terben@astro.uni-bonn.de)
# Last modified on:    05/02/2009
# Description:         convert ds9 region masks
# ----------------------------------------------------------------

# Script history information:
#
# 09.01.2009:
# Project started 
#
# 05.02.2009:
# Comment lines at the begiining of the original region file
# are propagated now; previously all comment lines were deleted.

# define THELI_DEBUG because of the '-u' script flag
# (the use of undefined variables will be treated as errors!)
# THELI_DEBUG is used in the cleanTmpFiles function.
THELI_DEBUG=${THELI_DEBUG:-""}

##
## function definitions:
##
function printUsage 
{
  echo "SCRIPT NAME:"
  echo "    convert_mask.sh - converts ds9 region masks between RA/DEC and pix"
  echo ""
  echo "ARGUMENTS:"
  echo "    1. input ds9 region mask"
  echo "    2. conversion method (see below)"
  echo "    3. FITS image or header with astrometric"
  echo "       information for mask conversion "
  echo "   (4. FITS image or header with astrometric"
  echo "       information for mask conversion; for certain"
  echo "       conversion methds only)"
  echo "    n. name of output mask"
  echo ""
  echo "DESCRIPTION:"
  echo "    The script takes a SAOimage/ds9 region file mask (first argument)"
  echo "    as input and converts it to a new mask. The conversion is either"
  echo "    from RA/DEC to pixel coordinates (conversion method RADECPIX),"
  echo "    from pixel to RA/DEC coordinates (conversion method PIXRADEC) or"
  echo "    from pixel coordinates of one image to pixel coordinates of"
  echo "    another image (conversion method PIXPIX).  In case of conversion"
  echo "    modes RADECPIX and PIXRADEC the third argument has to be a FITS"
  echo "    image or an image header giving astrometric information for the"
  echo "    input mask.  In case of mode PIXPIX a ssecond image/header with"
  echo "    information on the target image has to be provided.  If a FITS"
  echo "    header is given it needs to be in binary (not ASCII!) format. One"
  echo "    possibility to create such a header is the imcat utility 'imhead"
  echo "    (with option -b)'."
  echo ""
  echo "    The script propagates ONLY those comment lines (lines starting"
  echo "    with a \# and not containing any polygon) that are the beginning"
  echo "    of the file."
  echo ""
  echo "    If the target frame is pixel based only masks falling within this"
  echo "    frame (defined via the NAXIS1 and NAXIS2 header keywords) are"
  echo "    printed."
  echo "    "
  echo "IMPORTANT DEPENDENCIES:"
  echo "    Besids standard UNIX tools, the script needs the programs"
  echo "    'sky2xy', 'xy2sky', 'dfits' and 'fitsort' in your path. 'sky2xy' and"
  echo "    'xy2sky' are MODIFIED versions (by the script author) of Doug"
  echo "    Minsks wcstools.  These modified versions allow data input from"
  echo "    STDIN which this script makes use of."
  echo ""
  echo "EXAMPLES:"
  echo "    ./convert_mask.sh CFHTLS_W_i_022539-041200_T0005.reg RADECPIX \\"
  echo "                      W1p2p3_i.V1.7A.swarp.cut.fits \\"
  echo "                      W1p2p3.head W1p2p3_pix.reg"
  echo ""
  echo "AUTHOR:"
  echo "    Thomas Erben         (terben@astro.uni-bonn.de)"
  echo ""
}

# If the environment variable THELI_DEBUG is set temporary files
# are NOT cleaned up!
#
function cleanTmpFiles
{
  if [ -z ${THELI_DEBUG} ]; then
    echo "Cleaning temporary files for script $0"

    test -f ./vertices_$$.txt && rm ./vertices_$$.txt
  else
    echo "Variable THELI_DEBUG set!"
    echo "No cleaning of temp. files for script `basename $0`"    
  fi
}

# Handling of program interruption by CRTL-C
trap "echo 'Script $0 interrupted!! Cleaning up and exiting!'; \
      cleanTmpFiles; exit 1" INT

##
## initial sanity checks
##

# check validity of command line arguments:
# the first condition with '-lt 4' is necessary to prevent a possible
# '$2 unbound' error:
if [ $# -lt 4 ] ||\
   ( [ $# -ne 4 ] && ( [ $2 = "PIXPIX" ] && [ $# -ne 5 ] ) ); then
  printUsage
  exit 1
fi

# Existence of input files:
if [ "$2" = "PIXPIX" ]; then
  if [ ! -f $3 ] || [ ! -f $4 ]; then
    echo "Input/Output reference image/header not present!"
    echo "Exiting!"
    exit 1;
  fi
else
  if [ ! -f $3 ]; then
    echo "Input reference image/header not present!"
    echo "Exiting!"
    exit 1;
  fi    
fi

##
## Main script tasks
##

# If we convert to some pixel grid the output catalogue
# will contain only masks within the target image. The
# following variables serve this purpose:
SIZEXOUT=0
SIZEYOUT=0
if [ "$2" = "PIXPIX" ]; then
  SIZEXOUT=`dfits $4 | fitsort -d NAXIS1 | awk '{print $2}'`
  SIZEYOUT=`dfits $4 | fitsort -d NAXIS2 | awk '{print $2}'`
elif [ "$2" = "RADECPIX" ]; then
  SIZEXOUT=`dfits $3 | fitsort -d NAXIS1 | awk '{print $2}'`
  SIZEYOUT=`dfits $3 | fitsort -d NAXIS2 | awk '{print $2}'`
fi  

test -f ./vertices_$$.txt && rm ./vertices_$$.txt

# The conversion is done in a 'larger' gawk program.
# For the conversion from RA-DEC to pixels we need to
# go through some C-code. We found that treating one
# polygon after the other leads to prohibitively long
# execution times. We therefore collect ALL vertices
# from ALL polygons and go through 'sky2xy-xy2sky' only
# once. The administration of the polygons then leads
# to a bit of an administration effort :-)

# the 'sed' gets rid of empty lines in the region files;
# those would screw up the result at the end.
sed 's/  *//' $1 |\
gawk -v convmethod="$2" \
    'BEGIN {FS="[ (),]"; npoly = 0; ncomm = 0; getcomm = 0 ;} {
    # save complete comment lines at beginning of region file:
    if ($0 ~ /^#/ ) { 
      if (NR == 1) {
        getcomm = 1;
      }
      if (getcomm == 1) {
        comm[NR] = $0;
        ncomm++;
      }
    } else if ($1 ~ /[Pp][Oo][Ll][Yy][Gg][Oo][Nn]/) {
      getcomm = 0;
      # first get the comment at the end of line (if any)
      comment[npoly] = ""; commentcount = 10000
      for(i = 2; i <= NF; i++) {
        if($i ~ /#/) { 
          commentcount = i;
        }
        if(i >= commentcount) { 
          comment[npoly] = comment[npoly] " " $i;
        }
      }  
      nvertexcurr = 0;
      for(i = 2; i <= NF; i++) {
        if(i%2 == 0 && $i ~ /^[-+0-9][0-9]*/ ) {
          j = i + 1; 
          print $i, $j >> "vertices_'$$'.txt"
          nvertexcurr ++;
        }
      } 
      nvertex[npoly] = nvertexcurr;
      npoly++; 
    }
  } 
  END {
    close("vertices_'$$'.txt");
    # print comment lines at start of region file:
    for (i = 1; i <= ncomm; i++) {
      print comm[i];
    }
    if (convmethod == "PIXPIX" ) {
      command = "cat vertices_'$$'.txt | xy2sky -n 6 -d "
      command = command "'$3' @STDIN | sky2xy '$4' @STDIN";
      fieldcol1 = 5;
      fieldcol2 = 6;
    }
    if (convmethod == "PIXRADEC" ) {
      command = "cat vertices_'$$'.txt | xy2sky -n 6 -d '$3' @STDIN";
      fieldcol1 = 1;
      fieldcol2 = 2;
      # identify the polygon file as a WCS file:
      print "fk5";
    }
    if (convmethod == "RADECPIX" ) {
      command = "cat vertices_'$$'.txt | sky2xy '$3' @STDIN";
      fieldcol1 = 5;
      fieldcol2 = 6;
    }
    npolyread = 0; 
    nread = 0; 
    polygonstring = "polygon("
    while ((command | getline tmp) > 0) {
      split(tmp, tmpfields, " ");
      nread ++;
      if(nread < nvertex[npolyread]) {
        polygonstring = polygonstring tmpfields[fieldcol1] "," 
        polygonstring = polygonstring tmpfields[fieldcol2] ",";

        if (convmethod == "PIXPIX" || convmethod == "RADECPIX") {
          if ((tmpfields[fieldcol1] > 0 && 
               tmpfields[fieldcol1] < '${SIZEXOUT}') && 
              (tmpfields[fieldcol2] > 0 && 
               tmpfields[fieldcol2] < '${SIZEYOUT}')) {
            usepolygon = 1;
          }
        }
      } else {
        if(usepolygon == 1 || convmethod == "PIXRADEC") { 
          printstring = polygonstring tmpfields[fieldcol1] "," 
          printstring = printstring tmpfields[fieldcol2] ") " comment[npolyread];
          print printstring;
        }
        polygonstring = "polygon(";
        nread = 0;
        npolyread ++;
        usepolygon = 0;
      }          
    }
  }' > ${!#}

echo "Mask ${!#} written!"

# clean up temporary files and bye
cleanTmpFiles

exit 0;
