#!/usr/bin/env python

# script to convert angles from sexagesimal to decimal notation and
# vice-versa. Its functionality unifies the four programs hmstodecimal,
# decimaltohms etc. from the imcat package.

# HISTORY COMMENTS:
# =================
#
# 21.04.2015:
# I corrected a bug for conversion of negative angles (such as
# '-00:03:30' to positive decimal degrees!

"""
USAGE: convertangles.py Ra/Dec sexa/dec angle  OR
       convertangles.py Ra/Dec sexa/dec column file1 file2 ....

DESCRIPTION:
The program converts angles from sexagesimal to decimal notation and
vice versa. In the first version a single angle is converted.
In the second version all values in a column of files are converted.
All output goes to STDOUT.

NOTE:
No real testing about the validity of input data takes place!

EXAMPLES:
(1) convertangles.py Ra sexa "21:13:45.5"

    convert the input Ra value "21:13:45.5" from sexagesimal to
    decimal notation

(2) convertangles.py Ra dec 2 test.txt

    The second column in file test.txt represents Ra values in decimal
    notation. Convert them to sexagesimal notation.

(3) convertangles.py Ra dec 2 test.txt | convertangles.py Dec dec 3 -

    The second column in file test.txt represents Ra values in decimal
    notation and the third column represents Dec values in decimal
    notation. Convert them to sexagesimal notation.

AUTHOR:
    Thomas Erben (terben@astro.uni-bonn.de)
"""

import math
import sys
import fileinput

# converts an angle from degrees to sexagesimal notation
def angtosexa(ang):
    sign = 1
    if (ang < 0.0):
        sign = -1

    ang = ang * sign
    (rem, deg) = math.modf(ang)
    (sec, min) = math.modf(60.0 * rem)
    sec = 60.0 * sec

    if sign == 1:
        sexa = "%02d:%02d:%07.4f" % ((int)(deg), (int)(min), sec)
    else:
        sexa = "-%02d:%02d:%07.4f" % ((int)(deg), (int)(min), sec)

    return sexa

# converts an angle from sexagesimal notation to degrees
def angtodec(ang):
    # we need to test for the sign before we convert individual
    # numbers in a string like '12:15:37' to floats. If we did
    # this after the conversion the 'negative' angle '-00:03:30'
    # would be converted to a positive angle!!
    sign = 1
    if ang[0] == '-':
        sign = -1
    (hours, mins, secs) = map(float, ang.split(":"))

    hours = sign * hours
    dec = sign * (hours + mins / 60. + secs / 3600.)

    return dec

# driver function for angtosexa and angtodec
def convertang(sourcetype, sourceform, ang):
    if sourcetype == 'Ra':
        if sourceform == 'sexa':
            return "%.10f" % (angtodec(ang) * 15.0)
        else:
            return angtosexa((float)(ang) / 15.0)
    else:
        if sourceform == 'sexa':
            return "%.10f" % (angtodec(ang))
        else:
            return angtosexa((float)(ang))

# main program
if len(sys.argv) < 4:
    print __doc__
    sys.exit(1)

sourcetype = sys.argv[1]
sourceform = sys.argv[2]

if len(sys.argv) == 4:
    print convertang(sourcetype, sourceform, sys.argv[3])
else:
    column = (int)(sys.argv[3])
    for line in fileinput.input(sys.argv[4:]):
        linesplit = line.split()
        
        if len(linesplit) < column:
            sys.stderr.write("File %s does not have enough columns!\n" %
                             (fileinput.filename()))
            sys.exit(1)

        linesplit[column - 1] = convertang(sourcetype, sourceform,
                                           linesplit[column - 1])

        # Join the line again and print the result:
        print "   ".join(map(str, linesplit))

# and bye:
sys.exit(0)
