#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include "utils_globals.h"
#include "utils_macros.h"
#include "lists_globals.h"
#include "options_globals.h"
#include "files_cat_globals.h"
#include "make_set_vers.h"

#ifndef MAXCHAR
#define MAXCHAR  256
#endif

/*
 * 15.07.01:
 * modifications to avoid compiler warnings
 */

int	global_set;

typedef struct {
    char	infile[MAXCHAR], outfile[MAXCHAR];
    int		quiet;
} control_struct;
/**/
void usage(char *me)
/****i* usage ***************************************************
 *
 *  NAME
 *      usage - Output on stderr the program usage information
 *
 *  SYNOPSIS
 *      void usage(char *me)
 *
 *  FUNCTION
 *      User interface information printed to standard error. This allows
 *      for a more pleasant  
 *
 *  INPUTS
 *      me - The calling name of the program
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *      This routine exits the program with an error code 1
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 01-08-1996
 *
 ******************************************************************************/
{
    fprintf(stderr,"Usage: %s\n", me);
    fprintf(stderr," 	-i incat -o outcat [-q]\n\n");
    fprintf(stderr," This program adds a set sequence number to a night files catalog\n");
    exit(1);
}
/**/
void init(int argc, char *argv[], control_struct *control)
/****i* init ***************************************************
 *
 *  NAME
 *      init - Parse the command line arguments and fill the control structure
 *
 *  SYNOPSIS
 *      void init(int argc, char *argv[], control_struct *c)
 *
 *  FUNCTION
 *      This routine reads the command line arguments, interprets them
 *      in terms of program usage and fills the control structure with
 *      the appropriate values from the arguments.
 *
 *  INPUTS
 *      argc   - The number of command line arguments
 *      argv   - The list of command line argument strings
 *
 *  RESULTS
 *      c      - The control structure the receives the argument information
 *
 *  RETURNS
 *      None
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *      In case of command line errors, the usage routine is called.
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 01-08-1996
 *
 ******************************************************************************/
{
/*
 *  Default file name is calling name with .conf extension
 */
    int i;

    strcpy(control->infile, "");
    strcpy(control->outfile, "");
    control->quiet = 1;
    if (argc == 1) usage(argv[0]);
    for (i=1; i<argc; i++) {
      if (argv[i][0] == '-') {
         switch(tolower((int)argv[i][1])) {
         case 'i': strcpy(control->infile,argv[++i]);
                   break;
         case 'o': strcpy(control->outfile,argv[++i]);
                   break;
         case 'q': control->quiet = 0; 
                   break;
         default:  syserr(WARNING,"init","Command line syntax error\n");
                   usage(argv[0]);
         }
      } else {
         syserr(WARNING,"init","Command line syntax error\n");
         usage(argv[0]);
      }
   }
   if (control->infile[0] == '\0' || control->outfile[0] == '\0') {
      syserr(WARNING,"init","input or output file missing\n");
      usage(argv[0]);
   }
}


int main(int argc, char *argv[])
{
/****** make_sets ***************************************************
 *
 *  NAME	
 *      make_sets - Identify and mark individual sets in a night's catalog
 *
 *  SYNOPSIS
 *      make_sets -i incat -o outcat
 *               [-q]
 *
 *  FUNCTION
 *      This program adds a set sequence number to a night files catalog.
 *      The sets are defined in terms of consecutive frames in a strip or
 *      consecutive frames in photometric sequence. 
 *
 *  INPUTS
 *      -i incat   - The input catalog as created by make_files_cat
 *      -o outcat  - The output catalog with a setnumber column added.
 *      -q         - Make the program verbose
 *
 *  RESULTS
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *      Currently there is a parameter OCTAGONAL that has the value 8 to
 *      allow distinction of consecutive photometric sequences of the same
 *      standard star.
 *
 *      It is necessary to run this program after a successful completion
 *      of the make_files_cat program. It has to be run only once on a
 *      night's catalog.
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 01-08-1996
 *
 ******************************************************************************/
    int		n, m, o, p, *numbers, i, j, k, l, ll, type, set_nr;
    char	**std_names, *std_list, **filts, **cams, dum[MAXCHAR];
    file_id	*files;
    control_struct control;
    char	header[MAXCHAR];

    init(argc, argv, &control);
 
    sprintf(header,"\n   %s Version: %s (%s)\n",NAME,VERS,DATE);
    VPRINTF(NORMAL,header);

    p = o = n = read_files_tab(control.infile);
    numbers = uniq_int(get_number_nrs(), &n);
    cams  = uniq_strings(get_cameras(0), CAMLEN, &p);
    filts = uniq_strings(get_filters(0), FILTLEN, &o);
    for (j=0;j<o;j++) {
    global_set = 0;
    for (i=0;i<n;i++) {
       switch ((type = obstype_nr(numbers[i]))) {
       case FLAG:
          break;
       case CALIB:
          std_list = get_all_names(numbers[i], OBJLEN, &m);
          std_names = uniq_strings(std_list, OBJLEN, &m);
          for (k=0;k<m;k++) {
             if (std_names[k][0] != '\0') {
                files = get_files(numbers[i], filts[j], FILTLEN,
                                  std_names[k], OBJLEN, type, &l);
                set_nr = 0; global_set++;
                if (control.quiet)
                   printf("Calib %d %s %d %s\n", numbers[k], std_names[k], 
                       global_set,filts[j]);
                for (ll=0;ll<l;ll++) {
                   if ((files[ll].seq_nr % OCTAGONAL) == 1) {
                      if (set_nr != 0) {
                         global_set++;
                         if (control.quiet)
                            printf("Calib %d %s %d %s\n", numbers[i], 
                                   std_names[k], global_set, filts[j]);
                      }
                      set_nr++;
                   }
                   set_set_nrs(files[ll].filename, global_set);
                }
                for (ll=0;ll<l;ll++) 
                   ED_FREE(files[ll].filename);
                ED_FREE(std_names[k]);
             }
          }
          ED_FREE(std_names);
          break;
       case STRIP:
          global_set++;
          if (control.quiet)
             printf("Strip %d %d %s\n", numbers[i], global_set, filts[j]);
          files = get_files(numbers[i], filts[j],FILTLEN,
                            dum, OBJLEN, type, &l);
          for (ll=0;ll<l;ll++) {
             strcpy(dum, strip_name(numbers[i]));
             set_set_nrs(files[ll].filename, global_set);
          }
          for (ll=0;ll<l;ll++) 
             ED_FREE(files[ll].filename);
          break;
       }
    }
    }
    ED_FREE(numbers); 
    paste_set_nrs(control.outfile, "make_sets");

    return 0;
}

