#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <math.h>
#include "utils_globals.h"
#include "utils_macros.h"
#include "lists_globals.h"
#include "options_globals.h"
#include "files_cat_globals.h"
#include "set_info_vers.h"

#define NPROCS     18
#define NSETS      0
#define CHANNELS   1
#define ROOT_NAME  2
#define BLEND_LIST 3
#define SET_LIST   4
#define SET_TYPE   5
#define OBJS_DIR   6
#define ASTROM_NAME 7
#define NFIELDS    8
#define STD_NAME   9
#define FLAG_FILE  10
#define STRIP_NAME 11
#define EIS_LIST   12
#define NUMBER     13
#define FLAT_FILE  14
#define CAMERAS    15
#define INSTRUMENT 16
#define WEIGHT_FILE 17

/* 15.07.01:
 * modifications to avoid compiler warnings
 */

static char *process_names[NPROCS] = {"NSETS", "CHANNELS", "ROOT_NAME",
                                      "BLEND_LIST", "SET_LIST", "SET_TYPE",
                                      "OBJS_DIR", "ASTROM_NAME", "NFIELDS",
                                      "STD_NAME", "FLAG_FILE", "STRIP_NAME",
                                      "EIS_LIST", "NUMBER", "FLAT_FILE",
                                      "CAMERAS", "INSTRUMENT", "WEIGHT_FILE"};
#ifndef MAXCHAR
#define MAXCHAR  256
#endif

typedef struct {
    char	process_type[MAXCHAR], catalog[MAXCHAR], 
                filter_name[MAXCHAR], camera_name[MAXCHAR];
    int		setnr, equal;
} control_struct;

int process_type(char *s)
{
    int		i;

    for (i=0;i<NPROCS;i++) {
       if (strcmp(to_upper(s),process_names[i]) == 0) {
          return (i);
       }
    }
    syserr(FATAL, "process_type", "Unknown process type %s\n", s);
    return UNKNOWN;
}


void usage(char *me)
{
    fprintf(stderr,"Usage: %s \n",me);
    fprintf(stderr,"		-r report_type\n");
    fprintf(stderr,"		-i night_catalog\n");
    fprintf(stderr,"		[-n set_number]\n\n");
    fprintf(stderr,"		[-f filter_name]\n\n");
    fprintf(stderr,"		[-c camera_name]\n\n");
    fprintf(stderr,"		[-e] (equal filter and camera)\n\n");
    fprintf(stderr,"Possible report_type:\n");
    fprintf(stderr,"	NSETS:      Report number of sets in the file\n");
    fprintf(stderr,"	CHANNELS:   Report available filter names for a given set\n");
    fprintf(stderr,"	CAMERAS:   Report available channel names for a given set\n");
    fprintf(stderr,"	SET_TYPE:   Report the type of a given set\n");
    fprintf(stderr,"	OBJS_DIR:   Report objects dirname for a given set\n");
    fprintf(stderr,"	SET_LIST:   Produce listing of all channel setnames\n");
    fprintf(stderr,"	ROOT_NAME:  Provide root file name for a given set and channel\n");
    fprintf(stderr,"	ASTROM_NAME:Provide astrom file name for a given set\n");
    fprintf(stderr,"	STD_NAME:   Provide photometric standard name for a given set\n");
    fprintf(stderr,"	NFIELDS:    Report number of fields for a given set and channel\n");
    fprintf(stderr,"	BLEND_LIST: Produce listing for blend input list for a given set\n");
    fprintf(stderr,"	            and channel\n");
    fprintf(stderr,"	FLAG_FILE:  Report the path to the flag file for (set, channel)\n");
    fprintf(stderr,"	FLAT_FILE:  Report the path to the flat file for (set, channel)\n");
    fprintf(stderr,"	STRIP_NAME: Report the name of the strip for a given set\n");
    fprintf(stderr,"	EIS_LIST:   Produce listing for sex input list for a given set\n");
    fprintf(stderr,"	NUMBER:     Report the running number  for a given set\n");
    fprintf(stderr,"	INSTRUMENT: Report the instrument name of a given set\n");
    fprintf(stderr,"	WEIGHT_FILE: Report the path to the weight file for (set, channel)\n");

    exit(1);
}

void init(int argc, char *argv[], control_struct *control)
{
/*
 *  Default file name is calling name with .conf extension
 */
    int i;

    if (argc == 1) usage(argv[0]);
    control->setnr = UNKNOWN;
    control->equal = FALSE;
    strcpy(control->filter_name,"");
    strcpy(control->camera_name,"");
    for (i=1; i<argc; i++) {
      if (argv[i][0] == '-') {
         switch(tolower((int)argv[i][1])) {
         case 'r': strcpy(control->process_type,argv[++i]);
                   break;
         case 'i': strcpy(control->catalog,argv[++i]);
                   break;
         case 'n': control->setnr = atoi(argv[++i]);
                   break;
         case 'f': strcpy(control->filter_name,argv[++i]);
                   break;
         case 'c': strcpy(control->camera_name,argv[++i]);
                   break;
         case 'e': control->equal = TRUE;
                   i++;
                   break;
         default:  syserr(WARNING,"init","Command line syntax error\n");
                   usage(argv[0]);
         }
      } else {
         syserr(WARNING,"init","Command line syntax error\n");
         usage(argv[0]);
      }
   }
   if (control->process_type[0] == '\0' || control->catalog[0] == '\0')
      syserr(FATAL, "init", "No report type or input catalog specified\n");
}


int main(int argc, char *argv[])
{
/****** set_info ***************************************************
 *
 *  NAME	
 *      set_info - Provide the caller with specific night set information
 *
 *  SYNOPSIS
 *      set_info -r report_type
 *               -i night_catalog
 *              [-n set_number]
 *              [-f filter_name]
 *              [-c camera_name]
 *
 *  FUNCTION
 *      This program reports several types of information to the caller
 *      depending on the argument list in tha call. The possible 
 *      report_types are:
 *          NSETS:      Report number of sets in the file
 *          CHANNELS:   Report available filter names for a given set
 *          CAMERAS:    Report available camera names for a given set
 *          SET_TYPE:   Report the type of a given set
 *          OBJS_DIR:   Report objects dirname for a given set
 *          SET_LIST:   Produce listing of all channel setnames
 *          ROOT_NAME:  Provide root file name for a given set and channel
 *          ASTROM_NAME:Provide astrom file name for a given set
 *          STD_NAME:   Provide photometric standard name for a given set
 *          NFIELDS:    Report number of fields for a given set and channel
 *          FLAG_FILE:  Return the full pathname of the FLAG file (set,channel)
 *          FLAT_FILE:  Return the full pathname of the FLAT file (set,channel)
 *          STRIP_NAME: Return the full name of the Strip for a set
 *          NUMBER:     Return the running number for a given set.
 *          BLEND_LIST: Produce listing for blend input list for a given set
 *                      and channel
 *          EIS_LIST:   Produce listing for EIS sex input list for a given set
 *                      and channel
 *
 *  INPUTS
 *      -r report_type   - The type of information needed
 *      -i night_catalog - The name of the input catalog
 *     [-n set_number]   - The number of the set (if required)
 *     [-f filter_name] - The name of the filter (if required)
 *     [-c camera_name] - The name of the camera (if required)
 *
 *  RESULTS
 *      On standard output the program responds with the requested
 *      information.
 *
 *  RETURNS
 *      TRUE  if the requested information could be given
 *      FALSE if the requested information could not be given or the
 *            command line arguments were incomplete
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *      This program is the main driver for guiding and constructing the 
 *      input to the LDAC pipeline routines. 
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 01-08-1996
 *
 ******************************************************************************/
    int		n, o, p, i, set_type;
    int		proc_type;
    char	**filts, **cams;
    control_struct control;
    char	header[MAXCHAR];

    init(argc, argv, &control);
 
    sprintf(header,"\n   %s Version: %s (%s)\n",NAME,VERS,DATE);
    VPRINTF(NORMAL,header);

    p = o = n = read_files_tab(control.catalog);
    proc_type = process_type(control.process_type);
    filts = uniq_strings(get_filters(control.setnr), FILTLEN, &o);
    cams  = uniq_strings(get_cameras(control.setnr), CAMLEN, &p);
/*
 *  First checkout set globals things
 */
    switch (proc_type) {
    case NSETS:
       printf("%d\n",number_of_sets());
       break;
    case CHANNELS:   
       for (i=0;i<o;i++) printf("%s ",filts[i]);
       printf("\n");
       break;
    case CAMERAS:   
       for (i=0;i<p;i++) printf("%s ",cams[i]);
       printf("\n");
       break;
    default:
/*
 *     Specific processing for a given set number
 */
       if (control.setnr < 0) {
          syserr(FATAL, "main", "Please specify set number\n");
       }
       if (control.setnr > number_of_sets()) {
          syserr(FATAL, "main", "Set number too large\n");
       }
       set_type = obstype_set(control.setnr);
       switch (proc_type) {
       case SET_TYPE:   
          switch(set_type) {
          case CALIB: 
             printf("Calib\n");
             break;
          case STRIP:
             printf("Strip\n");
             break;
          }
          break;
       case STRIP_NAME:
          switch(set_type) {
          case CALIB: 
             exit(1);
             break;
          case STRIP:
             get_strip_name(control.setnr);
             break;
          default:
             get_strip_name(control.setnr);
             break;
          }
          break;
       case OBJS_DIR:   
          make_objs_dir(control.setnr);
          break;
       case SET_LIST:
          make_set_names(control.setnr, control.equal);
          break;
       case ASTROM_NAME:
          make_astrom_name(control.setnr);
          break;
       case STD_NAME:
          make_std_name(control.setnr);
          break;
       case NUMBER:
          make_number(control.setnr);
          break;
       case INSTRUMENT:
          make_instrument(control.setnr);
          break;
       default:
/*
 *        Remaining functions need a filter name specification
 */
          if (control.filter_name[0] == '\0') {
             syserr(FATAL,"main", "No filter name provided\n");
          }
          for (i=0;i<o;i++) {
             if (strcmp(to_upper(filts[i]), 
                        to_upper(control.filter_name)) == 0) {
                break;
             }
          }
          if (i == o) {
             syserr(FATAL,"main", "No filter name called %s\n",
                                  control.filter_name);
          }
          if (control.camera_name[0] == '\0') {
             syserr(FATAL,"main", "No camera name provided\n");
          }
          for (i=0;i<p;i++) {
             if (strcmp(to_upper(cams[i]), 
                        to_upper(control.camera_name)) == 0) {
                break;
             }
          }
          if (i == p) {
             syserr(FATAL,"main", "No camera name called %s\n",
                                  control.camera_name);
          }
          switch (proc_type) {
          case ROOT_NAME:
             make_root_name(control.setnr, control.filter_name,
                            control.camera_name);
             printf("\n");
             break;
          case EIS_LIST:
             make_eis_list(control.setnr, control.filter_name, 
                           control.camera_name);
             break;
          case BLEND_LIST:
             make_blend_list(control.setnr, control.filter_name,
                            control.camera_name);
             break;
          case NFIELDS:
             make_field_count(control.setnr, control.filter_name,
                            control.camera_name);
             break;
          case FLAG_FILE:
             make_file_path(control.setnr, control.filter_name,
                            control.camera_name, FLAG);
             printf("\n");
             break;
          case WEIGHT_FILE:
             make_file_path(control.setnr, control.filter_name,
                              control.camera_name, WEIGHT);
             printf("\n");
             break;
          case FLAT_FILE:
             make_file_path(control.setnr, control.filter_name,
                            control.camera_name, FLAT);
             printf("\n");
             break;
          }
       }
    }

    return 0;
}
