 /*
 				header.c

*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*
*	Part of:	SWarp
*
*	Author:		E.BERTIN (IAP)
*
*	Contents:	Provide additional FITS header management.
*
*	Last modify:	08/10/2002
*
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

/*
 * 03.03.2004:
 * function writefitsinfo_outfield:
 * we now add 20 dummy keywords at the end
 * of the output header: this is for adding
 * additional information to the image header
 * after the coaddition process.
 *
 * 26.06.2004:
 * function writefitsinfo_outfield:
 * several COMMENT lines are only written if no external
 * header is given. If we first create a header for
 * the output image and do coaddition later several COMMENT
 * appeared twice.
 *
 * 05.07.2004:
 * function read_aschead:
 * I added the keywords RADECSYS, EQUINOX, CUNIT1 and CUNIT2
 * to the list of keywords not touched by this routine.
 * When they are treated by the routine the final headers
 * are no longer FITS conform (to be revisited as the problem
 * is not cleanly solved in this way)
 *
 * 29.01.2005:
 * function writefitsinfo_outfield:
 * I added HISTORY information to the FITS header.
 * This was done to be consistent with the other programs.
 * The original header already contains very good HISTORY
 * and Logging information.
 * 
 * 14.10.2007:
 * I increased the number of DUMMY jeyword from 20 to 50.
 *
 * 03.09.2010:
 * The time put into the history of generated images was in GMT
 * I changed that to the local time. The program name put into the
 * history is now swarp_theli instaead of swarp.
 */

#ifdef HAVE_CONFIG_H
#include	"config.h"
#endif

#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>

#include "define.h"
#include "globals.h"
#include "data.h"
#include "fits/fitscat.h"
#include "fitswcs.h"
#include "field.h"
#include "header.h"
#include "key.h"
#include "prefs.h"
#include "wcs/wcs.h"


/****** read_aschead ********************************************************
PROTO	int	read_aschead(char *filename, int frameno, tabstruct *tab)
PURPOSE	Read a ASCII header file and update the current field's tab
INPUT	Name of the ASCII file,
	Frame number (if extensions),
	Tab structure.
OUTPUT	RETURN_OK if the file was found, RETURN_ERROR otherwise.
NOTES	-.
AUTHOR	E. Bertin (IAP)
VERSION	08/06/2002
 ***/
int	read_aschead(char *filename, int frameno, tabstruct *tab)
  {
   static char	keyword[MAXCHAR],data[MAXCHAR],comment[MAXCHAR];
   FILE		*file;
   h_type	htype;
   t_type	ttype;
   int		i, pos;

  if (file=fopen(filename, "r"))
    {
/*- Skip previous ENDs in multi-FITS extension headers */
    for (i=(frameno?(frameno-1):0); i--;)
      while (fgets(gstr, MAXCHAR, file)
		&& strncmp(gstr,"END ",4)
		&& strncmp(gstr,"END\n",4));
    pos = fitsfind(tab->headbuf, "END     ");
    memset(gstr, ' ', 80);
    while (fgets(gstr, 81, file) && strncmp(gstr,"END ",4)
				&& strncmp(gstr,"END\n",4))
      {
      fitspick(gstr, keyword, data, &htype, &ttype, comment);
/*---- Block critical keywords */
      if (!wstrncmp(keyword, "SIMPLE  ", 8)
	||!wstrncmp(keyword, "BITPIX  ", 8)
	||!wstrncmp(keyword, "NAXIS   ", 8)
	||!wstrncmp(keyword, "BSCALE  ", 8)
	||!wstrncmp(keyword, "RADECSYS", 8)
	||!wstrncmp(keyword, "EQUINOX ", 8)
	||!wstrncmp(keyword, "CUNIT1  ", 8)
	||!wstrncmp(keyword, "CUNIT2  ", 8)
	||!wstrncmp(keyword, "BZERO   ", 8))
        continue;
      
/*---- Always keep a one-line margin */
      if ((pos+1)*80>=tab->headnblock*FBSIZE)
        {
        tab->headnblock++;
        QREALLOC(tab->headbuf, char, tab->headnblock*FBSIZE);
        }
      fitsadd(tab->headbuf, keyword, comment);
      if (fitsfind(tab->headbuf, "END     ")>=pos)
        pos++;
      fitswrite(tab->headbuf, keyword, data, htype, ttype);
      memset(gstr, ' ', 80);
      }
    fclose(file);
/*-- Update the tab data */
    readbasic_head(tab);
    return RETURN_OK;
    }
  else
    return RETURN_ERROR;
  }


/****** readfitsinfo_field ****************************************************
PROTO	void	readfitsinfo_field(fieldstruct *field, tabstruct *tab)
PURPOSE	Read additional information from a FITS header and update the current
	field's tab
INPUT	Pointer to the field structure,
	Tab structure.
OUTPUT	-.
NOTES	-.
AUTHOR	E. Bertin (IAP)
VERSION	17/08/2001
 ***/
void	readfitsinfo_field(fieldstruct *field, tabstruct *tab)
  {
#define	FITSREADF(buf, k, val, def) \
		{if (fitsread(buf,k, &val, H_FLOAT,T_DOUBLE) != RETURN_OK) \
		   val = def; \
		}

#define	FITSREADI(buf, k, val, def) \
		{if (fitsread(buf,k, &val, H_INT,T_LONG) != RETURN_OK) \
		   val = def; \
		}

#define	FITSREADS(buf, k, str, def) \
		{if (fitsread(buf,k,str, H_STRING,T_STRING) != RETURN_OK) \
		   strcpy(str, (def)); \
		}
   char		*buf;
   wcsstruct	*wcs;
   int		l;

  buf = tab->headbuf;
  wcs = field->wcs;
  if (!buf || !wcs)
    return;
  for (l=0; l<tab->naxis; l++)
    {
    sprintf(gstr, "COMIN%-3d", l+1);
    FITSREADI(buf, gstr, wcs->outmin[l], 1);
    wcs->outmax[l] = wcs->outmin[l] + wcs->naxisn[l] - 1;
    }
  FITSREADF(buf, "BACKMEAN", field->backmean, 0.0);
  FITSREADF(buf, "BACKSIG ", field->backsig, 0.0);
/* Set the flux scale */
  FITSREADF(buf,prefs.fscale_keyword, field->fscale, field->fscale);
/* Set the conversion factor */
  FITSREADF(buf, prefs.gain_keyword, field->gain, field->gain);

  return;

#undef FITSREADF
#undef FITSREADI
#undef FITSREADS
  }


/****** writefitsinfo_outfield ************************************************
PROTO	void writefitsinfo_outfield(fieldstruct *field)
PURPOSE	Add relevant information to the output field FITS header.
INPUT	Pointer to the field.
OUTPUT	-.
NOTES	Global preferences are used.
AUTHOR	E. Bertin (IAP)
VERSION	08/10/2002
 ***/
void	writefitsinfo_outfield(fieldstruct *field)

  {
   struct passwd	*pwd;
   struct tm		*tm;
   time_t		thetime;
   tabstruct		*tab;
   extern pkeystruct	key[];
   extern char		keylist[][16];
   char			*pstr;
   int			d, n;

  tab = field->tab;
/* Keep some margin */
  /* the '55' at the end is for 50 dummy keywords,
     two corresponding COMMENT lines and 3 HISTORY
     keywords
  */
  n = ((80*(15+tab->naxis*3+prefs.ninfield*9+55))/FBSIZE)+1;
  QREALLOC(tab->headbuf, char, FBSIZE*(tab->headnblock+n));

  if(field->headflag)
  {
    fitsadd(tab->headbuf, "COMMENT ", "");
  }
  fitsadd(tab->headbuf, "SOFTNAME", "The software that processed those data");
  fitswrite(tab->headbuf, "SOFTNAME", BANNER, H_STRING, T_STRING);
  fitsadd(tab->headbuf, "SOFTVERS", "Version of the software");
  fitswrite(tab->headbuf, "SOFTVERS", VERSION, H_STRING, T_STRING);
  fitsadd(tab->headbuf, "SOFTDATE", "Release date of the software");
  fitswrite(tab->headbuf, "SOFTDATE", DATE, H_STRING, T_STRING);
  fitsadd(tab->headbuf, "SOFTAUTH", "Maintainer of the software");
  fitswrite(tab->headbuf, "SOFTAUTH", COPYRIGHT, H_STRING, T_STRING);
  fitsadd(tab->headbuf, "SOFTINST", "Institute");
  fitswrite(tab->headbuf, "SOFTINST", INSTITUTE, H_STRING, T_STRING);

  if(field->headflag)
  {
    fitsadd(tab->headbuf, "COMMENT ", "");
  }

/* User name */
  if (pwd=getpwuid((uid_t)geteuid()))
    {
    fitsadd(tab->headbuf, "AUTHOR", "Who ran the software");
    fitswrite(tab->headbuf, "AUTHOR", strtok(pwd->pw_name,","),
		H_STRING, T_STRING);
    }

/* Host name */
  if (!gethostname(gstr, 80))
    {
    fitsadd(tab->headbuf, "ORIGIN", "Where it was done");
    fitswrite(tab->headbuf, "ORIGIN", gstr, H_STRING, T_STRING);
    }

/* Obs dates */
  thetime = time(NULL);
  tm = gmtime(&thetime);
  sprintf(gstr,"%04d-%02d-%02dT%02d:%02d:%02d",
                        tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday,
                        tm->tm_hour, tm->tm_min, tm->tm_sec);
  fitsadd(tab->headbuf, "DATE    ", "When it was started (GMT)");
  fitswrite(tab->headbuf, "DATE    ", gstr, H_STRING, T_STRING);

/* Config parameters */
/* COMBINE_TYPE */
  fitsadd(tab->headbuf,"COMBINET","COMBINE_TYPE config parameter for " BANNER);
  fitswrite(tab->headbuf, "COMBINET", key[findkeys("COMBINE_TYPE", keylist,
	FIND_STRICT)].keylist[prefs.coadd_type], H_STRING, T_STRING);

/* Axis-dependent config parameters */
  if(field->headflag)
  {
    fitsadd(tab->headbuf, "COMMENT ", "");
    fitsadd(tab->headbuf, "COMMENT ", "Axis-dependent config parameters");
  }

  for (d=0; d<tab->naxis; d++)
    {
/*-- RESAMPLING_TYPE */
    sprintf(gstr, "RESAMPT%1d", d+1);
    fitsadd(tab->headbuf, gstr,
		"RESAMPLING_TYPE config parameter");
    fitswrite(tab->headbuf, gstr, key[findkeys("RESAMPLING_TYPE", keylist,
	FIND_STRICT)].keylist[prefs.resamp_type[d]], H_STRING, T_STRING);
/*-- CENTER_TYPE */
    sprintf(gstr, "CENTERT%1d", d+1);
    fitsadd(tab->headbuf, gstr, "CENTER_TYPE config parameter");
    fitswrite(tab->headbuf, gstr, key[findkeys("CENTER_TYPE", keylist,
	FIND_STRICT)].keylist[prefs.center_type[d]], H_STRING, T_STRING);
/*-- PIXELSCALE_TYPE */
    sprintf(gstr, "PSCALET%1d", d+1);
    fitsadd(tab->headbuf, gstr, "PIXELSCALE_TYPE config parameter");
    fitswrite(tab->headbuf, gstr, key[findkeys("PIXELSCALE_TYPE", keylist,
	FIND_STRICT)].keylist[prefs.pixscale_type[d]], H_STRING, T_STRING);
    }

/* Image-dependent config parameters */
  if (prefs.writefileinfo_flag)
    {
    fitsadd(tab->headbuf, "COMMENT ", "");
    fitsadd(tab->headbuf, "COMMENT ", "Image-dependent config parameters");
    for (n=0; n<prefs.ninfield; n++)
      {
/*---- Input images */
      sprintf(gstr, "FILE%04d", n+1);
      fitsadd(tab->headbuf, gstr, "Input filename");
/*---- A short, "relative" version of the filename */
      if (!(pstr = strrchr(prefs.infield_name[n], '/')))
        pstr = prefs.infield_name[n];
      else
        pstr++;
      fitswrite(tab->headbuf, gstr, pstr, H_STRING, T_STRING);
/*---- Input weights */
      if (n<prefs.ninwfield)
        {
        sprintf(gstr, "WGHT%04d", n+1);
        fitsadd(tab->headbuf, gstr, "Input weight-map");
/*------ A short, "relative" version of the filename */
        if (!(pstr = strrchr(prefs.inwfield_name[n], '/')))
          pstr = prefs.inwfield_name[n];
        else
          pstr++;
        fitswrite(tab->headbuf, gstr, pstr, H_STRING, T_STRING);
/*------ WEIGHT_TYPE */
        sprintf(gstr, "WGTT%04d", n+1);
        fitsadd(tab->headbuf, gstr,
		"WEIGHT_TYPE config parameter");
        fitswrite(tab->headbuf, gstr, key[findkeys("WEIGHT_TYPE", keylist,
	  FIND_STRICT)].keylist[prefs.weight_type[n]], H_STRING, T_STRING);
        }
/*---- Interpolation flag */
      sprintf(gstr, "INTF%04d", n+1);
      fitsadd(tab->headbuf, gstr, "INTERPOLATE config flag");
      fitswrite(tab->headbuf, gstr, &prefs.interp_flag[n], H_BOOL, T_LONG);
/*---- Background-subtraction flag */
      sprintf(gstr, "SUBF%04d", n+1);
      fitsadd(tab->headbuf, gstr, "SUBTRACT_BACK config flag");
      fitswrite(tab->headbuf, gstr, &prefs.subback_flag[n], H_BOOL, T_LONG);
/*---- BACK_TYPE */
      sprintf(gstr, "BCKT%04d", n+1);
      fitsadd(tab->headbuf, gstr, "BACK_TYPE config parameter");
      fitswrite(tab->headbuf, gstr, key[findkeys("BACK_TYPE", keylist,
        FIND_STRICT)].keylist[prefs.back_type[n]], H_STRING, T_STRING);
/*---- BACK_DEFAULT */
      if (prefs.back_type[n]==BACK_ABSOLUTE)
        {
        sprintf(gstr, "BCKD%04d", n+1);
        fitsadd(tab->headbuf, gstr,"BACK_DEFAULT config parameter");
        fitswrite(tab->headbuf, gstr, &prefs.back_default[n], H_EXPO,T_DOUBLE);
        }
/*---- BACK_SIZE */
      sprintf(gstr, "BCKS%04d", n+1);
      fitsadd(tab->headbuf, gstr,"BACK_SIZE config parameter");
      fitswrite(tab->headbuf, gstr, &prefs.back_size[n], H_INT, T_LONG);
/*---- BACK_FILTERSIZE */
      sprintf(gstr, "BKFS%04d", n+1);
      fitsadd(tab->headbuf, gstr,"BACK_FILTERSIZE config parameter");
      fitswrite(tab->headbuf, gstr, &prefs.back_fsize[n], H_INT, T_LONG);
      }
    }
/* add some dummy keywords for later use */
  if(field->headflag)
  {
    fitsadd(tab->headbuf, "COMMENT ", "");
    fitsadd(tab->headbuf, "COMMENT ", "DUMMY keywords for later use");
  }
  d=0;
  for(n = 0; n < 50; n++)
  {
    sprintf(gstr, "DUMMY%d", n);
    fitsadd(tab->headbuf, gstr,"DUMMY keyword");
    fitswrite(tab->headbuf, gstr, &d, H_INT, T_LONG);    
  }

/* to be consistent with the other software's history information */
  fitsadd(tab->headbuf, "HISTORY ", "");
  sprintf(gstr, "swarp: %s", __PIPEVERS__);
  fitsadd(tab->headbuf, "HISTORY ", gstr);
  thetime = time(NULL);
  tm = localtime(&thetime);
  sprintf(gstr,"swarp_theli: called at: %04d-%02d-%02dT%02d:%02d:%02d",
                        tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday,
                        tm->tm_hour, tm->tm_min, tm->tm_sec);
  fitsadd(tab->headbuf, "HISTORY ", gstr);

/* Keep only what's necessary */
  tab->headnblock = ((fitsfind(tab->headbuf, "END     ")+36)*80)/FBSIZE;
  QREALLOC(tab->headbuf, char, tab->headnblock*FBSIZE);

  /* 
     ensure that the Header from the END mark up to
     the end is filled with spaces
  */
  n=(fitsfind(tab->headbuf, "END     ")+1)*80;
  memset(tab->headbuf+n, ' ', (size_t)(tab->headnblock*FBSIZE-n));

  return;
  }


/****** writefitsinfo_field **************************************************
PROTO	void writefitsinfo_field(fieldstruct *field)
PURPOSE	Add relevant information to a field FITS header.
INPUT	Pointer to the field.
OUTPUT	-.
NOTES	Global preferences are used.
AUTHOR	E. Bertin (IAP)
VERSION	13/04/2002
 ***/
void	writefitsinfo_field(fieldstruct *field)

  {
   struct passwd	*pwd;
   struct tm		*tm;
   time_t		thetime;
   tabstruct		*tab;
   wcsstruct		*wcs;
   extern pkeystruct	key[];
   extern char		keylist[][16];
   int			d, n, nf;

  nf = field->fieldno;
  tab = field->tab;
/* Keep some margin */
  n = 80*(15+tab->naxis*5+13);
  QREALLOC(tab->headbuf, char, FBSIZE*tab->headnblock+n);
  memset(tab->headbuf+FBSIZE*tab->headnblock, ' ', (size_t)n);

  fitsadd(tab->headbuf, "COMMENT ", "");
  fitsadd(tab->headbuf, "SOFTNAME", "The software that processed those data");
  fitswrite(tab->headbuf, "SOFTNAME", BANNER, H_STRING, T_STRING);
  fitsadd(tab->headbuf, "SOFTVERS", "Version of the software");
  fitswrite(tab->headbuf, "SOFTVERS", VERSION, H_STRING, T_STRING);
  fitsadd(tab->headbuf, "SOFTDATE", "Release date of the software");
  fitswrite(tab->headbuf, "SOFTDATE", DATE, H_STRING, T_STRING);
  fitsadd(tab->headbuf, "SOFTAUTH", "Maintainer of the software");
  fitswrite(tab->headbuf, "SOFTAUTH", COPYRIGHT, H_STRING, T_STRING);
  fitsadd(tab->headbuf, "SOFTINST", "Institute");
  fitswrite(tab->headbuf, "SOFTINST", INSTITUTE, H_STRING, T_STRING);
  fitsadd(tab->headbuf, "COMMENT ", "");

/* User name */
  if (pwd=getpwuid((uid_t)geteuid()))
    {
    fitsadd(tab->headbuf, "AUTHOR", "Who ran the software");
    fitswrite(tab->headbuf, "AUTHOR", strtok(pwd->pw_name,","),
		H_STRING, T_STRING);
    }

/* Host name */
  if (!gethostname(gstr, 80))
    {
    fitsadd(tab->headbuf, "ORIGIN", "Where it was done");
    fitsadd(tab->headbuf, "ORIGIN", "Where it was done");
    fitswrite(tab->headbuf, "ORIGIN", gstr, H_STRING, T_STRING);
    }

/* Obs dates */
  thetime = time(NULL);
  tm = gmtime(&thetime);
  sprintf(gstr,"%04d-%02d-%02dT%02d:%02d:%02d",
                        tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday,
                        tm->tm_hour, tm->tm_min, tm->tm_sec);
  fitsadd(tab->headbuf, "DATE    ", "When it was started (GMT)");
  fitswrite(tab->headbuf, "DATE    ", gstr, H_STRING, T_STRING);

/* Config parameters */
/* COMBINE_TYPE */
  fitsadd(tab->headbuf,"COMBINET","COMBINE_TYPE config parameter for " BANNER);
  fitswrite(tab->headbuf, "COMBINET", key[findkeys("COMBINE_TYPE", keylist,
	FIND_STRICT)].keylist[prefs.coadd_type], H_STRING, T_STRING);

/* Axis-dependent config parameters */
  fitsadd(tab->headbuf, "COMMENT ", "");
  fitsadd(tab->headbuf, "COMMENT ", "Axis-dependent SWarp parameters");
  for (d=0; d<tab->naxis; d++)
    {
    if ((wcs=field->wcs) && wcs->outmax[d])
      {
      sprintf(gstr, "COMIN%-3d", d+1);
      fitsadd(tab->headbuf, gstr, "Output minimum position of image");
      fitswrite(tab->headbuf, gstr, &wcs->outmin[d], H_INT, T_LONG);
      }
/*-- RESAMPLING_TYPE */
    sprintf(gstr, "RESAMPT%1d", d+1);
    fitsadd(tab->headbuf, gstr,
		"RESAMPLING_TYPE config parameter");
    fitswrite(tab->headbuf, gstr, key[findkeys("RESAMPLING_TYPE", keylist,
	FIND_STRICT)].keylist[prefs.resamp_type[d]], H_STRING, T_STRING);
/*-- CENTER_TYPE */
    sprintf(gstr, "CENTERT%1d", d+1);
    fitsadd(tab->headbuf, gstr, "CENTER_TYPE config parameter");
    fitswrite(tab->headbuf, gstr, key[findkeys("CENTER_TYPE", keylist,
	FIND_STRICT)].keylist[prefs.center_type[d]], H_STRING, T_STRING);
/*-- PIXELSCALE_TYPE */
    sprintf(gstr, "PSCALET%1d", d+1);
    fitsadd(tab->headbuf, gstr, "PIXELSCALE_TYPE config parameter");
    fitswrite(tab->headbuf, gstr, key[findkeys("PIXELSCALE_TYPE", keylist,
	FIND_STRICT)].keylist[prefs.pixscale_type[d]], H_STRING, T_STRING);
    }

/* Image-dependent config parameters */
  fitsadd(tab->headbuf, "COMMENT ", "");
  fitsadd(tab->headbuf, "COMMENT ", "Image-dependent SWarp parameters");
/* Flux scales */
  fitsadd(tab->headbuf, prefs.fscale_keyword,
	"Relative flux scaling from photometry");
  fitswrite(tab->headbuf,prefs.fscale_keyword, &field->fscale,H_EXPO,T_DOUBLE);
  fitsadd(tab->headbuf, "FLASCALE",
	"Relative flux scaling from astrometry");
  fitswrite(tab->headbuf, "FLASCALE", &field->fascale, H_EXPO,T_DOUBLE);
/* Gain (conversion factor) */
  fitsadd(tab->headbuf, prefs.gain_keyword,
	"Effective conversion factor in e-/ADU");
  fitswrite(tab->headbuf,prefs.gain_keyword, &field->gain, H_EXPO,T_DOUBLE);

/* Background */
  fitsadd(tab->headbuf, "BACKMEAN", "Measured background level");
  fitswrite(tab->headbuf, "BACKMEAN", &field->backmean, H_EXPO,T_DOUBLE);
  fitsadd(tab->headbuf, "BACKSIG ", "Measured background RMS");
  fitswrite(tab->headbuf, "BACKSIG ", &field->backsig, H_EXPO,T_DOUBLE);
/* Input image */
  fitsadd(tab->headbuf, "ORIGFILE", "Input filename");
  fitswrite(tab->headbuf, "ORIGFILE", field->rfilename, H_STRING, T_STRING);
/* Interpolation flag */
  fitsadd(tab->headbuf, "INTERPF ", "INTERPOLATE config flag");
  fitswrite(tab->headbuf, "INTERPF ", &prefs.interp_flag[nf], H_BOOL, T_LONG);
/* Background-subtraction flag */
  fitsadd(tab->headbuf, "BACKSUBF", "SUBTRACT_BACK config flag");
  fitswrite(tab->headbuf, "BACKSUBF", &prefs.subback_flag[nf], H_BOOL, T_LONG);
/* BACK_TYPE */
  fitsadd(tab->headbuf, "BACKTYPE", "BACK_TYPE config parameter");
  fitswrite(tab->headbuf, "BACKTYPE", key[findkeys("BACK_TYPE", keylist,
        FIND_STRICT)].keylist[prefs.back_type[nf]], H_STRING, T_STRING);
/* BACK_DEFAULT */
  if (field->back_type==BACK_ABSOLUTE)
    {
    fitsadd(tab->headbuf, "BACKDEF ", "BACK_DEFAULT config parameter");
    fitswrite(tab->headbuf, "BACKDEF ", &prefs.back_default[nf],
		H_EXPO,T_FLOAT);
    }
/* BACK_SIZE */
  fitsadd(tab->headbuf, "BACKSIZE", "BACK_SIZE config parameter");
  fitswrite(tab->headbuf, "BACKSIZE", &prefs.back_size[nf], H_INT, T_LONG);
/* BACK_FILTERSIZE */
  fitsadd(tab->headbuf, "BACKFSIZ", "BACK_FSIZE config parameter");
  fitswrite(tab->headbuf, "BACKFSIZ", &prefs.back_fsize[nf], H_INT, T_LONG);
/* Keep only what's necessary */
  tab->headnblock = ((fitsfind(tab->headbuf, "END     ")+36)*80)/FBSIZE;
  QREALLOC(tab->headbuf, char, tab->headnblock*FBSIZE);

  return;
  }

