/*---------------------------------------------------------------------------

   File name    :       calib_photoz_data.c
   Authors      :       P. Capak, T. Erben
   Created on   :       10.08.2008 (first version of T. Erben)
   Description  :       calibration of ZPs (and templates) for photo-z studies

*--------------------------------------------------------------------------*/

/*
        $Id: calib_photoz_data.c,v 1.6 2008/11/28 23:03:13 thomas Exp $
        $Author: thomas $
        $Date: 2008/11/28 23:03:13 $
        $Revision: 1.6 $
 */

/*---------------------------------------------------------------------------
                               General Comments
   ---------------------------------------------------------------------------*/

/*
   This file originated from Peter Capaks original 'first_fit.c' C-code
   to estimate magnitude offsets for photo-z studies

   compile with: gcc -o calib_photoz_data calib_photoz_data.c -lm -O3

   Before the first check-in into CVS the following changes were made  to
   Peters original code:

   - In final offsets, photometric offsets were 'normalised' to
     filter '4'; I changed this to ref_filter!

   - I deleted the superfluous ABS function (fabs is in math.h)

   - Photometric offsets are estimated iteratively with a maximum
     of 10 iterations. Iterations are stopped if all offsets are less
     than 0.02 mag.

   - The effective wavelength of filters does not need to be given
     in the filters file anymore but it is estimated directly from
     the filter data.

   - I 'corrected' inconsistencies in flux <-> mag conversion
     (units of Janskys, nanoJanskys); however, the corrections
     (additions of constants) do not have any effect on final results.

   - I removed existing memory leaks

   - Wavelengths for SEDs and filter curves need to be provided in
     Angstroems now. The conversion to m is done within the program.
     This enables us to simply reuse typical SED and filter data.

   - I commented the program while going through it to understand
     what is being done.

   - arguments are collected via command line options

   - I added user documentaion
 */


/*---------------------------------------------------------------------------
                               Remarks and Todos
   ---------------------------------------------------------------------------*/

/*
   Remarks and TODOS (TE):
   - The program works for AB magnitudes!!
   - ensure consistent use of filters to be fitted (variable
     N_fit_filter in the main routine and associated variables)
   - The program seems to implicitely assume ordering of the
     filters from blue to red
 */

/*---------------------------------------------------------------------------
                               History comments
   ---------------------------------------------------------------------------*/

/*
   10.08.2008:
   First version of Thomas Erben

   13.08.2008:
   Removal of compiler warnings

   14.08.2008:
   Usage information added

   15.08.2008:
   The number of SED interpolations can be changed via
   a command line option now (-n)

   28.11.2008:
   New command line options (-e) and (-m). '-e' gives a minimum
   photometric error to all objects (added quadratically to existing
   errors); '-m' controls the number of iterations for zeropoint 
   correction.
 */

/*---------------------------------------------------------------------------
                                   Includes and Defines
   ---------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <unistd.h>

/* include the following for memory tracing */
/* #include "xmemory.h" */

#define __PROGRAM_NAME__       "calib_photoz_data"
#define __PROGRAM_VERSION__    "1.0.0"

#define MAXCHAR                2048 /* maximum number of characters in file lines */
#define BIG                    9e99

/*---------------------------------------------------------------------------
                                   Subroutines
   ---------------------------------------------------------------------------*/

/* Just print the usage message of the program */
void print_usage(void);

/* this routine caculates the necessary values for a spline
   (adapted from numerical recipies)
 */
void spline(double *x, double *y,
            int N,
            double yp0, double ypN,
            double *y2, double MAX)
{
  int    i, k;  /* counters */
  double p, qn, sig, un, *u;

  u = (double *)calloc(N, sizeof(double));

  /* set the boundry condition for the first array element */
  if (yp0 > MAX) {
    y2[0] = u[0] = 0.0;
  } else {
    y2[0] = -0.5;
    u[0]  = (3.0 / (x[1] - x[0])) * ((y[1] - y[0]) / (x[1] - x[0]) - yp0);
  }

  /* start the tri-diagonal decomposition algorithm to figure out derivatives
     y2 and u used for temporary storage
   */
  for (i = 1; i < (N - 1); i++) {
    sig   = (x[i] - x[i - 1]) / (x[i + 1] - x[i - 1]);
    p     = sig * y2[i - 1] + 2.0;
    y2[i] = (sig - 1.0) / p;
    u[i]  = (y[i + 1] - y[i]) / (x[i + 1] - x[i]) -
            (y[i] - y[i - 1]) / (x[i] - x[i - 1]);
    u[i] = (6.0 * u[i] / (x[i + 1] - x[i - 1]) - sig * u[i - 1]) / p;
  }

  /* set the last boundry condition to be natural or first derivative */
  if (ypN > MAX) {
    qn = un = 0.0;
  } else {
    qn = 0.5;
    un = (3.0 / (x[N - 1] - x[N - 2])) *
         (ypN - (y[N - 1] - y[N - 2]) / (x[N - 1] - x[N - 2]));
  }

  y2[N - 1] = (un - qn * u[N - 2]) / (qn * y2[N - 2] + 1.0);

  for (k = N - 2; k >= 0; k--) {
    y2[k] = y2[k] * y2[k + 1] + u[k];
  }

  free(u);
}

/* routine to get the value once spline has been calculated
   adapted from numerical recipies
 */
double splint(double *xa, double *ya, double *y2a, int N, double x)
{
  int    klo, khi, k;
  double h, b, a, y;

  klo = 0;
  khi = N - 1;

  /* search by bisection */
  while (khi - klo > 1) {
    k = (khi + klo) >> 1;

    if (xa[k] > x) {
      khi = k;
    } else {
      klo = k;
    }
  }

  h = xa[khi] - xa[klo];

  if (h == 0.0) {
    return(0.0);
  } else {
    a = (xa[khi] - x) / h;
    b = (x - xa[klo]) / h;
    y = a * ya[klo] + b * ya[khi] +
        ((a * a * a - a) * y2a[klo] +
         (b * b * b - b) * y2a[khi]) * (h * h) / 6.0;
    return(y);
  }
}

/* routine to get the value once spline has been calculated
   adapted from numerical recipies
 */
double splint2(double *xa, double *ya, double *y2a, int N, double x)
{
  int    klo, khi, k;
  double h, b, a, y;

  klo = 0;
  khi = N - 1;

  if (x <= xa[0]) {
    return(ya[0]);
  }

  if (x >= xa[N - 1]) {
    return(ya[N - 1]);
  }

  /* search by bisection */
  while (khi - klo > 1) {
    k = (khi + klo) >> 1;

    if (xa[k] > x) {
      khi = k;
    } else {
      klo = k;
    }
  }

  h = xa[khi] - xa[klo];

  if (h == 0.0) {
    return(0.0);
  } else {
    a = (xa[khi] - x) / h;
    b = (x - xa[klo]) / h;
    y = a * ya[klo] + b * ya[khi] +
        ((a * a * a - a) * y2a[klo] +
         (b * b * b - b) * y2a[khi]) * (h * h) / 6.0;
    return(y);
  }
}

/* this routine does a linear interpolation */
double lininterp(double x0, double x1, double y0, double y1, double x)
{
  double m, b;

  m = (y1 - y0) / (x1 - x0);
  b = y0 - (m * x0);

  return((m * x) + b);
}

/* The Madau A parameter */
double madau_A(double z)
{
  double Damax = 1170 * (1 + z);
  double Damin = 1050 * (1 + z);

  double dla, Da, A;
  int    i;

  dla = (Damax - Damin) / 100;
  A   = 0;

  for (i = 0; i < 100; i++) {
    Da  = Damin + dla * i;
    Da /= 1216;
    Da  = pow(Da, 3.46);
    Da *= -3.6E-3;
    Da  = exp(Da);
    Da *= dla;

    A += Da;
  }

  A *= (1 / (120 * (1 + z)));
  return(A);
}

/* The Madau B parameter */
double madau_B(double z)
{
  const double Dbmax = 1015 * (1 + z);
  const double Dbmin = 920 * (1 + z);

  double dlb, Lb, Db, h1, h2, h3, B;
  int    i;

  dlb = (Dbmax - Dbmin) / 100;
  B   = 0;

  for (i = 0; i < 100; i++) {
    Lb = Dbmin + dlb * i;

    h1  = Lb / 1026;
    h1  = pow(h1, 3.46);
    h1 *= -1.7E-3;

    h2  = Lb / 972.5;
    h2  = pow(h2, 3.46);
    h2 *= -1.2E-3;

    h3  = Lb / 950.0;
    h3  = pow(h3, 3.46);
    h3 *= -9.3E-4;

    Db  = h1 + h2 + h3;
    Db  = exp(Db);
    Db *= dlb;

    B += Db;
  }

  B *= (1 / (95 * (1 + z)));

  return(B);
}

/* The Calzetti et. al. 2000 reddening law */
double calzetti_redden(double wave, double Av)
{
  const double Rv = 4.05;  /* Set by Calzetti */
  double       micron, r = 0.;

  micron = wave * 1e6;  /* convert to micron */

  if (micron <= 0.12) {
    r = lininterp(0.110000, 0.120000, 13.258139, 12.119376, micron);
  }

  if ((micron > 0.12) && (micron <= 0.63)) {
    r  = 0.011 / micron;
    r += -0.198;
    r /= micron;
    r += 1.509;
    r /= micron;
    r += -2.156;
    r *= 2.659;
    r += Rv;
  }

  if ((micron > 0.63) && (micron <= 2.20)) {
    r  = 1.040;
    r /= micron;
    r += -1.857;
    r *= 2.659;
    r += Rv;
  }

  if (micron > 2.20) {
    r = lininterp(2.190000, 2.200000, 0.374958, 0.369219, micron);
  }

  r *= Av;
  r /= Rv;
  r /= -2.5;
  r  = pow(10, r);

  return(r);
}

/* this routine mixes templates */
void mix_template(double ratio,
                  int N1,
                  double *wave,
                  double *flux, double *wave1,
                  double *flux1, int N2,
                  double *wave2, double *flux2)
{
  int    s_counter, i;
  double f_hold;

  s_counter = 1;

  for (i = 0; i < N1; i++) {
    while ((wave2[s_counter] < wave1[i]) && (s_counter < N2)) {
      s_counter++;
    }

    f_hold = lininterp(wave2[s_counter - 1],
                       wave2[s_counter],
                       flux2[s_counter - 1],
                       flux2[s_counter],
                       wave1[i]) * ratio;

    flux[i]  = (1 - ratio) * flux1[i];
    flux[i] += f_hold;
    wave[i]  = wave1[i];
  }
}

/* this routine reddens a template */
void redden_template(double Av, int N, double *wave, double *flux)
{
  int i;

  for (i = 0; i < N; i++) {
    flux[i] *= calzetti_redden(wave[i], Av);
  }
}

/* This routine will redshift a template and apply the Madau correction */
void redshift_template(double z, int N, double *wave,
                       double *flux, double *out_wave,
                       double *out_flux)
{
  int    i;
  double l, A, B;

  A = madau_A(z);
  B = madau_B(z);

  for (i = 0; i < N; i++) {
    out_wave[i] = wave[i] * (1 + z);
    l           = out_wave[i] * 1e10;

    if (l < 920) {
      out_flux[i] = 0;
    } else if ((l >= 920) && (l <= 1015)) {
      out_flux[i] = flux[i] * B;
    } else if ((l > 1015) && (l <= 1170)) {
      out_flux[i] = flux[i] * A;
    } else {
      out_flux[i] = flux[i];
    }
  }
}

/* This routine will calculate a magnitude from an SED and filter profile */
double calc_magnitude(int Ns,           /* points in the SED */
                      double *wave_s,   /* wavelength data of SED (im m) */
                      double *flux_s,   /* flux data of SED (in units of
                                           wavelength */
                      int Nf,           /* points in the filter curve */
                      double *wave_f,   /* wavelength data of filter (im m) */
                      double *trans_f)  /* transmission of filter */
{
  const double c = 3e8;

  double  nu_f_max, nu_f_min; /* limits for integration */
  double *nu_s, *nu_f;        /* arrays for the frequency */
  double *fnu_s;              /* array for f_nu */
  int     s_min, s_max;       /* limits of integration in the SED */
  int     i, f_counter;       /* counter for integration, and filter */
  double  t;                  /* transmission of filter at this point */
  double  iflux, itrans;      /* ntegrated flux and transmission */
  double  mag;                /* conversion to magnitudes */

  /* allocate the Nu array */
  nu_s  = (double *)calloc(Ns, sizeof(double));
  fnu_s = (double *)calloc(Ns, sizeof(double));
  nu_f  = (double *)calloc(Nf, sizeof(double));

  /* convert from lambda to nu */
  for (i = 0; i < Ns; i++) {
    nu_s[i] = c / wave_s[i];

    /* the provided SED flux data is in units of wavelength;
       to convert wanvelength flux to frequency flux the formula
       to apply is:

       F(nu) = F(lambda) * lambda**2 / c

       The lambda**2 / c originates from the differential
       d(lambda) / d(mu)
     */
    fnu_s[i] = flux_s[i] * wave_s[i] * wave_s[i] / c;
  }

  for (i = 0; i < Nf; i++) {
    nu_f[i] = c / wave_f[i];
  }

  /* we implicitely assume that the SED/filter data are ordered
     with increasing wavelength, i.e. decreasing frequency
   */
  nu_f_max = nu_f[0];
  nu_f_min = nu_f[Nf - 1];

  /* find the counter for the SED */
  s_min = 0;
  while (nu_s[s_min] > nu_f_min) {
    s_min++;
  }

  s_max = 0;

  while (nu_s[s_max] > nu_f_max) {
    s_max++;
  }

  /* setup the filter counter */
  f_counter = 1;

  /* initialize the sums */
  iflux  = 0;
  itrans = 0;

  /* go from the max nu to the min (sed sorted in decending nu) */
  for (i = s_max; i < s_min; i++) {
    while (nu_f[f_counter] > nu_s[i]) {
      f_counter++;
    }

    t = lininterp(nu_f[f_counter - 1],
                  nu_f[f_counter],
                  trans_f[f_counter - 1],
                  trans_f[f_counter],
                  nu_s[i]);

    if (t < 0) {  /* prevent negative flux */
      t = 0;
    }

    if (fnu_s[i] > 0) {  /* prevent negative flux */
      iflux += fnu_s[i] * t / nu_s[i];
    }

    itrans += t / nu_s[i];
  }

  /* the used formula to go from flux to AB magnitudes is:
     mag(AB) = 2.5 * (32 - log10(flux(nano Jansky))) - 48.6
             = 31.4 - 2.5 * log(flux(nano Jansky));

     the relation of the calculated flux (iflux etc.)
     to nanoJansky is not yet completely clear to me!
     However, addition/subtraction of constants does
     not matter for the final result of photometry offsets.
   */
  if (iflux > 0) {
    mag = iflux / itrans;
    mag = 31.4 - 2.5 * (log10(mag) + 32);
  } else {
    mag = 99;
  }

  free(nu_s);
  free(fnu_s);
  free(nu_f);

  return(mag);
}

/* converts a magnitude to a flux */
double mag2flux(double mag)
{
  double hold;

  if (mag > 0) {
    if (mag != 99) {
      hold = pow(10, ((mag - 31.4) / -2.5));
    } else {
      hold = 0;
    }
  } else {
    if (mag != -99) {
      hold = -1.0 * pow(10, (((-1.0 * mag) - 31.4) / -2.5));
    } else {
      hold = -99;
    }
  }
  return(hold);
}

/* converts mag error to flux error */
double dmag2dflux(double flux, double dm)
{
  if (flux > 0) {
    return(flux * dm / (2.5 / log(10)));
  } else {
    return(-1.0 * flux * dm / (2.5 / log(10)));
  }
}

/* calculates the chi-sq/deg freedom for an object */
double fit_object(int N, int N_fit,
                  double *template,
                  double *phot, double *phot_err,
                  double *limits, double *mu,
                  double *obj_offset)
{
  int     i, Nphot;
  double  magsum, errsum, hold;
  double *temp_flux, *flux, *flux_err, *flux_limit;  /* fluxes */
  double  temp_err;

  magsum = 0;
  errsum = 0;

  /* calculate the mean offset for the template */
  for (i = 0; i < N_fit; i++) {
    if ((template[i] != 99) && (phot[i] != 99) && (phot[i] > 0)) {
      hold    = pow(phot_err[i], 2);
      magsum += (phot[i] - template[i]) / hold;
      errsum += 1 / hold;
    }
  }
  *mu = magsum / errsum;

  /* allocate flux arrays */
  temp_flux  = (double *)calloc(N, sizeof(double));
  flux       = (double *)calloc(N, sizeof(double));
  flux_err   = (double *)calloc(N, sizeof(double));
  flux_limit = (double *)calloc(N, sizeof(double));

  for (i = 0; i < N; i++) {
    temp_flux[i]  = mag2flux(template[i] + *mu);
    temp_err      = sqrt((phot_err[i] * phot_err[i]));
    flux[i]       = mag2flux(phot[i]);
    flux_err[i]   = dmag2dflux(flux[i], temp_err);
    flux_limit[i] = mag2flux(limits[i]);
  }

  magsum = 0;
  Nphot  = 0;
  for (i = 0; i < N; i++) {
    if (flux[i] != -99) {
      if (flux[i] > flux_limit[i]) {
        magsum += pow(((flux[i] - temp_flux[i]) / flux_err[i]), 2);
      } else {
        magsum += pow(((flux[i] - temp_flux[i]) / flux_limit[i]), 2);
      }
      /* output offsets to fix the template */
      if (flux[i] > (5 * flux_limit[i])) {
        obj_offset[i] = phot[i] - template[i] - *mu;
      } else {
        obj_offset[i] = 99;
      }
      Nphot++;
    } else {
      obj_offset[i] = 99;
    }
  }

  /* decrement degree of freedom by one because of scale factor */
  Nphot -= 1;

  free(temp_flux);
  free(flux);
  free(flux_err);
  free(flux_limit);

  return(magsum / Nphot);
}

void stats(int N,
           double *data,
           double *wt,
           double *mean, double *clipmean, double *sigma)
{
  double h_wt;
  int    i;

  *mean = 0;
  h_wt  = 0;

  for (i = 0; i < N; i++) {
    if (data[i] < 99) {
      *mean += data[i] / (wt[i] * wt[i]);
      h_wt  += 1 / (wt[i] * wt[i]);
    }
  }
  *mean /= h_wt;

  *sigma = 0;
  h_wt   = 0;

  for (i = 0; i < N; i++) {
    if (data[i] < 99) {
      *sigma += ((*mean - data[i]) * (*mean - data[i])) / (wt[i] * wt[i]);
      h_wt   += 1 / (wt[i] * wt[i]);
    }
  }

  *sigma /= h_wt;
  *sigma  = sqrt(*sigma);

  *clipmean = 0;
  h_wt      = 0;

  for (i = 0; i < N; i++) {
    if (data[i] < 99) {
      if ((data[i] < ((*sigma * 3) + *mean)) &&
          (data[i] > (*mean - (*sigma * 3)))) {
        *clipmean += data[i] / (wt[i] * wt[i]);
        h_wt      += 1 / (wt[i] * wt[i]);
      }
    }
  }
  *clipmean /= h_wt;
}

void bin_data(double *x, double *y, double *dy,
              int Ndata, double *bins,
              double *bx, double *by, double *bdy,
              int Nbins)
{
  int     i, j;
  double *wt;

  wt = (double *)calloc(Nbins, sizeof(double));

  for (i = 0; i < Ndata; i++) {
    if ((x[i] > bins[0]) && (x[i] < bins[Nbins])) {
      j = 0;
      while (x[i] > bins[j]) {
        j++;
      }

      if (y[i] != 99) {
        bx[j] += x[i] / (dy[i] * dy[i]);
        by[j] += y[i] / (dy[i] * dy[i]);
        wt[j] += 1 / (dy[i] * dy[i]);
      }
    }
  }

  for (i = 0; i < Nbins; i++) {
    if (wt[i] > 0) {
      bx[i] /= wt[i];
      by[i] /= wt[i];
      bdy[i] = sqrt(1 / wt[i]);
    }
  }

  free(wt);
}

/* bin data by weight to a maximum number of bins */
void wt_bin_data(double wt_min, double xstep,
                 double *x, double *y, double *dy, int Ndata,
                 double *bx, double *by, double *bdy,
                 int Nbins, int *outbins)
{
  int     i, j;
  double *wt, xmax;

  wt = (double *)calloc(Nbins, sizeof(double));

  /* set bin wt's to big */
  for (i = 0; i < Nbins; i++) {
    wt[i] = 0;
    bx[i] = 0;
    by[i] = 0;
  }

  j    = 0;
  xmax = pow(10, (log10(x[0]) + xstep));
  for (i = 0; i < Ndata; i++) {
    /* check if the error is small enough */
    if (((1 / sqrt(wt[j])) < wt_min) && (j < Nbins) && (x[i] > xmax)) {
      j++;
      xmax = pow(10, (log10(x[i]) + xstep));
    }

    bx[j] += x[i] / (dy[i] * dy[i]);
    by[j] += y[i] / (dy[i] * dy[i]);
    wt[j] += 1 / (dy[i] * dy[i]);
  }

  for (i = 0; i <= j; i++) {
    bx[i] /= wt[i];
    by[i] /= wt[i];
    bdy[i] = sqrt(1 / wt[i]);
  }

  *outbins = j;

  free(wt);
}

void calc_phot_offset(double **offset_phot, double **offset_err,
                      double *zobj, double *lambda_filter,
                      int Nobj, int N_filter, double *phot_new_off)
{
  /* variable for offsets by color */
  double **color_offset, **color_err, **color_lambda;
  /* bins for data */
  double *bins, **lambda_bin, **offset_bin, **err_bin;
  /* array of offsets between the filters */
  double **temp_offset, **temp_offset_wt;
  int      Nbins;                /* number of bins */
  double   Bmin, Bmax, dlogBin;  /* min and max bin, log spacing */
  /* variable for size of offsets by color array */
  int *   color_size;
  int     i, j, k;            /* counters */
  double *phot_new_err;
  /* temporary variable, carried no more than a few lines */
  double hold;

  Nbins = 100;
  Bmin  = 1000 / 1e10;    /* set min bin at 1000 AA */
  Bmax  = 100000 / 1e10;  /* set max bin at 25000 AA (K band) */

  /* allocate offset array */
  temp_offset    = (double **)calloc(N_filter, sizeof(double *));
  temp_offset_wt = (double **)calloc(N_filter, sizeof(double *));

  for (i = 0; i < N_filter; i++) {
    temp_offset[i]    = (double *)calloc(N_filter, sizeof(double));
    temp_offset_wt[i] = (double *)calloc(N_filter, sizeof(double));
  }

  /* allocate bins */
  lambda_bin = (double **)calloc(N_filter, sizeof(double *));
  offset_bin = (double **)calloc(N_filter, sizeof(double *));
  err_bin    = (double **)calloc(N_filter, sizeof(double *));
  /* allocate bins array one larger than the rest for bottom of last bin */
  bins = (double *)calloc((Nbins + 1), sizeof(double));
  for (i = 0; i < N_filter; i++) {
    lambda_bin[i] = (double *)calloc(Nbins, sizeof(double));
    offset_bin[i] = (double *)calloc(Nbins, sizeof(double));
    err_bin[i]    = (double *)calloc(Nbins, sizeof(double));
  }

  /* setup the bins */
  dlogBin = (log10(Bmax) - log10(Bmin)) / Nbins;
  bins[0] = Bmin;
  for (j = 1; j < (Nbins + 1); j++) {
    bins[j] = pow(10, (log10(bins[j - 1]) + dlogBin));
  }

  /* allocate arrays */
  color_offset = (double **)calloc(N_filter, sizeof(double *));
  color_err    = (double **)calloc(N_filter, sizeof(double *));
  color_lambda = (double **)calloc(N_filter, sizeof(double *));
  color_size   = (int *)calloc(N_filter, sizeof(int));
  for (i = 0; i < N_filter; i++) {
    color_offset[i] = (double *)calloc(Nobj, sizeof(double));
    color_err[i]    = (double *)calloc(Nobj, sizeof(double));
    color_lambda[i] = (double *)calloc(Nobj, sizeof(double));
    color_size[i]   = 0;
  }

  /* fill color array by going through offset array
     loop over objects
   */
  for (i = 0; i < Nobj; i++) {
    /* loop over filters */
    for (j = 0; j < N_filter; j++) {
      /* check if this is an OK offset */
      if ((offset_phot[i][j] > -1) && (offset_phot[i][j] < 1)) {
        /* fill in the new array */
        color_offset[j][color_size[j]] = offset_phot[i][j];
        color_err[j][color_size[j]]    = offset_err[i][j];
        color_lambda[j][color_size[j]] = lambda_filter[j] / (1 + zobj[i]);
        /* increment the color_size counter for new data */
        color_size[j]++;
      }
    }
  }

  /* bin up the data */
  for (j = 0; j < N_filter; j++) {
    bin_data(color_lambda[j],
             color_offset[j],
             color_err[j],
             color_size[j],
             bins, lambda_bin[j],
             offset_bin[j],
             err_bin[j], Nbins);
  }

  /* calculate offset matrix */

  /* initialize the matrix */
  for (i = 0; i < N_filter; i++) {
    for (j = 0; j < N_filter; j++) {
      temp_offset[i][j]    = 0;
      temp_offset_wt[i][j] = 0;
    }
  }

  /* calculate offset matrix */
  for (i = 0; i < N_filter; i++) {
    for (j = i; j < N_filter; j++) {
      for (k = 0; k < Nbins; k++) {
        if (err_bin[i][k] > 0) {
          hold = (err_bin[i][k] * err_bin[i][k]) +
                 (err_bin[j][k] * err_bin[j][k]);
          temp_offset[i][j] += (offset_bin[i][k] - offset_bin[j][k]) /
                               hold;
          temp_offset_wt[i][j] += 1 / hold;
        }
      }
    }
  }

  /* allocate phot error array */
  phot_new_err = (double *)calloc(N_filter, sizeof(double));

  /* calculate offsets for each filter from other filters */
  phot_new_off[0] = 0;
  phot_new_err[0] = 1.0 / sqrt(temp_offset_wt[0][0] * 2);

  for (i = 1; i < N_filter; i++) {
    phot_new_off[i] = 0;
    hold            = 0;
    for (j = (i - 1); j >= 0; j--) {
      if (temp_offset_wt[j][i] > 0) {
        phot_new_off[i] += temp_offset[j][i] +
                           phot_new_off[j] * temp_offset_wt[j][i];
        hold += temp_offset_wt[j][i];
      }
    }
    phot_new_off[i] /= hold;
    phot_new_err[i]  = 1.0 / sqrt(hold);
  }

  /* free offsets array */
  for (i = 0; i < N_filter; i++) {
    free(temp_offset[i]);
    free(temp_offset_wt[i]);
  }
  free(temp_offset);
  free(temp_offset_wt);
  free(phot_new_err);

  /* free data */
  for (i = 0; i < N_filter; i++) {
    free(color_offset[i]);
    free(color_err[i]);
    free(color_lambda[i]);
  }
  free(color_offset);
  free(color_err);
  free(color_lambda);
  free(color_size);

  /* free bins */
  for (i = 0; i < N_filter; i++) {
    free(lambda_bin[i]);
    free(offset_bin[i]);
    free(err_bin[i]);
  }
  free(lambda_bin);
  free(offset_bin);
  free(err_bin);
  free(bins);
}

void calc_sed_correction(int ref_filter,
                         double *filter_lambda,
                         double *zobj,
                         double **offset_phot,
                         double **offset_phot_err,
                         int dsed, int Nsed, int N_raw_sed,
                         int Nobj, int N_filter,
                         int *best_sed, double *best_chisq,
                         double *phot_new_off,
                         double **template_fix,
                         double **template_fix_lambda,
                         int *template_fix_size, int Nbins)
{
  int i, j, k, p;                                  /* counters */
  /* raw sed and mix counter for mixing */
  int    rawsed, mixsed;
  double mix;                                      /* mixing value */
  /* temporary variable carred no more than a few lines */
  double   hold;
  double **lambda_bin, **offset_bin, **err_bin;     /* bins for data */
  double **lambda_fin, **offset_fin, **spline_bin;  /* final corrections */
  /* Number of Sed's in each bin and the divisions between SED types */
  int *Nsed_bins, *Sed_bins, *Sed_counter;
  /* array to hold which sed bin the object is in */
  int *obj_binned_sed;
  /* array to hold offset data */
  double **offset_data, **wavelength_data, **offset_err;
  /* array to hold type of sed in offset and size of sed correction array */
  int **offset_type, *offset_Nbins;

  /* allocate the SED bins */
  Nsed_bins   = (int *)calloc(N_raw_sed, sizeof(int));
  Sed_bins    = (int *)calloc(N_raw_sed, sizeof(int));
  Sed_counter = (int *)calloc(N_raw_sed, sizeof(int));

  /* allocate sed_binned_sed; */
  obj_binned_sed = (int *)calloc(Nobj, sizeof(int));

  /* allocate the bins and final output */
  lambda_bin = (double **)calloc(N_raw_sed, sizeof(double *));
  offset_bin = (double **)calloc(N_raw_sed, sizeof(double *));
  lambda_fin = (double **)calloc(N_raw_sed, sizeof(double *));
  offset_fin = (double **)calloc(N_raw_sed, sizeof(double *));
  spline_bin = (double **)calloc(N_raw_sed, sizeof(double *));
  err_bin    = (double **)calloc(N_raw_sed, sizeof(double *));

  /* allocate bins array one larger than the rest for bottom of last bin */
  for (i = 0; i < N_raw_sed; i++) {
    lambda_bin[i] = (double *)calloc(Nbins, sizeof(double));
    offset_bin[i] = (double *)calloc(Nbins, sizeof(double));
    err_bin[i]    = (double *)calloc(Nbins, sizeof(double));
    offset_fin[i] = (double *)calloc(Nbins, sizeof(double));
    lambda_fin[i] = (double *)calloc(Nbins, sizeof(double));
    spline_bin[i] = (double *)calloc(Nbins, sizeof(double));
  }

  /* initialize the final data array */
  for (i = 0; i < N_raw_sed; i++) {
    for (j = 0; j < Nbins; j++) {
      lambda_fin[i][j] = 0;
      offset_fin[i][j] = 0;
    }
  }

  /* rebin data by sed correction */

  /* initialize Nsed_bins */
  for (i = 0; i < N_raw_sed; i++) {
    Nsed_bins[i] = 0;
  }

  /* setup the SED bins */
  Sed_bins[0] = dsed / 2;
  for (i = 1; i < N_raw_sed; i++) {
    Sed_bins[i] = Sed_bins[i - 1] + dsed;
  }

  for (i = 0; i < Nobj; i++) {
    /* figure out which bin we are in */
    k = 0;
    while ((best_sed[i] > Sed_bins[k]) && (k < (N_raw_sed - 1))) {
      k++;
    }

    /* remember where we are and increment counter */
    obj_binned_sed[i] = k;
    Nsed_bins[k]++;
  }

  /* allocate the array for the raw offsets */
  offset_data     = (double **)calloc(N_raw_sed, sizeof(double *));
  offset_err      = (double **)calloc(N_raw_sed, sizeof(double *));
  wavelength_data = (double **)calloc(N_raw_sed, sizeof(double *));
  offset_type     = (int **)calloc(N_raw_sed, sizeof(int *));
  offset_Nbins    = (int *)calloc(N_raw_sed, sizeof(int *));
  for (i = 0; i < N_raw_sed; i++) {
    offset_data[i]     = (double *)calloc(Nobj * N_filter, sizeof(double));
    offset_err[i]      = (double *)calloc(Nobj * N_filter, sizeof(double));
    wavelength_data[i] = (double *)calloc(Nobj * N_filter, sizeof(double));
    offset_type[i]     = (int *)calloc(Nobj * N_filter, sizeof(int));
    offset_Nbins[i]    = 0;
  }


  /* initialize Sed_counter */
  for (i = 0; i < N_raw_sed; i++) {
    Sed_counter[i]        = 0;
    wavelength_data[i][0] = BIG;
  }

  /* re-organize the data into new data array by SED rather than by
     object and filter
     sort as we go by lambda
   */
  for (i = 0; i < Nobj; i++) {
    for (j = 0; j < N_filter; j++) {
      /* check if we want this data in the array, make sure it isn't bad */
      if ((offset_phot[i][j] > -1) && (offset_phot[i][j] < 1)) {
        /* check to see where the data belongs */
        hold = filter_lambda[j] / (1 + zobj[i]);
        p    = 0;
        if (Sed_counter[obj_binned_sed[i]] == 0) {
          p = 0;
        } else {
          if (hold > wavelength_data[obj_binned_sed[i]][Sed_counter[obj_binned_sed[i]] - 1]) {
            p = Sed_counter[obj_binned_sed[i]];
          } else {
            while (hold > wavelength_data[obj_binned_sed[i]][p]) {
              p++;
            }
          }

          /* make space */
          for (k = Sed_counter[obj_binned_sed[i]]; k > p; k--) {
            wavelength_data[obj_binned_sed[i]][k] =
              wavelength_data[obj_binned_sed[i]][k - 1];
            offset_data[obj_binned_sed[i]][k] =
              offset_data[obj_binned_sed[i]][k - 1];
            offset_err[obj_binned_sed[i]][k] =
              offset_err[obj_binned_sed[i]][k - 1];
            offset_type[obj_binned_sed[i]][k] =
              offset_type[obj_binned_sed[i]][k - 1];
          }
        }

        /* put the data in place */
        wavelength_data[obj_binned_sed[i]][p] = filter_lambda[j] /
                                                (1 + zobj[i]);
        offset_data[obj_binned_sed[i]][p] = offset_phot[i][j] -
                                            phot_new_off[j] +
                                            phot_new_off[ref_filter];
        /* TODO put in something to consider the chisq */
        offset_err[obj_binned_sed[i]][p] = offset_phot_err[i][j] *
                                           exp(best_chisq[i] - 1);
        offset_type[obj_binned_sed[i]][p] = best_sed[i];
        Sed_counter[obj_binned_sed[i]]++;
      }
    }
  }

  /* loop over to undo mixing */
  for (k = 0; k < (dsed * dsed * 20); k++) {
    /* create binned offsets */
    for (i = 0; i < N_raw_sed; i++) {
      wt_bin_data(0.02, 0.04, wavelength_data[i],
                  offset_data[i], offset_err[i],
                  Sed_counter[i], lambda_bin[i],
                  offset_bin[i], err_bin[i], Nbins,
                  &offset_Nbins[i]);
    }

    /* add new offsets to final offsets and create spline arrays */
    for (i = 0; i < N_raw_sed; i++) {
      for (j = 0; j < offset_Nbins[i]; j++) {
        lambda_fin[i][j]  = lambda_bin[i][j];
        offset_fin[i][j] += offset_bin[i][j];
      }
      spline(lambda_bin[i],
             offset_bin[i],
             offset_Nbins[i],
             0, 0,
             spline_bin[i], BIG);
    }

    /* offset the photometry */
    for (i = 0; i < N_raw_sed; i++) {
      for (j = 0; j < Sed_counter[i]; j++) {
        /* determine the offset */
        rawsed = offset_type[i][j] / dsed;
        mixsed = offset_type[i][j] % dsed;
        mix    = 1.0 * mixsed / dsed;

        /* check if best_sed is too high */
        if (rawsed >= (N_raw_sed - 1)) {
          hold = splint2(lambda_bin[N_raw_sed - 1],
                         offset_bin[N_raw_sed - 1],
                         spline_bin[N_raw_sed - 1],
                         offset_Nbins[N_raw_sed - 1],
                         wavelength_data[i][j]);
          /* fprintf(stderr,"%d %lf\n",k,hold); */
          offset_data[i][j] -= hold;
        } else {
          /* fprintf(stderr,"%d %d %d %d %d %lf %e\n",
                             rawsed,
                             offset_type[i][j],
                             mixsed,i,j, mix,
                             wavelength_data[i][j]);
           */
          hold = ((1 - mix) * splint2(lambda_bin[rawsed],
                                      offset_bin[rawsed],
                                      spline_bin[rawsed],
                                      offset_Nbins[rawsed],
                                      wavelength_data[i][j])) +
                 (mix * splint2(lambda_bin[rawsed + 1],
                                offset_bin[rawsed + 1],
                                spline_bin[rawsed + 1],
                                offset_Nbins[rawsed],
                                wavelength_data[i][j]));
          /* fprintf(stderr,"%d %lf\n",k,hold); */
          offset_data[i][j] -= hold;
        }
      }
    }
  }

  /* output the final results */
  for (i = 0; i < N_raw_sed; i++) {
    template_fix_size[i] = offset_Nbins[i];
    for (j = 0; j < Nbins; j++) {
      template_fix_lambda[i][j] = lambda_fin[i][j];
      template_fix[i][j]        = offset_fin[i][j];
    }
  }

  /* free up the memory */
  for (i = 0; i < N_raw_sed; i++) {
    free(lambda_bin[i]);
    free(offset_bin[i]);
    free(err_bin[i]);
    free(offset_fin[i]);
    free(lambda_fin[i]);
    free(spline_bin[i]);
    free(offset_data[i]);
    free(offset_err[i]);
    free(wavelength_data[i]);
    free(offset_type[i]);
  }
  free(lambda_bin);
  free(offset_bin);
  free(err_bin);
  free(offset_fin);
  free(lambda_fin);
  free(spline_bin);
  free(offset_data);
  free(offset_err);
  free(wavelength_data);
  free(offset_type);

  free(offset_Nbins);
  free(Nsed_bins);
  free(Sed_bins);
  free(Sed_counter);
  free(obj_binned_sed);
}

void fix_seds(int N_filter,
              double *filter_lambda,
              int Nobj, int N_raw_sed, int Nsed,
              int dsed, int N_red,
              double ****template_phot,
              double ****template_new_phot,
              double *zobj,
              int Nfix,
              double **template_fix,
              double **template_fix_lambda,
              int Nbins)
{
  int    i, j, k, l, r;  /* counters */
  double mix, *fcorr, **spline_data;


  /* allocate the array for the corrections */
  fcorr = (double *)calloc(N_filter, sizeof(double));
  /* allocate array for spline info */
  spline_data = (double **)calloc(Nfix, sizeof(double *));

  for (i = 0; i < Nfix; i++) {
    spline_data[i] = (double *)calloc(Nbins, sizeof(double));
    spline(template_fix_lambda[i],
           template_fix[i],
           Nbins,
           0, 0,
           spline_data[i], BIG);
  }
  /* loop through all the sed's for all the objects */
  for (i = 0; i < Nobj; i++) {
    for (j = 0; j < (N_raw_sed - 1); j++) {
      for (k = 0; k < dsed; k++) {
        mix = 1.0 * k / dsed;

        /* generate the correction */
        if (j < (Nfix - 1)) {
          for (l = 0; l < N_filter; l++) {
            /* fprintf(stderr,"%d %d %d %d\n",i,j,k,l); */
            fcorr[l] = (1 - mix) * splint2(template_fix_lambda[j],
                                           template_fix[j],
                                           spline_data[j],
                                           Nbins,
                                           filter_lambda[l] / (1 + zobj[i])) +
                       (mix * splint2(template_fix_lambda[j + 1],
                                      template_fix[j + 1],
                                      spline_data[j + 1],
                                      Nbins,
                                      filter_lambda[l] / (1 + zobj[i])));
          }
        } else {
          for (l = 0; l < N_filter; l++) {
            fcorr[l] = splint2(template_fix_lambda[Nfix - 1],
                               template_fix[Nfix - 1],
                               spline_data[Nfix - 1], Nbins,
                               filter_lambda[l] / (1 + zobj[i]));
          }
        }

        /* go through all the redening levels */
        for (r = 0; r < N_red; r++) {
          for (l = 0; l < N_filter; l++) {
            template_new_phot[i][j * dsed + k][r][l] =
              template_phot[i][j * dsed + k][r][l] + fcorr[l];
          }
        }
      }
    }

    /* do the last template */
    for (l = 0; l < N_filter; l++) {
      fcorr[l] = splint2(template_fix_lambda[Nfix - 1],
                         template_fix[Nfix - 1],
                         spline_data[Nfix - 1],
                         Nbins, filter_lambda[l] / (1 + zobj[i]));
    }

    for (r = 0; r < N_red; r++) {
      for (l = 0; l < N_filter; l++) {
        template_new_phot[i][Nsed - 1][r][l] =
          template_phot[i][Nsed - 1][r][l] + fcorr[l];
      }
    }
  }

  for (i = 0; i < Nfix; i++) {
    free(spline_data[i]);
  }
  free(spline_data);

  free(fcorr);
}

void calc_new_offsets(int Nobj,
                      int N_filter,
                      int *best_sed,
                      int *best_red,
                      double *mu,
                      double ****template_phot,
                      double **obj_phot,
                      double **offset)
{
  int i, j;

  for (i = 0; i < Nobj; i++) {
    for (j = 0; j < N_filter; j++) {
      offset[i][j] = obj_phot[i][j] -
                     template_phot[i][best_sed[i]][best_red[i]][j] - mu[i];
    }
  }
}

/* add offsets to the photometry */
void fix_photometry(int ref_filter,
                    int N_filter, int Nobj,
                    double *new_offset, double **phot)
{
  int i, j;

  for (i = 0; i < Nobj; i++) {
    for (j = 0; j < N_filter; j++) {
      phot[i][j] += new_offset[j] - new_offset[ref_filter];
    }
  }
}

void sum_sed_repair(double **template_fix,
                    double **template_fix_lambda,
                    int *template_fix_size,
                    double **template_final,
                    double **template_final_lambda,
                    int Nbins, int Nfix)
{
  int      i, j;
  double **spline_data, hold;

  /* allocate array for spline info */
  spline_data = (double **)calloc(Nfix, sizeof(double *));
  for (i = 0; i < Nfix; i++) {
    spline_data[i] = (double *)calloc(template_fix_size[i], sizeof(double));
    spline(template_fix_lambda[i],
           template_fix[i],
           template_fix_size[i], 0, 0, spline_data[i], BIG);
  }

  for (i = 0; i < Nfix; i++) {
    for (j = 0; j < Nbins; j++) {
      hold = splint2(template_fix_lambda[i],
                     template_fix[i],
                     spline_data[i],
                     template_fix_size[i],
                     template_final_lambda[i][j]);

      template_final[i][j] += hold;
      /* fprintf(stderr,"%d %e %lf\n",i,template_final_lambda[i][j],hold); */
    }
  }
  free(spline_data);
}

/*---------------------------------------------------------------------------
                                   Main
   ---------------------------------------------------------------------------*/

/* This program will read in some SED's, filter profiles,
   and redshifts, and output magnitudes
 */
int main(int argc, char *argv[])
{
  int i, j, k, r, iteration_count, max_iteration_count;
  /* sed counters */
  int N_raw_sed, *raw_sed_size, dsed, dsed_todo, Nsed;
  /* data array for the raw SED's */
  double **raw_sed_wave_data, **raw_sed_data;
  /* filter info */
  int N_filter, *filter_size, N_fit_filter, ref_filter;
  /* data array for the filters */
  double **filter_wave_data, **filter_data, *filter_limit, *filter_lambda;
  /* temporary variable for estimating effective wavelengths of filters */
  double filter_trans;
  /* counter for objects and ID's for objects */
  int Nobj, *objID;
  /* redshift data for objects */
  double *zobj;
  /* object photometry */
  double **phot, **phot_err, *phot_init_off;
  /* minimum photometric error */
  double min_phot_err;
  /* counter for reddening */
  int N_red;
  /* Redening parameters */
  double Av_max, dAv;
  /* the size of the templates in the data cube */
  int **size_sed_data;
  /* Data cube for reddened, interpolated SED's */
  double ***sed_wave_data, ***sed_data;
  /* temporary arrays for red-shifting */
  double *hold_wave_data, *hold_data;
  /* file handles and strings for file names */
  FILE *f_sed_list, *f_filter_list, *f_obj_list, *f_read;
  FILE *f_output, *f_calib;
  char  incat_name[MAXCHAR]       = "incat.txt";
  char  sed_file_name[MAXCHAR]    = "seds.txt";
  char  filter_file_name[MAXCHAR] = "filters.txt";
  char  output_name[MAXCHAR]      = "output.txt";
  char  calib_file_name[MAXCHAR]  = "recalib.txt";
  /* temp arrays for IO */
  char line[MAXCHAR], filename[MAXCHAR], *success;
  char filter_filename[MAXCHAR][MAXCHAR];
  /* template photometry data */
  double ****template_phot, ****template_new_phot;
  /* chisq numbers */
  double chisq, newchisq, mu;
  /* average photometry offsets */
  double **offset_phot, **offset_wt, *final_offset;
  double * new_offset, max_offset, *obj_offset;
  int      large_offsets;
  /* output stuff */
  int *   best_red, *best_sed;
  double *best_mu, *best_chisq;
  /* corrections to the SED */
  double **template_fix, **template_fix_lambda;
  double **final_template, **final_template_lambda;
  /* definintion of correction matrix in lambda space */
  double corr_min, corr_max, corr_step;
  /* bins in SED correction */
  int Nbins, *template_fix_size;

  /* number of filters to use when fixing SEDS;
     if this is smaller than the number of input filters,
     the first Nfix filters are used; the variable currently
     has no effect; for fixing photometry all filters
     are used irrespective of Nfix
   */
  int Nfix;
  /* for parsing command line options */
  int option;

  /* some hard coded properties and default values: */
  min_phot_err = 0.0;
  max_iteration_count = 10;

  /* number of interpolations in the SED */
  dsed = 2;

  /* setup of reddening parameters */
  Av_max = 2;
  dAv    = 0.2;
  N_red  = Av_max / dAv;

  /* number of SED correction bins */
  Nbins     = 100;
  Nfix      = 5;
  corr_min  = log10(1000 / 1e10);
  corr_max  = log10(100000 / 1e10);
  corr_step = 1.0 * (corr_max - corr_min) / Nbins;

  ref_filter = 0;

  /* parse command line options */
  while ((option = getopt(argc, argv, "c:f:hi:n:o:r:s:")) != -1) {
    switch (option) {
    case 'c':
      strncpy(calib_file_name, optarg, MAXCHAR - 1);
      break;
    case 'e':
      min_phot_err = atof(optarg);
      break;
    case 'f':
      strncpy(filter_file_name, optarg, MAXCHAR - 1);
      break;
    case 'h':
      print_usage();
      exit(0);
      break;
    case 'i':
      strncpy(incat_name, optarg, MAXCHAR - 1);
      break;
    case 'm':
      max_iteration_count = atoi(optarg);
      break;
    case 'n':
      dsed = atoi(optarg);
      break;
    case 'o':
      strncpy(output_name, optarg, MAXCHAR - 1);
      break;
    case 'r':
      ref_filter = atoi(optarg);
      break;
    case 's':
      strncpy(sed_file_name, optarg, MAXCHAR - 1);
      break;
    default:
      print_usage();
      exit(0);
      break;
    }
  }

  /* print start message */
  fprintf(stdout, "\n%s (V%s) - zeropoint corrections for photo-z studies\n\n",
          __PROGRAM_NAME__, __PROGRAM_VERSION__);

  /* first sanity checks */
  if ((f_sed_list = fopen(sed_file_name, "r")) == NULL) {
    fprintf(stderr, "Cannot open SED file file %s\n", sed_file_name);
    fprintf(stderr, "\ntype '%s -h' to obtain a help message!\n", argv[0]);
    exit(1);
  }

  if ((f_filter_list = fopen(filter_file_name, "r")) == NULL) {
    fprintf(stderr, "Cannot open filter file file %s\n", filter_file_name);
    fprintf(stderr, "\ntype '%s -h' to obtain a help message!\n", argv[0]);
    exit(1);
  }

  if ((f_obj_list = fopen(incat_name, "r")) == NULL) {
    fprintf(stderr, "Cannot open object file file %s\n", incat_name);
    fprintf(stderr, "\ntype '%s -h' to obtain a help message!\n", argv[0]);
    exit(1);
  }

  if (dsed < 1) {
    dsed = 1;
    fprintf(stderr, "Warning: set number of SED interpolations to 1!\n");
  }

  if (min_phot_err < 0.0) {
    min_phot_err = 0.0;
    fprintf(stderr, "Warning: set minimum photometric error to 0.0!\n");
  }

  if (max_iteration_count < 1) {
    max_iteration_count = 1;
    fprintf(stderr, "Warning: set ZP correction iterations to 1!\n");
  }

  /*
     STUFF NEEDED FOR TODOS !!

     N_fit_filter = atoi(argv[5]);
     fprintf(stderr, "First %d Filters will determine best SED fit\n",
                  N_fit_filter);
   */

  /* read in the number of SED's from the SED list */
  fprintf(stdout, "Program run parameters:\n");
  if ((success = fgets(line, MAXCHAR, f_sed_list)) == line) {
    sscanf(line, "%d", &N_raw_sed);
    fprintf(stdout, "%d raw SED's\n", N_raw_sed);
    fprintf(stdout, "%d levels of interpolation\n", dsed);
    fprintf(stdout, "%d levels of reddening\n", N_red);
    fprintf(stdout, "A total of %d SEDs to fit\n", (((N_raw_sed - 1) *
                                                     dsed) + 1) * N_red);
  } else {
    fprintf(stderr, "First line of SED file must contain number of SEDs\n");
    fprintf(stderr, "\ntype '%s -h' to obtain a help message!\n", argv[0]);
    exit(1);
  }

  /* allocate an array for the size of the raw sed's */
  raw_sed_size = (int *)calloc(N_raw_sed, sizeof(int));

  /* allocate space for the raw sed's */
  raw_sed_wave_data = (double **)calloc(N_raw_sed, sizeof(double *));
  raw_sed_data      = (double **)calloc(N_raw_sed, sizeof(double *));

  /* Allocate memory and read in SED's */
  for (i = 0; i < N_raw_sed; i++) {
    if ((success = fgets(line, MAXCHAR, f_sed_list)) == line) {
      sscanf(line, "%s %d", filename, &raw_sed_size[i]);

      raw_sed_wave_data[i] = (double *)calloc(raw_sed_size[i],
                                              sizeof(double));
      raw_sed_data[i] = (double *)calloc(raw_sed_size[i],
                                         sizeof(double));

      if ((f_read = fopen(filename, "r")) == NULL) {
        fprintf(stderr, "Cannot open SED file %s\n", filename);
        fprintf(stderr, "\ntype '%s -h' to obtain a help message!\n", argv[0]);
        exit(1);
      }
      for (j = 0; j < raw_sed_size[i]; j++) {
        if ((success = fgets(line, MAXCHAR, f_read)) == line) {
          sscanf(line, "%lf %lf", &raw_sed_wave_data[i][j],
                 &raw_sed_data[i][j]);

          /* convert from Angstroem to m */
          raw_sed_wave_data[i][j] *= 1e-10;
        } else {
          break;
        }
      }
      fclose(f_read);
    } else {
      break;
    }
  }
  fclose(f_sed_list);

  /* allocate the SED corrections */
  template_fix          = (double **)calloc(N_raw_sed, sizeof(double *));
  template_fix_lambda   = (double **)calloc(N_raw_sed, sizeof(double *));
  template_fix_size     = (int *)calloc(N_raw_sed, sizeof(int));
  final_template        = (double **)calloc(N_raw_sed, sizeof(double *));
  final_template_lambda = (double **)calloc(N_raw_sed, sizeof(double *));

  for (i = 0; i < N_raw_sed; i++) {
    template_fix[i]          = (double *)calloc(Nbins, sizeof(double));
    template_fix_lambda[i]   = (double *)calloc(Nbins, sizeof(double));
    final_template[i]        = (double *)calloc(Nbins, sizeof(double));
    final_template_lambda[i] = (double *)calloc(Nbins, sizeof(double));
    for (j = 0; j < Nbins; j++) {
      final_template[i][j]        = 0;
      final_template_lambda[i][j] = pow(10, corr_min + j * corr_step);
    }
  }

  /* read in the number of Filters from Filter list */
  if ((success = fgets(line, MAXCHAR, f_filter_list)) == line) {
    sscanf(line, "%d", &N_filter);
    fprintf(stdout, "Number of Filters = %d\n", N_filter);

    /* for the time being we set N_fit_filter to N_filter;
       the variable is not used right now where we only estimate zeropoint
       offsets but no template corrections:
     */
    N_fit_filter = N_filter;
  } else {
    fprintf(stderr,
            "First line of filter file must contain number of filters\n");
    fprintf(stderr, "\ntype '%s -h' to obtain a help message!\n", argv[0]);
    exit(1);
  }

  /* allocate an array for the size of the filter profiles and filter limits */
  filter_size   = (int *)calloc(N_filter, sizeof(int));
  filter_limit  = (double *)calloc(N_filter, sizeof(double));
  filter_lambda = (double *)calloc(N_filter, sizeof(double));
  phot_init_off = (double *)calloc(N_filter, sizeof(double));

  /* allocate data array for the filters */
  filter_wave_data = (double **)calloc(N_filter, sizeof(double *));
  filter_data      = (double **)calloc(N_filter, sizeof(double *));

  /* read filter data */
  for (i = 0; i < N_filter; i++) {
    if ((success = fgets(line, MAXCHAR, f_filter_list)) == line) {
      sscanf(line, "%s %d %lf %lf", filter_filename[i],
             &filter_size[i],
             &filter_limit[i],
             &phot_init_off[i]);

      filter_wave_data[i] = (double *)calloc(filter_size[i], sizeof(double));
      filter_data[i]      = (double *)calloc(filter_size[i], sizeof(double));

      if ((f_read = fopen(filter_filename[i], "r")) == NULL) {
        fprintf(stderr, "Cannot open Filter file %s\n", filter_filename[i]);
        fprintf(stderr, "\ntype '%s -h' to obtain a help message!\n", argv[0]);
        exit(1);
      }

      filter_trans = 0;    /* for estimating effective filter wavelength */
      for (j = 0; j < filter_size[i]; j++) {
        if ((success = fgets(line, MAXCHAR, f_read)) == line) {
          sscanf(line, "%lf %lf", &filter_wave_data[i][j], &filter_data[i][j]);

          /* convert from Angstroem to m */
          filter_wave_data[i][j] *= 1.0e-10;

          filter_lambda[i] += filter_wave_data[i][j] * filter_data[i][j];
          filter_trans     += filter_data[i][j];

          /*printf("%e %lf\n",filter_wave_data[i][j], filter_data[i][j]); */
        } else {
          break;
        }
      }
      filter_lambda[i] /= filter_trans;
      fclose(f_read);
    } else {
      break;
    }
  }
  fclose(f_filter_list);

  fprintf(stdout, "Using reference filter: %s\n", filter_filename[ref_filter]);

  /* read in the number of objects from object file */
  if ((success = fgets(line, MAXCHAR, f_obj_list)) == line) {
    sscanf(line, "%d", &Nobj);
    fprintf(stdout, "Number of Objects = %d\n\n", Nobj);
  } else {
    fprintf(stderr,
            "First line of object file must contain number of objects\n");
    fprintf(stderr, "\ntype '%s -h' to obtain a help message!\n", argv[0]);
    exit(1);
  }

  objID = (int *)calloc(Nobj, sizeof(int));
  zobj  = (double *)calloc(Nobj, sizeof(double));

  /* allocate the array for the filter */
  phot     = (double **)calloc(Nobj, sizeof(double *));
  phot_err = (double **)calloc(Nobj, sizeof(double *));
  for (i = 0; i < Nobj; i++) {
    phot[i]     = (double *)calloc(N_filter, sizeof(double));
    phot_err[i] = (double *)calloc(N_filter, sizeof(double));
  }

  for (i = 0; i < Nobj; i++) {
    if ((success = fgets(line, MAXCHAR, f_obj_list)) == line) {
      /* read contents from each object line */
      objID[i] = atoi(strtok(line, " \t\n"));
      zobj[i]  = atof(strtok((char *)NULL, " \t\n"));

      for (j = 0; j < N_filter; j++) {
        phot[i][j] = atof(strtok((char *)NULL, " \t\n"));
      }

      for (j = 0; j < N_filter; j++) {
        phot_err[i][j] = atof(strtok((char *)NULL, " \t\n"));
        phot_err[i][j] = sqrt(phot_err[i][j] * phot_err[i][j] + 
                              min_phot_err * min_phot_err);
      }
    } else {
      break;
    }
  }

  fclose(f_obj_list);

  /* apply zeropoint offsets to the phtometry phot_init_off */
  for (i = 0; i < Nobj; i++) {
    for (j = 0; j < N_filter; j++) {
      if ((phot[i][j] != 99) && (phot[i][j] != -99)) {
        if (phot[i][j] > 0) {
          phot[i][j] += phot_init_off[j];
        } else {
          phot[i][j] -= phot_init_off[j];
        }
      }
    }
  }

  /* allocate a part of the data cube for the SED's
     do by SED first then by reddening
   */
  Nsed  = ((N_raw_sed - 1) * dsed) + 1;
  N_red = Av_max / dAv;

  size_sed_data = (int **)calloc(Nsed, sizeof(int *));
  sed_wave_data = (double ***)calloc(Nsed, sizeof(double **));
  sed_data      = (double ***)calloc(Nsed, sizeof(double **));
  for (i = 0; i < Nsed; i++) {
    size_sed_data[i] = (int *)calloc(N_red, sizeof(int));
    sed_wave_data[i] = (double **)calloc(N_red, sizeof(double *));
    sed_data[i]      = (double **)calloc(N_red, sizeof(double *));
  }

  /* ok let's fill the SED array */
  for (i = 0; i <= (N_raw_sed - 1); i++) {
    /* when arriving at the last template we do not need
       to do any interpolation in the loop */
    if (i < (N_raw_sed - 1)) {
      dsed_todo = dsed;
    } else {
      dsed_todo = 1;
    }
    for (j = 0; j < dsed_todo; j++) {
      for (r = 0; r < N_red; r++) {
        /* allocate the SED */
        size_sed_data[(i * dsed) + j][r] = raw_sed_size[i];
        sed_wave_data[(i * dsed) + j][r] = (double *)calloc(raw_sed_size[i],
                                                            sizeof(double));
        sed_data[(i * dsed) + j][r] = (double *)calloc(raw_sed_size[i],
                                                       sizeof(double));
        /* fill the new array */

        /* copy if no interpolation, else mix */
        if (j == 0) {
          for (k = 0; k < raw_sed_size[i]; k++) {
            sed_wave_data[(i * dsed) + j][r][k] = raw_sed_wave_data[i][k];
            sed_data[(i * dsed) + j][r][k]      = raw_sed_data[i][k];
          }
        } else {
          mix_template((1.0 * j / dsed),
                       raw_sed_size[i],
                       sed_wave_data[(i * dsed) + j][r],
                       sed_data[(i * dsed) + j][r],
                       raw_sed_wave_data[i], raw_sed_data[i],
                       raw_sed_size[i + 1],
                       raw_sed_wave_data[i + 1],
                       raw_sed_data[i + 1]);
        }
        /* redden the template as needed */
        if (r > 0) {
          redden_template((r * dAv),
                          raw_sed_size[i],
                          sed_wave_data[(i * dsed) + j][r],
                          sed_data[(i * dsed) + j][r]);
        }
      }
    }
  }

  /* Done with the raw SED's we can clear them out */
  for (i = 0; i < N_raw_sed; i++) {
    free(raw_sed_wave_data[i]);
    free(raw_sed_data[i]);
  }
  free(raw_sed_wave_data);
  free(raw_sed_data);
  free(raw_sed_size);

  /* allocate an array for the output photometry */
  template_phot     = (double ****)calloc(Nobj, sizeof(double ***));
  template_new_phot = (double ****)calloc(Nobj, sizeof(double ***));
  for (i = 0; i < Nobj; i++) {
    template_phot[i]     = (double ***)calloc(Nsed, sizeof(double **));
    template_new_phot[i] = (double ***)calloc(Nsed, sizeof(double **));
    for (j = 0; j < Nsed; j++) {
      template_phot[i][j]     = (double **)calloc(N_red, sizeof(double *));
      template_new_phot[i][j] = (double **)calloc(N_red, sizeof(double *));
      for (r = 0; r < N_red; r++) {
        template_phot[i][j][r] = (double *)calloc(N_filter,
                                                  sizeof(double));
        template_new_phot[i][j][r] = (double *)calloc(N_filter,
                                                      sizeof(double));
      }
    }
  }

  fprintf(stdout, "Generating templates for %d objects\n", Nobj);
  /* OK now we can do our red-shifting and integrations */
  for (i = 0; i < Nobj; i++) {
    for (j = 0; j < Nsed; j++) {
      for (r = 0; r < N_red; r++) {
        /* allocate array for temp data */
        hold_wave_data = (double *)calloc(size_sed_data[j][r], sizeof(double));
        hold_data      = (double *)calloc(size_sed_data[j][r], sizeof(double));

        /* redshift the template, output to temp data */
        redshift_template(zobj[i],
                          size_sed_data[j][r],
                          sed_wave_data[j][r],
                          sed_data[j][r],
                          hold_wave_data,
                          hold_data);

        /* integrate the filter profiles, and output the data */
        for (k = 0; k < N_filter; k++) {
          template_phot[i][j][r][k] = calc_magnitude(size_sed_data[j][r],
                                                     hold_wave_data,
                                                     hold_data,
                                                     filter_size[k],
                                                     filter_wave_data[k],
                                                     filter_data[k]);
        }

        if (i % 100 == 0) {
          fprintf(stdout, "templates for %d objects created\r", i);
        }

        /* free temp data array */
        free(hold_wave_data);
        free(hold_data);
      }
    }
  }

  /* free the SED and filter data since we don't need it anymore */
  for (i = 0; i < Nsed; i++) {
    for (r = 0; r < N_red; r++) {
      free(sed_wave_data[i][r]);
      free(sed_data[i][r]);
    }
    free(sed_wave_data[i]);
    free(sed_data[i]);
    free(size_sed_data[i]);
  }

  free(size_sed_data);
  free(sed_wave_data);
  free(sed_data);

  for (i = 0; i < N_filter; i++) {
    free(filter_wave_data[i]);
    free(filter_data[i]);
  }

  free(filter_wave_data);
  free(filter_data);
  free(filter_size);

  /* allocate an array for the photometry outputs */
  obj_offset  = (double *)calloc(N_filter, sizeof(double));
  offset_wt   = (double **)calloc(Nobj, sizeof(double *));
  offset_phot = (double **)calloc(Nobj, sizeof(double *));

  for (i = 0; i < Nobj; i++) {
    offset_wt[i]   = (double *)calloc(N_filter, sizeof(double));
    offset_phot[i] = (double *)calloc(N_filter, sizeof(double));

    for (j = 0; j < N_filter; j++) {
      offset_wt[i][j]   = 99;
      offset_phot[i][j] = 99;
    }
  }

  /* allocate array for photometry offsets; with calloc
     they are initialised to zero automatically
   */
  final_offset = (double *)calloc(N_filter, sizeof(double));
  new_offset   = (double *)calloc(N_filter, sizeof(double));

  /* allocate output arrays */
  best_red   = (int *)calloc(Nobj, sizeof(int));
  best_sed   = (int *)calloc(Nobj, sizeof(int));
  best_mu    = (double *)calloc(Nobj, sizeof(double));
  best_chisq = (double *)calloc(Nobj, sizeof(double));

  /* calculate zero order new phot */
  for (i = 0; i < Nobj; i++) {
    for (j = 0; j < Nsed; j++) {
      for (r = 0; r < N_red; r++) {
        for (k = 0; k < N_filter; k++) {
          template_new_phot[i][j][r][k] = template_phot[i][j][r][k];
        }
      }
    }
  }

  if (Nfix > N_raw_sed) {
    Nfix = N_raw_sed;
  }

  /* improve photometric offsets iteratively; do a maximum of 10
     iterations but stop if all the offsets would be corrected
     by 0.02 mag or less
   */
  iteration_count = 0;
  max_offset      = 0.;
  large_offsets   = 1;

  while (iteration_count < max_iteration_count && large_offsets == 1) {
    for (i = 0; i < N_filter; i++) {
      new_offset[i] = 0;
    }

    fprintf(stdout, "\nItertation %d for estimating photometric offsets:\n",
            iteration_count + 1);
    fprintf(stdout, "Fitting templates\n");

    /* fit the templates to the data */
    for (i = 0; i < Nobj; i++) {
      chisq = 1e99;
      for (j = 0; j < Nsed; j++) {
        for (r = 0; r < N_red; r++) {
          newchisq = fit_object(N_filter, N_fit_filter,
                                template_new_phot[i][j][r],
                                phot[i], phot_err[i],
                                filter_limit, &mu,
                                obj_offset);

          /* if best fit so far record the data */
          if (newchisq < chisq) {
            chisq         = newchisq;
            best_sed[i]   = j;
            best_red[i]   = r;
            best_mu[i]    = mu;
            best_chisq[i] = chisq;
            for (k = 0; k < N_filter; k++) {
              offset_phot[i][k] = obj_offset[k];
              offset_wt[i][k]   = phot_err[i][k];
            }
          }
        }
      }
    }

    fprintf(stdout, "Calculating photometric offsets\n");
    calc_phot_offset(offset_phot,
                     offset_wt,
                     zobj, filter_lambda,
                     Nobj, N_filter,
                     new_offset);

    /* check whether we need the new offsets at all */
    for (i = 0; i < N_filter; i++) {
      if (fabs(new_offset[i]) > max_offset) {
        max_offset = fabs(new_offset[i]);
      }
    }

    if (max_offset < 0.02) {
      large_offsets = 0;
      fprintf(stdout,
           "Iteration ends with a potential maximum photometric shift of %f\n",
           max_offset);
    } else {
      fprintf(stdout, "Filter \t Offset\n");
      for (i = 0; i < N_filter; i++) {
        /* the last term was new_offset[4] in the original code! */
        final_offset[i] += new_offset[i] - new_offset[ref_filter];
        fprintf(stdout, "%d \t %lf %lf\n", i, final_offset[i], new_offset[i]);
      }

      calc_new_offsets(Nobj, N_filter,
                       best_sed, best_red, best_mu,
                       template_new_phot, phot, offset_phot);

      fprintf(stdout, "Offseting the photometry\n");
      fix_photometry(ref_filter, N_filter, Nobj, new_offset, phot);
    }

    /* fprintf(stderr,"Calculating SED repair function\n");
       calc_sed_correction(ref_filter,
                           filter_lambda,
                           zobj,
                           offset_phot,
                           offset_wt,
                           dsed,
                           Nsed, Nfix, Nobj, N_filter,
                           best_sed, best_chisq,
                           new_offset,
                           template_fix, template_fix_lambda, template_fix_size,
                           Nbins);

       fprintf(stderr,"Repairing the SED's\n");

       sum_sed_repair(template_fix, template_fix_lambda, template_fix_size,
                      final_template, final_template_lambda, Nbins, Nfix);

       fix_seds(N_filter, filter_lambda, Nobj, N_raw_sed, Nsed,
                dsed, N_red, template_phot, template_new_phot,
                zobj,Nfix,final_template, final_template_lambda, Nbins);
     */

    max_offset = 0.;
    iteration_count++;
  }

  /* print output object catalogue */
  if ((f_output = fopen(output_name, "w")) == NULL) {
    fprintf(stderr, "Cannot open output catalogue file %s for writing\n",
            output_name);
    fprintf(stderr, "\ntype '%s -h' to obtain a help message!\n", argv[0]);
    exit(1);
  }

  for (i = 0; i < Nobj; i++) {
    for (j = 0; j < N_filter; j++) {
      fprintf(f_output, "%d %f %f %f %d %d %d %f %f %f\n",
              objID[i],
              best_chisq[i],
              zobj[i],
              filter_lambda[j] * 1e10,
              j, best_sed[i],
              best_red[i],
              template_new_phot[i][best_sed[i]][best_red[i]][j] + best_mu[i],
              phot[i][j],
              phot_err[i][j]);
    }
  }

  fclose(f_output);

  /* print final calibration numbers */
  if ((f_calib = fopen(calib_file_name, "w")) == NULL) {
    fprintf(stderr, "Cannot open calibration output file %s for writing\n",
            calib_file_name);
    fprintf(stderr, "\ntype '%s -h' to obtain a help message!\n", argv[0]);
    exit(1);
  }

  for (i = 0; i < N_filter; i++) {
    fprintf(f_calib, "%d %s %f\n", i, filter_filename[i], final_offset[i]);
  }

  fclose(f_calib);

  /*
     STUFF NEEDED FOR TODOS

     for(j=0;j<Nfix;j++){
     for(i=0;i<Nbins;i++){
      printf("%d %e %e\n",j,final_template_lambda[j][i], final_template[j][i]);
     }
     }
   */

  /* free up remaining memory */
  for (i = 0; i < Nobj; i++) {
    for (j = 0; j < Nsed; j++) {
      for (r = 0; r < N_red; r++) {
        free(template_phot[i][j][r]);
        free(template_new_phot[i][j][r]);
      }
      free(template_phot[i][j]);
      free(template_new_phot[i][j]);
    }
    free(template_phot[i]);
    free(template_new_phot[i]);
  }
  free(template_phot);
  free(template_new_phot);

  for (i = 0; i < Nobj; i++) {
    free(offset_wt[i]);
    free(offset_phot[i]);
  }

  free(offset_wt);
  free(offset_phot);

  free(objID);
  free(zobj);

  for (i = 0; i < Nobj; i++) {
    free(phot[i]);
    free(phot_err[i]);
  }

  free(phot);
  free(phot_err);

  for (i = 0; i < N_raw_sed; i++) {
    free(template_fix[i]);
    free(template_fix_lambda[i]);
    free(final_template[i]);
    free(final_template_lambda[i]);
  }

  free(template_fix);
  free(template_fix_lambda);
  free(final_template);
  free(final_template_lambda);
  free(template_fix_size);

  free(obj_offset);
  free(filter_limit);
  free(filter_lambda);
  free(phot_init_off);

  free(final_offset);
  free(new_offset);

  free(best_red);
  free(best_sed);
  free(best_mu);
  free(best_chisq);

  /* include the following for memory tracing */
  /* xmemory_status(); */

  return(0);
}

/* just print the programs usage/help message */
void print_usage(void)
{
  fprintf(stdout, "PROGRAMNAME\n");
  fprintf(stdout, "        calib_photoz_data - calibration of ZPs for photo-z studies\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "SYNOPSIS\n");
  fprintf(stdout, "        calib_photoz_data -i -o -c -f -s -r\n");
  fprintf(stdout, "              [ -i      # input object catalogue  (default: incat.txt)     ]\n");
  fprintf(stdout, "              [ -o      # output object catalogue (default: output.txt)    ]\n");
  fprintf(stdout, "              [ -c      # output calibration info (default: recalib.txt)   ]\n");
  fprintf(stdout, "              [ -f      # file with filter info   (default: filters.txt)   ]\n");
  fprintf(stdout, "              [ -s      # file with SED info      (default: seds.txt)      ]\n");
  fprintf(stdout, "              [ -r      # number of ref. filter   (default: 0)             ]\n");
  fprintf(stdout, "              [ -n      # number of SED interpolations       (default: 2)  ]\n");
  fprintf(stdout, "              [ -e      # minimum photometric error          (default: 0.) ]\n");
  fprintf(stdout, "              [ -m      # max. number of ZP corr. iterations (default: 10) ]\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "SHORT DESCRIPTION\n");
  fprintf(stdout, "   The program uses an object catalogue with spectroscopic redshift\n");
  fprintf(stdout, "   information and magnitudes (plus errors) from a multi-colour observation\n");
  fprintf(stdout, "   to estimate possible magnitude zeropoint calibration errors.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "   Similar to photometric redshift estimation it determines the best-fit SED\n");
  fprintf(stdout, "   to photometric data using the known spectroscopic redshift. The magnitudes\n");
  fprintf(stdout, "   predicted by the best fit SED are compared to the input magnitudes and\n");
  fprintf(stdout, "   offsets are determined.\n");
  fprintf(stdout, "   (SOME MORE DETAILS TO BE ADDED)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "INPUT FILES\n");
  fprintf(stdout, "   All input file have to 'precisely' follow the formats below. There\n");
  fprintf(stdout, "   must NOT be any comment lines in the files.\n");
  fprintf(stdout, " \n");
  fprintf(stdout, "   - The object catalogue (option -i) is an ASCII file with the format:\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "     ID z_spec mag_1 mag_2 .... magerr_1 magerr_2 ....\n");
  fprintf(stdout, "     .\n");
  fprintf(stdout, "     .\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "     with ID a running number, z_spec the spectroscopic redshift.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "   - The filter info file (option -f) is an ASCII file with the format:\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "     Total_number_of_filters\n");
  fprintf(stdout, "     path_to_filter_1  lines_in_1  limiting_mag_1 offset_in_1  \n");
  fprintf(stdout, "     .\n");
  fprintf(stdout, "     .\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "     The first line gives the number of filters; following lines give\n");
  fprintf(stdout, "     the names of the files with filter curves, the number of lines in\n");
  fprintf(stdout, "     the filter curve files, the limiting magnitudes and a known intial\n");
  fprintf(stdout, "     zeropoint offset.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "     The filter transmission files themselves have two columns:\n");
  fprintf(stdout, "     Wanvelength in Angstroems and throughput in arbitrary units.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "     (REMARK: It seems that the program implicitely assumes ordering\n");
  fprintf(stdout, "      of filters from blue to red!)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "   - The SED info file (option -s) is an ASCII file with the format:\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "     Total_number_of_SEDs\n");
  fprintf(stdout, "     path_to_SED_1  lines_in_1   \n");
  fprintf(stdout, "     .\n");
  fprintf(stdout, "     .\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "     The first line gives the number of SEDs; following lines give\n");
  fprintf(stdout, "     the names of the files with SED data and the number of lines in\n");
  fprintf(stdout, "     each SED data file.\n");
  fprintf(stdout, "    \n");
  fprintf(stdout, "     The SED data themselfs consist of two columns with the first being \n");
  fprintf(stdout, "     wavelength in Angstroem and the second quantum efficiency.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "OTHER INPUT PARAMETERS\n");
  fprintf(stdout, "   - The reference option (-r) gives the filter with the best photometric\n");
  fprintf(stdout, "     calibration. It is used as reference with a zero final offset.\n");
  fprintf(stdout, "     Counting of filters corresponds to the order of filters\n");
  fprintf(stdout, "     in the filter info file and starts with zero! \n");
  fprintf(stdout, "\n");
  fprintf(stdout, "   - The program can estimate intermediate galaxy types by\n");
  fprintf(stdout, "     interpolating between filters (N SEDs are interpolated between\n");
  fprintf(stdout, "     two basic types with N given by option (-n)); N=1 means no\n");
  fprintf(stdout, "     interpolation and the provided template set is used.  If\n");
  fprintf(stdout, "     interpolation is turned on the order of filters should correspond\n");
  fprintf(stdout, "     to the Hubble sequence.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "   - A minimum photometric error can be added quadratically to the\n");
  fprintf(stdout, "     input photometric errors (option (-e))\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "   - Zeropoint corrections are estimated iteratively. The default\n");
  fprintf(stdout, "     maximum number of iterations is 10. It can be adapted with \n");
  fprintf(stdout, "     option (-m)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "OUTPUT FILES\n");
  fprintf(stdout, "   - The output object catalogue (option -o) has the format:\n");
  fprintf(stdout, "     \n");
  fprintf(stdout, "     ID chi2 z_spec CWL filter_id reddening spectral_type model_mag mag magerr  \n");
  fprintf(stdout, "\n");
  fprintf(stdout, "     chi2 is the chi-squared per degree of freedom for the best fit\n");
  fprintf(stdout, "     SED at spectroscopic redshift z_spec. CWL is the central wavelength\n");
  fprintf(stdout, "     of the filter 'filter_id' model_mag is the expected magnitude\n");
  fprintf(stdout, "     from the SED representing spectral_type. The program reddens each \n");
  fprintf(stdout, "     template with the Calzetti et al. (2000) reddening law with Av ranging\n");
  fprintf(stdout, "     from 0 to 2 with a step of 0.2; 'reddening' gives the best fit reddening\n");
  fprintf(stdout, "     value (Av = 'reddening' * 0.2)\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "     A useful plot is (mag-mag_model) vs. restframe wavelength \n");
  fprintf(stdout, "     (CWL/(1+z_spec)) to verify whether magnitude offsets depend on this\n");
  fprintf(stdout, "     quanitity.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "   - The output calibration info (option -c) lists the filters together with\n");
  fprintf(stdout, "     estimated magnitude shifts.\n");
  fprintf(stdout, "     \n");
  fprintf(stdout, "AUTHOR\n");
  fprintf(stdout, "        Peter Capak \n");
  fprintf(stdout, "        (original program with name 'first_fit')\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "        current version maintained by\n");
  fprintf(stdout, "        Thomas Erben: terben@astro.uni-bonn.de\n");
}
