 /*
 				main.c

*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*
*	Part of:	LDACTest
*
*	Author:		E.BERTIN, DeNIS/LDAC
*
*	Contents:	parsing and main loop.
*
*	Last modify:	15/01/97
*
*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
*/

#include	<stdio.h>
#include	<stdlib.h>
#include	<ctype.h>

#include	"fitscat_defs.h"
#include	"fitscat.h"
#include        "plot_glob.h"
#include        "options_globals.h"
#include        "utils_globals.h"
#include        "utils_macros.h"


#define		SYNTAX \
"ldacshow \n\
	-i Catalog1 Catalog2 ... \n\
	[-m minimum_val]\n\
	[-p photometric_parameter] \n\
	[-l xmin xmax ymin ymax] \n\
	[options]\n\n\
Options are:	-q (Quiet flag: defaulted to verbose!)\n\
                -s (Use sky coordinates)\n"

#define		MAXCHAR 256

#define         XMAX  0.0
#define         YMAX  0.0

catstruct	*incat, *outcat;
int		qflag;

/********************************** main ************************************/
main(int argc, char *argv[])

/****** ldacshow ***************************************************
 *
 *  NAME	
 *      ldacshow - Display table information graphically
 *
 *  SYNOPSIS
 *      ldacshow -i incat1 incat2 ... -n # [-m min_val]
 *              [-l xmin xmax ymin ymax] [-q]
 *
 *  FUNCTION
 *      This program displays the objects from a catalog graphically. It
 *      opens an X-display (through the plplot package) and shows the
 *      position and bounding ellips of each object as stored in a specific
 *      fields OBJECTS table section.
 *
 *  INPUTS
 *      -i incat1 incat2 .. - A list of input catalog files
 *      -n #                - The field number to be displayed
 *     [-m min_val]         - The minimum photometric parameter value for
 *                            plotted objects
 *     [-l xmin xmax ymin ymax]
 *                          - Limiting pixel boundary values for the plot
 *                            area. Objects outside the area are clipped.
 *
 *  RESULTS
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *      plplot library required.
 *
 *  BUGS
 *
 *  AUTHOR
 *      E. Bertin - dd 01-08-1996
 *      E.R. Deul - dd 26-06-1996 Adapted to LDAC pipeline convention
 *                                Added boundary condition arguments
 *      E. Bertin - dd 15-01-1996 Added handling of errors
 *
 ******************************************************************************/
  {
   tabstruct	*tab, *ftab;
   keystruct	*key;
   char	str[MAXCHAR], outfilename[MAXCHAR], phot[MAXCHAR];
   int	a,ab,ae, i,j, ni, nobj, n;
   float minp, xmin,xmax,ymin,ymax;
   int *xwid, *ywid;
   int sky_coord, numb;

  if (argc<2)
    error(EXIT_FAILURE,"SYNTAX: ", SYNTAX);

/*default parameters */
  qflag = 1;
  n = 1;
  sky_coord = 0;
  minp = 0.0;
  xmin = ymin = 0.0;
  xmax = XMAX; ymax = YMAX;
  strcpy(phot,"FLUX_APER");

  sprintf(outfilename, "default_test.cat");

  ab=ae=ni=0;
  for (a=1; a<argc; a++)
    switch((int)tolower((int)argv[a][1]))
      {
      case 'i':	for(ab = ++a; (a<argc) && (argv[a][0]!='-'); a++);
		ae = a--;
		ni = ae - ab;
	        break;
      case 'n':	n = atoi(argv[++a]);
	        break;
      case 'o':	sprintf(outfilename, "%s", argv[++a]);
	        break;
      case 'p':	sprintf(phot, "%s", argv[++a]);
	        break;
      case 'l':	xmin = atof(argv[++a]);
                xmax = atof(argv[++a]);
                ymin = atof(argv[++a]);
                ymax = atof(argv[++a]);
	        break;
      case 'q': qflag = 1;
	        break;
      case 's': sky_coord = 1;
	        break;
      case 'm': minp = atof(argv[++a]);
	        break;
      default : error(EXIT_FAILURE,"SYNTAX: ", SYNTAX);
      }

  if (!ni)
    error(EXIT_SUCCESS,"SYNTAX: ", SYNTAX);

/*Load the input catalogs*/
  QFPRINTF(OUTPUT, "Reading catalog(s)");
  if (!(incat = read_cat(argv[ab])))
    error(EXIT_FAILURE, "*Error*: catalog not found: ", argv[ab]);
  about_cat(incat, stdout);
  nobj = read_field(tab = name_to_tab(incat,"OBJECTS", 0), NULL, NULL, 0, n,
		(ftab = name_to_tab(incat,"FIELDS", 0)));
  if (xmax == XMAX) {
     ED_GETKEY(ftab, key, "MAPNAXS1", xwid, numb, "ldacshow");
     xmax = (float) xwid[n-1];
  }
  if (ymax == YMAX) {
     ED_GETKEY(ftab, key, "MAPNAXS2", ywid, numb, "ldacshow");
     ymax = (float) ywid[n-1];
  }
  plotfield(tab, nobj,"",  phot, sky_coord, minp, xmin, xmax, ymin, ymax, argv[ab]);

  free_cat(incat, 1);

  QFPRINTF(OUTPUT, "All done");
  QPRINTF(OUTPUT,"\n");

  QFREE(xwid);
  QFREE(ywid);
  return EXIT_SUCCESS;
  }
