// Program to determine the mean of quantities using the
// N nearest neughbours of an object in a user defined 
// space.
//
// The program is largely a copy of sigmaellip.cpp
//

// HISTORY INFORMATION
//
// 10.01.2000: I had to remove the freeing of keynames at the
// end of the program; otherwise a crash on OSF (reason unknown)
//
// 15.07.2001:
// included "tabutil_globals.h" to declare add_key_to tab
//
// 05.05.03:
// included the tsize.h file
//
// 10.01.05:
// I removed compiler warnings under gcc with optimisation
// enabled (-O3).
//
// 29.01.2005:
// I now add HISTORY information to the output catalog.
//
// 13.08.2008:
// I changed the inclusion of iostream.h to iostream (the first
// is no longer supported by gcc 4.3)

#include <iostream>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "ANN.h"
#include "fitscat.h"
#include "tsize.h"
#include "utils_globals.h"
#include "tabutil_globals.h"

using namespace std;

#define __PROGRAMMVERSION__ "1.0"
#define __PROGRAMMNAME__ "nextneighbour"

#define __USAGE__ "\n\
USAGE:\n\
       nextneighbour -i input catalog\n\
                     -o output catalog\n\
                     -c position coordinates\n\
                     -n quantities from the neighbour\n\
                   [ -t table with the objects (OBJECTS) ]\n\n\
DESCRIPTION:\n\
The program calculates the distance of the next neighbour\n\
for all objects in table specified with '-t' and position\n\
coordinates specified by '-c' (any dimensional). The table\n\
get's a new key 'next_dist' containing the distances. In\n\
addition other quantities from the neighbours can be overtaken.\n\n\
EXAMPLE:\n\
  nextneighbour -i tmp.cat -o tmp1.cat -c Xpos Ypos -n snratio\n\n\
This calculates the distances using the position coordinates\n\
'Xpos' and 'Ypos' and the snratio of the neighbour is stored\n\
as 'next_snratio'.\n\n\
AUTHOR:\n\
      Thomas Erben\n"

#ifndef MAXSTRINGLENGTH
#define MAXSTRINGLENGTH 256
#endif

int main(int argc, char **argv)
{
  char incatname[MAXSTRINGLENGTH];
  char outcatname[MAXSTRINGLENGTH]="out.cat";
  char tablename[MAXSTRINGLENGTH]="OBJECTS";
  char tmpstring[MAXSTRINGLENGTH];

  char **keynames;
  int startcoords = 0;
  int ncoords=0;     // the number of coordinated defininge
                     // the position (2)
  int startneighbourquan = 0;  
  int nneighbourquan=0;   // the number of quantities from
                     // the next neighbour going to the objects
 
  catstruct *incat;
  catstruct *outcat;
  tabstruct *tab;
  keystruct **keys;

  double **keyptr;
  double **neighbourquan;

  // variables fro the next neighbour search
  int nobj;  // the number of objects in the catalog
  int nneighbours = 2; // the number of next neighbours
  // (hast to be 2 as the object itself is the
  //  nearest neighbour)

  ANNpointArray data_pts;
  ANNpoint      query_pt;
  ANNidxArray   nn_idx;
  ANNdistArray  dists;
  ANNkd_tree    *the_tree;

  // other variables
  int i, j, k;

  // read command line arguments
  for (i=1; i<argc; i++) 
  {
    if (argv[i][0] == '-') 
    {
       switch(tolower(argv[i][1])) 
       {
         case 'i': strncpy(incatname, argv[++i], MAXSTRINGLENGTH-1);
                   break;
         case 'o': strncpy(outcatname, argv[++i], MAXSTRINGLENGTH-1);
                   break;
         case 't': strncpy(tablename, argv[++i], MAXSTRINGLENGTH-1);
                   break;
         case 'c': startcoords=i+1;
                   while (i+1<argc && argv[i+1][0] != '-') 
                   {
                      i++;
                      ncoords++;
                   }
                   break;
         case 'n': startneighbourquan=i+1;
                   while (i+1<argc && argv[i+1][0] != '-') 
                   {
                      i++;
                      nneighbourquan++;
                   }
                   break;
 
         default:  
                   break;
       }
    } 
    else 
    {
       syserr(FATAL, "main", "Wrong Command line syntax\n");
    }
  }

  // print start message:
  cout << "\n" << __PROGRAMMNAME__ << "  V" << __PROGRAMMVERSION__ 
       << " (" << __DATE__ << ";" << __TIME__ << ")" << endl << endl;

  if(argc < 8)
  {
    cout << __USAGE__ << endl;
    exit(1);
  }

  if(ncoords < 1)
  {
    syserr(FATAL, "main", "no position coordinate given\n");
  }

  // we need the coordinate keys and the keys for the quantity
  // to calculate the mean (median)

  ED_CALLOC(keynames, "main", char *, ncoords+nneighbourquan);

  for (i=0;i<ncoords+nneighbourquan;i++) 
  {
    ED_CALLOC(keynames[i], "main", char, MAXSTRINGLENGTH);
  }

  for(i=0; i< ncoords; i++)
  {
    strncpy(keynames[i], argv[startcoords+i], MAXSTRINGLENGTH-1);
  }

  for(i=0; i< nneighbourquan; i++)
  {
    strncpy(keynames[ncoords+i], argv[startneighbourquan+i], MAXSTRINGLENGTH-1);
  }

  // read catalog and quantities:
  if((incat=read_cat(incatname))==NULL)
  {
    syserr(FATAL, "main", "Error opening input catalog\n");
  }

  outcat=new_cat(1);
  inherit_cat(incat, outcat);
  copy_tabs(incat,outcat);
  
  if((tab=name_to_tab(outcat,tablename,0))==NULL)
  {
    syserr(FATAL, "main", "No %s table in catalog\n", tablename);
  }
  
  // read keys:
  ED_CALLOC(keys, "main", keystruct *, ncoords+nneighbourquan);
  ED_CALLOC(keyptr, "main", double *, ncoords+nneighbourquan);

  read_keys(tab, keynames, keys, ncoords+nneighbourquan, NULL);
  
  for(j=0; j<ncoords+nneighbourquan; j++)
  {
    if(!keys[j])
    {
      syserr(FATAL, "main", "No key %s in table\n", keynames[j]);
    }

    switch (keys[j]->ttype)
    {
    case T_FLOAT:
      ED_GETVEC(keyptr[j], "main", keys[j], 
                keys[j]->name, double, float);  
      break;
    case T_DOUBLE:
      ED_GETVEC(keyptr[j], "main", keys[j], 
                keys[j]->name, double, double);
      break;
    case T_LONG:
      ED_GETVEC(keyptr[j], "main", keys[j], 
                keys[j]->name, double, long);
      break;
    case T_SHORT:
      ED_GETVEC(keyptr[j], "main", keys[j], 
                keys[j]->name, double, short);
      break;
    default:
       syserr(FATAL, "main", "Unsupported key type key type in %s\n", keys[j]->name);        
      break;
    }
  }

  nobj = keys[0]->nobj;

  // prepare stuff for the next neighbour search:
  query_pt = annAllocPt(ncoords);
  data_pts = annAllocPts(nobj, ncoords);
  nn_idx = new ANNidx[nneighbours];
  dists = new ANNdist[nneighbours];

  nneighbourquan++; // for the distances
  ED_CALLOC(neighbourquan, "main", double*, nneighbourquan);

  for(i=0; i< nneighbourquan; i++)
  {
    ED_CALLOC(neighbourquan[i], "main", double, nobj);
  }

  // copy the coordinates into the required matrix structure
  // of the ANNpointArray

  for(i=0; i< nobj; i++)
  {
    for(j=0; j< ncoords; j++)
    {
      data_pts[i][j] = keyptr[j][i];
    }
  }

  cout << "building search tree ..." << endl;

  the_tree = new ANNkd_tree((ANNpointArray)(data_pts), nobj, ncoords);

  for(i=0; i< nobj; i++)
  {
    for(j=0; j<ncoords; j++)
    {
      query_pt[j] = keyptr[j][i];
    }

    the_tree->annkSearch(query_pt, nneighbours, nn_idx, dists, 0.0);

    //for(j=0; j< nneighbours; j++)
    //{
      neighbourquan[0][i] = sqrt(dists[1]);
      for(k=1; k < nneighbourquan; k++)
      {
        neighbourquan[k][i] = keyptr[ncoords+k-1][nn_idx[1]];
      }
    //}
  }

  cout << "adding quantities to catalogs ...." << endl;

  add_key_to_tab(tab, "next_dist", nobj, neighbourquan[0], T_DOUBLE,
                        1, "", "");

  for(i=1; i< nneighbourquan; i++)
  {  
    strcpy(tmpstring, "next_");
    // The '-5' comes from "next_"
    strncat(tmpstring, keynames[ncoords+i-1], MAXSTRINGLENGTH-1-5); 
    add_key_to_tab(tab, tmpstring, nobj, neighbourquan[i], T_DOUBLE,
                          1, "", "");
  }

  /* add some history information to the output catalog */
  historyto_cat(outcat, "nextneighbour", argc, argv);
  
  save_cat(outcat, outcatname);

  // free memory and bye
  delete [] nn_idx;
  delete [] dists;

  annDeallocPt(query_pt);
  annDeallocPts(data_pts);

  delete the_tree;

  for(i=0; i< nneighbourquan; i++)
  {
    ED_FREE(neighbourquan[i]);
  }
  ED_FREE(neighbourquan);

  /*for (i=0;i<ncoords+nneighbourquan;i++) 
  {
    ED_FREE(keynames[i]);
  }

  ED_FREE(keynames);*/

  /*free_cat(incat, 1);*/
  /*free_cat(outcat, 1);*/
    
  return 0;
}  


