#include <math.h>
#include "options_globals.h"
#include "utils_globals.h"

/* 15.07.01:
 * changed the var type of ndat (function avevar) from
 * unsigned long to int to avoid compiler warnings.
 *
 * 02.03.2004:
 * function moments:
 * if the variance is zero, only a warning (not an error)
 * is issued and skewness and kurtosis are set to zero.
 *
 * 20.07.2004:
 * - I removed compiler warnings due to possibly
 *   undefined variables (compilation with optimisation
 *   turned on)
 * - I removed the definition of the SQR function and
 *   calculate the sqaure by hand (the definition led
 *   to compiler warnings)
 *
 *  09.03.2006:
 *  I removed the last argument from the NR free_... functions.
 *  That argument is not used in the functions.
 *
 */

void moments(float *data, int n, float *ave, float *adev, 
             float *sdev, float *svar, float *skew, float *curt)
{
    int j;
    float s,p;

    if (n <= 0) 
       syserr(FATAL,"moments", "n must be at least 1\n");
    s=0.0;
    for (j=0;j<n;j++) s += data[j];
    *ave=s/n;
    *adev=(*svar)=(*skew)=(*curt)=0.0;
    for (j=0;j<n;j++) {
       *adev += fabs(s=data[j]-(*ave));
       *svar += (p=s*s);
       *skew += (p *= s);
       *curt += (p *= s);
    }
    *adev /= n;
    if(n>1)
    {
      *svar /= (n-1);
      *sdev=sqrt(*svar);
    }
    if (*svar) {
       *skew /= (n*(*svar)*(*sdev));
       *curt=(*curt)/(n*(*svar)*(*svar))-3.0;
    } 
    else
    { 
      syserr(WARNING,"moments", "No skew/kurtosis when variance = 0\n");
      *skew=0.0;
      *curt=0.0;
    }
}

#define POTN 1.571000
#define BIG 1.0e30
#define PI 3.14159265
#define ACC 1.0e-3

int nn;
float *xx,*yy,*sx,*sy,*ww,aa,offs;

void fitexy(float lx[], float ly[], int ndat, float lsigx[], float lsigy[],
     float *a, float *b, float *siga, float *sigb, float *chi2, float *q)
{
    void avevar(float data[], int n, float *ave, float *var);
    float brent(float ax, float bx, float cx, 
         float (*f)(float), float tol, float *xmin);
    float chixy(float bang);
    void fit(float x[], float y[], int ndata, float sig[], int mwt,
     float *a, float *b, float *siga, float *sigb, float *chi2, float *q);
    float gammq(float a, float x);
    void mnbrak(float *ax, float *bx, float *cx, float *fa, float *fb,
        float *fc, float (*func)(float));
    float zbrent(float (*func)(float), float x1, float x2, float tol);
    int j;
    float swap,amx,amn,varx,vary,ang[7],ch[7],scale,bmn,bmx,d1,d2,r2,
       dum1,dum2,dum3,dum4,dum5;
    float *x,*y,*sigx,*sigy;

    x   =vector(1,ndat);
    y   =vector(1,ndat);
    sigx=vector(1,ndat);
    sigy=vector(1,ndat);

    for (j=1;j<=ndat;j++) {
       x[j]=lx[j-1];
       y[j]=ly[j-1];
       sigx[j]=lsigx[j-1];
       sigy[j]=lsigy[j-1];
       VPRINTF(DEBUG,"%f %f %f %f\n", x[j],y[j],sigx[j],sigy[j]);
    }
    xx=vector(1,ndat);
    yy=vector(1,ndat);
    sx=vector(1,ndat);
    sy=vector(1,ndat);
    ww=vector(1,ndat);
    avevar(x,ndat,&dum1,&varx);
    avevar(y,ndat,&dum1,&vary);
    scale=sqrt(varx/vary);
    nn=ndat;
    for (j=1;j<=ndat;j++) {
       xx[j]=x[j];
       yy[j]=y[j]*scale;
       sx[j]=sigx[j];
       sy[j]=sigy[j]*scale;
       ww[j]=sqrt((sx[j]*sx[j])+(sy[j]*sy[j]));
    }
    fit(xx,yy,nn,ww,1,&dum1,b,&dum2,&dum3,&dum4,&dum5);
    offs=ang[1]=0.0;
    ang[2]=atan(*b);
    ang[4]=0.0;
    ang[5]=ang[2];
    ang[6]=POTN;
    for (j=4;j<=6;j++)
       ch[j]=chixy(ang[j]);
    mnbrak(&ang[1],&ang[2],&ang[3],&ch[1],&ch[2],&ch[3],chixy);
    *chi2=brent(ang[1],ang[2],ang[3],chixy,ACC,b);
    *chi2=chixy(*b);
    *a=aa;
    *q=gammq(0.5*(nn-2),*chi2*0.5);
    for (r2=0.0,j=1;j<=nn;j++) r2 += ww[j];
    r2=1.0/r2;
    bmx=BIG;
    bmn=BIG;
    offs=(*chi2)+1.0;
    for (j=1;j<=6;j++) {
       if (ch[j] > offs) {
          d1=fabs(ang[j]-(*b));
          while (d1 >= PI) d1 -= PI;
          d2=PI-d1;
          if (ang[j] < *b) {
             swap=d1;
             d1=d2;
             d2=swap;
          }
          if (d1 < bmx) bmx=d1;
          if (d2 < bmn) bmn=d2;
       }
    }
    if (bmx < BIG &&
        chixy(*b)*chixy(*b+bmx)<0.0&&
        chixy(*b)*chixy(*b-bmn)<0.0) {
       bmx=zbrent(chixy,*b,*b+bmx,ACC)-(*b);
       amx=aa-(*a);
       bmn=zbrent(chixy,*b,*b-bmn,ACC)-(*b);
       amn=aa-(*a);
       *sigb=sqrt(0.5*(bmx*bmx+bmn*bmn))/(scale*(cos(*b)*cos(*b)));
       *siga=sqrt(0.5*(amx*amx+amn*amn)+r2)/scale;
    } else 
       (*sigb)=(*siga)=BIG;
    *a /= scale;
    *b=tan(*b)/scale;
    free_vector(x,1);
    free_vector(y,1);
    free_vector(sigx,1);
    free_vector(sigy,1);
    free_vector(ww,1);
    free_vector(sy,1);
    free_vector(sx,1);
    free_vector(yy,1);
    free_vector(xx,1);
}

float chixy(float bang)
{
    int j;
    float ans,avex=0.0,avey=0.0,sumw=0.0,b;

    b=tan(bang);
    for (j=1;j<=nn;j++) {
       ww[j] = (b*sx[j]*b*sx[j])+(sy[j]*sy[j]);
       sumw += (ww[j] = (ww[j] == 0.0 ? BIG : 1.0/ww[j]));
       avex += ww[j]*xx[j];
       avey += ww[j]*yy[j];
    }
    if (sumw == 0.0) sumw = BIG;
    avex /= sumw;
    avey /= sumw;
    aa=avey-b*avex;
    for (ans=-offs,j=1;j<=nn;j++)
       ans += ww[j]*(yy[j]-aa-b*xx[j])*(yy[j]-aa-b*xx[j]);
    return ans;
}


void avevar(float data[], int n, float *ave, float *svar)
{
	int j;
	float s;

	*ave=(*svar)=0.0;
	for (j=1;j<=n;j++) *ave += data[j];
	*ave /= n;
	for (j=1;j<=n;j++) {
		s=data[j]-(*ave);
		*svar += s*s;
	}
	*svar /= (n-1);
}

void fit(float x[], float y[], int ndata, float sig[], int mwt,
         float *a, float *b, float *siga, float *sigb, float *chi2, float *q)
{
	int i;
	float wt,t,sxoss,sx=0.0,sy=0.0,st2=0.0,ss,sigdat;
        float gammq(float a, float x);

	*b=0.0;
	if (mwt) {
		ss=0.0;
		for (i=1;i<=ndata;i++) {
			wt=1.0/(sig[i]*sig[i]);
			ss += wt;
			sx += x[i]*wt;
			sy += y[i]*wt;
		}
	} else {
		for (i=1;i<=ndata;i++) {
			sx += x[i];
			sy += y[i];
		}
		ss=ndata;
	}
	sxoss=sx/ss;
	if (mwt) {
		for (i=1;i<=ndata;i++) {
			t=(x[i]-sxoss)/sig[i];
			st2 += t*t;
			*b += t*y[i]/sig[i];
		}
	} else {
		for (i=1;i<=ndata;i++) {
			t=x[i]-sxoss;
			st2 += t*t;
			*b += t*y[i];
		}
	}
	*b /= st2;
	*a=(sy-sx*(*b))/ss;
	*siga=sqrt((1.0+sx*sx/(ss*st2))/ss);
	*sigb=sqrt(1.0/st2);
	*chi2=0.0;
	if (mwt == 0) {
		for (i=1;i<=ndata;i++)
			*chi2 += (y[i]-(*a)-(*b)*x[i])*(y[i]-(*a)-(*b)*x[i]);
		*q=1.0;
		sigdat=sqrt((*chi2)/(ndata-2));
		*siga *= sigdat;
		*sigb *= sigdat;
	} else {
		for (i=1;i<=ndata;i++)
			*chi2 += ((y[i]-(*a)-(*b)*x[i])/sig[i])*
                                 ((y[i]-(*a)-(*b)*x[i])/sig[i]);
		*q=gammq(0.5*(ndata-2),0.5*(*chi2));
	}
}

void fit0(float x0[], float y0[], int ndata, float sig0[], int mwt,
         float *a, float *b, float *siga, float *sigb, float *chi2, float *q)
{
    float	*x, *y, *sig;
    int		i;

    x   = vector(1,ndata);
    y   = vector(1,ndata);
    sig = vector(1,ndata);
    for (i=1;i<=ndata;i++) {
       x[i]   = x0[i-1];
       y[i]   = y0[i-1];
       sig[i] = sig0[i-1];
       VPRINTF(DEBUG,"%f %f %f \n", x[i], y[i], sig[i]);
    }
    fit(x, y, ndata, sig, mwt, a, b, siga, sigb, chi2, q);
    free_vector(x,1);
    free_vector(y,1);
    free_vector(sig,1);
}

float gammq(float a, float x)
{
	float gamser,gammcf,gln;
        void gser(float *gamser, float a, float x, float *gln);
        void gcf(float *gammcf, float a, float x, float *gln);

	if (x < 0.0 || a <= 0.0) {
            syserr(WARNING,"gammq","Invalid arguments in routine GAMMQ");
            return 0.0;
        }
	if (x < (a+1.0)) {
		gser(&gamser,a,x,&gln);
		return 1.0-gamser;
	} else {
		gcf(&gammcf,a,x,&gln);
		return gammcf;
	}
}

#define ITMAX 100
#define EPS 3.0e-7

void gser(float *gamser, float a, float x, float *gln)
{
	int n;
	float sum,del,ap;
	float gammln(float xx);

	*gln=gammln(a);
	if (x <= 0.0) {
		if (x < 0.0) 
                   syserr(WARNING,"gser","x less than 0 in routine GSER");
		*gamser=0.0;
		return;
	} else {
		ap=a;
		del=sum=1.0/a;
		for (n=1;n<=ITMAX;n++) {
			ap += 1.0;
			del *= x/ap;
			sum += del;
			if (fabs(del) < fabs(sum)*EPS) {
				*gamser=sum*exp(-x+a*log(x)-(*gln));
				return;
			}
		}
		syserr(WARNING,"gser","a too large, ITMAX too small in routine GSER");
                *gamser=1.0;
		return;
	}
}

#undef ITMAX
#undef EPS


#define ITMAX 100
#define EPS 3.0e-7

void gcf(float *gammcf, float a, float x, float *gln)
{
	int n;
	float gold=0.0,g,fac=1.0,b1=1.0;
	float b0=0.0,anf,ana,an,a1,a0=1.0;
	float gammln(float xx);

	*gln=gammln(a);
	a1=x;
	for (n=1;n<=ITMAX;n++) {
		an=(float) n;
		ana=an-a;
		a0=(a1+a0*ana)*fac;
		b0=(b1+b0*ana)*fac;
		anf=an*fac;
		a1=x*a0+anf*a1;
		b1=x*b0+anf*b1;
		if (a1) {
			fac=1.0/a1;
			g=b1*fac;
			if (fabs((g-gold)/g) < EPS) {
				*gammcf=exp(-x+a*log(x)-(*gln))*g;
				return;
			}
			gold=g;
		}
	}
	syserr(WARNING,"gcf","a too large, ITMAX too small in routine GCF");
        *gammcf=0.0;
}

#undef ITMAX
#undef EPS


float gammln(float xx)
{
	double x,tmp,ser;
	static double cof[6]={76.18009173,-86.50532033,24.01409822,
		-1.231739516,0.120858003e-2,-0.536382e-5};
	int j;

	x=xx-1.0;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.0;
	for (j=0;j<=5;j++) {
		x += 1.0;
		ser += cof[j]/x;
	}
	return -tmp+log(2.50662827465*ser);
}

#undef MAX
#define GOLD 1.618034
#define GLIMIT 100.0
#define TINY 1.0e-20
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#define SIGN(a,b) ((b) > 0.0 ? fabs(a) : -fabs(a))
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);

void mnbrak(float *ax, float *bx, float *cx, float *fa, float *fb, float *fc,
            float (*func)(float))
{
	float ulim,u,r,q,fu,dum;

	*fa=(*func)(*ax);
	*fb=(*func)(*bx);
	if (*fb > *fa) {
		SHFT(dum,*ax,*bx,dum)
		SHFT(dum,*fb,*fa,dum)
	}
	*cx=(*bx)+GOLD*(*bx-*ax);
	*fc=(*func)(*cx);
	while (*fb > *fc) {
		r=(*bx-*ax)*(*fb-*fc);
		q=(*bx-*cx)*(*fb-*fa);
		u=(*bx)-((*bx-*cx)*q-(*bx-*ax)*r)/
			(2.0*SIGN(MAX(fabs(q-r),TINY),q-r));
		ulim=(*bx)+GLIMIT*(*cx-*bx);
		if ((*bx-u)*(u-*cx) > 0.0) {
			fu=(*func)(u);
			if (fu < *fc) {
				*ax=(*bx);
				*bx=u;
				*fa=(*fb);
				*fb=fu;
				return;
			} else if (fu > *fb) {
				*cx=u;
				*fc=fu;
				return;
			}
			u=(*cx)+GOLD*(*cx-*bx);
			fu=(*func)(u);
		} else if ((*cx-u)*(u-ulim) > 0.0) {
			fu=(*func)(u);
			if (fu < *fc) {
				SHFT(*bx,*cx,u,*cx+GOLD*(*cx-*bx))
				SHFT(*fb,*fc,fu,(*func)(u))
			}
		} else if ((u-ulim)*(ulim-*cx) >= 0.0) {
			u=ulim;
			fu=(*func)(u);
		} else {
			u=(*cx)+GOLD*(*cx-*bx);
			fu=(*func)(u);
		}
		SHFT(*ax,*bx,*cx,u)
		SHFT(*fa,*fb,*fc,fu)
	}
}

#undef GOLD
#undef GLIMIT
#undef TINY
#undef MAX
#undef SIGN
#undef SHFT


#define ITMAX 100
#define CGOLD 0.3819660
#define ZEPS 1.0e-10
#define SIGN(a,b) ((b) > 0.0 ? fabs(a) : -fabs(a))
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);

float brent(float ax, float bx, float cx, float (*f)(float), 
            float tol, float *xmin)
{
	int iter;
	float a,b,d=0.0,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
	float e=0.0;

	a=((ax < cx) ? ax : cx);
	b=((ax > cx) ? ax : cx);
	x=w=v=bx;
	fw=fv=fx=(*f)(x);
	for (iter=1;iter<=ITMAX;iter++) {
		xm=0.5*(a+b);
		tol2=2.0*(tol1=tol*fabs(x)+ZEPS);
		if (fabs(x-xm) <= (tol2-0.5*(b-a))) {
			*xmin=x;
			return fx;
		}
		if (fabs(e) > tol1) {
			r=(x-w)*(fx-fv);
			q=(x-v)*(fx-fw);
			p=(x-v)*q-(x-w)*r;
			q=2.0*(q-r);
			if (q > 0.0) p = -p;
			q=fabs(q);
			etemp=e;
			e=d;
			if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
				d=CGOLD*(e=(x >= xm ? a-x : b-x));
			else {
				d=p/q;
				u=x+d;
				if (u-a < tol2 || b-u < tol2)
					d=SIGN(tol1,xm-x);
			}
		} else {
			d=CGOLD*(e=(x >= xm ? a-x : b-x));
		}
		u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
		fu=(*f)(u);
		if (fu <= fx) {
			if (u >= x) a=x; else b=x;
			SHFT(v,w,x,u)
			SHFT(fv,fw,fx,fu)
		} else {
			if (u < x) a=u; else b=u;
			if (fu <= fw || w == x) {
				v=w;
				w=u;
				fv=fw;
				fw=fu;
			} else if (fu <= fv || v == x || v == w) {
				v=u;
				fv=fu;
			}
		}
	}
	syserr(FATAL,"brent","Too many iterations in BRENT");
	*xmin=x;
	return fx;
}

#undef ITMAX
#undef CGOLD
#undef ZEPS
#undef SIGN


#define ITMAX 100
#define EPS 3.0e-8

float zbrent(float (*func)(float), float x1, float x2, float tol)
{
	int iter;
	float a=x1,b=x2,c=0.0,d=0.0,e=0.0,min1,min2;
	float fa=(*func)(a),fb=(*func)(b),fc,p,q,r,s,tol1,xm;

	if (fb*fa > 0.0) syserr(FATAL,"zbrent","Root must be bracketed in ZBRENT");
	fc=fb;
	for (iter=1;iter<=ITMAX;iter++) {
		if (fb*fc > 0.0) {
			c=a;
			fc=fa;
			e=d=b-a;
		}
		if (fabs(fc) < fabs(fb)) {
			a=b;
			b=c;
			c=a;
			fa=fb;
			fb=fc;
			fc=fa;
		}
		tol1=2.0*EPS*fabs(b)+0.5*tol;
		xm=0.5*(c-b);
		if (fabs(xm) <= tol1 || fb == 0.0) return b;
		if (fabs(e) >= tol1 && fabs(fa) > fabs(fb)) {
			s=fb/fa;
			if (a == c) {
				p=2.0*xm*s;
				q=1.0-s;
			} else {
				q=fa/fc;
				r=fb/fc;
				p=s*(2.0*xm*q*(q-r)-(b-a)*(r-1.0));
				q=(q-1.0)*(r-1.0)*(s-1.0);
			}
			if (p > 0.0)  q = -q;
			p=fabs(p);
			min1=3.0*xm*q-fabs(tol1*q);
			min2=fabs(e*q);
			if (2.0*p < (min1 < min2 ? min1 : min2)) {
				e=d;
				d=p/q;
			} else {
				d=xm;
				e=d;
			}
		} else {
			d=xm;
			e=d;
		}
		a=b;
		fa=fb;
		if (fabs(d) > tol1)
			b += d;
		else
			b += (xm > 0.0 ? fabs(tol1) : -fabs(tol1));
		fb=(*func)(b);
	}
	syserr(FATAL,"zbrent","Maximum number of iterations exceeded in ZBRENT");
        return b;
}

#undef ITMAX
#undef EPS
