/* 21.06.01:
   removed most (up to one) compiler warnings under DEC

   09.09.01:
   resolved short int, int, long int, LONG issues in the global
   variables:
   fieldfpos, rows, cols,refxpos, refypos
   and the var. isum in make_reference_zps

   10.01.03:
   removed most (up to one) compiler warnings under Linux

   10.01.05:
   fixed a potential bug in read_fixed; it is potential as
   the functionality of getting the field pos from the FITS
   filename was never used. A char (instead of a 'char *'
   was given to the strncmp function comparing two strings.

   28.02.2005:
   routine 'derive_abs_phot':
   the 'fixed' field to be considered is now determined directly
   from the fieldfpos instead from row, col. Thus, the routine
   works correctly if the EIS concept of rows and columns is not
   valid for the data under consideration.

   13.03.2005:
   function derive_abs_phot:
   We now fill two arrays indicating whether a vertain frame is
   under the photometric ones and we also save a corrected zeropoint
   for these frames (the corrected zeropoint is the zeropoint at zero 
   airmass corrected for extinction and estimated relative zeropoint.
   Hence, ideally, these corrected zeropoints should all coincide.

   Changes to avoid compiler warnings with -Wall -W -pedantic

   09.03.2006:
   I removed the last argument from the NR free_... functions.
   That argument is not used in the functions.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "fitscat.h"
#include "options_globals.h"
#include "utils_globals.h"
#include "lists_globals.h"
#include "relphotom_globals.h"
#include "fixed_phot_globals.h"
#include "sv_globals.h"

static int      nfixed;
static float    *fixedcfs;
static float    *fixedzps;
static keystruct *fposkey;
static short int *fieldfpos;
static char	*fitsfile;

static LONG     *rows, *cols;
static int      nfields;
static double   *ra, *dec;
static double   *airmass;
static char	*filenames;

static float    *refzps;
static LONG     *refxpos, *refypos;
static int       nref;

extern control_struct control;


/**/
int in_list(int k, short int *list, int n)
/****** in_list ***************************************************
 *
 *  NAME	
 *      in_list - Get the index of an indexed list
 *
 *  SYNOPSIS
 *      int in_list(int k, int *list, int n)
 *
 *  FUNCTION
 *      Get the index to the element of list that has the value of k
 *
 *  INPUTS
 *      k    - The value of list to find
 *      list - The array to search in
 *      n    - The length of the array list
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      The index in the list of the element that equals k.
 *      Or -1 is there is no element with value k
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 03-03-1998
 *
 ******************************************************************************/
{
   int		i;

   for (i=0;i<n;i++) 
      if (list[i] == k)
         return i;

   return -1;
}
/**/
void read_fixed(catstruct *cat)
/****** read_fixed ***************************************************
 *
 *  NAME	
 *      read_fixed - Read the FIXED table 
 *
 *  SYNOPSIS
 *      void read_fixed()
 *
 *  FUNCTION
 *      Read the appropriate information from the FIXED table of the
 *      user provided fixed photometric data. The FIXEd table contains
 *      a list of field numbers for which the user knows that they have
 *      been observed in photometric conditions. The list also contains
 *      the estimated extinction coefficients.
 *
 *  INPUTS
 *
 *  RESULTS
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *      fieldfpos - The list of the field position numbers
 *      fixedcfs  - The list of extinction coefficients
 *      nfixed    - The number of fixed field entries
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E. R. Deul - dd 03-03-1998
 *
 ******************************************************************************/
{
    catstruct	*f_cat;
    tabstruct	*f_tab;
    keystruct	**keys, *key;
    char	**keynames;
    int		i, j, FKEYS = 3;
/*
 *  Loading of the FIXED table information (if it has been given)
 */
    if ((f_cat = read_cat(control.fixedfile)) == NULL) {
       syserr(FATAL,"read_fixed", "Could not read catalog %s\n", 
          control.fixedfile);
    }
    if ((f_tab = name_to_tab(f_cat, "FIXED",0))) {
       ED_CALLOC(keynames, "read_fixed", char *,      FKEYS);
       ED_CALLOC(keys,     "read_fixed", keystruct *, FKEYS);
       for (i=0;i<FKEYS;i++) {
          ED_CALLOC(keynames[i], "read_fixed", char,      MAXCHAR);
       }
       if ((fposkey = name_to_key(f_tab, control.fpos))) 
       {
         strcpy(keynames[0], control.fpos);
       }
       else
       {
         strcpy(keynames[0], control.fitsfile);
       }
       strcpy(keynames[1], control.coef);
       strcpy(keynames[2], control.zp);
       read_keys(f_tab, keynames, keys, FKEYS, NULL);
       for (i=0;i<FKEYS;i++) {
          if (!keys[i])
            syserr(FATAL,"read_fixed", "No key %s in table FIXED\n",
              keynames[i]);
       }
       nfixed = keys[0]->nobj;
       if (fposkey)
         fieldfpos = keys[0]->ptr;
       else {
         fitsfile = keys[0]->ptr;
         ED_CALLOC(fieldfpos, "read_fixed", short int, nfixed);
         if (!(key=name_to_key(name_to_tab(cat,"FIELDS",0),control.fitsfile)))
            syserr(FATAL, "read_fixed", "No key %s in FIXED table\n",
                          control.fitsfile);
         for (i=0;i<nfixed;i++) {
           fieldfpos[i] = -1;
           for (j=0;j<key->nobj;j++)
	   {
               if (!strncmp(&fitsfile[i*keys[0]->nbytes],
                            &filenames[j*key->nbytes], 
                            MIN(keys[0]->nbytes,key->nbytes)))
	       {
                 fieldfpos[i] = j;
	       }
	       
	   }
           if (fieldfpos[i] == -1)
             syserr(FATAL, "read_fixed", 
                    "No %s with value %s in FIELDS table of input catalog\n", 
                    control.fitsfile, fitsfile[i*keys[0]->nbytes]);
           }
       }
       fixedcfs = keys[1]->ptr;
       fixedzps = keys[2]->ptr;
       for (i=0;i<FKEYS;i++) ED_FREE(keynames[i]);
       ED_FREE(keynames);
    } else {
       nfixed = 0;
    }
    nref = control.nref = 0;
}
/**/
int check_fixed(int *ix, int *iy, int j, int i, float *x, float *sx, 
                pair_struct *pairs)
/****** check_fixed ***************************************************
 *
 *  NAME	
 *      check_fixed - Check if the current field is a fixed one
 *
 *  SYNOPSIS  
 *      int check_fixed(int *ix, int *iy, int i, float *x, float *sx,
 *                      pair_struct *pairs)
 *
 *  FUNCTION
 *      This function checks whether the current field is present in the
 *      list of FIXED fields. If so it saves this information to the
 *      arguments and notifies the caller.
 *
 *  INPUTS
 *      ix       - List of row numbers of frame
 *      iy       - List of column number of frame
 *      j        - Index of current field in (ix,iy)
 *      i        - Field number of current field in pairs
 *      pairs    - Pointer to the primary pairing information
 *
 *  RESULTS
 *      x        - List of indices
 *      sx       - List of zp values
 *
 *  RETURNS
 *      FALSE if current field is not in FIXED list
 *      TRUE  if current field is in FIXED list
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 03-03-1998
 *
 ******************************************************************************/
{
    int		retcode = FALSE;
    int		fix, fiy;

    if ((fix = in_list(ix[j], fieldfpos, nfixed)) != -1 &&
        (fiy = in_list(iy[j], fieldfpos, nfixed)) != -1 &&
        fixedzps[fix] != 0 && fixedzps[fiy] != 0) {
/*
 * Substitute an overlap instead of adding a new one
 *
       x[j+1]  = - fixedcfs[fiy] * airmass[fiy] +
                 fixedcfs[fix] * airmass[fix];
       sx[j+1] = sx[j]/control.fixed_fact;
       ix[j+1] = pairs->iind[i];
       iy[j+1] = pairs->jind[i];
       VPRINTF(VERBOSE,"Fix: %f %f %d %d\n",
                          x[j+1], sx[j+1], ix[j+1], iy[j+1]);
 */
       x[j]  = fixedcfs[fiy] * airmass[fiy] -
               fixedcfs[fix] * airmass[fix];
       sx[j] = sx[j]/control.fixed_fact;
       ix[j] = pairs->iind[i];
       iy[j] = pairs->jind[i];
       VPRINTF(VERBOSE,"Fix: %f %f %d %d\n",
                          x[j], sx[j], ix[j], iy[j]);
       retcode = TRUE;
    }
    return retcode;
}
/**/
int read_fields(catstruct *i_cat)
/****** read_fields ***************************************************
 *
 *  NAME	
 *      read_fields - Read the position/airmass FIELD information
 *
 *  SYNOPSIS
 *      void read_fields(catstruct *i_cat)
 *
 *  FUNCTION
 *      This function reads the row/col numbers of the individual fields
 *      in the area, reads the airmass as recorded in the FIELDS table
 *      and stores that information in a number of global variables
 *
 *  INPUTS
 *      i_cat   - Pointer to the input catalog
 *
 *  RESULTS
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *      rows    - The list of row numbers by field
 *      cols    - The list of column numbers by field
 *      airmass - Thelist of airmasses by field
 *      nfields - The number of fields
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 03-03-1998
 *
 ******************************************************************************/
{
    tabstruct	*i_tab;
    keystruct	**keys;
    char	**keynames;
    int		i, MKEYS = 4;
/*
 *  Loading of catalog FIELDS data
 */
    VPRINTF(NORMAL,"Reading catalog %s\n", control.listfile);
    if (!(i_tab = name_to_tab(i_cat, "FIELDS", 0))) 
       syserr(FATAL,"read_fields", 
          "No FIELDS table in catalog %s\n", control.listfile);
    ED_CALLOC(keynames, "read_fields", char *,      MKEYS);
    ED_CALLOC(keys,     "read_fields", keystruct *, MKEYS);
    for (i=0;i<MKEYS;i++) {
       ED_CALLOC(keynames[i], "read_fields", char,      MAXCHAR);
    }
    if (!name_to_key(i_tab, control.row))
      strcpy(keynames[0], control.crval2);
    else
      strcpy(keynames[0], control.row);
    if (!name_to_key(i_tab, control.col))
      strcpy(keynames[1], control.crval1);
    else
      strcpy(keynames[1], control.col);
    strcpy(keynames[2], control.airmass);
    strcpy(keynames[3], control.fitsfile);
    read_keys(i_tab, keynames, keys, MKEYS, NULL);
    for (i=0;i<MKEYS;i++) {
       if (!keys[i])
          syserr(FATAL,"read_fields", "No key %s in table FIELDS\n",
           keynames[i]);
    }
    nfields   = keys[0]->nobj;
    if (!name_to_key(i_tab, control.row)) {
      ED_CALLOC(rows, "read_fields", LONG, nfields);
      ra        = keys[0]->ptr;
    } else
      rows      = keys[0]->ptr;
    if (!name_to_key(i_tab, control.col)) {
      ED_CALLOC(cols, "read_fields", LONG, nfields);
      dec       = keys[1]->ptr;
    } else
      cols      = keys[1]->ptr;
    airmass   = keys[2]->ptr;
    filenames = keys[3]->ptr;

    return nfields;
}
/**/
void derive_abs_phot(float *p, float *params, 
                     float *framephot, float *corrzp, int npol)
{
/****** derive_abs_phot ***************************************************
 *
 *  NAME	
 *      derive_abs_phot - Derive absolute photemtric zeropoint
 *
 *  SYNOPSIS
 *      derive_abs_phot(float *p, float *params, int npol)
 *
 *  FUNCTION
 *      This routine derives a best fitting polynomial surface to the
 *      relative zeropoint using the list of fixed frames and their
 *      true zeropoints.
 *
 *  INPUTS
 *      p      - The polynomail terms of the fit
 *      params - The original relative zeropoints
 *      npol   - Number of polynomial parameters
 *
 *  RESULTS
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 06-01-1998
 *
 ******************************************************************************/
    int		i,k;
    /* int       j; */
    float	*x, *y, *s, **u, **v, *w, chisq;
/*
 *  Derive number of free parameters
 */
    x = vector(1, nref);
    y = vector(1, nref);
    s = vector(1, nref);
    for (i=1;i<=nref;i++) 
    {
      /*
       for (j=0,k=-1;j<nfields;j++) 
           if (rows[j] == refypos[i-1] && cols[j] == refxpos[i-1]) 
              k = j;
       if (k == -1) 
          syserr(FATAL, "derive_abs_phot", 
             "No field definition at (%d,%d) in FIELDS table of input catalog",
             refxpos[i-1], refypos[i-1]);
      */

      /* the fieldfpos starts with '1', the index of params with '0', hence
         we have to query params[k-1] !!!!! */
       k = fieldfpos[i-1];
       x[i] = (float) (i - 1);
       y[i] = refzps[i - 1] - params[k-1];
       s[i] = 0.00002;
       framephot[k-1] = 1.;
       corrzp[k-1] = y[i];
    }
    u = matrix(1, nref, 1, npol);
    v = matrix(1, npol, 1, npol);
    w = vector(1, npol);

    svdfit(x, y, s, nref, p, npol, u, v, w, &chisq, polfit);

    free_vector(w, 1);
    free_matrix(v, 1, npol, 1);
    free_matrix(u, 1, nref, 1);
    free_vector(s, 1);
    free_vector(y, 1);
    free_vector(x, 1);
}
/**/
void polfit(float x, float *p, int np)
{
    int		k[] = {0, 0, 0, 1, 1}, l[] = {0, 0, 1, 0, 1}, i;

    
    for (i=1;i<=np;i++)
    {
          p[i] = 1.0 * pow((double)refxpos[(int)x],(double)k[i]) *
                       pow((double)refypos[(int)x],(double)l[i]);
    }
      
}
/**/
float polval(float x, float *p, int np)
{
    int		k[] = {0, 0, 0, 1, 1}, l[] = {0, 0, 1, 0, 1}, i;
    float	val = 0.0;

    val = 0.0;
    for (i=1;i<=np;i++) 
       val += p[i] * pow((double)cols[(int)x],(double)k[i]) *
                     pow((double)rows[(int)x],(double)l[i]);
    return val;
}
/**/
void uniq_frames(pair_struct *pairs)
/****** uniq_frames ***************************************************
 *
 *  NAME	
 *      uniq_frames - Calculate the number of uniq frames in pairs
 *
 *  SYNOPSIS
 *      void uniq_frames(pair_struct *pairs)
 *
 *  FUNCTION
 *      This routine derives the number of unique frames in the pairs
 *      structure. It updates the nc and nover values in the pairs
 *      structure together with the indexing information.
 *
 *  INPUTS
 *      pairs   - A pointer to the pairing information
 *
 *  RESULTS
 *      pairs   - the nc, nover and indexing elements are updated
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 03-03-1998
 *
 ******************************************************************************/
{
    int         i, j, k;
    int         maxif = 0, maxit = 0;
    int         nc;

    VPRINTF(NORMAL, "Deriving total number of frames\n");
    for (i=0;i<pairs->nobjs;i++) {
        maxif = MAX(maxif, pairs->from_f[i]);
        maxit = MAX(maxit, pairs->to_f[i]);
    }
    nc = MAX(maxif, maxit);
    pairs->nc = nc;
/*
 *  Make the index maps to relate the overlaps
 */
    ED_CALLOC(pairs->indx, "read_pairing", int, (nc*nc));
    ED_CALLOC(pairs->iind, "read_pairing", int, (nc*nc));
    ED_CALLOC(pairs->jind, "read_pairing", int, (nc*nc));
    k = 1;
    for (i=0;i<nc;i++) {
       for (j=i+1;j<nc;j++) {
          pairs->indx[j+i*nc] = pairs->indx[i+j*nc] = k;
          pairs->iind[k] = i+1;
          pairs->jind[k] = j+1;
          k++;
       }
    }
    pairs->nover= k;
}
/**/
int make_reference_zps(pair_struct *refs, int nrefcats, int kkk)
{
    int		i, ii, j;
    int		nc, ndiag, ndata, nmaxref, nref_block, from_chan=FALSE;
    double      *sum, *ave, temp_f, temp_t;
    LONG        *isum;

    nref = 0;
    nmaxref = 0;
    nref_block = 20;
    for (j=0;j<nrefcats;j++) {
       ndiag  = refs[j].nover;
       nc     = refs[j].nc;
       sum    = dvector(1, ndiag);
       ave    = dvector(1, ndiag);
       isum   = lvector(1, ndiag);

       for (i=1;i<=ndiag;i++) {
          sum[i] = 0;
          isum[i] = 0;
       }
       ndata = refs[j].nobjs;
/*
 *     The local from is always the observed data
 *     The local to   is always the reference data
 */
       for (i=0;i<ndata;i++) {
          if (refs[j].from_chan[i] == 0) {
             from_chan = TRUE;
             temp_f = refs[j].from_m[kkk][i];
             temp_t = refs[j].to_m[kkk][i];
             if (refs[j].from_f[i]>refs[j].to_f[i]) 
               ii = refs[j].indx[refs[j].from_f[i]-1 + (refs[j].to_f[i]-1)*nc];
             else
               ii = refs[j].indx[refs[j].to_f[i]-1 + (refs[j].from_f[i]-1)*nc];
          } else {
             from_chan = FALSE;
             temp_f = refs[j].to_m[kkk][i];
             temp_t = refs[j].from_m[kkk][i];
             if (refs[j].from_f[i]>refs[j].to_f[i]) 
               ii = refs[j].indx[refs[j].from_f[i]-1 + (refs[j].to_f[i]-1)*nc];
             else
               ii = refs[j].indx[refs[j].to_f[i]-1 + (refs[j].from_f[i]-1)*nc];
          }
          if ((temp_f - control.maxmag) > 0.0 ||
              (control.minmag - temp_f) > 0.0) continue;
          if (control.use_flag == YES && 
              ((refs[j].from_flag[i]&control.flag_mask) ||
               (refs[j].to_flag[i]&control.flag_mask))) continue;
          sum[ii+1] += (temp_t - temp_f);
          isum[ii+1]++;
       }
       for (i=1;i<=ndiag;i++) {
          if (isum[i] != 0.0) {
             ave[i] = sum[i]/(double)isum[i];
          }
          sum[i] = 0;
          isum[i] = 0;
       }
       for (i=0;i<ndata;i++) {
          if (refs[j].from_chan[i] == 0) {
             temp_f = refs[j].from_m[kkk][i];
             temp_t = refs[j].to_m[kkk][i];
             if (refs[j].from_f[i]>refs[j].to_f[i]) 
               ii = refs[j].indx[refs[j].from_f[i]-1 + (refs[j].to_f[i]-1)*nc];
             else
               ii = refs[j].indx[refs[j].to_f[i]-1 + (refs[j].from_f[i]-1)*nc];
          } else {
             temp_f = refs[j].to_m[kkk][i];
             temp_t = refs[j].from_m[kkk][i];
             if (refs[j].from_f[i]>refs[j].to_f[i]) 
               ii = refs[j].indx[refs[j].from_f[i]-1 + (refs[j].to_f[i]-1)*nc];
             else
               ii = refs[j].indx[refs[j].to_f[i]-1 + (refs[j].from_f[i]-1)*nc];
          }
          if ((temp_f - control.maxmag) > 0.0 ||
              (control.minmag - temp_f) > 0.0 ||
              fabs((temp_t - temp_f)-ave[ii]) > control.magdif) continue;
          if (control.use_flag == YES && 
              ((refs[j].from_flag[i]&control.flag_mask) ||
               (refs[j].to_flag[i]&control.flag_mask))) continue;
          sum[ii+1] += (temp_t - temp_f);
          isum[ii+1]++;
       }
/*
 *     Append the derived zeropoints to the list
 */
       for (i=1;i<=ndiag;i++) {
         if (isum[i] != 0.0) {
            VPRINTF(VERBOSE,"Aver: %f, (%d)\n",sum[i]/isum[i], i);
            if (nref == 0) {
               nmaxref += nref_block;
               ED_CALLOC(refzps,  "make_reference_zps", float, nmaxref);
               ED_CALLOC(refxpos, "make_reference_zps", LONG,   nmaxref);
               ED_CALLOC(refypos, "make_reference_zps", LONG,   nmaxref);
            }
            if (nref == nmaxref) {
               nmaxref += nref_block;
               ED_REALLOC(refzps,  "make_reference_zps", float, nmaxref);
               ED_REALLOC(refxpos, "make_reference_zps", LONG,   nmaxref);
               ED_REALLOC(refypos, "make_reference_zps", LONG,   nmaxref);
            }
            refzps[nref]  = sum[i]/isum[i];
            if (from_chan) {
               refxpos[nref] = cols[refs[j].jind[i]-1];
               refypos[nref] = rows[refs[j].jind[i]-1];
            } else {
               refxpos[nref] = cols[refs[j].iind[i]-1];
               refypos[nref] = rows[refs[j].iind[i]-1];
            }
            VPRINTF(VERBOSE,"ZP(%d,%d) = %f\n", refxpos[nref], refypos[nref],
                           refzps[nref]);
            nref++;
         }
       }
       free_dvector(sum, 1);
       free_dvector(ave, 1);
       free_lvector(isum, 1);
    }
/*
 *  Append the fixed points to the list
 */
    for (i=0;i<nfixed;i++) {
       if (nref == 0) {
          nmaxref += nref_block;
          ED_CALLOC(refzps,  "make_reference_zps", float, nmaxref);
          ED_CALLOC(refxpos, "make_reference_zps", LONG,   nmaxref);
          ED_CALLOC(refypos, "make_reference_zps", LONG,   nmaxref);
       }
       if (nref == nmaxref) {
          nmaxref += nref_block;
          ED_REALLOC(refzps,  "make_reference_zps", float, nmaxref);
          ED_REALLOC(refxpos, "make_reference_zps", LONG,   nmaxref);
          ED_REALLOC(refypos, "make_reference_zps", LONG,   nmaxref);
       }
       if (fixedzps[i] != 0.0) {
          refxpos[nref] = cols[fieldfpos[i]-1];
          refypos[nref] = rows[fieldfpos[i]-1];
          refzps[nref]  = fixedzps[i] - airmass[fieldfpos[i]-1] *
                          fixedcfs[i];

          nref++;
       }
    }

    return nref;
}


