#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "stdcat_globals.h"
#include "stdcat_types.h"
#include "fitscat.h"
#include "lists_globals.h"
#include "utils_globals.h"
#include "sort_globals.h"
#include "options_globals.h"
#include "utils_macros.h"
#include "precess_globals.h"
#include "bin_search_globals.h"

/*
   27.03.2006:
   I renamed sorting and indexing functions to the new naming scheme
   described in the sort_globals.h file
*/

#define QUAL_LEN      1
#define CATS_LEN      3
#define Q_LEN         3
#define SPECTRUM_LEN 11
#define CROSS_ID_LEN 22
#define IREF_LEN      3
#define JREF_LEN      3
#define KREF_LEN      3

double	*ra, *dec;
char	**std_names;
float 	*epoch, *Imag, *Jmag, *Kmag;
int	n_stds = 0, *idx;
long int	*SeqNr;

extern control_struct control;

void read_stdcat(char *filename, char **stdnames, int nstd)
{
    int		i, n, NAMES_LEN;
    char	*Names;
    double	raa, decc;

    catstruct	*cat;
    tabstruct	*tab;
    keystruct	*key;

    if (nstd != 3) 
       syserr(FATAL,"read_stdcat", "Currently only three STDMAGs allowed\n");
    if ((cat = read_cat(filename)) == NULL) {
       syserr(FATAL, "read_stdcat", "Could not read catalog: %s\n", filename);
    }
    tab = name_to_tab(cat, "STDTAB", 0);
    ED_GETKEY(tab, key, "Name",  Names, n, "read_stdcat");
    NAMES_LEN = key->naxisn[0];
    ED_MALLOC(std_names, "read_stdcat", char *, n);
    for (i=0;i<n;i++) {
       ED_MALLOC(std_names[i], "read_stdcat", char, NAMES_LEN+1);
       strncpy(std_names[i], &Names[i*NAMES_LEN], NAMES_LEN);
    }
    ED_GETKEY(tab, key, "SeqNr", SeqNr, n, "read_stdcat");
    ED_GETKEY(tab, key, "Ra",    ra,    n, "read_stdcat");
    ED_GETKEY(tab, key, "Dec",   dec,   n, "read_stdcat");
    if (stdnames) {
       ED_GETKEY(tab, key, stdnames[0], Imag,  n, "read_stdcat");
       ED_GETKEY(tab, key, stdnames[1], Jmag,  n, "read_stdcat");
       ED_GETKEY(tab, key, stdnames[2], Kmag,  n, "read_stdcat");
    }
    ED_GETKEY(tab, key, "Epoch", epoch, n, "read_stdcat");
    for (i=0;i<n;i++) {
        raa = ra[i]; decc = dec[i];
        precess(raa, decc, epoch[i], &ra[i], &dec[i], 2000.0);
    }
    ED_MALLOC(idx, "read_stdcat", int, n);
    dindex(n, ra, idx);
    n_stds = n;
    close_cat(cat);
}

int select_std_star(double ref_ra, double ref_dec, float tol, 
                    float *I, float *J, float *K, char *name, int *seqnr)
{
    int i, j, ibeg, iend, retcode = FALSE;
    float least_dist = 99999.99, curr_dist;

    if (n_stds == 0)
       syserr(FATAL, "Cannot select star, because standard star catalog has not been opened yet\n");
    tol = tol / 3600.0;
    dbin_search_indexed(n_stds, ra, idx, (float) ref_ra, (float) tol, 
                        &ibeg, &iend, 0);
    j = -1;
    for (i=ibeg;i<=iend;i++) {
       VPRINTF(DEBUG,"Ref: %lf %lf Check: %lf %lf %f\n",ref_ra, ref_dec, 
                          ra[idx[i]],dec[idx[i]],tol);
       if (fabs(ref_ra  - ra[idx[i]])  < tol && 
           fabs(ref_dec - dec[idx[i]]) < tol) {
          retcode = TRUE;
          curr_dist = ref_ra  - ra[idx[i]] * ref_ra  - ra[idx[i]] +
                      ref_dec - dec[idx[i]]* ref_dec - dec[idx[i]];
          if (curr_dist < least_dist) {
             j = idx[i];
             least_dist = curr_dist;
          }
       }
    }
    if (j != -1) {
       *I      = Imag[j];
       *J      = Jmag[j];
       *K      = Kmag[j]; 
       strcpy(name, std_names[j]); 
       *seqnr  = SeqNr[j];
    }
    return retcode;
}

float epoch_std_star(char *name)
{
    int		i, j, l;
    char	*t, *s;

    if (n_stds == 0)
       syserr(FATAL, "Cannot select star, because standard star catalog has not been opened yet\n");
    ED_MALLOC(t, "epoch_std_star", char, MAXCHAR);
    ED_MALLOC(s, "epoch_std_star", char, MAXCHAR);
    l = strlen(name);
    for (i=0;i<l;i++) {
       if (name[i] != '_' && name[i] != ' ') break;
    }
    for (j=l;j<0;j--) {
       if (name[j] != '_') break;
    }
    strncpy(t, &name[i], j-i+1);
    sprintf(s,"*%s*",t);
    for (i=0;i<n_stds;i++) {
       if (match(std_names[i],s)) {
          break;
       }
    }
    if (i == n_stds)
       syserr(FATAL,"epoch_std_star", "No standard star named %s found in catalog\n", t);
    ED_FREE(t);
    ED_FREE(s);

    return epoch[i];
}


int get_std_star(int seqnr, float *I, float *J, float *K, char *name)
{
    int		i;

    for (i=0;i<n_stds;i++) {
       if (SeqNr[i] == seqnr) {
          *I = Imag[i];
          *J = Jmag[i];
          *K = Kmag[i];
          strcpy(name, std_names[i]);
          return TRUE;
       }
    }
    *I = *J = *K = -99.0;
    return FALSE;
}


void close_stdcat()
{
    if (n_stds == 0)
       syserr(FATAL, "Cannot close unopened standard star catalog\n");
    ED_FREE(idx);
    ED_FREE(SeqNr);
    ED_FREE(ra);
    ED_FREE(dec);
    ED_FREE(Imag);
    ED_FREE(Jmag);
    ED_FREE(Kmag);
    n_stds = 0;
}
