/* History:
   12.01.2000:
   In the main program two ED_CALLOC's are
   only made if control.rlist > 0

   21.06.01:
   removed compiler warnings under DEC

   26.10.01:
   introduced minobj to the control structure.
   It gives the minimum number of objects that have
   to be in common in two overlapping frames so that
   this overlap is considered in the relphotom
   process

   05.05.03:
   included the tsize.h file

   10.01.05:
   removed compiler warnings under gcc with optimisation
   (-O3) enabled.

   16.01.2005:
   replaced occurenced of 'long int' by LONG

   29.01.2005:
   - I added history information to the output catalog
   - I added argc and argv to the arguments of function
     init_relphotom to write HISTORY info to output catalogs.

   13.03.2005:
   Changes to avoid compiler warnings with -W -Wall -pedantic

   11.04.2005:
   I increased the maximum number of the MINOBJECTS config. variable
   from 10 to 100.
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "fitscat.h"
#include "tsize.h"
#include "solve.h"
#include "options_globals.h"
#include "utils_globals.h"
#include "lists_globals.h"
#include "relphotom_globals.h"
#include "relphotom_vers.h"
#include "fixed_phot_globals.h"
#include "derive_relzps_globals.h"
#include "relphotom_tab_globals.h"

char    header[MAXCHAR];

static int	nmags;
static char	**mag_names, **magerr_names;

control_struct control;

option my_opts[] = {
            {"AIRMASS", TEXT, &control.airmass, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "2MUAIRM",
             {"", "", "", "", ""}, 0,
            "The name of the FIELDS table column containing the airmass value"},
            {"ROW", TEXT, &control.row, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "EISROW",
             {"", "", "", "", ""},0,
            "The name of the FIELDS table column containing the row number"},
            {"COL", TEXT, &control.col, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "EISCOL",
             {"", "", "", "", ""}, 0,
            "The name of the FIELDS table column containing the column number"},
            {"CRVAL1", TEXT, &control.crval1, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "CRVAL1",
             {"", "", "", "", ""}, 0,
            "The name of the FIELDS table column containing the center RA"},
            {"CRVAL2", TEXT, &control.crval2, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "CRVAL2",
             {"", "", "", "", ""}, 0,
            "The name of the FIELDS table column containing the center DEC"},
            {"XPOS", TEXT, &control.xpos, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "Xpos",
             {"", "", "", "", ""},0,
            "The name of the OBJECTS table column containing the object x position"},
            {"YPOS", TEXT, &control.ypos, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "Ypos",
             {"", "", "", "", ""}, 0, 
            "The name of the OBJECTS table column containing the object y position"},
            {"FIELD_POS", TEXT, &control.fpos, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "FIELD_POS",
             {"", "", "", "", ""}, 0,
            "The name of the OBJECTS table column containing the object field position"},
            {"ZP", TEXT, &control.zp, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "ZP",
             {"", "", "", "", ""}, 0,
            "The name of the RELPHOTOM table column containing the field zero point"},
            {"COEF", TEXT, &control.coef, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "COEF",
             {"", "", "", "", ""}, 0, 
            "The name of the RELPHOTOM table column containing the field extinction coefficient"},
            {"FIXED_FACT", FLOAT, &control.fixed_fact, 0, 0, 0.1, 10.0,
              0, 2.0, NO, "",
             {"", "", "", "", ""}, 0,
            ""},
            {"MINMAG", FLOAT, &control.minmag, 0, 0, -30.0, 30.0,
              0, 20.0, NO, "",
             {"", "", "", "", ""}, 0,
            "The minimum object magnitude value allowed in the determination"},
            {"MAXMAG", FLOAT, &control.maxmag, 0, 0, -30.0, 30.0,
              0, 20.0, NO, "",
             {"", "", "", "", ""}, 0, 
            "The maximum object magnitude value allowed in the determination"},
            {"MAGDIFF", FLOAT, &control.magdif, 0, 0, 0.0, 5.0,
              0, 0.5, NO, "",
             {"", "", "", "", ""}, 0,
            "The maximum allowed magnitude difference for pairs"},
            {"MINOBJECTS", INT, &control.minobj, 0, 100, 0.0, 0.0,
              5, 0.0, NO, "",
             {"", "", "", "", ""}, 0,
            "The number of iterations per magnitude estimator"},
            {"NITER", INT, &control.niter, 0, 10, 0.0, 0.0,
              3, 0.0, NO, "",
             {"", "", "", "", ""}, 0,
            "The number of iterations per magnitude estimator"},
            {"MAG", TEXTVEC, &control.mags, 0, 10, 0.0, 0.0,
              0, 0.0, NO, "Mag",
             {"", "", "", "", ""}, 0,
            "The name of the PAIRS table vector for magnitude"},
            {"MAGERR", TEXTVEC, &control.magerrs, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "MagErr",
             {"", "", "", "", ""}, 0,
            "The name of the PAIRS table vector for magnitude error"},
            {"FIELD", TEXT, &control.field, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "Field",
             {"", "", "", "", ""}, 0,
            "The name of the PAIRS table vector to store field numbers"},
            {"FLAG", TEXT, &control.flag, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "Flag",
             {"", "", "", "", ""}, 0,
            "The name of the OBJECTS table column containing the extraction flag"},
            {"FLAG_MASK", INT, &control.flag_mask, 0, 64768, 0.0, 0.0,
              249, 0.0, NO, "",
             {"", "", "", "", ""}, 0,
            "The bitwise mask use for selecting FLAGged objects"},
            {"IMAFLAGS", TEXT, &control.imaflags, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "IMAFLAGS_ISO",
             {"", "", "", "", ""}, 0,
            "The name of the OBJECTS table column containing the IMAFLAGS_ISO flag"},
            {"IMAFLAG_MASK", INT, &control.imaflags_mask, 0, 64768, 0.0, 0.0,
              255, 0.0, NO, "",
             {"", "", "", "", ""}, 0,
            "The bitwise mask use for selecting FLAGged objects"},
            {"USE_FLAG", BOOLEAN, &control.use_flag, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "NO",
             {"YES", "NO", "", "", ""}, 0,
            "The boolean to identify selection of unflagged objects"},
            {"NPOL_PARAM", INT, &control.npol_param, 1, 4, 0.0, 0.0,
              3, 0.0, NO, "",
             {"", "", "", "", ""}, 0,
            "Polynomial terms of the fit. NPOL_PARAM = 1 for DC only, NPOL_PARAM = 2 for X slope + DC, NPOL_PARAM = 3 for Y slope + DC, NPOL_PARAM = 4 for X and Y slope + DC"},
            {"FITSFILE", TEXT, &control.fitsfile, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "FITSFILE",
             {"", "", "", "", ""}, 0,
            "The name of the FIELDS table column containing the fits filename"},
            {"METHOD", STRING, &control.method, 0, 0, 0.0, 0.0,
              0, 0.0, NO, "BICONGRAD",
             {"BICONGRAD", "CHOLESKI", "SVD", "", ""}, 0,
            "The name of the method to use to solve the linear equations"},
            {"", BOOLEAN, NULL, 0, 0, 0.0, 0.0, 0, 0.0, NO, "",
             {"YES", "NO", "", "", ""}, 0,
              "dummy key to end structure my_opts"}
             };
/**/
void usage(char *me, char *head)
{
    fprintf(stderr,"%s \n\n",head);
    fprintf(stderr,"Usage: %s \n",me);
    fprintf(stderr,"		-i incat 		(input catalog name)\n");
    fprintf(stderr,"		-o outcat		(output catalog name)\n");
    fprintf(stderr,"		-p paircat		(pairs catalog name)\n");
    fprintf(stderr,"		[-r refcat1 [refcat2 [refcat3 ...]]]\n");
    fprintf(stderr,"					(sequence of reference pair catalogs)\n");
    fprintf(stderr,"		[-f fixedfile]		(name of fixed table catalog)\n");
    fprintf(stderr,"		[-c conffile]\n");
    fprintf(stderr,"		[-OPTION1 OPTVAL1 [-OPTION OPTVAL2] ...]\n\n");
    exit(1);
}
/**/
void init(int argc, char *argv[], char *head)
{
    int i, j, k;
 
    if (argc == 1) usage(argv[0], head);
    init_options(0, my_opts);
    control.fixed = FALSE;
    strcpy(control.conffile,"");
    for (i=1; i<argc; i++) {
       if (argv[i][0] == '-') {
          switch(argv[i][1]) {
          case 'c': strcpy(control.conffile,argv[++i]);
                    break;
          case 'f': strcpy(control.fixedfile,argv[++i]);
                    control.fixed = TRUE;
                    break;
          case 'i': strcpy(control.listfile,argv[++i]);
                    control.ilist = 1;
                    break;
          case 'o': strcpy(control.outfile,argv[++i]);
                    break;
          case 'p': strcpy(control.pairfile,argv[++i]);
                    break;
          case 'r': control.rlist = 0;
                    strcpy(control.reffile[control.rlist],argv[++i]);
                    while (i+1<argc && argv[i+1][0] != '-') {
                       strcpy(control.reffile[++control.rlist],argv[++i]);
                    }
                    ++control.rlist;
                    break;
          case '@': print_options(argv[0]);
                    exit(0);
          default:  parse_options(control.conffile);
                    j = i++; k = i;
                    if (argv[j] && argv[k])
                      set_command_option(argv[j], argv[k]);
                    else 
                      usage(argv[0], head);
                    break;
          }
       } else {
          syserr(WARNING,"init","Command line syntax error\n");
          usage(argv[0], head);
       }
   }
   if (control.ilist == 0)
      usage(argv[0], head);
   finish_options(control.conffile);
   VPRINTF(NORMAL,"\n   %s\n\n",head);
   mag_names    = copy_textvec(control.mags,    &nmags);
   magerr_names = copy_textvec(control.magerrs, &i);
   if (i != nmags)
      syserr(FATAL, "init", "Number if %s parameters differs from %s\n",
              control.magerrs, control.mags);
}
/**/
int read_pairing(catstruct *p_cat, char *name)
/****** read_pairing ***************************************************
 *
 *  NAME	
 *      read_pairing - Read the primary pairing information
 *
 *  SYNOPSIS
 *      void read_pairing(catstruct *p_cat, char *name)
 *
 *  FUNCTION
 *      Read and interpret the pairing information available in the primary
 *      pairing table. Information retrieved are: The list of from/to
 *      field numbers and the associated flags. From the list of field
 *      numbers the total number of fields and the number of overlaps
 *      are derived.
 *
 *  INPUTS
 *      p_cat   - Pointer to the catalog of pairs
 *      name    - The name of the column containing the field numbers
 *
 *  RESULTS
 *      pairs   - Pointer to the structure recieving the output info.
 *
 *  RETURNS
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R Deul - dd 03-03-1998
 *
 ******************************************************************************/
{
    tabstruct	*tab;
    keystruct	**keys;
    char	**keynames;
    pair_struct	pairs;
    int		j, len, NKEYS = 1;
/*
 *  Read in the relavent pairing information
 */
    VPRINTF(NORMAL,"Reading catalog %s\n", control.pairfile);
    if (!(tab = name_to_tab(p_cat, "PAIRS", 0))) 
       syserr(FATAL,"read_pairing", 
          "No PAIRS table in catalog %s\n", control.pairfile);
    ED_CALLOC(keynames, "read_pairing", char *,      NKEYS);
    ED_CALLOC(keys,     "read_pairing", keystruct *, NKEYS);
    for (j=0;j<NKEYS;j++) {
       ED_CALLOC(keynames[j], "read_pairing", char,      MAXCHAR);
    }
    strcpy(keynames[0], name);
    read_keys(tab, keynames, keys, NKEYS, NULL);
    for (j=0;j<NKEYS-1;j++) {
       if (!keys[j])
          syserr(FATAL,"read_pairing", "No key %s in table PAIRS\n",
              keynames[j]);
    }
    if (control.use_flag == YES && ! keys[NKEYS-1])
       syserr(FATAL,"read_pairing", "No key %s in table PAIRS\n",
              keynames[NKEYS-1]);
    len = strlen(keynames[0]);
    keynames[0][len+1] = 1;
    ED_GETVEC(pairs.from_f, "read_pairing", keys[0], keynames[0], 
              int, short int);
    len = strlen(keynames[0]);
    keynames[0][len+1] = 2;
    ED_GETVEC(pairs.to_f  ,"read_pairing", keys[0], keynames[0], \
              int, short int);
    pairs.nobjs = keys[0]->nobj;
    for (j=0;j<NKEYS;j++) ED_FREE(keynames[j]);
    ED_FREE(keynames);
    ED_FREE(keys);

    uniq_frames(&pairs);

    ED_FREE(pairs.from_f);
    ED_FREE(pairs.to_f);
    ED_FREE(pairs.indx);
    ED_FREE(pairs.iind);
    ED_FREE(pairs.jind);
    return pairs.nc;
}
/**/
catstruct *init_relphotom(catstruct *p_cat, catstruct *i_cat, char *outname,
                          char *name, int *nfields, int argc, char **argv)
{
/****** init_relphotom ***************************************************
 *
 *  NAME	
 *      init_relphotom - Initialize the relphot table structure
 *
 *  SYNOPSIS
 *      catstruct *init_relphotom(catstruct *p_cat, catstruct *i_cat,
 *                                char *outname, char *name)
 *
 *  FUNCTION
 *      This routine initializes the RELPHOTOM table structure, opens
 *      an I/O port to write line by line the results to the catalog and
 *      does the required pre-work to allow initialization.
 *
 *  INPUTS
 *      p_cat     - A pointer to the pairs catalog data structure
 *      i_cat     - A pointer to the input catalog data structure
 *      outname   - Name of the output catalog
 *      name      - The name of the fields column in the PAIRS file
 *      nfields   - The number of fields in the input catalog
 *
 *  RESULTS
 *
 *  RETURNS
 *      A pointer to the output catalog
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 30-07-1997
 *
 ******************************************************************************/
    catstruct	*out_cat;
    int		nc;

    init_relphotom_tab();
    nc = read_pairing(p_cat, name);
    *nfields = read_fields(i_cat);
    if (control.fixed) read_fixed(i_cat);
/*
 *  Create the output catalog
 */
    out_cat = new_cat(1);
    inherit_cat(i_cat, out_cat);
    historyto_cat(out_cat, "relphotom", argc, argv);
    strcpy(out_cat->filename, outname);
    if (open_cat(out_cat, WRITE_ONLY) != RETURN_OK)
      syserr(FATAL, "init_relphotom","Cannot open output catalog %s\n",
             outname);
/*
 *  Save the primary FITS header
 */
    save_tab(out_cat, out_cat->tab);
    return out_cat;
}
/**/
int load_mags(catstruct *p_cat, pair_struct *pairs, char *name, char *fname,
                      char *imafname, char *mag, char *magerr, int *numags)
{
/****** load_mags ***************************************************
 *
 *  NAME	
 *      load_mags - Load the photometric magnitudes 
 *
 *  SYNOPSIS
 *      int load_mags(char *mag, char *magerr)
 *
 *  FUNCTION
 *      This routine loads the magnitude and its error into the globals
 *      structures
 *
 *  INPUTS
 *      p_cat   - A pointer to the pairs catalog
 *      pairs   - Pointer to the pairs data structure
 *      name    - The name of the field column
 *      fname   - The name of the flag column
 *      imafname- The name of the imaflags_iso column
 *      mag     - The name of the magnitude column to be read
 *      magerr  - The name of the magnitude error column to be read
 *
 *  RESULTS
 *      numags  - The number of individual elements in the mag vector
 *
 *  RETURNS
 *      The number of object read is returned
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 30-07-1997
 *
 ******************************************************************************/
    tabstruct	*tab;
    keystruct	**keys;
    char	**keynames;
    int		i, len, NKEYS = 6;
    float	*fptr;
    int		*iptr;
    short int	*sptr;

    VPRINTF(NORMAL,"Loading magnitude parameter %s\n", mag);
    ED_CALLOC(keynames, "relphotom", char *,      NKEYS);
    ED_CALLOC(keys,     "relphotom", keystruct *, NKEYS);
    for (i=0;i<NKEYS;i++) {
       ED_CALLOC(keynames[i], "relphotom", char,      MAXCHAR);
    }
    strcpy(keynames[0], mag);
    strcpy(keynames[1], magerr);
    strcpy(keynames[2], name);
    strcpy(keynames[3], fname);
    strcpy(keynames[4], imafname);
    strcpy(keynames[5], "CHANNEL_NR");
    if (!(tab = name_to_tab(p_cat, "PAIRS", 0))) 
       syserr(FATAL,"load_mags", "No PAIRS table pairs catalog\n");
    read_keys(tab, keynames, keys, NKEYS, NULL);
    for (i=0;i<NKEYS-1;i++) {
       if (!keys[i])
         syserr(FATAL,"load_photo_fields", "No key %s in table PAIRS\n",
           keynames[i]);
    }
    pairs->nobjs = keys[0]->nobj;
    *numags = 1;
/*
 *  The first dimension is the from-to (dimension 2). The next dimensions
 *  are from the data istself
 */
    for (i=1;i<keys[0]->naxis;i++) *numags *= keys[0]->naxisn[i];

    ED_CALLOC(pairs->from_m,  "load_strip_phots", float *, (*numags));
    ED_CALLOC(pairs->to_m,    "load_strip_phots", float *, (*numags));
    ED_CALLOC(pairs->from_e,  "load_strip_phots", float *, (*numags));
    ED_CALLOC(pairs->to_e,    "load_strip_phots", float *, (*numags));
    for (i=0;i<*numags;i++) {
       len = strlen(keynames[0]);
       keynames[0][len+1] = i+1;
       ED_GETVEC(fptr, "relphotom", keys[0], keynames[0], float, float);
       pairs->from_m[i] = fptr;
       len = strlen(keynames[0]);
       keynames[0][len+1] = i+1 + (*numags);
       ED_GETVEC(fptr, "relphotom", keys[0], keynames[0], float, float);
       pairs->to_m[i] = fptr;
       len = strlen(keynames[1]);
       keynames[1][len+1] = i+1;
       ED_GETVEC(fptr , "relphotom", keys[1], keynames[1], float, float);
       pairs->from_e[i] = fptr;
       len = strlen(keynames[1]);
       keynames[1][len+1] = i+1 + (*numags);
       ED_GETVEC(fptr, "relphotom", keys[1], keynames[1], float, float);
       pairs->to_e[i] = fptr;
    }
    len = strlen(keynames[2]);
    keynames[2][len+1] = 1;
    ED_GETVEC(iptr, "read_pairing", keys[2], keynames[2], int, short int);
    pairs->from_f = iptr;
    len = strlen(keynames[2]);
    keynames[2][len+1] = 2;
    ED_GETVEC(iptr ,"read_pairing", keys[2], keynames[2], int, short int);
    pairs->to_f = iptr;
    if (control.use_flag == YES) {
       len = strlen(keynames[3]);
       keynames[3][len+1] = 1;
       if (keys[3]->ttype == T_SHORT) {
          ED_GETVEC(iptr, "read_pairing", keys[3],
                    keynames[3], int, short int);
       } else {
          ED_GETVEC(iptr, "read_pairing", keys[3],
                    keynames[3], int, LONG);
       }
       pairs->from_flag = iptr;
       len = strlen(keynames[3]);
       keynames[3][len+1] = 2;
       if (keys[3]->ttype == T_SHORT) {
          ED_GETVEC(iptr, "read_pairing", keys[3],
                    keynames[3], int, short int);
       } else {
          ED_GETVEC(iptr, "read_pairing", keys[3],
                    keynames[3], int, LONG);
       }
       pairs->to_flag = iptr;
/*
 *     And now the imaflags
 */
       len = strlen(keynames[4]);
       keynames[4][len+1] = 1;
       if (keys[4]->ttype == T_SHORT) {
          ED_GETVEC(iptr, "read_pairing", keys[4],
                    keynames[4], int, short int);
       } else {
          ED_GETVEC(iptr, "read_pairing", keys[4],
                    keynames[4], int, LONG);
       }
       pairs->from_imaflag = iptr;
       len = strlen(keynames[4]);
       keynames[4][len+1] = 2;
       if (keys[4]->ttype == T_SHORT) {
          ED_GETVEC(iptr, "read_pairing", keys[4],
                    keynames[4], int, short int);
       } else {
          ED_GETVEC(iptr, "read_pairing", keys[4],
                    keynames[4], int, LONG);
       }
       pairs->to_imaflag = iptr;
    }
    if (keys[5]) {
       len = strlen(keynames[5]);
       keynames[5][len+1] = 1;
       ED_GETVEC(sptr, "read_pairing", keys[5], keynames[5], 
                 short int, short int);
       pairs->from_chan = sptr;
       len = strlen(keynames[5]);
       keynames[5][len+1] = 2;
       ED_GETVEC(sptr, "read_pairing", keys[5], keynames[5], 
                 short int, short int);
       pairs->to_chan = sptr;
    }
    for (i=0;i<NKEYS;i++) {
       if (keys[i])
          ED_FREE(keys[i]->ptr);
       ED_FREE(keynames[i]);
    }
    ED_FREE(keynames);
    ED_FREE(keys);
    uniq_frames(pairs);
    return pairs->nobjs;
}
/**/
void close_outcat(catstruct *i_cat, catstruct *out_cat, int nfields)
/****** close_outcat ***************************************************
 *
 *  NAME
 *      close_outcat - Close the output table of relphotom
 *
 *  SYNOPSIS
 *      void close_outcat(catstruct *i_cat, catstruct *out_cat)
 *
 *  FUNCTION
 *      This function finished the object by object writing and
 *      closes the output table.
 *
 *  INPUTS
 *      i_cat   - Pointer to the input catalog
 *      out_cat - Pointer to output catalog
 *      nfields - The number of fields in the input catalog
 *
 *  RESULTS
 *      None
 *
 *  RETURNS
 *      None
 *
 *
 *  MODIFIED GLOBALS
 *
 *  NOTES
 *
 *  BUGS
 *
 *  AUTHOR
 *      E.R. Deul - dd 30-07-1997
 *
 ******************************************************************************/
{
    catstruct	*f_cat;
    char	**tabnames;
    int		i, ntabs;

    save_relphotom_tab(out_cat, nfields);
    save_options(name_to_tab(out_cat, "RELPHOTOM", 0), my_opts, 0);
/*
 *  Copy the remaining input catalog tables to the output and save them
 */
    copy_tabs(i_cat, out_cat);
    tabnames = tabs_list(out_cat, &ntabs);
    for (i=1;i<ntabs;i++) {
       save_tab(out_cat, name_to_tab(out_cat, tabnames[i],0));
       ED_FREE(tabnames[i]);
    }
    ED_FREE(tabnames);
/*
 *  Copy the fixed catalog table(s) to the output
 */
    if (control.fixed) {
       if ((f_cat = read_cat(control.fixedfile)) == NULL) {
          syserr(FATAL,"relphotom", "Could not read catalog %s\n", 
             control.fixedfile);
       }
       copy_tabs(f_cat, out_cat);
       tabnames = tabs_list(f_cat, &ntabs);
       for (i=1;i<ntabs;i++) {
          save_tab(out_cat, name_to_tab(out_cat, tabnames[i],0));
          ED_FREE(tabnames[i]);
       }
       ED_FREE(tabnames);
    }
    close_cat(out_cat);
}
/**/
int main(int argc, char *argv[])
{
    catstruct	*p_cat, **r_cats=NULL, *in_cat, *out_cat;
    int		i, j, imags, idum, nfields;
    pair_struct	*pairs, *refs=NULL;
  
    sprintf(header,"%s Version: %s (%s)",NAME,VERS,DATE);
    init(argc, argv, header);

    VPRINTF(NORMAL,"Reading catalog input catalogs\n");
    if ((p_cat = read_cat(control.pairfile)) == NULL) {
       syserr(FATAL,"relphotom", "Could not read catalog %s\n", 
          control.pairfile);
    }
    ED_CALLOC(pairs, "relphotom", pair_struct, 1);
    if (!(in_cat = read_cat(control.listfile))) {
       syserr(FATAL,"relphotom", "Could not read catalog %s\n",
            control.listfile);
    }

    if(control.rlist > 0)
    {
      ED_CALLOC(r_cats,  "relphotom", catstruct *, control.rlist);
    }
    for (i=0;i<control.rlist;i++) 
    {
       if (!(r_cats[i] = read_cat(control.reffile[i]))) 
       {
          syserr(FATAL,"relphotom", "Could not read catalog %s\n",
               control.reffile[i]);
       }
    }

    if(control.rlist > 0)
    {
      ED_CALLOC(refs,    "relphotom", pair_struct, control.rlist);
    }

    out_cat = init_relphotom(p_cat, in_cat, control.outfile, control.field,
              &nfields, argc, argv);

    for (i=0;i<nmags;i++) {
       load_mags(p_cat, pairs, control.field, control.flag, control.imaflags, 
                 mag_names[i], magerr_names[i], &imags);
       for (j=0;j<control.rlist;j++) 
          load_mags(r_cats[j], &refs[j], control.field, control.flag,
                    control.imaflags, 
                    mag_names[i], magerr_names[i], &idum);
       derive_relzps(pairs, refs, control.rlist, mag_names[i], imags,
                     nfields);
    }


    ED_FREE(pairs->from_e);
    ED_FREE(pairs->to_e);
    ED_FREE(pairs->from_m);
    ED_FREE(pairs->to_m);
    ED_FREE(pairs->from_f);
    ED_FREE(pairs->to_f);
    ED_FREE(pairs->indx);
    ED_FREE(pairs);

    VPRINTF(NORMAL,"Saving output catalogs\n");
    close_outcat(in_cat, out_cat, nfields);
    return 0;
}
