#!/usr/bin/env python

"""
SCRIPT NAME:
  detect_OMEGACAM_bad_chip_23.py

  identify and mark chip 23 OMEGACAM data suffering from gain jumps

ARGUMENTS:
  1. main directory     (e.g. /data/thomas/OMEGACAM/run_11_11_f)
  2. science directory  (e.g. SCIENCE_i_SDSS)
  3. image ending       (e.g. OFCSF)

DESCRIPTION:
  Chip 23 (THELI numbering) of OMEGACAM images shows severe discrete
  gain jumps in an unpredictable way (the problem can occur spontaneously
  in an exposure ans alo vanish again as quickly). The signature
  are discrete broad regions of varying sky background in an exposure.

  We detect the problem in the following way:
 
  (1) determine the median of each row
  (2) determine the derivative of the median vector of the previous
      step.
  (3) the gain jump problem manifests itself in a larger number of high
      absolute values of the derivative determined in the previous step.
      The numer of high derivative values and the 'high value' itself
      are the free parameters of the algorithm.
 
  Data idetified as bad get the BADCCD keyword set to 1.

AUTHOR:
  Thomas Erben (terben@astro.uni-bonn.de)
"""

# HISTORY COMMENTS:
#
# 04.08.2015:
# project started: Thomas Erben
#
# 07.08.2015:
# A status image was not correct - I forgot to mention the image under
# consideration. Fixed!

import sys
import os
import glob
import multiprocessing
import numpy as np
import datetime as dt
import astropy.io.fits as aif


def test_bad_chip23(image):
    # The threshold of high derivative values that we consider as problematic:
    deriv_threshold = 20

    # The number of high derivative values an image must have to be considered
    # bad:
    n_high_deriv = 5

    print "Testing image %s" % (image)

    try:
        data, header = aif.getdata(image, header=True)

        # do nothing if the chip is already marked as bad; probably
        # this script was run already:
        if 'BADCCD' not in header or header['BADCCD'] == 0:
            # determine median vector of rows and their derivative
            median = np.median(data, axis=1)
            deriv = median[1:] - median[:-1]

            # If the image is identified as suffering from discrete
            # GAIN jumps we set the BADCCD keyword to 1:
            if deriv[np.abs(deriv) > deriv_threshold].size >= n_high_deriv:
                print "Image %s identified as bad!" % (image)

                if 'BADCCD' not in header:
                    print "Warning! Image %s has no BADCCD keyword. I add it!"

                header['BADCCD'] = 1

                # add History information to the header:
                currdate = dt.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
                header['HISTORY'] = ""
                header['HISTORY'] = "%s: called at %s" % \
                                    (os.path.basename(sys.argv[0]), currdate)
                header['HISTORY'] = "image identified to suffer " \
                                    "from gain jumps!"
            
                # just overwrite the old file:
                aif.writeto(image, data, header, clobber=True)
        else:
            print "BADCCD keyword of image %s already set to 1 " \
                  "I skip testing this image" % (image)

    except IOError:
        print "I cannot open file %s. I skip it!" % (image)

# Here the main script starts:
if __name__ == "__main__":

    if len(sys.argv) != 4:
        print __doc__
        sys.exit(1)

    # main directory
    MD = sys.argv[1]
    # science directory
    SD = sys.argv[2]
    # image ending
    ending = sys.argv[3]

    # collect lists of files to process:
    imagepath = "%s/%s/*_23%s.fits" % (MD, SD, ending)

    images = glob.glob(imagepath)

    # do we need to do something at all?
    if len(images) == 0:
        print "No images of chip 23 with ending %s found in %s/%s" % \
            (ending, SD, WD)
        sys.exit(1)

    # do the testing of images in parallel if possible:

    # Initialise process pool:
    pool = multiprocessing.Pool()

    # execute the conversion with a 'pool-map' command:
    pool.map(test_bad_chip23, images)

    sys.exit(0)
